/*
 * Copyright (c) 2004
 *	Hartmut Brandt
 * 	All rights reserved.
 *
 * Author: Hartmut Brandt <harti@freebsd.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Begemot: atmsupport/ats_cc/kern.c,v 1.2 2004/08/17 16:32:20 brandt Exp $
 */
#include <sys/types.h>
#include <sys/queue.h>
#include <sys/socket.h>

#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <err.h>

#include <netnatm/unimsg.h>
#include <netnatm/msg/unistruct.h>
#include <netnatm/msg/unimsglib.h>
#include <netnatm/sig/unidef.h>
#include <netnatm/api/unisap.h>
#include <netnatm/api/atmapi.h>
#include <netnatm/api/ccatm.h>

#include <netgraph.h>
#include <netgraph/atm/ng_uni.h>
#include <netgraph/atm/ng_ccatm.h>

#include "ats.h"
#include "testlib.h"
#include "test_cc.h"

#define	MAX_MESSAGE	10000
#define	MAX_DATA	60000

/* all current UNI ports and their list */
struct port {
	LIST_ENTRY(port) link;
	u_int	portno;
	char	hook[NG_HOOKSIZ];
};
static LIST_HEAD(, port) ports = LIST_HEAD_INITIALIZER(ports);

/* all current endpoints */
struct user {
	LIST_ENTRY(user) link;
	char	hook[NG_HOOKSIZ];
};
static LIST_HEAD(, user) users = LIST_HEAD_INITIALIZER(users);

/* netgraph sockets */
static int dsock, csock;

/* true if we have the manage hook */
static int have_manage;

/*
 *        [ats_cc]
 *           |dump
 *           |dump
 *        [ccatm]
 */
static struct stack def[] = {
	{ STACK_MKPEER,  ".",  "ccatm",  "dump",  "dump",    "ccatm" },
	{ STACK_END,     NULL, NULL,     NULL,     NULL,     NULL }
};

static void kern_destroy_user(const char *);
static void kern_destroy_port(u_int);

/*
 * Ascii-fy a control message. If we are here, the message is unexpected.
 */
static void
dump_cmsg(struct ng_mesg *m, const char *path)
{
	struct ng_mesg *a = (struct ng_mesg *)(void *)m->data;

	if (NgSendMsg(csock, path, NGM_GENERIC_COOKIE, NGM_BINARY2ASCII,
	    m, sizeof(*m) + m->header.arglen) == -1)
		err(1, "NgSendMsg: BINARY2ASCII");
	if (NgRecvMsg(csock, m, MAXMSG, NULL) == -1)
		err(1, "NgRecvMsg: BINARY2ASCII");

	printf("received %s '%s'(%d) from '%s':\n",
	    (a->header.flags & NGF_RESP) ? "response" : "request",
	    a->header.cmdstr, a->header.cmd, path);

	if (*a->data != '\0')
		printf("Args: '%s'\n", a->data);
}

/*
 * Read a control message
 */
static struct ng_mesg *
read_msg(char *path)
{
	struct ng_mesg *msg;

	msg = malloc(MAX_MESSAGE);
	if (msg == NULL)
		err(1, NULL);
	if (NgRecvMsg(csock, msg, MAX_MESSAGE, path) == -1)
		err_file("reading control message: %s", strerror(errno));

	return (msg);
}

/*
 * Something happend on the control socket.
 */
static void
ready_cmsg(int fd __unused, void *uap __unused)
{
	char path[NG_PATHSIZ];
	struct ng_mesg *msg;

	msg = read_msg(path);
	dump_cmsg(msg, path);
	free(msg);
}

/*
 * Wait for a reponse
 */
static struct ng_mesg *
wait_resp(int token)
{
	struct ng_mesg *resp;
	char path[NG_PATHSIZ];

	for (;;) {
		resp = read_msg(path);
		if ((u_int)token == resp->header.token)
			break;
		dump_cmsg(resp, path);
	}
	return (resp);
}

/*
 * CC output to a UNI port
 */
static void
port_input(struct port *port, void *buf, size_t len)
{
	struct uni_arg *uarg;
	struct uni_msg *msg;

	if (len < sizeof(*uarg))
		errx(1, "port output too short %zu", len);
	uarg = (struct uni_arg *)buf;

	msg = uni_msg_alloc(len - sizeof(*uarg));
	if (msg == NULL)
		err(1, NULL);

	memcpy(msg->b_wptr, uarg->data, len - sizeof(*uarg));
	msg->b_wptr += len - sizeof(*uarg);

	handle_uni_data(port->portno, uarg->sig, uarg->cookie, msg);
}

/*
 * CC output to an EP port
 */
static void
user_input(struct user *user, void *buf, size_t len)
{
	struct ccatm_op op;
	struct atm_resp resp;

	if (len < sizeof(op))
		errx(1, "user output too short %zu", len);
	memcpy(&op, buf, sizeof(op));
	buf = (u_char *)buf + sizeof(op);
	len -= sizeof(op);

	if (op.op != ATMOP_RESP) {
		handle_api_data(user->hook, op.op, buf, len);
		return;
	}
	if (len < sizeof(resp))
		errx(1, "user response too short %zu", len);
	memcpy(&resp, buf, sizeof(resp));
	buf = (u_char *)buf + sizeof(resp);
	len -= sizeof(resp);

	handle_api_response(user->hook, resp.resp, resp.data, buf, len);
}

/*
 * Input at the data socket
 */
static void
ready_dmsg(int fd __unused, void *uap __unused)
{
	char hook[NG_HOOKSIZ];
	void *buf;
	ssize_t ret;
	struct port *port;
	struct user *user;

	/* Get the message */
	if ((buf = malloc(MAX_DATA)) == NULL)
		err(1, NULL);
	if ((ret = NgRecvData(dsock, buf, MAX_DATA, hook)) == -1)
		err(1, "NgRecvData");

	if (strcmp(hook, "manage") == 0)
		err_file("unexpected input on management hook");

	if (strcmp(hook, "dump") == 0) {
		printf("%.*s", (int)ret, (char *)buf);
		free(buf);
		return;
	}

	LIST_FOREACH(port, &ports, link)
		if (strcmp(hook, port->hook) == 0) {
			port_input(port, buf, (size_t)ret);
			free(buf);
			return;
		}

	LIST_FOREACH(user, &users, link)
		if (strcmp(hook, user->hook) == 0) {
			user_input(user, buf, (size_t)ret);
			free(buf);
			return;
		}

#if 0
size_t r;

printf("%zu:", (size_t)ret);
for (r = 0; r < (size_t)ret; r++)
printf("%02x ", ((u_char *)buf)[r]);
printf("\n");
#endif

	err_file("unexpected input from hook '%s'", hook);
}

/*
 * Set log level
 */
static void
kern_set_log(u_int log_level)
{
	uint32_t data = log_level;
	int token;
	struct ng_mesg *resp;

	if ((token = NgSendMsg(csock, ".:dump", NGM_CCATM_COOKIE,
	    NGM_CCATM_SETLOG, &data, sizeof(data))) == -1)
		err_file("NGM_CCATM_SETLOG: %s", strerror(errno));

	resp = wait_resp(token);
	free(resp);
}

/*
 * Instantiate a user space stack
 */
static void
kern_init(u_int log_level)
{

	build_stack(&csock, &dsock, def);

	ats_register_fd(csock, ready_cmsg, NULL);
	ats_register_fd(dsock, ready_dmsg, NULL);

	kern_set_log(log_level);
}

/*
 * Reset the cc
 */
static void
kern_reset(void)
{
	struct port *port;
	struct user *user;
	void *buf;
	ssize_t ret;

	while ((port = LIST_FIRST(&ports)) != NULL)
		kern_destroy_port(port->portno);

	while ((user = LIST_FIRST(&users)) != NULL)
		kern_destroy_user(user->hook);

	if (NgSendAsciiMsg(csock, ".:dump", "reset") == -1)
		err(1, "resetting cc");

	/* flush messages that are in the ng_socket's buffer */
	if ((buf = malloc(MAX_DATA)) == NULL)
		err(1, NULL);
	for (;;) {
		ret = recv(dsock, buf, MAX_DATA, MSG_DONTWAIT);
		if (ret == -1) {
			if (errno == EAGAIN)
				break;
			err(1, "recv(MSG_DONTWAIT)");
		}
	}
	free(buf);
}

/*
 * Start the CC
 */
static void
kern_start(void)
{
	struct ngm_connect msg;

	if (!have_manage) {
		strcpy(msg.path, "ccatm:");
		strcpy(msg.ourhook, "manage");
		strcpy(msg.peerhook, "manage");

		if (NgSendMsg(csock, ".", NGM_GENERIC_COOKIE, NGM_CONNECT,
		    &msg, sizeof(msg)) == -1)
			err(1, "NgSendMsg: NGM_CONNECT");

		have_manage = 1;
	}
}

/*
 * Handle events
 */
static void
kern_work(void)
{
}

/*
 * Dump state
 */
static void
kern_dump(void)
{

	if (NgSendAsciiMsg(csock, ".:dump", "dump") == -1)
		err(1, "dumping");
}

/*
 * Send a message from the lower layer to CC
 */
static void
kern_uni_to_cc(u_int portno, u_int sig, u_int cookie, struct uni_msg *msg)
{
	struct port *port;
	struct uni_arg *uarg;
	ssize_t ret;

	LIST_FOREACH(port, &ports, link)
		if (port->portno == portno)
			break;

	if (port == NULL)
		err_file("bad port %u", portno);

	if ((uarg = malloc(sizeof(*uarg) + uni_msg_len(msg))) == NULL)
		err(1, NULL);

	uarg->sig = sig;
	uarg->cookie = cookie;
	memcpy(uarg->data, msg->b_rptr, uni_msg_len(msg));

	if ((ret = NgSendData(dsock, port->hook, (const u_char *)uarg,
	    sizeof(*uarg) + uni_msg_len(msg))) == -1)
		err(1, "NgSendData(%s)", port->hook);

	uni_msg_destroy(msg);
	free(uarg);
}

/*
 * Send a message from the user to CC
 */
static void
kern_api_to_cc(const char *name, u_int sig, struct uni_msg *msg)
{
	struct ccatm_op *uarg;
	struct user *user;
	ssize_t ret;

	LIST_FOREACH(user, &users, link)
		if (strcmp(user->hook, name) == 0)
			break;

	if (user == NULL)
		err_file("bad endpoint '%s'", name);

	if ((uarg = malloc(sizeof(*uarg) + uni_msg_len(msg))) == NULL)
		err(1, NULL);

	uarg->op = sig;
	memcpy(uarg->data, msg->b_rptr, uni_msg_len(msg));

	if ((ret = NgSendData(dsock, user->hook, (const u_char *)uarg,
	    sizeof(*uarg) + uni_msg_len(msg))) == -1)
		err(1, "NgSendData(%s)", user->hook);

	uni_msg_destroy(msg);
	free(uarg);
}

/*
 * Create a new user/endpoint
 */
static void *
kern_create_user(const char *ep)
{
	struct user *user;
	struct ngm_connect msg;

	user = ats_alloc(sizeof(*user));
	if (strlcpy(user->hook, ep, sizeof(user->hook)) >= sizeof(user->hook))
		err_file("endpoint name too long '%s'", ep);

	strcpy(msg.path, "ccatm:");
	strcpy(msg.ourhook, ep);
	strcpy(msg.peerhook, ep);

	if (NgSendMsg(csock, ".", NGM_GENERIC_COOKIE, NGM_CONNECT,
	    &msg, sizeof(msg)) == -1)
		err(1, "NgSendMsg: NGM_CONNECT");

	LIST_INSERT_HEAD(&users, user, link);
	return (user);
}

/*
 * Destroy an endpoint/user
 */
static void
kern_destroy_user(const char *ep)
{
	struct user *user;
	struct ngm_rmhook msg;

	LIST_FOREACH(user, &users, link)
		if (strcmp(user->hook, ep) == 0)
			break;

	if (user == NULL)
		err_file("bad endpoint '%s'", ep);

	strcpy(msg.ourhook, ep);

	if (NgSendMsg(csock, ".", NGM_GENERIC_COOKIE, NGM_RMHOOK,
	    &msg, sizeof(msg)) == -1)
		err(1, "NgSendMsg: NGM_RMHOOK");

	LIST_REMOVE(user, link);
	free(user);
}

/*
 * Create a new UNI port
 */
static void *
kern_create_port(u_int portno)
{
	struct port *port;
	struct ngm_connect msg;

	LIST_FOREACH(port, &ports, link)
		if (port->portno == portno)
			err_file("port %u already exists", portno);

	port = ats_alloc(sizeof(*port));
	port->portno = portno;
	snprintf(port->hook, sizeof(port->hook), "uni%u", portno);

	strcpy(msg.path, "ccatm:");
	strcpy(msg.ourhook, port->hook);
	strcpy(msg.peerhook, port->hook);

	if (NgSendMsg(csock, ".", NGM_GENERIC_COOKIE, NGM_CONNECT,
	    &msg, sizeof(msg)) == -1)
		err(1, "NgSendMsg: NGM_CONNECT");

	LIST_INSERT_HEAD(&ports, port, link);
	return (port);
}

/*
 * Disconnect a port
 */
static void
kern_destroy_port(u_int portno)
{
	struct port *port;
	struct ngm_rmhook msg;

	LIST_FOREACH(port, &ports, link)
		if (port->portno == portno)
			break;

	if (port == NULL)
		err_file("bad port %u", portno);

	strcpy(msg.ourhook, port->hook);

	if (NgSendMsg(csock, ".", NGM_GENERIC_COOKIE, NGM_RMHOOK,
	    &msg, sizeof(msg)) == -1)
		err(1, "NgSendMsg: NGM_RMHOOK");

	LIST_REMOVE(port, link);
	free(port);
}

/*
 * Clear addresses
 */
static void
kern_clear_port(u_int portno)
{
	struct port *port;
	struct ngm_ccatm_port msg;

	LIST_FOREACH(port, &ports, link)
		if (port->portno == portno)
			break;

	if (port == NULL)
		err_file("bad port %u", portno);

	msg.port = portno;

	if (NgSendMsg(csock, ".:dump", NGM_CCATM_COOKIE, NGM_CCATM_CLEAR,
	    &msg, sizeof(msg)) == -1)
		err(1, "NgSendMsg: NGM_CCATM_CLEAR");
}

/*
 * Set port parameters
 */
static void
kern_port_set_info(const struct atm_port_info *info)
{

	if (NgSendMsg(csock, ".:dump", NGM_CCATM_COOKIE,
	    NGM_CCATM_SET_PORT_PARAM, info, sizeof(*info)) == -1)
		err(1, "NgSendMsg: NGM_CCATM_SET_PORT_PARAM");
}

/*
 * Retrieve port info
 */
static void
kern_port_get_info(u_int portno, struct atm_port_info *info)
{
	struct port *port;
	struct ngm_ccatm_port msg;
	int token;
	struct ng_mesg *resp;

	LIST_FOREACH(port, &ports, link)
		if (port->portno == portno)
			break;

	if (port == NULL)
		err_file("bad port %u", portno);

	msg.port = portno;

	if ((token = NgSendMsg(csock, ".:dump", NGM_CCATM_COOKIE,
	    NGM_CCATM_GET_PORT_PARAM, &msg, sizeof(msg))) == -1)
		err(1, "NgSendMsg: NGM_CCATM_GET_PORT_PARAM");

	resp = wait_resp(token);
	if (resp->header.arglen != sizeof(*info))
		err(1, "port_info: got %u, need %zu", resp->header.arglen,
		    sizeof(*info));
	memcpy(info, resp->data, sizeof(*info));
	free(resp);
}

/*
 * Return list of ports
 */
static void
kern_list_ports(u_int *nports, u_int **pp)
{
	int token;
	struct ng_mesg *resp;
	struct ngm_ccatm_portlist *list;
	u_int i;

	if ((token = NgSendMsg(csock, ".:dump", NGM_CCATM_COOKIE,
	    NGM_CCATM_GET_PORTLIST, NULL, 0)) == -1)
		err(1, "NgSendMsg: NGM_CCATM_GET_PORTLIST");

	resp = wait_resp(token);
	if (resp->header.arglen < sizeof(*list))
		err(1, "portlist header too short");
	list = (struct ngm_ccatm_portlist *)(void *)resp->data;
	if (resp->header.arglen != sizeof(*list) + list->nports *
	    sizeof(list->ports[0]))
		err(1, "portlist too short");

	*nports = list->nports;
	if ((*pp = malloc(sizeof(**pp) * list->nports)) == NULL)
		err(1, NULL);
	for (i = 0; i < list->nports; i++)
		(*pp)[i] = list->ports[i];
	free(resp);
}

/*
 * Start/stop port
 */
static void
kern_admin_port(u_int portno, int flag)
{
	struct port *port;
	struct ngm_ccatm_port msg;

	LIST_FOREACH(port, &ports, link)
		if (port->portno == portno)
			break;

	if (port == NULL)
		err_file("bad port %u", portno);

	msg.port = portno;

	if (flag) {
		if (NgSendMsg(csock, ".:dump", NGM_CCATM_COOKIE,
		    NGM_CCATM_START, &msg, sizeof(msg)) == -1)
			err(1, "NgSendMsg: NGM_CCATM_START");
	} else {
		if (NgSendMsg(csock, ".:dump", NGM_CCATM_COOKIE,
		    NGM_CCATM_STOP, &msg, sizeof(msg)) == -1)
			err(1, "NgSendMsg: NGM_CCATM_STOP");
	}
}

/*
 * Register an address
 */
static int
kern_register_addr(u_int portno, const struct uni_addr *addr)
{
	struct port *port;
	struct ngm_ccatm_addr_req msg;

	LIST_FOREACH(port, &ports, link)
		if (port->portno == portno)
			break;

	if (port == NULL)
		err_file("bad port %u", portno);

	msg.port = port->portno;
	msg.addr = *addr;

	if (NgSendMsg(csock, ".:dump", NGM_CCATM_COOKIE,
	    NGM_CCATM_ADDRESS_REGISTERED, &msg, sizeof(msg)) == -1)
		return (errno);
	return (0);
}

/*
 * Unregister and address
 */
static int
kern_unregister_addr(u_int portno, const struct uni_addr *addr)
{
	struct port *port;
	struct ngm_ccatm_addr_req msg;

	LIST_FOREACH(port, &ports, link)
		if (port->portno == portno)
			break;

	if (port == NULL)
		err_file("bad port %u", portno);

	msg.port = port->portno;
	msg.addr = *addr;

	if (NgSendMsg(csock, ".:dump", NGM_CCATM_COOKIE,
	    NGM_CCATM_ADDRESS_UNREGISTERED, &msg, sizeof(msg)) == -1)
		return (errno);
	return (0);
}

/*
 * Get address list from port or all ports
 */
static void
kern_query_addr_list(u_int portno, struct uni_addr **addrs, u_int **pp,
    u_int *naddrs)
{
	struct ngm_ccatm_port msg;
	struct ng_mesg *resp;
	struct ngm_ccatm_get_addresses *list;
	int token;
	u_int i;

	msg.port = portno;

	if ((token = NgSendMsg(csock, ".:dump", NGM_CCATM_COOKIE,
	    NGM_CCATM_GET_ADDRESSES, &msg, sizeof(msg))) == -1)
		err(1, "NgSendMsg: NGM_CCATM_GET_ADDRESSES");

	resp = wait_resp(token);
	if (resp->header.arglen < sizeof(*list))
		err(1, "address list header too short");
	list = (struct ngm_ccatm_get_addresses *)(void *)resp->data;
	if (resp->header.arglen != sizeof(*list) + list->count *
	    sizeof(list->addr[0]))
		err(1, "address list too short");

	*naddrs = list->count;
	if ((*pp = malloc(sizeof(**pp) * list->count)) == NULL)
		err(1, NULL);
	if ((*addrs = malloc(sizeof(**addrs) * list->count)) == NULL)
		err(1, NULL);

	for (i = 0; i < list->count; i++) {
		(*pp)[i] = list->addr[i].port;
		(*addrs)[i] = list->addr[i].addr;
	}
	free(resp);
}

/*
 * Unmanage
 */
static void
kern_unmanage(void)
{
	struct ngm_rmhook msg;

	if (!have_manage)
		err_file("already unmanaged");
	have_manage = 0;

	strcpy(msg.ourhook, "manage");

	if (NgSendMsg(csock, ".", NGM_GENERIC_COOKIE, NGM_RMHOOK,
	    &msg, sizeof(msg)) == -1)
		err(1, "NgSendMsg: NGM_RMHOOK");
}

/*
 * Retrieve extended status information
 */
static struct exstat *
kern_get_exstat(void)
{
	struct ng_mesg *resp;
	struct atm_exstatus *s;
	struct exstat *e;
	int token;
	size_t offs;

	if ((token = NgSendMsg(csock, ".:dump", NGM_CCATM_COOKIE,
	    NGM_CCATM_GET_EXSTAT, NULL, 0)) == -1)
		err(1, "NgSendMsg: NGM_CCATM_GET_EXSTAT");

	resp = wait_resp(token);
	if (resp->header.arglen < sizeof(*s))
		err(1, "exstat header too short");
	s = (struct atm_exstatus *)(void *)resp->data;

	offs = sizeof(*s) + s->neps * sizeof(*e->ep) +
	    s->nports * sizeof(*e->port) + s->nconns * sizeof(*e->conn) +
	    s->nparties * sizeof(*e->party);
	if (resp->header.arglen != offs)
		err(1, "exstat too short: %u, need %zu", resp->header.arglen,
		    offs);

	if ((e = ats_alloc(sizeof(*e))) == NULL)
		err(1, NULL);

	memcpy(&e->status, resp->data, sizeof(e->status));
	offs = sizeof(e->status);

	e->ep = (struct atm_exstatus_ep *)(resp->data + offs);
	offs += sizeof(*e->ep) * s->neps;

	e->port = (struct atm_exstatus_port *)(void *)(resp->data + offs);
	offs += sizeof(*e->port) * s->nports;

	e->conn = (struct atm_exstatus_conn *)(void *)(resp->data + offs);
	offs += sizeof(*e->conn) * s->nconns;

	e->party = (struct atm_exstatus_party *)(void *)(resp->data + offs);
	offs += sizeof(*e->party) * s->nparties;

	return (e);
}

/*
 * Free the status
 */
static void
kern_free_exstat(struct exstat *ex)
{
	free((u_char *)ex->ep - offsetof(struct exstat, ep) -
	    offsetof(struct ng_mesg, data));
	free(ex);
}

/*
 * Function vector
 */
const struct cc kernel_cc = {
	kern_init,
	kern_reset,
	kern_start,
	kern_work,
	kern_dump,
	kern_uni_to_cc,
	kern_api_to_cc,
	kern_create_user,
	kern_destroy_user,
	kern_create_port,
	kern_destroy_port,
	kern_clear_port,
	kern_port_set_info,
	kern_port_get_info,
	kern_list_ports,
	kern_admin_port,
	kern_register_addr,
	kern_unregister_addr,
	kern_query_addr_list,
	kern_unmanage,
	kern_get_exstat,
	kern_free_exstat,
	kern_set_log,
};
