/*
 * Copyright (c) 2004
 *	Hartmut Brandt
 * 	All rights reserved.
 *
 * Author: Hartmut Brandt <harti@freebsd.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Begemot: atmsupport/ats_cc/main.c,v 1.8 2004/08/17 18:15:47 brandt Exp $
 */

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <netgraph.h>

#define	index foo_index
#define	err foo_err
#include <tcl.h>
#undef index
#undef err

#include <netnatm/unimsg.h>
#include <netnatm/msg/unistruct.h>
#include <netnatm/msg/unimsglib.h>
#include <netnatm/sig/unidef.h>
#include <netnatm/api/unisap.h>
#include <netnatm/api/atmapi.h>
#include <unitcl.h>

#include "ats.h"
#include "test_cc.h"

/* library context for parsing and formating */
static struct unicx unicx;

/* TCL interpreter context for parsing and formating */
static Tcl_Interp *tcl;

/* the UNI messages we are waiting for */
static struct waitlist *wait_uni_data;

/* the API messages we are waiting for */
static struct waitlist *wait_api_data;

/* which stack to use */
static const struct cc *cc = &kernel_cc;

/* log level */
static u_int log_level;

/* signals to the UNI */
static struct eventlist *waiting_to_uni;

/* signals to the API */
static struct eventlist *waiting_to_api;

static void def_var(char *);

int
main(int argc, char *argv[])
{
	int opt;
	u_long val;
	char *end;
	int ng_debug = 0;

#ifdef FreeBSD
	_malloc_options = "JA";
#endif

	while ((opt = getopt(argc, argv, "l:nqt:uv:")) != -1)
		switch (opt) {

		  case 'l':
			errno = 0;
			val = strtoul(optarg, &end, 0);
			if (*end != '\0' || errno != 0 || end == optarg ||
			    val > UINT_MAX) {
				fprintf(stderr, "bad log value '%s'\n", optarg);
				exit(1);
			}
			log_level = val;
			break;

		  case 'n':
			NgSetDebug(++ng_debug);
			break;

		  case 'q':
			if (report_level > 0)
				report_level--;
			break;

		  case 't':
			errno = 0;
			val = strtoul(optarg, &end, 10);
			if (*end != '\0' || errno != 0 || end == optarg ||
			    val > UINT_MAX) {
				fprintf(stderr, "bad trace value '%s'\n",
				    optarg);
				exit(1);
			}
			trace_level = val;
			break;

		  case 'u':
			cc = &user_cc;
			break;

		  case 'v':
			while (*optarg != '\0') {
				SETVERB(*optarg);
				optarg++;
			}
			break;
		}

	uni_initcx(&unicx);
	unicx.q2932 = 2;

	if ((tcl = Tcl_CreateInterp()) == NULL) {
		perror("Tcl_CreateInterp");
		exit(1);
	}

	test_init();

	wait_uni_data = waitlist_create("UNI");
	wait_api_data = waitlist_create("API");
	waiting_to_uni = eventlist_create("TO UNI", wait_uni_data);
	waiting_to_api = eventlist_create("TO API", wait_api_data);

	cc->init(log_level);

	while (optind < argc) {
		if (strchr(argv[optind], '=') != NULL)
			def_var(argv[optind]);
		else {
			input_open_file(argv[optind]);
			test_exec();
		}
		optind++;
	}

	if (ISVERB('F'))
		funcs_dump();

	exit(report_total() ? 1 : 0);
}

static void
def_var(char *str)
{
	char *eq;
	struct var *v;

	eq = strchr(str, '=');
	*eq++ = '\0';
	v = var_find(str, NULL);
	var_setn(v, eq, strlen(eq));
}

/*
 * Return non-zero if we are waiting for something
 */
int
test_wait_cb(void)
{

//	cc->work();
	return (0);
}

/*
 * Parse a UNIAPI message
 */
static void
parse_uniapi_msg(const char *word, u_int *portno, u_int *sig,
    u_int *cookie, union uniapi_all *u, size_t *len)
{
	int ac;
	const char **av;
	char *end;

	if (Tcl_SplitList(tcl, word, &ac, &av) != TCL_OK)
		err_file("%s -- %s", curr_key, tcl->result);

	if (ac < 3)
		err_file("%s -- missing args", curr_key);

	memset(u, 0, sizeof(*u));
	if (parse_uniapi_sig(tcl, av[0], sig) != TCL_OK)
		err_file("%s -- %s", curr_key, tcl->result);

	errno = 0;
	*portno = strtoul(av[1], &end, 10);
	if (*end != '\0' || errno != 0 || end == av[1])
		err_file("%s -- bad port", curr_key);

	errno = 0;
	*cookie = strtoul(av[2], &end, 10);
	if (*end != '\0' || errno != 0 || end == av[2])
		err_file("%s -- bad cookie", curr_key);

	if (parse_uniapi(tcl, ac - 3, av + 3, *sig, u, len, &unicx) != TCL_OK)
		err_file("%s -- %s", curr_key, tcl->result);

	Tcl_Free((char *)av);
}

/*
 * Data input to CC from UNI
 */
static void
act_from_uni(struct line *l)
{
	char *word;
	struct uni_msg *msg;
	u_int sig, cookie, portno;
	union uniapi_all u;
	size_t len;

	if ((word = skip_blanks(l)) == NULL)
		err_file("%s -- missing message", curr_key);

	trace(3, "from UNI: %s", word);
	parse_uniapi_msg(word, &portno, &sig, &cookie, &u, &len);

	if ((msg = uni_msg_alloc(len)) == NULL) {
		perror(NULL);
		exit(1);
	}
	memcpy(msg->b_wptr, &u, len);
	msg->b_wptr += len;

	cc->uni_to_cc(portno, sig, cookie, msg);
}

/*
 * Wait for a message to the UNI from CC
 */
static void
act_wait_to_uni(struct line *l)
{

	waitlist_put(wait_uni_data, l);
	test_wait_events();
}
static void
act_grp_wait_to_uni(struct line *l)
{

	waitlist_put(wait_uni_data, l);
}

static void
dump_msg(const u_char *ptr, size_t len)
{
	u_int i;

	for (i = 0; i < len; i++) {
		if (i % 16 == 0)
			printf("%04u", i);
		if (i % 8 == 0)
			printf(" ");
		printf(" %02x", ptr[i]);
		if (i % 16 == 15)
			printf("\n");
	}
	if (i % 16 != 0 && i != 0)
		printf("\n");
}

/*
 * handle message from CC to lower layer
 */
void
handle_uni_data(u_int portno, u_int sig, u_int cookie, struct uni_msg *msg)
{
	Tcl_DString str;
	char sbuf[100];

	Tcl_DStringInit(&str);
	if (fmt_uniapi_sig(tcl, &str, sig) != TCL_OK) {
		fprintf(stderr, "uniapi signal -- %s\n", tcl->result);
		exit(1);
	}

	sprintf(sbuf, "%u", portno);
	Tcl_DStringAppendElement(&str, sbuf);

	sprintf(sbuf, "%u", cookie);
	Tcl_DStringAppendElement(&str, sbuf);

	if (msg != NULL) {
		if (ISVERB('D'))
			dump_msg(msg->b_rptr, uni_msg_len(msg));
		if (fmt_uniapi(tcl, &str, sig, msg->b_rptr, uni_msg_len(msg))
		    != TCL_OK) {
			fprintf(stderr, "uniapi message -- %s\n", tcl->result);
			exit(1);
		}
		uni_msg_destroy(msg);
	}

	eventlist_put(waiting_to_uni, Tcl_DStringValue(&str));
	Tcl_DStringFree(&str);
}

/*
 * Wait for a message from CC to the API
 */
static void
act_wait_to_api(struct line *l)
{

	waitlist_put(wait_api_data, l);
	test_wait_events();
}
static void
act_grp_wait_to_api(struct line *l)
{

	waitlist_put(wait_api_data, l);
}

/*
 * Parse a ATMAPI message
 */
static void
parse_api_msg(const char *word, char *epname, u_int *sig,
    u_char *buf, size_t *len)
{
	int ac;
	const char **av;

	if (Tcl_SplitList(tcl, word, &ac, &av) != TCL_OK)
		err_file("%s -- %s", curr_key, tcl->result);

	if (ac < 2)
		err_file("%s -- missing args", curr_key);

	if (parse_atmapi_sig(tcl, av[0], sig) != TCL_OK)
		err_file("%s -- %s", curr_key, tcl->result);

	if (strlen(av[1]) >= ATM_EPNAMSIZ)
		err_file("%s -- endpoint name too long", curr_key);
	strcpy(epname, av[1]);

	if (parse_atmapi(tcl, ac - 2, av + 2, *sig, buf, len, &unicx) != TCL_OK)
		err_file("%s -- %s", curr_key, tcl->result);

#if 0
	{
		u_int i;

		for (i = 0; i < *len; i++)
			printf("%02x ", buf[i]);
		printf("\n");
	}
#endif

	Tcl_Free((char *)av);
}

/*
 * Data input to CC from API
 */
static void
act_from_api(struct line *l)
{
	char *word;
	u_int sig;
	char epname[ATM_EPNAMSIZ];
	size_t len;
	struct uni_msg *msg;

	if ((word = skip_blanks(l)) == NULL)
		err_file("%s -- missing message", curr_key);

	trace(3, "api to cc: %s", word);

	if (ISVERB('A')) {
		char *str = ats_strdup(word);
		compress_blanks(str);
		printf("CC<-API: %s\n", str);
		free(str);
	}

	if ((msg = uni_msg_alloc(10000)) == NULL) {
		perror(NULL);
		exit(1);
	}
	parse_api_msg(word, epname, &sig, msg->b_wptr, &len);
	msg->b_wptr += len;

	cc->api_to_cc(epname, sig, msg);
}

/*
 * handle data output to API
 */
void
handle_api_data(const char *epname, u_int op, void *arg, size_t len)
{
	Tcl_DString str;

	if (ISVERB('D'))
		dump_msg(arg, len);

	Tcl_DStringInit(&str);
	if (fmt_atmapi_sig(tcl, &str, op) != TCL_OK) {
		fprintf(stderr, "atmapi signal -- %s\n", tcl->result);
		exit(1);
	}

	Tcl_DStringAppendElement(&str, epname);

	if (fmt_atmapi(tcl, &str, op, arg, len) != TCL_OK) {
		fprintf(stderr, "atmapi message -- %s\n", tcl->result);
		exit(1);
	}

	if (ISVERB('A'))
		printf("CC->API: %s\n", Tcl_DStringValue(&str));

	eventlist_put(waiting_to_api, Tcl_DStringValue(&str));
	Tcl_DStringFree(&str);
}

/*
 * handle response to API
 */
void
handle_api_response(const char *epname, int e, u_int d, void *arg, size_t len)
{
	Tcl_DString str;
	struct uni_msg *msg;

	if (ISVERB('D'))
		dump_msg(arg, len);

	msg = uni_msg_alloc(len + sizeof(struct atm_resp));
	if (msg == NULL) {
		perror(NULL);
		exit(1);
	}
	uni_msg_rptr(msg, struct atm_resp *)->resp = e;
	uni_msg_rptr(msg, struct atm_resp *)->data = d;
	memcpy(msg->b_rptr + sizeof(struct atm_resp), arg, len);
	msg->b_wptr += sizeof(struct atm_resp) + len;

	Tcl_DStringInit(&str);
	if (fmt_atmapi_sig(tcl, &str, ATMOP_RESP) != TCL_OK) {
		fprintf(stderr, "atmapi signal -- %s\n", tcl->result);
		exit(1);
	}

	Tcl_DStringAppendElement(&str, epname);

	if (fmt_atmapi(tcl, &str, ATMOP_RESP, msg->b_rptr,
	    uni_msg_len(msg)) != TCL_OK) {
		fprintf(stderr, "atmapi response -- %s\n", tcl->result);
		exit(1);
	}
	uni_msg_destroy(msg);

	if (ISVERB('A'))
		printf("CC->API: %s\n", Tcl_DStringValue(&str));

	eventlist_put(waiting_to_api, Tcl_DStringValue(&str));
	Tcl_DStringFree(&str);
}

/*
 * Reset everything
 */
void
test_reset_cb(int tout __unused)
{

	cc->reset();
}

/*
 * Callback to start a test
 */
void
test_start_cb(void)
{

	cc->start();
}

static void
act_create_user(struct line *l)
{
	char *word;

	if ((word = get_word(l)) == NULL)
		err_file("%s -- missing endpoint name", curr_key);
	if (strlen(word) >= ATM_EPNAMSIZ)
		err_file("%s -- endpoint name too long", curr_key);

	if (cc->create_user(word) == NULL)
		err_file("%s -- cannot create user", curr_key);

	if (skip_blanks(l) != NULL)
		err_file("%s -- extra args", curr_key);
}

static void
act_destroy_user(struct line *l)
{
	char *word;

	if ((word = get_word(l)) == NULL)
		err_file("%s -- missing endpoint name", curr_key);
	if (strlen(word) >= ATM_EPNAMSIZ)
		err_file("%s -- endpoint name too long", curr_key);

	cc->destroy_user(word);

	if (skip_blanks(l) != NULL)
		err_file("%s -- extra args", curr_key);
}

static u_int
get_port(struct line *l)
{
	char *word, *end;
	u_long portno;

	if ((word = get_word(l)) == NULL)
		err_file("%s -- missing port number", curr_key);
	errno = 0;
	portno = strtoul(word, &end, 10);
	if (*end != '\0' || errno != 0 || portno == 0 || portno > UINT32_MAX)
		err_file("%s -- bad port number", curr_key);

	if (get_word(l) != NULL)
		err_file("%s -- extra args", curr_key);

	return (portno);
}

/*
 * Create/destroy port
 */
static void
act_create_port(struct line *l)
{

	if (cc->create_port(get_port(l)) == NULL)
		err_file("%s -- could not create port", curr_key);
}
static void
act_destroy_port(struct line *l)
{

	cc->destroy_port(get_port(l));
}
static void
act_start_port(struct line *l)
{

	cc->admin_port(get_port(l), 1);
}
static void
act_stop_port(struct line *l)
{

	cc->admin_port(get_port(l), 0);
}
static void
act_clear_port(struct line *l)
{

	cc->clear_port(get_port(l));
}

static void
get_addr(struct line *l, int *exerr, u_int *portno, struct uni_addr *addr)
{
	char *word, *end;
	const char *av[2];
	u_long n;
	long ll;

	if ((word = get_word(l)) == NULL)
		err_file("%s -- missing error code", curr_key);
	if (strcmp(word, "EISCONN") == 0)
		ll = EISCONN;
	else {
		errno = 0;
		ll = strtol(word, &end, 10);
		if (errno != 0 || *end != '\0')
			err_file("%s -- bad error code", curr_key);
	}
	*exerr = ll;

	if ((word = get_word(l)) == NULL)
		err_file("%s -- missing port number", curr_key);
	errno = 0;
	n = strtoul(word, &end, 10);
	if (*end != '\0' || errno != 0 || n == 0 || n > UINT32_MAX)
		err_file("%s -- bad port number", curr_key);
	*portno = n;

	if ((word = get_word(l)) == NULL)
		err_file("%s -- missing address type number", curr_key);
	av[0] = word;
	if ((word = get_word(l)) == NULL)
		err_file("%s -- missing address", curr_key);
	av[1] = word;

	if (parse_addr(tcl, av, addr) != TCL_OK)
		err_file("%s -- %s", tcl->result, curr_key);

	if (get_word(l) != NULL)
		err_file("%s -- extra args", curr_key);
}

static void
act_register_addr(struct line *l)
{
	u_int portno;
	struct uni_addr addr;
	int error, exerr;

	get_addr(l, &exerr, &portno, &addr);
	if ((error = cc->register_addr((u_int)portno, &addr)) != exerr)
		err_file("%s -- error %s", strerror(error), curr_key);
}
static void
act_unregister_addr(struct line *l)
{
	u_int portno;
	struct uni_addr addr;
	int error, exerr;

	get_addr(l, &exerr, &portno, &addr);
	if ((error = cc->unregister_addr((u_int)portno, &addr)) != exerr)
		err_file("%s -- error %s", strerror(error), curr_key);
}

static void
act_query_addr(struct line *l)
{
	const char *word;
	char *end;
	u_long portno;
	struct uni_addr *addrs;
	u_int naddrs, i, *ports;
	Tcl_DString str;
	char buf[20];

	if ((word = get_word(l)) == NULL)
		err_file("%s -- missing port number", curr_key);

	errno = 0;
	portno = strtoul(word, &end, 10);
	if (*end != '\0' || errno != 0 || end == word)
		err_file("%s -- bad port", curr_key);
	
	if ((word = skip_blanks(l)) == NULL)
		word = "";

	cc->query_addr_list((u_int)portno, &addrs, &ports, &naddrs);

	Tcl_DStringInit(&str);
	for (i = 0; i < naddrs; i++) {
		sprintf(buf, "%u", ports[i]);
		Tcl_DStringAppendElement(&str, buf);
		Tcl_DStringStartSublist(&str);
		if (fmt_addr(tcl, &str, &addrs[i]) != TCL_OK)
			err_file("%s -- %s", curr_key, tcl->result);
		Tcl_DStringEndSublist(&str);
	}
	if (match(word, Tcl_DStringValue(&str)) != 0)
		report_problem("address list does not match");

	Tcl_DStringFree(&str);
	free(addrs);
}

static void
act_unmanage(struct line *l)
{

	if (get_word(l) != NULL)
		err_file("%s -- extra args", curr_key);
	cc->unmanage();
}

static void
act_list_ports(struct line *l)
{
	u_int nports, *ports, i;
	Tcl_DString str;
	char buf[100];
	const char *word;

	cc->list_ports(&nports, &ports);

	Tcl_DStringInit(&str);
	for (i = 0; i < nports; i++) {
		sprintf(buf, "%u", ports[i]);
		Tcl_DStringAppendElement(&str, buf);
	}

	if ((word = skip_blanks(l)) == NULL)
		word = "";

	if (match(word, Tcl_DStringValue(&str)) != 0)
		report_problem("port list does not match");

	Tcl_DStringFree(&str);
	free(ports);
}

static void
act_set_port(struct line *l)
{
	const char *word;
	int ac;
	const char **av;
	struct atm_port_info info;

	if ((word = skip_blanks(l)) == NULL)
		err_file("%s -- no port info", curr_key);

	if (Tcl_SplitList(tcl, word, &ac, &av) != TCL_OK)
		err_file("%s -- %s", curr_key, tcl->result);

	if (parse_port_info(tcl, ac, av, &info) != TCL_OK)
		err_file("%s -- %s", curr_key, tcl->result);

	cc->port_set_info(&info);

	Tcl_Free((char *)av);
}

static void
act_get_port(struct line *l)
{
	const char *word;
	u_long portno;
	struct atm_port_info info;
	Tcl_DString str;

	if ((word = skip_blanks(l)) == NULL)
		err_file("%s -- missing port number", curr_key);

	errno = 0;
	portno = strtoul(word, NULL, 10);
	if (errno != 0)
		err_file("%s -- bad port", curr_key);
	
	cc->port_get_info(portno, &info);

	Tcl_DStringInit(&str);
	if (fmt_port_info(&str, &info) != TCL_OK)
		err_file("%s -- %s", curr_key, tcl->result);

	if (match(word, Tcl_DStringValue(&str)) != 0)
		report_problem("port info does not match");

	Tcl_DStringFree(&str);
}

static void
act_work(struct line *l)
{

	if (get_word(l) != NULL)
		err_file("%s -- extra args", curr_key);
	cc->work();
}

static void
act_dump(struct line *l)
{

	if (get_word(l) != NULL)
		err_file("%s -- extra args", curr_key);
	cc->dump();
}

static void
act_exstat(struct line *l)
{
	const char *word;
	struct exstat *ex;
	Tcl_DString str;

	ex = cc->get_exstat();
	Tcl_DStringInit(&str);
	if (atmapi_fmt_exstatus(tcl, &str, &ex->status, ex->ep,
	    ex->port, ex->conn, ex->party) != TCL_OK)
		err_file("%s -- %s", curr_key, tcl->result);

	if ((word = skip_blanks(l)) == NULL)
		word = "";

	if (match(word, Tcl_DStringValue(&str)) != 0)
		report_problem("extended status does not match");

	Tcl_DStringFree(&str);
	cc->free_exstat(ex);
}

static void
act_flags(struct line *l)
{
	const char *w, *a;
	char *end;
	u_long val;

	while ((w = get_word(l)) != NULL) {
		while (*w != '\0') {
			switch (*w++) {

			  case 'l':
				if ((a = get_word(l)) == NULL)
					err_file("'l' requires argument");
				errno = 0;
				val = strtoul(a, &end, 0);
				if (*end != '\0' || errno != 0 || end == a ||
				    val > UINT_MAX)
					err_file("bad log value '%s'\n", a);
				log_level = val;
				cc->set_log(log_level);
				break;

			  case 't':
				if ((a = get_word(l)) == NULL)
					err_file("'a' requires argument");
				errno = 0;
				val = strtoul(a, &end, 10);
				if (*end != '\0' || errno != 0 || end == a ||
				    val > UINT_MAX)
					err_file("bad trace value '%s'\n", a);
				trace_level = val;
				break;

			  case 'v':
				if ((a = get_word(l)) == NULL)
					err_file("'v' requires argument");
				while (*a != '\0') {
					SETVERB(*a);
					a++;
				}
				break;

			  default:
				err_file("bad flag '%c'", w[-1]);
			}
		}
	}
}

/* user defined keys */
const struct key usrkeys[] = {
	{ "*>",		ACT_RUNNING, 1, 1, act_from_uni },
	{ "*<",		ACT_RUNNING, 0, 1, act_wait_to_uni },
	{ "*+",		ACT_RUNNING, 1, 1, act_create_port },
	{ "*-",		ACT_RUNNING, 1, 1, act_destroy_port },
	{ "P+",		ACT_RUNNING, 1, 1, act_start_port },
	{ "P-",		ACT_RUNNING, 1, 1, act_stop_port },
	{ "*C",		ACT_RUNNING, 0, 1, act_clear_port },
	{ "*S",		ACT_RUNNING, 1, 1, act_set_port },
	{ "*G",		ACT_RUNNING, 0, 1, act_get_port },
	{ "*L",		ACT_RUNNING, 0, 1, act_list_ports },
	{ "*A+",	ACT_RUNNING, 1, 1, act_register_addr },
	{ "*A-",	ACT_RUNNING, 1, 1, act_unregister_addr },
	{ "*AQ",	ACT_RUNNING, 0, 1, act_query_addr },
	{ "@>",		ACT_RUNNING, 1, 1, act_from_api },
	{ "@<",		ACT_RUNNING, 0, 1, act_wait_to_api },
	{ "@+",		ACT_RUNNING, 1, 1, act_create_user },
	{ "@-",		ACT_RUNNING, 1, 1, act_destroy_user },
	{ "^-",		ACT_RUNNING, 1, 1, act_unmanage },
	{ "W",		ACT_RUNNING, 1, 1, act_work },
	{ "D",		ACT_RUNNING, 1, 1, act_dump },
	{ "EX",		ACT_RUNNING, 1, 1, act_exstat },
	{ "%%",		ACT_RUNNING, 1, 0, act_flags },
	{ NULL,		0,	     0, 0, NULL }
};
const struct key usrgrpkeys[] = {
	{ "*",		ACT_RUNNING, 0, 1, act_grp_wait_to_uni },
	{ "@",		ACT_RUNNING, 0, 1, act_grp_wait_to_api },
	{ NULL,		0,	     0, 0, NULL }
};

