/*
 * Copyright (c) 2004
 *	Hartmut Brandt
 * 	All rights reserved.
 *
 * Author: Hartmut Brandt <harti@freebsd.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Begemot: atmsupport/ats_cc/user.c,v 1.6 2004/07/16 18:57:10 brandt Exp $
 */
#include <sys/types.h>
#include <sys/queue.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <unistd.h>
#include <ctype.h>
#include <limits.h>

#include <netnatm/unimsg.h>
#include <netnatm/msg/unistruct.h>
#include <netnatm/msg/unimsglib.h>
#include <netnatm/saal/sscfudef.h>
#include <netnatm/sig/unidef.h>
#include <netnatm/api/unisap.h>
#include <netnatm/api/atmapi.h>
#include <netnatm/api/ccatm.h>

#include "ats.h"
#include "test_cc.h"

static void user_send_user(struct ccuser *, void *, u_int, void *, size_t);
static void user_resp_user(struct ccuser *, void *, int, u_int, void *, size_t);
static void user_send_uni(struct ccconn *, void *, u_int, u_int,
    struct uni_msg *);
static void user_send_uni_glob(struct ccport *, void *, u_int, u_int,
    struct uni_msg *);
static void user_log(const char *, ...);

static struct cc_funcs cc_funcs = {
	user_send_user,
	user_resp_user,
	user_send_uni,
	user_send_uni_glob,
	user_log
};

/*
 * User arg
 */
struct uarg {
	LIST_ENTRY(uarg) link;
	char	name[ATM_EPNAMSIZ];
	struct ccuser *user;
};

/*
 * Port arg
 */
struct parg {
	LIST_ENTRY(parg) link;
	u_int	portno;
	struct ccport *port;
};

static struct ccdata *cc;
static LIST_HEAD(, uarg) users = LIST_HEAD_INITIALIZER(users);
static LIST_HEAD(, parg) ports = LIST_HEAD_INITIALIZER(ports);

static void
user_init(u_int log_level)
{

	if ((cc = cc_create(&cc_funcs)) == NULL)
		fprintf(stderr, "cannot create cc instance: %s\n",
		    strerror(errno));
	cc_set_log(cc, log_level);
}

/*
 * Reset the cc
 */
static void
user_reset(void)
{

	cc_reset(cc);
}

/*
 * Start the CC
 */
static void
user_start(void)
{
}

/*
 * Handle events
 */
static void
user_work(void)
{
	cc_work(cc);
}

/*
 * Dump state
 */
static int
user_dump_func(struct ccdata *c __unused, void *uarg __unused, const char *buf)
{

	printf("%s", buf);
	return (0);
}
static void
user_dump(void)
{

	cc_dump(cc, BUFSIZ, user_dump_func, NULL);
}

/*
 * Send a message from the lower layer to CC
 */
static void
user_uni_to_cc(u_int portno, u_int sig, u_int cookie, struct uni_msg *msg)
{
	struct parg *parg;
	int e;

	LIST_FOREACH(parg, &ports, link)
		if (parg->portno == portno)
			break;

	if (parg == NULL)
		err_file("bad port %u", portno);

	if (sig == UNIAPI_ERROR) {
		struct uniapi_error *r = uni_msg_rptr(msg,
		    struct uniapi_error *);

		e = cc_uni_response(parg->port, cookie, r->reason, r->state);
		if (e != 0)
			err_file("cc_uni_response returns %d", e);

		uni_msg_destroy(msg);
	} else {
		e = cc_uni_signal(parg->port, cookie, sig, msg);
		if (e != 0)
			err_file("cc_uni_signal returns %d", e);
	}
	cc_work(cc);
}

/*
 * Send a message from the user to CC
 */
static void
user_api_to_cc(const char *name, u_int sig, struct uni_msg *msg)
{
	struct uarg *uarg;
	int e;

	LIST_FOREACH(uarg, &users, link)
		if (strcmp(uarg->name, name) == 0)
			break;

	if (uarg == NULL)
		err_file("bad endpoint '%s'", name);

	if ((e = cc_user_signal(uarg->user, sig, msg)) != 0)
		err_file("user signal returns %s", strerror(e));
	cc_work(cc);
}

/*
 * send signal to API user
 */
static void
user_send_user(struct ccuser *user __unused, void *uarg, u_int op,
    void *arg, size_t len)
{
	struct uarg *u = uarg;

	handle_api_data(u->name, op, arg, len);
}

/*
 * send response to API user
 */
static void
user_resp_user(struct ccuser *user __unused, void *uarg, int e, u_int d,
    void *arg, size_t len)
{
	struct uarg *u = uarg;

	handle_api_response(u->name, e, d, arg, len);
}

/*
 * send signal to uni for connection
 */
static void
user_send_uni(struct ccconn *conn __unused, void *parg, u_int sig, u_int cookie,
    struct uni_msg *msg)
{
	struct parg *p = parg;

	handle_uni_data(p->portno, sig, cookie, msg);
}

/*
 * send global signal to uni
 */
static void
user_send_uni_glob(struct ccport *port __unused, void *parg, u_int sig,
    u_int cookie, struct uni_msg *msg)
{
	struct parg *p = parg;

	handle_uni_data(p->portno, sig, cookie, msg);
}

/*
 * log a message
 */
static void
user_log(const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	fprintf(stderr, "cc: ");
	vfprintf(stderr, fmt, ap);
	fprintf(stderr, "\n");
	va_end(ap);
}

static void *
user_create_user(const char *ep)
{
	struct uarg *uarg;
	struct ccuser *u;

	uarg = ats_alloc(sizeof(*uarg));
	strcpy(uarg->name, ep);
	u = cc_user_create(cc, uarg, ep);
	if (u == NULL) {
		free(uarg);
		return (NULL);
	}
	uarg->user = u;
	LIST_INSERT_HEAD(&users, uarg, link);
	return (u);
}

static void
user_destroy_user(const char *ep)
{
	struct uarg *uarg;

	LIST_FOREACH(uarg, &users, link)
		if (strcmp(uarg->name, ep) == 0)
			break;

	if (uarg == NULL)
		err_file("bad endpoint '%s'", ep);

	cc_user_destroy(uarg->user);
	LIST_REMOVE(uarg, link);
	free(uarg);
	cc_work(cc);
}

static void *
user_create_port(u_int portno)
{
	struct parg *parg;
	struct ccport *p;

	parg = ats_alloc(sizeof(*parg));
	parg->portno = portno;
	p = cc_port_create(cc, parg, portno);
	if (p == NULL) {
		free(parg);
		return (NULL);
	}
	parg->port = p;
	LIST_INSERT_HEAD(&ports, parg, link);
	return (p);
}

static void
user_destroy_port(u_int portno)
{
	struct parg *parg;

	LIST_FOREACH(parg, &ports, link)
		if (parg->portno == portno)
			break;

	if (parg == NULL)
		err_file("bad port %u", portno);

	cc_port_destroy(parg->port, 0);
	LIST_REMOVE(parg, link);
	free(parg);
	cc_work(cc);
}

static void
user_admin_port(u_int portno, int flag)
{
	struct parg *parg;

	LIST_FOREACH(parg, &ports, link)
		if (parg->portno == portno)
			break;

	if (parg == NULL)
		err_file("bad port %u", portno);

	if (flag)
		cc_port_start(cc, portno);
	else
		cc_port_stop(cc, portno);
}
static void
user_clear_port(u_int portno)
{
	struct parg *parg;

	LIST_FOREACH(parg, &ports, link)
		if (parg->portno == portno)
			break;

	if (parg == NULL)
		err_file("bad port %u", portno);

	cc_port_clear(cc, portno);
}

static void
user_list_ports(u_int *nports, u_int **pp)
{
	int e;

	if ((e = cc_port_getlist(cc, nports, pp)) != 0)
		err_file("port_list: %s", strerror(e));
}

static int
user_register_addr(u_int portno, const struct uni_addr *addr)
{
	struct parg *parg;

	LIST_FOREACH(parg, &ports, link)
		if (parg->portno == portno)
			break;

	if (parg == NULL)
		err_file("bad port %u", portno);

	return (cc_addr_register(cc, portno, addr));
}

static int
user_unregister_addr(u_int portno, const struct uni_addr *addr)
{
	struct parg *parg;

	LIST_FOREACH(parg, &ports, link)
		if (parg->portno == portno)
			break;

	if (parg == NULL)
		err_file("bad port %u", portno);

	return (cc_addr_unregister(cc, portno, addr));
}

static void
user_query_addr_list(u_int portno, struct uni_addr **addrs, u_int **pp,
    u_int *naddrs)
{

	if (cc_get_addrs(cc, portno, addrs, pp, naddrs) != 0)
		err_file("cannot get addrs");
}

static void
user_unmanage(void)
{

	cc_unmanage(cc);
}

static void
user_port_set_info(const struct atm_port_info *info)
{
	int e;

	if ((e = cc_port_set_param(cc, info)) != 0)
		err_file("cannot set port info: %s", strerror(e));
}

static void
user_port_get_info(u_int portno, struct atm_port_info *info)
{
	int e;

	if ((e = cc_port_get_param(cc, portno, info)) != 0)
		err_file("cannot get port info: %s", strerror(e));
}

static struct exstat *
user_get_exstat(void)
{
	struct exstat *ex;
	int e;

	ex = ats_alloc(sizeof(*ex));
	if ((e = cc_get_extended_status(cc, &ex->status, &ex->ep, &ex->port,
	    &ex->conn, &ex->party)) != 0)
		err_file("cannot get extended status: %s", strerror(e));
	return (ex);
}

static void
user_free_exstat(struct exstat *ex)
{
	free(ex->ep);
	free(ex->port);
	free(ex->conn);
	free(ex->party);
	free(ex);
}

static void
user_set_log(u_int log_level)
{
	cc_set_log(cc, log_level);
}

const struct cc user_cc = {
	user_init,
	user_reset,
	user_start,
	user_work,
	user_dump,
	user_uni_to_cc,
	user_api_to_cc,
	user_create_user,
	user_destroy_user,
	user_create_port,
	user_destroy_port,
	user_clear_port,
	user_port_set_info,
	user_port_get_info,
	user_list_ports,
	user_admin_port,
	user_register_addr,
	user_unregister_addr,
	user_query_addr_list,
	user_unmanage,
	user_get_exstat,
	user_free_exstat,
	user_set_log,
};
