/*
 * Copyright (c) 2003-2003
 *	Fraunhofer Institute for Open Communication Systems (FhG Fokus).
 * 	All rights reserved.
 *
 * Author: Hartmut Brandt <harti@freebsd.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Begemot: atmsupport/ats_sig/kern.c,v 1.4 2004/08/18 12:49:50 brandt Exp $
 */
#ifdef HAVE_NETGRAPH_H
#include <sys/types.h>
#include <sys/queue.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <unistd.h>
#include <ctype.h>
#include <limits.h>
#include <err.h>

#include <netnatm/unimsg.h>
#include <netnatm/msg/unistruct.h>
#include <netnatm/msg/unimsglib.h>
#include <netnatm/saal/sscopdef.h>
#include <netnatm/saal/sscfudef.h>
#include <netnatm/sig/unidef.h>
#include <netnatm/sig/uni.h>

#include <netgraph.h>
#include <netgraph/atm/ng_uni.h>
#include <netgraph/atm/ng_sscfu.h>

#include "ats.h"
#include "testlib.h"
#include "test_sig.h"

/* netgraph sockets */
static int dsock, csock;

/*
 *      [ ats_uni ]
 *   upper|    |lower
 *   upper|    |
 *      [uni]  |
 *   lower|    |
 *        +----+
 */
static struct stack def[] = {
	{ STACK_MKPEER,  ".",  "uni",  "upper", "upper", "uni" },
	{ STACK_CONNECT, ".",  "uni:", "lower",  "lower",  NULL },
	{ STACK_END,     NULL, NULL,     NULL,     NULL,     NULL }
};

/*
 * Reset the stack
 */
static void
kernel_reset(void)
{

	if (NgSendAsciiMsg(csock, ".:upper", "disable") == -1)
		err(1, "disable stack");
}

/*
 * Start the stack
 */
static void
kernel_start(const u_int *debug, const struct uni_config *config,
    uint32_t m1, uint32_t m2, uint32_t m3)
{
	u_int i;
	struct ngm_uni_debug dbg;
	struct ngm_uni_set_config cfg;
	struct ngm_uni_config_mask *mask;
	struct ng_mesg *resp;
	int token;

	(void)NgSendAsciiMsg(csock, ".:upper", "disable");

	memset(&dbg, 0, sizeof(dbg));
	for (i = 0; i < UNI_MAXFACILITY; i++)
		dbg.level[i] = debug[i];
	if (NgSendMsg(csock, ".:upper", NGM_UNI_COOKIE, NGM_UNI_SETDEBUG,
	    &dbg, sizeof(dbg)) == -1)
		err_file("NGM_UNI_SETDEBUG: %s", strerror(errno));

	memset(&cfg, 0, sizeof(cfg));
	cfg.config = *config;
	cfg.mask.mask = m1;
	cfg.mask.popt_mask = m2;
	cfg.mask.option_mask = m3;
	token = NgSendMsg(csock, ".:upper", NGM_UNI_COOKIE, NGM_UNI_SET_CONFIG,
	    &cfg, sizeof(cfg));
	if (token == -1)
		err_file("NGM_UNI_SET_CONFIG: %s", strerror(errno));

	resp = malloc(10000);
	if (resp == NULL)
		err(1, NULL);
	if (NgRecvMsg(csock, resp, 10000, NULL) == -1)
		err_file("NGM_UNI_SET_CONFIG response: %s", strerror(errno));
	if ((u_int)token != resp->header.token)
		err_file("NGM_UNI_SET_CONFIG response: bad token");
	mask = (struct ngm_uni_config_mask *)(void *)resp->data;
	
	if (mask->mask != 0 || mask->popt_mask != 0 || mask->option_mask != 0)
		err_file("could not set config %#x %#x %#x", mask->mask,
		    mask->popt_mask, mask->option_mask);

	free(resp);

	if (NgSendAsciiMsg(csock, ".:upper", "enable") == -1)
		err(1, "enable stack");
}

/*
 * Get configuration
 */
static void
kernel_getconfig(struct uni_config *config)
{
	struct ng_mesg *resp;
	int token;

	token = NgSendMsg(csock, ".:upper", NGM_UNI_COOKIE, NGM_UNI_GET_CONFIG,
	    NULL, 0);
	if (token == -1)
		err_file("NGM_UNI_GET_CONFIG: %s", strerror(errno));

	resp = malloc(10000);
	if (resp == NULL)
		err(1, NULL);
	if (NgRecvMsg(csock, resp, 10000, NULL) == -1)
		err_file("NGM_UNI_GET_CONFIG response: %s", strerror(errno));
	if ((u_int)token != resp->header.token)
		err_file("NGM_UNI_GET_CONFIG response: bad token");

	*config = *(struct uni_config *)(void *)resp->data;

	free(resp);
}

/*
 * From SAAL to UNI
 */
static void
kernel_saal_to_uni(struct uni_msg *msg)
{
	struct sscfu_arg *arg;
	int ret;

	arg = malloc(sizeof(*arg) + uni_msg_len(msg));
	if (arg == NULL)
		err(1, NULL);
	arg->sig = SAAL_DATA_indication;
	memcpy(arg->data, msg->b_rptr, uni_msg_len(msg));

	ret = NgSendData(dsock, "lower", (const u_char *)arg,
	    sizeof(*arg) + uni_msg_len(msg));
	if (ret == -1)
		err(1, "NgSendData(lower)");

	free(arg);
	uni_msg_destroy(msg);
}

static void
send_simple_sig(u_int sig)
{
	struct sscfu_arg *arg;
	int ret;

	arg = malloc(sizeof(*arg));
	if (arg == NULL)
		err(1, NULL);
	arg->sig = sig;

	ret = NgSendData(dsock, "lower", (const u_char *)arg, sizeof(*arg));
	if (ret == -1)
		err(1, "NgSendData(lower)");

	free(arg);
}

static void
kernel_saal_establish_ind(void)
{
	send_simple_sig(SAAL_ESTABLISH_indication);
}

static void
kernel_saal_establish_conf(void)
{
	send_simple_sig(SAAL_ESTABLISH_confirm);
}

static void
kernel_saal_release_ind(void)
{
	send_simple_sig(SAAL_RELEASE_indication);
}

static void
kernel_api_to_uni(u_int sig, u_int cookie, struct uni_msg *msg)
{
	struct uni_arg *arg;
	int ret;

	arg = malloc(sizeof(*arg) + uni_msg_len(msg));
	if (arg == NULL)
		err(1, NULL);
	arg->sig = sig;
	arg->cookie = cookie;
	memcpy(arg->data, msg->b_rptr, uni_msg_len(msg));

	ret = NgSendData(dsock, "upper", (const u_char *)arg,
	    sizeof(*arg) + uni_msg_len(msg));
	if (ret == -1)
		err(1, "NgSendData(upper)");

	free(arg);
	uni_msg_destroy(msg);
}

/*
 * Input at the data socket
 */
static void
ready_dmsg(int fd __unused, void *uap __unused)
{
	char hook[NG_HOOKLEN + 1];
	const size_t bufsiz = 60000;
	u_char *buf;
	ssize_t ret;
	struct sscfu_arg *larg;
	struct uni_arg *uarg;
	struct uni_msg *msg;

	if ((buf = malloc(bufsiz)) == NULL)
		err(1, NULL);

	ret = NgRecvData(dsock, buf, bufsiz, hook);
	if (ret == -1)
		err(1, "NgRecvData");

	if (strcmp(hook, "lower") == 0) {
		if ((size_t)ret < sizeof(*larg))
			errx(1, "lower layer input too short %zd", ret);
		larg = (struct sscfu_arg *)(void *)buf;

		switch (larg->sig) {

		  case SAAL_ESTABLISH_request:
			if ((size_t)ret > sizeof(*larg))
				report_problem("unexpected user data in "
				    "SAAL-ESTABLISH.request");
			handle_establish_request();
			break;

		  case SAAL_RELEASE_request:
			report_problem("unexpected SAAL-RELEASE.request");
			if ((size_t)ret > sizeof(*larg))
				report_problem("unexpected user data in "
				    "SAAL-RELEASE.request");
			break;

		  case SAAL_DATA_request:
			msg = uni_msg_alloc((size_t)ret - sizeof(*larg));
			if (msg == NULL)
				err(1, NULL);
			memcpy(msg->b_wptr, larg->data,
			    (size_t)ret - sizeof(*larg));
			msg->b_wptr += (size_t)ret - sizeof(*larg);
			handle_saal_data(msg);
			uni_msg_destroy(msg);
			break;

		  case SAAL_UDATA_request:
			report_problem("unexpected SAAL-UDATA.request");
			break;

		  default:
			err(1, "illegal SAAL signal from UNI: %u", larg->sig);
			break;
		}

	} else if (strcmp(hook, "upper") == 0) {
		if ((size_t)ret < sizeof(*uarg))
			errx(1, "upper layer input too short %zd", ret);
		uarg = (struct uni_arg *)(void *)buf;

		msg = uni_msg_alloc((size_t)ret - sizeof(*uarg));
		if (msg == NULL)
			err(1, NULL);
		memcpy(msg->b_wptr, uarg->data, (size_t)ret - sizeof(*uarg));
		msg->b_wptr += (size_t)ret - sizeof(*uarg);

		handle_api_data(uarg->sig, uarg->cookie, msg);
		uni_msg_destroy(msg);

	} else
		errx(1, "input from unknown hook '%s'", hook);
	free(buf);
}

/*
 * Something happend on the control socket. Try to read a message and to
 * dump it.
 */
static void
dump_cmsg(struct ng_mesg *m, const char *path)
{
	struct ng_mesg *a = (struct ng_mesg *)(void *)m->data;

	if (NgSendMsg(csock, path, NGM_GENERIC_COOKIE, NGM_BINARY2ASCII,
	    m, sizeof(*m) + m->header.arglen) < 0)
		err(1, "NgSendMsg: BINARY2ASCII");
	if (NgRecvMsg(csock, m, MAXMSG, NULL) < 0)
		err(1, "NgRecvMsg: BINARY2ASCII");

	printf("received %s '%s'(%d) from '%s':\n",
	    (a->header.flags & NGF_RESP) ? "response" : "request",
	    a->header.cmdstr, a->header.cmd, path);

	if (*a->data != '\0')
		printf("Args: '%s'\n", a->data);
}

static void
ready_cmsg(int fd __unused, void *uap __unused)
{
	char path[NG_PATHLEN + 1];
	u_char buf[MAXMSG];
	struct ng_mesg *m = (struct ng_mesg *)(void *)buf;

	if (NgRecvMsg(csock, m, sizeof(buf), path) < 0)
		err(1, "NgRecvMsg");
	dump_cmsg(m, path);
}

/*
 * Instantiate a user space stack
 */
static void
kernel_init(void)
{

	build_stack(&csock, &dsock, def);

	ats_register_fd(csock, ready_cmsg, NULL);
	ats_register_fd(dsock, ready_dmsg, NULL);
}

static void
kernel_work(void)
{
}

const struct ustack kernel_stack = {
	kernel_init,
	kernel_reset,
	kernel_start,
	kernel_saal_to_uni,
	kernel_saal_establish_ind,
	kernel_saal_establish_conf,
	kernel_saal_release_ind,
	kernel_work,
	kernel_api_to_uni,
	kernel_getconfig
};
#endif
