/*
 * Copyright (c) 2003-2003
 *	Fraunhofer Institute for Open Communication Systems (FhG Fokus).
 * 	All rights reserved.
 *
 * Author: Hartmut Brandt <harti@freebsd.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Begemot: atmsupport/ats_sig/main.c,v 1.27 2004/08/18 12:49:51 brandt Exp $
 */

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#define	index foo_index
#include <tcl.h>
#undef index

#include <netnatm/unimsg.h>
#include <netnatm/msg/unistruct.h>
#include <netnatm/msg/unimsglib.h>
#include <netnatm/sig/unidef.h>
#include <unitcl.h>

#include "ats.h"
#include "test_sig.h"

/* library context for parsing and formating */
static struct unicx unicx;

/* TCL interpreter context for parsing and formating */
static Tcl_Interp *tcl;

/* the SAAL messages we are waiting for */
static struct waitlist *wait_saal_data;

/* the API messages we are waiting for */
static struct waitlist *wait_api_data;

/* wait for establish request */
static int wait_establish_req;

/* which stack to use */
#ifdef HAVE_NETGRAPH_H
static const struct ustack *stack = &kernel_stack;
#else
static const struct ustack *stack = &user_stack;
#endif

/* ignore all API signals */
static u_int api_ignore;

static void act_from_saal(struct line *);
static void act_from_saal_patch(struct line *);
static void act_wait_to_saal(struct line *);
static void act_group_to_saal(struct line *);
static void act_saal_establish_ind(struct line *);
static void act_saal_establish_conf(struct line *);
static void act_saal_release_ind(struct line *);
static void act_wait_to_api(struct line *);
static void act_group_to_api(struct line *);
static void act_from_api(struct line *);
static void act_api_ignore_start(struct line *);
static void act_api_ignore_end(struct line *);
static void act_wait_establish_req(struct line *);
static void act_group_establish_req(struct line *);

/* user defined keys */
const struct key usrkeys[] = {
	{ "*>",		ACT_RUNNING, 1, 1, act_from_saal },
	{ "**>",	ACT_RUNNING, 1, 1, act_from_saal_patch },
	{ "*+>",	ACT_RUNNING, 1, 1, act_saal_establish_ind },
	{ "*!>",	ACT_RUNNING, 1, 1, act_saal_establish_conf },
	{ "*+<",	ACT_RUNNING, 1, 1, act_wait_establish_req },
	{ "*->",	ACT_RUNNING, 1, 1, act_saal_release_ind },
	{ "*<",		ACT_RUNNING, 0, 1, act_wait_to_saal },
	{ "@<",		ACT_RUNNING, 0, 1, act_wait_to_api },
	{ "@>",		ACT_RUNNING, 1, 1, act_from_api },
	{ "@+",		ACT_RUNNING, 1, 1, act_api_ignore_start },
	{ "@-",		ACT_RUNNING, 1, 1, act_api_ignore_end },
	{ NULL,		0,	     0, 0, NULL }
};
const struct key usrgrpkeys[] = {
	{ "*",		ACT_RUNNING, 0, 1, act_group_to_saal },
	{ "@",		ACT_RUNNING, 0, 1, act_group_to_api },
	{ "*+",		ACT_RUNNING, 1, 1, act_group_establish_req },
	{ NULL,		0,	     0, 0, NULL }
};

/* stack configuration */
static struct uni_config config;

static void make_opts(void);
static void def_var(char *);

static const char *const debug_fac[] = {
#define	UNI_DEBUG_DEFINE(D) [UNI_FAC_##D] = #D,
	UNI_DEBUG_FACILITIES
#undef UNI_DEBUG_DEFINE
	NULL
};

static u_int dbg_flags[UNI_MAXFACILITY];

int
main(int argc, char *argv[])
{
	int opt;
	u_long val;
	char *end;
	u_int i;

#ifdef FreeBSD
	_malloc_options = "JA";
#endif

	while ((opt = getopt(argc, argv, "d:qt:uv:")) != -1)
		switch (opt) {

		  case 'd':
			if ((end = strchr(optarg, '=')) != NULL)
				*end++ = '\0';
			for (i = 0; debug_fac[i] != NULL; i++)
				if (strcasecmp(optarg, debug_fac[i]) == 0)
					break;
			if (debug_fac[i] == NULL) {
				fprintf(stderr, "unknown facility '%s'\n",
				    optarg);
				continue;
			}
			if (end == 0)
				dbg_flags[i]++;
			else
				dbg_flags[i] = strtoul(end, NULL, 0);
			break;

		  case 'q':
			if (report_level > 0)
				report_level--;
			break;

		  case 't':
			errno = 0;
			val = strtoul(optarg, &end, 10);
			if (*end != '\0' || errno != 0 || end == optarg ||
			    val > UINT_MAX) {
				fprintf(stderr, "bad trace value '%s'\n",
				    optarg);
				exit(1);
			}
			trace_level = val;
			break;

		  case 'u':
			stack = &user_stack;
			break;

		  case 'v':
			while (*optarg != '\0') {
				SETVERB(*optarg);
				optarg++;
			}
			break;
		}

	uni_initcx(&unicx);
	unicx.q2932 = 2;

	if ((tcl = Tcl_CreateInterp()) == NULL) {
		fprintf(stderr, "Tcl_CreateInterp: %s", strerror(errno));
		exit(1);
	}

	test_init();

	wait_saal_data = waitlist_create("SAAL");
	wait_api_data = waitlist_create("API");

	stack->init();

	make_opts();

	while (optind < argc) {
		if (strchr(argv[optind], '=') != NULL)
			def_var(argv[optind]);
		else {
			input_open_file(argv[optind]);
			test_exec();
		}
		optind++;
	}

	if (ISVERB('F'))
		funcs_dump();

	exit(report_total() ? 1 : 0);
}

static void
def_var(char *str)
{
	char *eq;
	struct var *v;

	eq = strchr(str, '=');
	*eq++ = '\0';
	v = var_find(str, NULL);
	var_setn(v, eq, strlen(eq));
}

/*
 * Set a TIMER
 */
static void
timer_set(struct value *val, const char *str, size_t len)
{
	char sbuf[100];
	char *end;
	u_long j;
	uint32_t *t = (uint32_t *)(void *)val->val;

	if (len > 30)
		err_file("bad timer value");
	strncpy(sbuf, str, len);
	sbuf[len] = '\0';

	errno = 0;
	j = strtoul(str, &end, 0);
	if (errno != 0 || *end != '\0' || j > UINT32_MAX)
		err_file("bad timer value");

	*t = j;
}

static char *
timer_get(struct value *val)
{
	static char sbuf[100];
	uint32_t *t = (uint32_t *)(void *)val->val;

	sprintf(sbuf, "%u", *t);
	return (sbuf);
}

static void
popt_set(struct value *val, const char *str, size_t len)
{
	char sbuf[100];
	char *end;
	u_long j;
	uint32_t bit = (uint32_t)(uintptr_t)val->val;

	if (len > 30)
		err_file("bad option value");
	strncpy(sbuf, str, len);
	sbuf[len] = '\0';

	errno = 0;
	j = strtoul(str, &end, 0);
	if (errno != 0 || *end != '\0' || (j != 0 && j != 1))
		err_file("bad option value");

	if (j)
		config.popt |= bit;
	else
		config.popt &= ~bit;
}

static char *
popt_get(struct value *val)
{
	static char sbuf[100];
	uint32_t bit = (uint32_t)(uintptr_t)val->val;

	sprintf(sbuf, "%u", (config.popt & bit) != 0);
	return (sbuf);
}

/*
 * Make option variables
 */
static void
make_opts(void)
{
	struct var *v;

	stack->getconfig(&config);

	v = var_define("OPT_T397", NULL, 0, NULL);
	v->v->get_val = timer_get;
	v->v->set_val = timer_set;
	v->v->val = (char *)&config.timer397;

	v = var_define("OPT_T398", NULL, 0, NULL);
	v->v->get_val = timer_get;
	v->v->set_val = timer_set;
	v->v->val = (char *)&config.timer398;

	v = var_define("OPT_T399", NULL, 0, NULL);
	v->v->get_val = timer_get;
	v->v->set_val = timer_set;
	v->v->val = (char *)&config.timer399;

	v = var_define("OPT_SB_TB", NULL, 0, NULL);
	v->v->get_val = popt_get;
	v->v->set_val = popt_set;
	v->v->val = (char *)UNIPROTO_SB_TB;

#if 0
	u_int32_t	proto;		/* which protocol */
	u_int32_t	popt;		/* protocol option */
	u_int32_t	option;		/* other options */
	u_int32_t	timer301;	/* T301 */
	u_int32_t	timer303;	/* T303 */
	u_int32_t	init303;	/* T303 retransmission count */
	u_int32_t	timer308;	/* T308 */
	u_int32_t	init308;	/* T308 retransmission count */
	u_int32_t	timer309;	/* T309 */
	u_int32_t	timer310;	/* T310 */
	u_int32_t	timer313;	/* T313 */
	u_int32_t	timer316;	/* T316 */
	u_int32_t	init316;	/* T316 retransmission count */
	u_int32_t	timer317;	/* T317 */
	u_int32_t	timer322;	/* T322 */
	u_int32_t	init322;	/* T322 retransmission count */
#endif
}

/*
 * Return non-zero if we are waiting for something
 */
int
test_wait_cb(void)
{

	stack->work();
	if (wait_establish_req == 1)
		return (1);
	if (wait_establish_req == 2) {
		wait_establish_req = 0;
		return (-1);
	}
	return (0);
}

/*
 * Make a UNI message from a string
 */
static struct uni_msg *
parse_uni_msg(const char *who, const char *arg)
{
	int ac, ret;
	const char **av;
	struct uni_all uni;
	struct uni_msg *msg;

	if (Tcl_SplitList(tcl, arg, &ac, &av) != TCL_OK)
		err_file("%s -- %s", who, tcl->result);

	if (ac < 3)
		err_file("%s -- missing args", who);

	memset(&uni, 0, sizeof(uni));
	if (parse_msghdr(tcl, ac, av, &uni.mtype, &uni.u.hdr) != TCL_OK)
		err_file("%s -- %s", who, tcl->result);
	if (parse_msg(tcl, ac - 3, av + 3, &uni, &unicx) != TCL_OK)
		err_file("%s -- %s", who, tcl->result);

	Tcl_Free((char *)av);

	if ((msg = uni_msg_alloc(1000)) == NULL)
		err_file("%s", strerror(errno));
	if ((ret = uni_encode(msg, &uni, &unicx)) != 0)
		err_file("%s -- cannot encode (%d)", who, ret);

	return (msg);
}

/*
 * Data input to UNI from SAAL
 */
static void
act_from_saal(struct line *l)
{
	char *word;
	struct uni_msg *msg;
	u_int n;

	if ((word = skip_blanks(l)) == NULL)
		err_file("*> -- missing message");

	trace(3, "to saal: %s", word);
	msg = parse_uni_msg("*>", word);

	if (ISVERB('D')) {
		for (n = 0; n < uni_msg_len(msg); n++)
			printf("%02x ", msg->b_rptr[n]);
		printf("\n");
	}

	stack->saal_to_uni(msg);
}

/*
 * Patch a message
 */
static void
patch_msg(struct uni_msg *msg, char *patch)
{
	char *arg;
	int off, n, byte;

	while ((arg = strsep(&patch, ",")) != NULL) {
		if (sscanf(arg, "%i:%i%n", &off, &byte, &n) != 2 ||
		    strlen(arg) != (size_t)n)
			err_file("bad patch '%s'", arg);
		if (off < 0 || (u_int)off > uni_msg_len(msg))
			err_file("patch offset out of range");
		if (byte < 0 || byte > 255)
			err_file("byte out of range");
		msg->b_rptr[off] = byte;
	}
}

/*
 * Data input to UNI from SAAL
 */
static void
act_from_saal_patch(struct line *l)
{
	char *word, *patch;
	struct uni_msg *msg;
	u_int n;

	if ((patch = get_word(l)) == NULL)
		err_file("**> -- missing patch");
	if ((word = skip_blanks(l)) == NULL)
		err_file("**> -- missing message");

	trace(3, "patch to saal: %s", word);
	msg = parse_uni_msg("**>", word);

	patch_msg(msg, patch);

	if (ISVERB('D')) {
		for (n = 0; n < uni_msg_len(msg); n++)
			printf("%02x ", msg->b_rptr[n]);
		printf("\n");
	}

	stack->saal_to_uni(msg);
}

/*
 * Wait for a SAAL signal to appear from the stack
 */
static void
act_group_to_saal(struct line *l)
{

	waitlist_put(wait_saal_data, l);
}
static void
act_wait_to_saal(struct line *l)
{

	waitlist_put(wait_saal_data, l);
	test_wait_events();
}

/*
 * handle data output to SAAL
 */
void
handle_saal_data(struct uni_msg *msg)
{
	int res;
	struct uni_all uni;
	Tcl_DString str;
	u_int n;

	if (ISVERB('D')) {
		for (n = 0; n < uni_msg_len(msg); n++)
			printf("%02x ", msg->b_rptr[n]);
		printf("\n");
	}

	if ((res = uni_decode(msg, &uni, &unicx)) != 0)
		err_file("saal data -- decoding error: %d", res);

	Tcl_DStringInit(&str);
	if (fmt_msghdr(tcl, &str, uni.mtype, &uni.u.hdr) != TCL_OK)
		err_file("saal data -- %s", tcl->result);
	if (fmt_msg(tcl, &str, &uni) != TCL_OK)
		err_file("saal data -- %s", tcl->result);

	waitlist_match(wait_saal_data, Tcl_DStringValue(&str));

	Tcl_DStringFree(&str);
}

static void
parse_api_msg(const char *who, const char *word, u_int *sig, u_int *cookie,
    union uniapi_all *u, size_t *len)
{
	int ac;
	const char **av;
	char *end;

	if (Tcl_SplitList(tcl, word, &ac, &av) != TCL_OK)
		err_file("%s -- %s", who, tcl->result);

	if (ac < 2)
		err_file("%s -- missing args", who);

	memset(u, 0, sizeof(*u));
	if (parse_uniapi_sig(tcl, av[0], sig) != TCL_OK)
		err_file("%s -- %s", who, tcl->result);

	errno = 0;
	*cookie = strtoul(av[1], &end, 10);
	if (*end != '\0' || errno != 0 || end == av[1])
		err_file("%s -- bad cookie", who);

	if (parse_uniapi(tcl, ac - 2, av + 2, *sig, u, len, &unicx) != TCL_OK)
		err_file("%s -- %s", who, tcl->result);

	Tcl_Free((char *)av);
}

/*
 * Wait for an API signal to appear from the stack
 */
static void
act_group_to_api(struct line *l)
{

	waitlist_put(wait_api_data, l);
}
static void
act_wait_to_api(struct line *l)
{

	waitlist_put(wait_api_data, l);
	test_wait_events();
}

/*
 * Data input to UNI from API
 */
static void
act_from_api(struct line *l)
{
	char *word;
	u_int sig, cookie;
	union uniapi_all u;
	size_t len;
	struct uni_msg *msg;

	if ((word = skip_blanks(l)) == NULL)
		err_file("@> -- missing message");

	trace(3, "to api: %s", word);
	parse_api_msg("@>", word, &sig, &cookie, &u, &len);

	if ((msg = uni_msg_alloc(len)) == NULL)
		err_file("%s", strerror(errno));
	memcpy(msg->b_wptr, &u, len);
	msg->b_wptr += len;

	stack->api_to_uni(sig, cookie, msg);
}

/*
 * handle data output to API
 */
void
handle_api_data(u_int sig, u_int cookie, struct uni_msg *msg)
{
	Tcl_DString str;
	char sbuf[100];

	Tcl_DStringInit(&str);
	if (fmt_uniapi_sig(tcl, &str, sig) != TCL_OK)
		err_file("api signal -- %s", tcl->result);

	sprintf(sbuf, "%u", cookie);
	Tcl_DStringAppendElement(&str, sbuf);

	if (msg != NULL)
		if (fmt_uniapi(tcl, &str, sig, msg->b_rptr, uni_msg_len(msg))
		    != TCL_OK)
			err_file("api message -- %s", tcl->result);

	if (api_ignore)
		waitlist_match_ign(wait_api_data, Tcl_DStringValue(&str));
	else
		waitlist_match(wait_api_data, Tcl_DStringValue(&str));

	Tcl_DStringFree(&str);
}

/*
 * Reset everything
 */
void
test_reset_cb(int tout __unused)
{

	stack->reset();
}

/*
 * Callback to start a test
 */
void
test_start_cb(void)
{

	stack->start(dbg_flags, &config, UNICFG_ALLMASK, UNIPROTO_ALLMASK,
	    UNIOPT_ALLMASK);
}

/*
 * Send establish indication
 */
static void
act_saal_establish_ind(struct line *l)
{
	if (get_word(l) != NULL)
		warn_file("junk after *+> ignored");

	stack->saal_establish_ind();
}

/*
 * Send establish confirm
 */
static void
act_saal_establish_conf(struct line *l)
{
	if (get_word(l) != NULL)
		warn_file("junk after *!> ignored");

	stack->saal_establish_conf();
}

/*
 * Send release indication
 */
static void
act_saal_release_ind(struct line *l)
{
	if (get_word(l) != NULL)
		warn_file("junk after *-> ignored");

	stack->saal_release_ind();
}

static void
act_api_ignore_start(struct line *l)
{
	if (get_word(l) != NULL)
		warn_file("junk ignored");

	api_ignore++;
}

static void
act_api_ignore_end(struct line *l)
{
	if (get_word(l) != NULL)
		warn_file("junk ignored");

	api_ignore--;
}

void
handle_establish_request(void)
{
	if (!wait_establish_req)
		report_problem("unexpected SAAL-ESTABLISH.request");
	wait_establish_req = 2;
}

static void
act_group_establish_req(struct line *l)
{

	if (get_word(l) != NULL)
		warn_file("junk after *+< ignored");
	wait_establish_req = 1;
}

static void
act_wait_establish_req(struct line *l)
{

	act_group_establish_req(l);
	test_wait_events();
}
