/*
 * Copyright (c) 2003-2003
 *	Fraunhofer Institute for Open Communication Systems (FhG Fokus).
 * 	All rights reserved.
 *
 * Author: Hartmut Brandt <harti@freebsd.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Begemot: atmsupport/ats_sig/user.c,v 1.13 2004/07/08 13:33:26 brandt Exp $
 */
#include <sys/types.h>
#include <sys/queue.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <unistd.h>
#include <ctype.h>
#include <limits.h>

#include <netnatm/unimsg.h>
#include <netnatm/msg/unistruct.h>
#include <netnatm/msg/unimsglib.h>
#include <netnatm/saal/sscfudef.h>
#include <netnatm/sig/unidef.h>
#include <netnatm/sig/uni.h>

#include "ats.h"
#include "test_sig.h"

static struct uni *stack;

/*
 * Reset the stack
 */
static void
user_reset(void)
{

	uni_reset(stack);
}

/*
 * Start the stack
 */
static void
user_start(const u_int *debug, const struct uni_config *config,
    uint32_t m1, uint32_t m2, uint32_t m3)
{
	u_int i;

	uni_reset(stack);
	for (i = 0; i < UNI_MAXFACILITY; i++)
		uni_set_debug(stack, i, debug[i]);
	uni_set_config(stack, config, &m1, &m2, &m3);
	if (m1 != 0 || m2 != 0 || m3 != 0)
		err_file("could not set config %#x %#x %#x", m1, m2, m3);
}

/*
 * Get configuration
 */
static void
user_getconfig(struct uni_config *config)
{
	uni_get_config(stack, config);
}

/*
 * Start an uni timer
 */
static void *
start_timer(struct uni *uni __unused, void *arg __unused, u_int msec,
    void (*func)(void *), void *farg)
{

	return (ats_start_timer(msec, func, farg));
}

/*
 * Stop an uni timer
 */
static void
stop_timer(struct uni *uni __unused, void *arg __unused, void *tp)
{

	ats_stop_timer(tp);
}

/*
 * output to the SAAL
 */
static void
saal_output(struct uni *uni __unused, void *arg __unused, enum saal_sig sig,
    struct uni_msg *msg)
{

	switch (sig) {

	  case SAAL_ESTABLISH_request:
		if (msg != NULL) {
			report_problem("unexpected user data in "
			    "SAAL-ESTABLISH.request");
			uni_msg_destroy(msg);
		}
		handle_establish_request();
		break;

	  case SAAL_RELEASE_request:
		report_problem("unexpected SAAL-RELEASE.request");
		if (msg != NULL) {
			report_problem("unexpected user data in "
			    "SAAL-RELEASE.request");
			uni_msg_destroy(msg);
		}
		break;

	  case SAAL_DATA_request:
		handle_saal_data(msg);
		break;

	  case SAAL_UDATA_request:
		report_problem("unexpected SAAL-UDATA.request");
		goto done;

	  default:
		report_problem("illegal SAAL signal from UNI: %u", sig);
	}

  done:
	if (msg != NULL)
		uni_msg_destroy(msg);
}

/*
 * output to the upper layer
 */
static void
uni_output(struct uni *uni __unused, void *arg __unused, enum uni_sig sig,
    uint32_t cookie, struct uni_msg *msg)
{
	handle_api_data(sig, cookie, msg);
	if (msg != NULL)
		uni_msg_destroy(msg);
}

/*
 * verbosity
 */
static void
uni_verbose(struct uni *uni __unused, void *arg __unused,
    enum uni_verb fac, const char *fmt, ...)
{
	static const char *fnames[] = {
#define UNI_DEBUG_DEFINE(D) [UNI_FAC_##D] = #D,
		UNI_DEBUG_FACILITIES
#undef UNI_DEBUG_DEFINE
	};
	va_list ap;

	fprintf(stderr, "%s: ", fnames[fac]);
	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
	fprintf(stderr, "\n");
}

/*
 * From SAAL to UNI
 */
static void
user_saal_to_uni(struct uni_msg *msg)
{
	uni_saal_input(stack, SAAL_DATA_indication, msg);
}

static void
user_saal_establish_ind(void)
{
	uni_saal_input(stack, SAAL_ESTABLISH_indication, NULL);
}

static void
user_saal_establish_conf(void)
{
	uni_saal_input(stack, SAAL_ESTABLISH_confirm, NULL);
}

static void
user_saal_release_ind(void)
{
	uni_saal_input(stack, SAAL_RELEASE_indication, NULL);
}

static void
user_work(void)
{
	uni_work(stack);
}

static void
user_api_to_uni(u_int sig, u_int cookie, struct uni_msg *msg)
{
	uni_uni_input(stack, sig, cookie, msg);
}

static const struct uni_funcs uni_funcs = {
	uni_output,
	saal_output,
	uni_verbose,
	NULL,
	start_timer,
	stop_timer
};

/*
 * Instantiate a user space stack
 */
static void
user_init(void)
{

	if ((stack = uni_create(NULL, &uni_funcs)) == NULL)
		err_file("cannot create uni instance: %s", strerror(errno));
}


const struct ustack user_stack = {
	user_init,
	user_reset,
	user_start,
	user_saal_to_uni,
	user_saal_establish_ind,
	user_saal_establish_conf,
	user_saal_release_ind,
	user_work,
	user_api_to_uni,
	user_getconfig
};
