/*
 * Copyright (c) 2003-2003
 *	Fraunhofer Institute for Open Communication Systems (FhG Fokus).
 * 	All rights reserved.
 *
 * Author: Hartmut Brandt <harti@freebsd.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Begemot: atmsupport/ats_sscfu/main.c,v 1.5 2004/08/18 12:50:24 brandt Exp $
 */

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#define	index foo_index
#include <tcl.h>
#undef index

#include <netnatm/unimsg.h>
#include <netnatm/saal/sscopdef.h>
#include <netnatm/saal/sscfudef.h>
#include <unitcl.h>

#include "ats.h"
#include "ats_sscfu.h"

/* TCL interpreter context for parsing and formating */
static Tcl_Interp *tcl;

/* the messages to lower layer we are waiting for */
static struct waitlist *wait_lower;

/* the messages to upper layer we are waiting for */
static struct waitlist *wait_upper;

/* the messages to management we are waiting for */
static struct waitlist *wait_manage;

/* which stack to use */
#ifdef HAVE_NETGRAPH_H
static const struct sscfu_stack *stack = &kern_stack;
#else
static const struct sscfu_stack *stack = &user_stack;
#endif

/* ignore upper layer messages */
static struct ignlist *ignore_upper;

/* ignore lower layer messages */
static struct ignlist *ignore_lower;

/* debug flags */
static u_int debug;

struct input_list lower_input_list =
    TAILQ_HEAD_INITIALIZER(lower_input_list);

struct input_list upper_input_list =
    TAILQ_HEAD_INITIALIZER(upper_input_list);

struct input_list manage_input_list =
    TAILQ_HEAD_INITIALIZER(manage_input_list);

static void act_from_lower(struct line *);
static void act_wait_to_lower(struct line *);
static void act_group_to_lower(struct line *);

static void act_from_upper(struct line *);
static void act_wait_to_upper(struct line *);
static void act_group_to_upper(struct line *);

static void act_from_manage(struct line *);
static void act_wait_to_manage(struct line *);
static void act_group_to_manage(struct line *);

static void handle_lower(struct uni_msg *msg);
static void handle_upper(u_int sig, struct uni_msg *msg);
static void handle_manage(u_int, u_int, u_int, struct uni_msg *);

/* user defined keys */
const struct key usrkeys[] = {
	{ "*>",		ACT_RUNNING, 1, 1, act_from_lower },
	{ "*<",		ACT_RUNNING, 0, 1, act_wait_to_lower },
	{ "@>",		ACT_RUNNING, 1, 1, act_from_upper },
	{ "@<",		ACT_RUNNING, 0, 1, act_wait_to_upper },
	{ "^>",		ACT_RUNNING, 1, 1, act_from_manage },
	{ "^<",		ACT_RUNNING, 0, 1, act_wait_to_manage },
	{ NULL,		0,	     0, 0, NULL }
};
const struct key usrgrpkeys[] = {
	{ "*",		ACT_RUNNING, 0, 1, act_group_to_lower },
	{ "@",		ACT_RUNNING, 0, 1, act_group_to_upper },
	{ "^",		ACT_RUNNING, 0, 1, act_group_to_manage },
	{ NULL,		0,	     0, 0, NULL }
};

static const struct {
	const char	*name;
	u_int		val;
} debug_strings[] = {
	{ "usig",	SSCOP_DBG_USIG },
	{ "timer",	SSCOP_DBG_TIMER },
	{ "bug",	SSCOP_DBG_BUG },
	{ "insig",	SSCOP_DBG_INSIG },
	{ "state",	SSCOP_DBG_STATE },
	{ "pdu",	SSCOP_DBG_PDU },
	{ "err",	SSCOP_DBG_ERR },
	{ "exec",	SSCOP_DBG_EXEC },
	{ "flow",	SSCOP_DBG_FLOW },
};

int
main(int argc, char *argv[])
{
	int opt;
	u_long val;
	char *end;
	u_int i;

#ifdef FreeBSD
	_malloc_options = "JA";
#endif

	while ((opt = getopt(argc, argv, "d:qt:uv:")) != -1)
		switch (opt) {

		  case 'd':
			for (i = 0; i < sizeof(debug_strings) /
			    sizeof(debug_strings[0]); i++)
				if (strcmp(optarg, debug_strings[i].name) == 0) {
					debug |= debug_strings[i].val;
					break;
				}
			if (i == sizeof(debug_strings) /
			    sizeof(debug_strings[0]))
				fprintf(stderr, "bad debug string '%s'\n",
				    optarg);
			break;

		  case 'q':
			if (report_level > 0)
				report_level--;
			break;

		  case 't':
			errno = 0;
			val = strtoul(optarg, &end, 10);
			if (*end != '\0' || errno != 0 || end == optarg ||
			    val > UINT_MAX) {
				fprintf(stderr, "bad trace value '%s'\n",
				    optarg);
				exit(1);
			}
			trace_level = val;
			break;

		  case 'u':
			stack = &user_stack;
			break;

		  case 'v':
			while (*optarg != '\0') {
				SETVERB(*optarg);
				optarg++;
			}
			break;
		}

	if ((tcl = Tcl_CreateInterp()) == NULL) {
		fprintf(stderr, "Tcl_CreateInterp: %s", strerror(errno));
		exit(1);
	}

	test_init();

	wait_lower = waitlist_create("LOWER");
	wait_upper = waitlist_create("UPPER");
	wait_manage = waitlist_create("MANAGE");

	ignore_upper = ignlist_create("upper ignore");
	ignore_lower = ignlist_create("lower ignore");

	stack->init();
	stack->debug(debug);

	while (optind < argc) {
		input_open_file(argv[optind]);
		test_exec();
		optind++;
	}

	if (ISVERB('F'))
		funcs_dump();

	exit(report_total() ? 1 : 0);
}

/*
 * Return non-zero if we are waiting for something
 */
int
test_wait_cb(void)
{
	struct winput *ip;

	while ((ip = TAILQ_FIRST(&lower_input_list)) != NULL) {
		TAILQ_REMOVE(&lower_input_list, ip, link);
		handle_lower(ip->msg);
		uni_msg_destroy(ip->msg);
		free(ip);
	}
	while ((ip = TAILQ_FIRST(&upper_input_list)) != NULL) {
		TAILQ_REMOVE(&upper_input_list, ip, link);
		handle_upper(ip->sig, ip->msg);
		if (ip->msg != NULL)
			uni_msg_destroy(ip->msg);
		free(ip);
	}
	while ((ip = TAILQ_FIRST(&manage_input_list)) != NULL) {
		TAILQ_REMOVE(&manage_input_list, ip, link);
		handle_manage(ip->sig, ip->arg1, ip->arg2, ip->msg);
		if (ip->msg != NULL)
			uni_msg_destroy(ip->msg);
		free(ip);
	}

	return (0);
}

/*
 * Make an SAAL signal message from a string
 */
static struct uni_msg *
upper_msg(const char *who, const char *arg, u_int *sig)
{
	int ac;
	const char **av;
	struct uni_msg *msg;

	if (Tcl_SplitList(tcl, arg, &ac, &av) != TCL_OK)
		err_file("%s -- %s", who, tcl->result);

	if ((msg = uni_msg_alloc(100)) == NULL)
		err_file("%s", strerror(errno));
	if (parse_saal_sig(tcl, sig, ac, av, msg) != TCL_OK)
		err_file("%s -- %s", who, tcl->result);

	Tcl_Free((char *)av);

	return (msg);
}

/*
 * Data input to SAAL from upper layer
 */
static void
act_from_upper(struct line *l)
{
	char *word;
	struct uni_msg *msg;
	u_int n, sig;

	if ((word = skip_blanks(l)) == NULL)
		err_file("@> -- missing message");

	trace(3, "upper to saal: %s", word);
	msg = upper_msg("@>", word, &sig);

	if (ISVERB('D')) {
		printf("sig=%#02x", sig);
		if (msg != NULL)
			for (n = 0; n < uni_msg_len(msg); n++)
				printf(" %02x", msg->b_rptr[n]);
		printf("\n");
	}

	stack->upper_to_saal(sig, msg);
}

/*
 * Wait for a SAAL signal to appear at the upper layer
 */
static void
act_group_to_upper(struct line *l)
{

	waitlist_put(wait_upper, l);
}
static void
act_wait_to_upper(struct line *l)
{

	waitlist_put(wait_upper, l);
	test_wait_events();
}

/*
 * handle data output to upper layer
 */
static void
handle_upper(u_int sig, struct uni_msg *msg)
{
	Tcl_DString str;
	u_int n;

	if (ISVERB('D')) {
		printf("sig=%#02x", sig);
		if (msg != NULL)
			for (n = 0; n < uni_msg_len(msg); n++)
				printf(" %02x", msg->b_rptr[n]);
		printf("\n");
	}

	Tcl_DStringInit(&str);
	if (fmt_saal_sig(tcl, &str, sig, msg) != TCL_OK)
		err_file("saal upper -- %s\n", tcl->result);

	if (!ignlist_match(ignore_upper, Tcl_DStringValue(&str)))
		waitlist_match(wait_upper, Tcl_DStringValue(&str));

	Tcl_DStringFree(&str);
}

/*
 * Make an SSCOP management signal from a string
 */
static struct uni_msg *
manage_msg(const char *who, const char *arg, u_int *sig, u_int *arg1,
    u_int *arg2)
{
	int ac;
	const char **av;
	struct uni_msg *msg;

	if (Tcl_SplitList(tcl, arg, &ac, &av) != TCL_OK)
		err_file("%s -- %s", who, tcl->result);

	if (parse_sscop_msig(tcl, sig, arg1, arg2, &msg, ac, av) != TCL_OK)
		err_file("%s -- %s", who, tcl->result);

	Tcl_Free((char *)av);

	return (msg);
}

/*
 * Data input to SSCOP from management layer
 */
static void
act_from_manage(struct line *l)
{
	char *word;
	struct uni_msg *msg;
	u_int n, sig, arg1, arg2;

	if ((word = skip_blanks(l)) == NULL)
		err_file("^> -- missing message");

	trace(3, "manage to sscop: %s", word);
	msg = manage_msg("^>", word, &sig, &arg1, &arg2);

	if (ISVERB('D')) {
		printf("sig=%#02x arg1=%#02x arg2=%#02x", sig, arg1, arg2);
		if (msg != NULL)
			for (n = 0; n < uni_msg_len(msg); n++)
				printf(" %02x", msg->b_rptr[n]);
		printf("\n");
	}

	stack->manage_to_sscop(sig, arg1, arg2, msg);
}

/*
 * Wait for a SSCOP signal to appear at the manage layer
 */
static void
act_group_to_manage(struct line *l)
{

	waitlist_put(wait_manage, l);
}
static void
act_wait_to_manage(struct line *l)
{

	waitlist_put(wait_manage, l);
	test_wait_events();
}

/*
 * handle data output to management layer
 */
static void
handle_manage(u_int sig, u_int arg1, u_int arg2, struct uni_msg *msg)
{
	Tcl_DString str;
	u_int n;

	if (ISVERB('D')) {
		printf("sig=%#02x arg1=%#02x arg2=%#02x", sig, arg1, arg2);
		if (msg != NULL)
			for (n = 0; n < uni_msg_len(msg); n++)
				printf(" %02x", msg->b_rptr[n]);
		printf("\n");
	}

	Tcl_DStringInit(&str);
	if (fmt_sscop_msig(tcl, &str, sig, arg1, arg2, msg) != TCL_OK)
		err_file("sscop manage -- %s\n", tcl->result);

	waitlist_match(wait_manage, Tcl_DStringValue(&str));

	Tcl_DStringFree(&str);
}

/*
 * Make an SSCOP PDU
 */
static struct uni_msg *
lower_msg(const char *who, const char *arg)
{
	int ac;
	const char **av;
	struct uni_msg *msg;

	if (Tcl_SplitList(tcl, arg, &ac, &av) != TCL_OK)
		err_file("%s -- %s", who, tcl->result);

	if ((msg = uni_msg_alloc(100)) == NULL)
		err_file("%s", strerror(errno));
	if (parse_sscop(tcl, msg, ac, av) != TCL_OK)
		err_file("%s -- %s", who, tcl->result);

	Tcl_Free((char *)av);

	return (msg);
}

/*
 * Data input to SSCOP from lower layer
 */
static void
act_from_lower(struct line *l)
{
	char *word;
	struct uni_msg *msg;
	u_int n;

	if ((word = skip_blanks(l)) == NULL)
		err_file("*> -- missing message");

	trace(3, "lower to sscop: %s", word);
	msg = lower_msg("*>", word);

	if (ISVERB('D')) {
		for (n = 0; n < uni_msg_len(msg); n++)
			printf(" %02x", msg->b_rptr[n]);
		printf("\n");
	}

	stack->lower_to_sscop(msg);
}

/*
 * Wait for a PDU to appear at the lower layer
 */
static void
act_group_to_lower(struct line *l)
{

	waitlist_put(wait_lower, l);
}
static void
act_wait_to_lower(struct line *l)
{

	waitlist_put(wait_lower, l);
	test_wait_events();
}

/*
 * handle data output to lower layer
 */
static void
handle_lower(struct uni_msg *msg)
{
	Tcl_DString str;
	u_int n;

	if (ISVERB('D')) {
		for (n = 0; n < uni_msg_len(msg); n++)
			printf(" %02x", msg->b_rptr[n]);
		printf("\n");
	}

	Tcl_DStringInit(&str);
	if (fmt_sscop(tcl, &str, msg) != TCL_OK)
		err_file("sscop lower -- %s\n", tcl->result);

	if (!ignlist_match(ignore_lower, Tcl_DStringValue(&str)))
		waitlist_match(wait_lower, Tcl_DStringValue(&str));

	Tcl_DStringFree(&str);
}

/*
 * Reset everything
 */
void
test_reset_cb(int tout __unused)
{

	stack->reset();
}

/*
 * Callback to start a test
 */
void
test_start_cb(void)
{

	stack->start();
}
