/*
 * Copyright (c) 2003-2003
 *	Fraunhofer Institute for Open Communication Systems (FhG Fokus).
 * 	All rights reserved.
 *
 * Author: Hartmut Brandt <harti@freebsd.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Begemot: atmsupport/libats/ats_input.c,v 1.12 2004/07/08 13:33:35 brandt Exp $
 */

#include <stdlib.h>
#include <stdarg.h>
#include <stddef.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <inttypes.h>

#include "ats.h"
#include "atspriv.h"

/* function call stack */
struct input *input;

/* file level */
u_int input_file_level;

/*
 * Open the file and push it on top of the input stack
 */
struct input *
input_open_file(const char *fname)
{
	struct input *f;

	f = ats_alloc(sizeof(*f));

	f->type = INPUT_FILE;
	f->pos.name = ats_strdup(fname);
	input_file_level++;
	if ((f->file_fp = fopen(fname, "r")) == NULL)
		err_file("%s: %s", fname, strerror(errno));

	f->stack = input;
	input = f;

	return (f);
}

/*
 * Open a function and push into onto the input stack
 */
struct input *
input_open_func(struct func *func)
{
	struct input *f;

	f = ats_alloc(sizeof(*f));

	f->type = INPUT_FUNC;
	f->pos.name = ats_strdup(func->name);
	f->pos.lno = func->def.lno;

	f->func_func = func;
	f->func_curr = TAILQ_FIRST(&func->lines);

	f->stack = input;
	input = f;

	return (f);
}

/*
 * Open a loop and push into onto the input stack
 */
struct input *
input_open_loop(struct string *cond, struct string_list *block)
{
	struct input *f;

	f = ats_alloc(sizeof(*f));

	f->type = INPUT_LOOP;
	f->pos.name = ats_strdup("loop");

	TAILQ_INIT(&f->loop_block);
	TAILQ_CONCAT(&f->loop_block, block, link);
	f->loop_cond = cond;
	f->loop_curr = TAILQ_FIRST(&f->loop_block);

	f->stack = input;
	input = f;

	return (f);
}

/*
 * Read the next line from a real file
 */
#define	LINE_GROW	200
#define	LINE_MIN	100

static int
read_file_line(struct line *l)
{
	char *ptr;

	input->file_len = 0;
	l->line[0] = '\0';

	for (;;) {
		if (l->alloc - input->file_len < LINE_MIN) {
			l->alloc += LINE_GROW;
			ptr = ats_realloc(l->line, l->alloc);
			l->line = ptr;
		}
		ptr = fgets(l->line + input->file_len,
		    l->alloc - input->file_len, input->file_fp);
		if (ptr == NULL) {
			if (ferror(input->file_fp))
				err_file("%s", strerror(errno));
			if (input->file_len == 0)
				return (-1);
			warn_file("no newline on EOF");
			break;
		}

		input->file_len = strlen(l->line);
		if (input->file_len != 0 &&
		    l->line[input->file_len - 1] == '\n') {
			input->pos.lno++;
			l->line[--input->file_len] = '\0';
			if (input->file_len == 0)
				break;
			if (l->line[input->file_len - 1] != '\\')
				break;
			l->line[--input->file_len] = '\0';
		}
		/* append next chunk */
	}
	l->ptr = l->line;

	if (ISVERB('i'))
		printf("%s: '%s'\n", __func__, l->line);
	return (0);
}

/*
 * Push a line into the input buffer, extending the buffer if neccessary
 */
static void
push_input_line(struct line *l, const char *str)
{

	if (strlen(str) + 1 > l->alloc) {
		l->alloc = strlen(str) + 1;
		l->line = ats_realloc(l->line, l->alloc);
	}
	strcpy(l->line, str);
	l->ptr = l->line;
}

/*
 * Read next line from function
 */
static int
read_func_line(struct line *l)
{
	struct string *s;

	if ((s = input->func_curr) == NULL)
		return (-1);

	push_input_line(l, s->string);
	input->func_curr = TAILQ_NEXT(s, link);
	input->pos.lno = s->lno;

	if (ISVERB('i'))
		printf("%s: '%s'\n", __func__, l->line);

	return (0);
}

/*
 * Read next line from loop
 */
static int
read_loop_line(struct line *l)
{
	struct string *s;

	if ((s = input->loop_curr) == NULL)
		return (-1);

	push_input_line(l, s->string);
	input->loop_curr = TAILQ_NEXT(s, link);
	input->pos.lno = s->lno;

	if (ISVERB('i'))
		printf("%s: '%s'\n", __func__, l->line);

	return (0);
}

/*
 * Evaluate loop condition
 */
static intmax_t
loop_eval_cond(struct line *l)
{
	char *w, *end;
	uintmax_t v;
	int sign;

	push_input_line(l, input->loop_cond->string);
	input->pos.lno = input->loop_cond->lno;

	substitute_line(l);

	if ((w = get_word(l)) == NULL)
		err_file("no loop condition");

	sign = 1;
	if (*w == '-') {
		sign = -1;
		w++;
	}
	errno = 0;
	v = strtoumax(w, &end, 0);
	if (errno != 0)
		err_file("bad condition: %s (%s)", strerror(errno), w);
	if (*end != '\0')
		err_file("bad condition: '%s'", w);

	return (sign * v);
}


/*
 * Read next line from the input stack
 */
int
read_line(struct line *l)
{
	struct input *fp;

	while (input != NULL) {
		switch (input->type) {

		  case INPUT_FILE:
			/* real file */
			if (read_file_line(l) == 0)
				return (0);
			if (ISVERB('i'))
				printf("%s: EOF on file\n", __func__);
			fclose(input->file_fp);
			func_flush_static();
			input_file_level--;
			break;

		  case INPUT_FUNC:
			/* function */
			if (read_func_line(l) == 0)
				return (0);
			var_pop(var_level--);
			if (ISVERB('i'))
				printf("%s: EOF on func\n", __func__);
			break;

		  case INPUT_LOOP:
			/* loop */
			for (;;) {
				if (read_loop_line(l) == 0)
					return (0);
				if (!loop_eval_cond(l))
					break;
				input->loop_curr =
				    TAILQ_FIRST(&input->loop_block);
			}
			if (ISVERB('i'))
				printf("%s: EOF on loop\n", __func__);
			stringlist_free(&input->loop_block);
			free(input->loop_cond->string);
			free(input->loop_cond);
			break;

		  default:
			abort();
		}
		fp = input;
		input = fp->stack;

		free(fp->pos.name);
		free(fp);
	}
	return (-1);
}

/*
 * Skip blanks
 * This moves the input pointer to the first non-blank and returns it.
 */
char *
skip_blanks(struct line *l)
{
	int c;

	/* skip leading space */
	while ((c = *l->ptr) != '\0' && isascii(c) && isspace(c))
		l->ptr++;

	if (c == '\0')
		/* no more words */
		return (NULL);

	if (!isascii(c))
		err_file("bad character 0x%02x '%s'", c & 0xff, l->ptr);

	return (l->ptr);
}

/*
 * Skip to the end of the line.
 */
void
skip_line(struct line *l)
{

	l->ptr += strlen(l->ptr);
}

/*
 * Find the begin and the length of the next word
 */
char *
peek_word(struct line *l, size_t *lenp)
{
	char *end;
	int c;

	if ((end = skip_blanks(l)) == NULL)
		return (NULL);

	while ((c = *end) != '\0' && isascii(c) && !isspace(c))
		end++;

	if (c != '\0' && !isascii(c))
		err_file("bad character 0x%02x '%s'", c & 0xff, l->ptr);

	if (lenp != NULL)
		*lenp = end - l->ptr;

	return (l->ptr);
}

/*
 * Eat the last peeked word (and nul-terminate it)
 */
void
swallow_word(struct line *l, size_t len)
{

	if (l->ptr[len] != '\0') {
		l->ptr[len] = '\0';
		l->ptr += len + 1;
	} else
		l->ptr += len;
}

/*
 * Get the next word (nul-terminated).
 */
char *
get_word(struct line *l)
{
	size_t len;
	char *w;

	if ((w = peek_word(l, &len)) == NULL)
		return (NULL);
	swallow_word(l, len);
	return (w);
}

/*
 * Get the next (possibly quoted) word.
 */
char *
get_qword(struct line *l)
{
	char *end, *w;
	int c;

	if ((end = skip_blanks(l)) == NULL)
		return (NULL);

	if ((c = *end) != '"')
		return (get_word(l));

	w = ++l->ptr;
	end++;
	while ((c = *end) != '\0' && c != '"' && isascii(c))
		end++;

	if (c != '"')
		err_file("bad character 0x%02x '%s'", c & 0xff, l->ptr);
	*end++ = '\0';
	l->ptr = end;
	return (w);
}

/*
 * Get next variable name, move the input pointer just behind it.
 */
char *
get_name(struct line *l)
{
	char *w;

	if ((w = skip_blanks(l)) == NULL)
		return (NULL);
	l->ptr = w;

	while (*l->ptr != '\0') {
		if (!isascii(*l->ptr))
			err_file("bad char in variable name %#02x", *l->ptr);
		if (!isalnum(*l->ptr) && *l->ptr != '_')
			break;
		l->ptr++;
	}
	if (w == l->ptr)
		err_file("expecting variable name at '%s'", w);

	return (w);
}

/*
 * Replace all space sequences in the given line by exactly one space.
 * Remove all spaces before } and after {
 */
void
compress_blanks(char *str)
{
	char *in, *out;
	int skip_blank;

	in = str;
	skip_blank = 1;
	for (out = in; *in != '\0'; in++) {
		if (!isascii(*in) || !isspace(*in)) {
			*out++ = *in;
			skip_blank = 0;
 		} else {
			if (!skip_blank)
				*out++ = ' ';
			skip_blank = 1;
		}
	}
	while (out > str && isascii(out[-1]) && isspace(out[-1]))
		out--;
	*out = '\0';

	for (out = in = str; *in != '\0'; in++) {
		if (*in == '}') {
			if (out > str && out[-1] == ' ')
				out--;
			*out++ = '}';
		} else if (*in == '{') {
			if (in[1] == ' ')
				in++;
			*out++ = '{';
		} else
			*out++ = *in;
	}
	*out++ = '\0';
}

/*
 * Save the given string
 */
struct string *
string_save(const char *str, u_int lno)
{
	struct string *s;

	s = ats_alloc(sizeof(*s));
	s->lno = lno;
	s->string = ats_strdup(str);
	return (s);
}
struct string *
string_saven(const char *str, u_int lno, size_t len)
{
	struct string *s;

	s = ats_alloc(sizeof(*s));
	s->lno = lno;
	s->string = ats_alloc(len + 1);
	strncpy(s->string, str, len);
	s->string[len] = '\0';
	return (s);
}

/*
 * Replace the part [ptr, ptr + len] of line with repl.
 */
void
splice(struct line *l, size_t len, const char *repl)
{
	size_t repl_len;
	size_t rest_len;
	size_t need;
	char *new;

	repl_len = strlen(repl);
	rest_len = strlen(l->ptr + len);

	if (repl_len < len)
		/* string gets shorter - move rest nearer */
		memmove(l->ptr + repl_len, l->ptr + len, rest_len + 1);

	else if (repl_len > len) {
		/* string gets longer - check for space */
		need = (l->ptr - l->line) + repl_len + rest_len + 1;
		if (need < l->alloc) {
			/* too short - reallocate */
			need += 100;
			new = ats_realloc(l->line, need);
			/* relocate pointers */
			l->ptr = new + (l->ptr - l->line);
			l->line = new;
			l->alloc = need;
		}
		memmove(l->ptr + repl_len, l->ptr + len, rest_len + 1);
	}
	memcpy(l->ptr, repl, repl_len);
}

static struct line *
exp_subst(const char *str, const char **end)
{
	struct line *n;
	const char *p, *r;
	int level;

	/* collect until closing ')' */
	level = 0;
	for (p = str;; p++) {
		if (*p == '\0')
			err_file("eol in expression");
		if (*p == '(')
			level++;
		else if (*p == ')' && --level == 0) {
			p++;
			break;
		}
	}

	n = line_create(p - str + 1);
	strncpy(n->line, str, p - str);
	n->line[p - str] = '\0';

	/* run substitution on this string */
	substitute_line(n);

	/* compute expression */
	r = expr_subst(n->line, end);

	if (**end != '\0')
		err_file("ups - malformed expression");

	push_input_line(n, r);

	if (ISVERB('s'))
		printf("expr substitution: '%.*s' -> '%s'\n",
		    (int)(p - str + 1), str - 1, n->line);
	*end = p;
	return (n);
}

/*
 * Do ! and $ substitution on a line.
 *
 * We do this by copying the old line character-wise to a new one.
 */
void
substitute_line(struct line *l)
{
	ptrdiff_t off = l->ptr - l->line;
	struct line *n = line_create(0);
	struct line *s;
	int c;
	const char *e, *r;

	while ((c = *l->ptr) != '\0') {
		if (c == '!') {
			r = var_subst(l->ptr + 1, &e);
			if (*e == '!')
				e++;
			line_append_string(n, r);
			l->ptr += (e - l->ptr);

		} else if (c == '$') {
			if ((c = l->ptr[1]) == '(') {
				s = exp_subst(l->ptr + 1, &e);
				line_append_string(n, s->line);
				l->ptr += (e - l->ptr);
				line_destroy(s);

			} else if (isascii(c) && (c == '_' || isalpha(c))) {
				s = str_subst(l->ptr + 1, &e);
				line_append_string(n, s->line);
				l->ptr += (e - l->ptr);
				line_destroy(s);

			} else
				err_file("bad $ replacement");

		} else {
			l->ptr++;
			line_append_char(n, c);
		}
	}

	if (off + strlen(n->line) + 1 > l->alloc) {
		l->alloc = off + strlen(n->line) + 1;
		l->line = ats_realloc(l->line, l->alloc);
	}
	strcpy(l->line + off, n->line);
	l->ptr = l->line + off;

	line_destroy(n);
}

/*
 * Print the call stack.
 */
void
print_stack(const struct input *in)
{
	while (in != NULL) {
		printf("  called from %s, %u\n", in->pos.name, in->pos.lno);
		in = in->stack;
	}
}

/*
 * report error and exit
 */
void
err_file(const char *fmt, ...)
{
	va_list ap;

	if (input != NULL)
		fprintf(stderr, "%s, %u: ", input->pos.name, input->pos.lno);
	fprintf(stderr, "error: ");

	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
	fprintf(stderr, "\n");

	if (input != NULL)
		print_stack(input->stack);

	exit(1);
}

/*
 * report a warning
 */
void
warn_file(const char *fmt, ...)
{
	va_list ap;

	fprintf(stderr, "%s, %u: warning: ", input->pos.name, input->pos.lno);
	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
	fprintf(stderr, "\n");

	print_stack(input->stack);
}

#if 0
void
stringlist_dump(struct string_list *list)
{
	u_int i = 0;
	struct string *s;

	TAILQ_FOREACH(s, list, link)
		printf("%u: '%s'\n", i++, s->string);
}
#endif

/*
 * Free a string list
 */
void
stringlist_free(struct string_list *list)
{
	struct string *s;

	while (!TAILQ_EMPTY(list)) {
		s = TAILQ_FIRST(list);
		TAILQ_REMOVE(list, s, link);
		free(s->string);
		free(s);
	}
}

/*
 * Create a new line
 */
struct line *
line_create(size_t len)
{
	struct line *l;

	if (len < 1000)
		len = 1000;
	l = ats_alloc(sizeof(*l));
	l->alloc = len;
	l->line = ats_alloc(l->alloc);
	l->ptr = l->line;
	l->line[0] = '\0';
	return (l);
}

/*
 * Destroy a line
 */
void
line_destroy(struct line *l)
{
	free(l->line);
	free(l);
}

/*
 * Append string to line
 */
void
line_append_string(struct line *l, const char *str)
{
	char *new;
	size_t need;

	need = strlen(l->line) + strlen(str) + 1;
	if (need > l->alloc) {
		new = ats_realloc(l->line, need);
		l->alloc = need;
		l->ptr = new + (l->ptr - l->line);
		l->line = new;
	}
	strcat(l->line, str);
}

/*
 * Append char to line
 */
void
line_append_char(struct line *l, int c)
{
	char *new;
	size_t need;

	need = strlen(l->line) + 2;
	if (need > l->alloc) {
		new = ats_realloc(l->line, need);
		l->alloc = need;
		l->ptr = new + (l->ptr - l->line);
		l->line = new;
	}
	l->line[need - 2] = c;
	l->line[need - 1] = '\0';
}
