/*
 * Copyright (c) 2003-2003
 *	Fraunhofer Institute for Open Communication Systems (FhG Fokus).
 * 	All rights reserved.
 *
 * Author: Hartmut Brandt <harti@freebsd.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Begemot: atmsupport/libats/ats_test.c,v 1.24 2004/12/20 10:44:54 brandt Exp $
 */

#include <sys/types.h>
#include <sys/time.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include <inttypes.h>
#include <rpoll.h>

#include "ats.h"
#include "atspriv.h"

/* current test name */
char *current_test_name;

/* skip this test */
int skip_test;

/* current key */
const char *curr_key;

/* timeout for current test */
static int test_timer;

/* timeout for wait timer */
static int wait_timer;

/* list of all wait lists */
static LIST_HEAD(, waitlist) waitlist_list =
    LIST_HEAD_INITIALIZER(waitlist_list);

/* list of all hold lists */
static LIST_HEAD(, holdlist) holdlist_list =
    LIST_HEAD_INITIALIZER(holdlist_list);

/* list of all event lists */
static LIST_HEAD(, eventlist) eventlist_list =
    LIST_HEAD_INITIALIZER(eventlist_list);

/*
 * Initialize test stuff
 */
void
test_init(void)
{

	wait_timer = -1;
	test_timer = -1;

	var_create_special();
}

/*
 * Wait for events to happen
 */
void
test_wait_events(void)
{
	int ret;
	struct waitlist *w;
	struct holdlist *h;
	struct eventlist *e;
	struct string *s;

	while (current_test_name != NULL) {
		if ((ret = test_wait_cb()) < 0)
			return;
		if (ret > 0)
			ret = 0;
		else {
			ret = 1;
			LIST_FOREACH(h, &holdlist_list, link) {
				while ((s = TAILQ_FIRST(&h->hold)) != NULL) {
					TAILQ_REMOVE(&h->hold, s, link);
					TAILQ_INSERT_TAIL(&h->waitlist->wait,
					    s, link);
				}
			}
			LIST_FOREACH(e, &eventlist_list, link) {
				while ((s = TAILQ_FIRST(&e->events)) != NULL) {
					TAILQ_REMOVE(&e->events, s, link);
					waitlist_match(e->waitlist, s->string);
					free(s->string);
					free(s);
				}
			}
			LIST_FOREACH(w, &waitlist_list, link)
				if (!TAILQ_EMPTY(&w->wait)) {
					ret = 0;
					break;
				}
			if (ret && wait_timer >= 0)
				ret = 0;
		}
		if (ret)
			return;

		poll_dispatch(1);
	}
}

/*
 * Call a function
 */
static void
act_call(struct line *l)
{
	char *func, *word;
	struct string *str;
	struct string_list args;

	if ((word = get_word(l)) == NULL)
		err_file("!! needs a function name");
	func = ats_strdup(word);

	TAILQ_INIT(&args);

	while ((word = get_qword(l)) != NULL) {
		str = string_save(word, input->pos.lno);
		TAILQ_INSERT_TAIL(&args, str, link);
	}

	func_call(func, &args);

	while ((str = TAILQ_FIRST(&args)) != NULL) {
		TAILQ_REMOVE(&args, str, link);
		free(str->string);
		free(str);
	}
	free(func);
}

/*
 * Declare (and set) a variable
 */
static void
act_decl_var(struct line *l)
{
	const char *name, *value;

	if ((name = get_word(l)) == NULL)
		err_file(". -- need variable name");
	if ((value = peek_word(l, NULL)) == NULL)
		value = "";
	var_define(name, value, 0, NULL);
}

/*
 * Assign a variable
 */
static void
act_assign(struct line *l)
{
	const char *name, *value;
	struct var *var;

	if ((name = get_word(l)) == NULL)
		err_file(".= -- need variable name");
	if ((var = var_find(name, NULL)) == NULL)
		err_file(".= -- unknown variable '%s'", name);
	if ((value = peek_word(l, NULL)) == NULL)
		value = "";

	var_setn(var, value, strlen(value));
}
/*
 * Declare (and set) a global variable
 */
static void
act_decl_global(struct line *l)
{
	const char *name, *value;

	if ((name = get_word(l)) == NULL)
		err_file(". -- need variable name");
	if ((value = peek_word(l, NULL)) == NULL)
		value = "";

	var_define_global(name, value);
}

/*
 * Declare and set a global, resetable variable
 */
static void
act_decl_resetable(struct line *l)
{
	const char *name, *value;

	if ((name = get_word(l)) == NULL)
		err_file(". -- need variable name");
	if ((value = peek_word(l, NULL)) == NULL)
		value = "";

	var_define_resetable(name, value);
}

/*
 * Compute expressions
 */
static void
act_expr(struct line *l)
{
	char *word, *end;

	while ((word = get_word(l)) != NULL) {
		errno = 0;
		(void)strtoumax(word, &end, 10);
		if (*end != '\0' || errno != 0 || word == end)
			err_file(".. -- bad expression");
	}
}

/*
 * Assert
 */
static void
act_assert(struct line *l)
{
	char *word, *end;
	uintmax_t v;

	if ((word = get_word(l)) == NULL)
		err_file("?! -- need expression");

	errno = 0;
	v = strtoumax(word, &end, 10);
	if (*end != '\0' || errno != 0 || word == end)
		err_file("?! -- bad expression");
	if (v == 0) {
		if ((word = skip_blanks(l)) == NULL)
			report_problem("assertion failed %"FMT_INTMAX"u", v);
		else
			report_problem("assertion failed %"FMT_INTMAX"u: %s",
			    v, word);
	}
}

/*
 * Timeout function
 */
static void
wait_timeout(int tid __unused, void *uarg __unused)
{

	wait_timer = -1;
}

/*
 * Wait for some time and await nothing
 */
static void
act_wait(struct line *l)
{
	const char *word;
	u_long del;
	char *end;

	if ((word = get_word(l)) == NULL)
		err_file("$$ -- needs a delay");
	errno = 0;
	del = strtoul(word, &end, 10);
	if (*end != '\0' || errno != 0 || word == end)
		err_file("$$ -- bad delay '%s'", word);
	
	wait_timer = poll_start_timer(del, 0, wait_timeout, NULL);

	test_wait_events();
}

/*
 * Reset everything
 */
static void
test_reset(int tout)
{
	struct waitlist *w;
	struct eventlist *e;
	struct string *s;

	if (!skip_test && test_timer >= 0) {
		poll_stop_timer(test_timer);
		test_timer = -1;
	}

	report_test_result();

	if (!skip_test) {
		var_reset();
		test_reset_cb(tout);

		LIST_FOREACH(w, &waitlist_list, link) {
			/* free messages we have been waiting for */
			while ((s = TAILQ_FIRST(&w->wait)) != NULL) {
				TAILQ_REMOVE(&w->wait, s, link);
				if (tout)
					trace(3, "waiting for %s: '%s'",
					    w->name, s->string);
				free(s->string);
				free(s);
			}
		}
		LIST_FOREACH(e, &eventlist_list, link) {
			/* free messages we have been waiting for */
			while ((s = TAILQ_FIRST(&e->events)) != NULL) {
				TAILQ_REMOVE(&e->events, s, link);
				if (tout)
					trace(3, "event in %s: '%s'",
					    e->name, s->string);
				free(s->string);
				free(s);
			}
		}

		if (wait_timer >= 0) {
			if (tout)
				trace(3, "waiting for timer");
			poll_stop_timer(wait_timer);
			wait_timer = -1;
		}
	}

	free(current_test_name);
	current_test_name = NULL;

	skip_test = 0;
}

/*
 * Timeout function
 */
static void
test_timeout(int tid __unused, void *uarg __unused)
{
	char *w;
	struct line *l = line_create(0);

	test_timer = -1;

	report_problem("test did not complete");

	/* read until EOT */
	while (read_line(l) == 0) {
		if ((w = get_word(l)) != NULL && strcmp(w, "%}") == 0) {
			if (get_word(l) != NULL)
				warn_file("junk after %%} ignored");
			var_pop(var_level--);
			test_reset(1);
			line_destroy(l);
			return;
		}
	}
	err_file("EOF in test '%s'", current_test_name);
}

/*
 * Stop a test
 */
static void
act_stop_test(struct line *l)
{

	if (get_word(l) != NULL)
		warn_file("junk after %%} ignored");
	if (!skip_test)
		var_pop(var_level--);
	test_reset(0);
}

/*
 * Start a test
 */
static void
act_start_test(struct line *l)
{
	char *word, *end;
	u_long tout;

	if (current_test_name != NULL)
		err_file("test '%s' already running", current_test_name);

	if ((word = get_word(l)) == NULL)
		err_file("no name for test");
	current_test_name = ats_strdup(word);

	if ((word = get_word(l)) == NULL)
		err_file("no timeout for test");
	errno = 0;
	tout = strtoul(word, &end, 10);
	if (*end != '\0' || errno != 0 || word == end)
		err_file("bad timeout value");

	report_test_init();

	if ((word = get_word(l)) != NULL) {
		if (strcmp(word, "skip") == 0) {
			skip_test = 1;
		} else
			err_file("bad key word in test start '%s'", word);
	}

	if (skip_test)
		return;

	var_level++;
	test_start_cb();

	test_timer = poll_start_timer(1000 * tout, 0, test_timeout, NULL);
}

/*
 * Eat a block of lines into a string list. End with the given key.
 */
static void
read_block(struct line *l, struct string_list *lines, const char *end)
{
	char *w;
	size_t len;
	struct string *str;

	TAILQ_INIT(lines);
	while (read_line(l) == 0) {
		if ((w = peek_word(l, &len)) != NULL && len == strlen(end) &&
		    strncmp(w, end, len) == 0) {
			swallow_word(l, len);
			return;
		}
		str = string_save(l->line, input->pos.lno);
		TAILQ_INSERT_TAIL(lines, str, link);
	}
	err_file("EOF in block definition (%s)", end);
}

/*
 * Define a function
 */
static void
act_define_common(struct line *l, const char *end, int stat)
{
	char *w, *func;
	struct string *str;
	struct string_list lines, args;

	if ((func = get_word(l)) == NULL)
		err_file("missing function name");
	func = ats_strdup(func);

	TAILQ_INIT(&args);

	while ((w = get_word(l)) != NULL) {
		str = string_save(w, input->pos.lno);
		TAILQ_INSERT_TAIL(&args, str, link);
	}

	read_block(l, &lines, end);
	if (get_word(l) != NULL)
		warn_file("junk after >> ignored");
	if (ISVERB('f'))
		printf("%s: END\n", func);

	func_define(func, &input->pos, &lines, &args, stat);
}
static void
act_define(struct line *l)
{
	act_define_common(l, ">>", 0);
}
static void
act_define_stat(struct line *l)
{
	act_define_common(l, ">>.", 1);
}

/*
 * Loops
 */
static void
act_loop_start(struct line *l)
{
	char *w;
	struct string *cond;
	struct string_list lines;

	if ((w = skip_blanks(l)) == NULL)
		err_file("missing loop condition");
	cond = string_save(w, input->pos.lno);

	read_block(l, &lines, "@}");
	if (get_word(l) != NULL)
		warn_file("junk after @} ignored");

	input_open_loop(cond, &lines);
}

/*
 * Trace messages
 */
static void
act_trace1(struct line *l)
{
	char *word;

	if ((word = skip_blanks(l)) == NULL)
		err_file("missing trace message");

	trace(1, "%s", word);
}

static void
act_trace2(struct line *l)
{
	char *word;

	if ((word = skip_blanks(l)) == NULL)
		err_file("missing trace message");

	trace(2, "%s", word);
}

static const struct key stdgrpkeys[] = {
	{ "**",		ACT_EITHER,	1, 0, act_trace1 },
	{ "--",		ACT_EITHER,	1, 1, act_trace2 },
	{ NULL,		0,		0, 0, NULL }
};

/*
 * Wait for a group of messages
 */
static void
act_wait_group(struct line *l)
{
	const char *w;
	const struct key *k;

	if (get_word(l) != NULL)
		warn_file("junk after < ignored");

	while (read_line(l) == 0) {
		if ((w = get_word(l)) == NULL)
			/* empty line */
			continue;
		if (w[0] == '#')
			/* comment */
			continue;
		if (strcmp(w, ">") == 0) {
			if (get_word(l) != NULL)
				warn_file("junk after > ignored");
			/* wait */
			test_wait_events();
			return;
		}

		for (k = stdgrpkeys; k->key != NULL; k++)
			if (strcmp(k->key, w) == 0)
				break;
		if (k->key == NULL) {
			for (k = usrgrpkeys; k->key != NULL; k++)
				if (strcmp(k->key, w) == 0)
					break;
			if (k->key == NULL)
				err_file("bad key '%s' in group", w);
		}

		if (k->subst)
			substitute_line(l);

		curr_key = k->key;
		k->func(l);
	}
	err_file("EOF in group");
}

static void
act_wait_group_end(struct line *l __unused)
{
	err_file("> -- outside of group");
}

/*
 * Define a string replacement
 */
static void
act_replace_common(struct line *l, int stat)
{
	char *w, c;
	char *name, *name_end;
	char *arg, *arg_end;
	struct string_list args;
	struct string *str;

	if ((name = get_name(l)) == NULL)
		err_file("%s -- need function name", curr_key);
	name_end = l->ptr;

	if ((w = skip_blanks(l)) == NULL || *w != '(')
		err_file("%s -- expecting '('", curr_key);
	l->ptr++;

	*name_end = '\0';

	name = ats_strdup(name);

	TAILQ_INIT(&args);

	/* peek for a closing parenthesis */
	w = l->ptr;
	if ((arg = skip_blanks(l)) != NULL && *arg == ')') {
		l->ptr++;

	} else {
		for (;;) {
			if ((arg = get_name(l)) == NULL)
				err_file("%s -- eol in arg list", curr_key);
			arg_end = l->ptr;

			if ((w = skip_blanks(l)) == NULL)
				err_file("%s -- eol in arg list", curr_key);
			c = *l->ptr++;

			*arg_end = '\0';

			str = string_save(arg, input->pos.lno);
			TAILQ_INSERT_TAIL(&args, str, link);

			if (c == ')')
				break;
			if (c != ',')
				err_file("%s -- expecting ','", curr_key);
		}
	}

	if ((w = skip_blanks(l)) == NULL)
		replace_define(name, &input->pos, &args, "", stat);
	else
		replace_define(name, &input->pos, &args, w, stat);
}
static void
act_replace(struct line *l)
{
	act_replace_common(l, 0);
}
static void
act_replace_stat(struct line *l)
{
	act_replace_common(l, 1);
}

static const struct key stdkeys[] = {
	{ "<<",		ACT_EITHER,	1, 1, act_define },
	{ "<<.",	ACT_EITHER,	1, 1, act_define_stat },
	{ "**",		ACT_EITHER,	1, 0, act_trace1 },
	{ "--",		ACT_EITHER,	1, 1, act_trace2 },
	{ "<",		ACT_RUNNING,	1, 1, act_wait_group },
	{ ">",		ACT_RUNNING,	1, 1, act_wait_group_end },
	{ "%{",		ACT_STOPPED,	1, 0, act_start_test },
	{ "%}",		ACT_RUNNING,	1, 0, act_stop_test },
	{ "!!",		ACT_EITHER,	1, 1, act_call },
	{ "@{",		ACT_RUNNING,	0, 1, act_loop_start },
	{ "$$",		ACT_RUNNING,	1, 1, act_wait },
	{ ".",		ACT_EITHER,	1, 1, act_decl_var },
	{ ".=",		ACT_EITHER,	1, 1, act_assign },
	{ ".!",		ACT_EITHER,	1, 1, act_decl_global },
	{ ".!!",	ACT_EITHER,	1, 1, act_decl_resetable },
	{ "..",		ACT_EITHER,	1, 1, act_expr },
	{ "?!",		ACT_EITHER,	1, 1, act_assert },
	{ "<>",		ACT_EITHER,	0, 1, act_replace },
	{ "<.>",	ACT_EITHER,	0, 1, act_replace_stat },
	{ NULL,		0,		0, 0, NULL }
};

/*
 * Read and execute lines until the bottom of the stack is reached.
 */
void
test_exec(void)
{
	char *w;
	const struct key *k;
	struct line *l = line_create(0);

	while (read_line(l) == 0) {
		if ((w = get_word(l)) == NULL)
			continue;

		if (w[0] == '#')
			/* comment */
			continue;

		for (k = stdkeys; k->key != NULL; k++)
			if (strcmp(k->key, w) == 0)
				break;
		if (k->key == NULL) {
			for (k = usrkeys; k->key != NULL; k++)
				if (strcmp(k->key, w) == 0)
					break;
			if (k->key == NULL)
				err_file("bad key '%s'", w);
		}

		switch (k->run) {

		  case ACT_RUNNING:
			if (current_test_name == NULL)
				err_file("%s -- no test running", w);
			break;

		  case ACT_STOPPED:
			if (current_test_name != NULL)
				err_file("%s -- test %s already running",
				    w, current_test_name);
			break;
		}
		if (skip_test && k->skip)
			continue;
		if (k->subst)
			substitute_line(l);
		curr_key = k->key;
		k->func(l);
	}
	if (current_test_name != NULL)
		err_file("EOF in test '%s'", current_test_name);

	line_destroy(l);
}

/*
 * Match a string to a template after doing variable substitution on
 * the template.
 */
int
match(const char *templ, const char *s)
{
	const char *t, *val, *vn;
	const char *tend;
	const char *send;
	struct var *var;
	int eol;
	int must_match;
	int junk;
	struct line *t0 = line_create(strlen(templ) + 1);

	strcpy(t0->line, templ);
	substitute_line(t0);
	compress_blanks(t0->ptr);
	trace(4, "try '%s'", t0->ptr);

	t = t0->ptr;
	while (*t != '\0') {
		if (*t == '?') {
			eol = 0;
			vn = t + 1;
			must_match = 1;
			junk = 0;
			if (*vn == '?') {
				vn++;
				eol = 1;
			} else if (*vn == '=') {
				vn++;
				eol = 1;
				must_match = 0;
			} else if (*vn == '.') {
				vn++;
				must_match = 0;
			}
			/* get the variable */
			if (*vn == '*') {
				junk = 1;
				var = NULL;
				tend = ++vn;
			} else
				var = var_find(vn, &tend);

			/* collect the matching string (alnum) */
			send = s;
			if (eol)
				while (*send != '\0')
					send++;
			else
				while (*send != '\0' && isascii(*send) &&
				    isalnum(*send))
					send++;

			if (!junk) {
				val = var_getval(var);
				if (val[0] == '\0' || !must_match)
					var_setn(var, s, send - s);
				else if (strlen(val) != (size_t)(send - s) ||
				    strncmp(val, s, send - s) != 0)
					report_problem("%.*s does not match "
					    "(old='%s' new='%.*s')",
					    (int)(tend - (t + 1)), t + 1,
					    val, (int)(send - s), s);
			}
			t = tend;
			s = send;

		} else if (*s++ != *t++) {
			if (ISVERB('m')) {
				printf("expecting: '%s'\n", t - 1);
				printf("got:       '%s'\n", s - 1); 
			}
			line_destroy(t0);
			return (1);
		}
	}
	if (*s != '\0' && ISVERB('m')) {
		printf("expecting: '%s'\n", t - 1);
		printf("got:       '%s'\n", s - 1); 
	}
	line_destroy(t0);
	return (*s != '\0');
}

/*
 * create a waitlist
 */
struct waitlist *
waitlist_create(const char *name)
{
	struct waitlist *w;

	LIST_FOREACH(w, &waitlist_list, link)
		if (strcmp(w->name, name) == 0)
			err_file("waitlist %s already exists", name);
	w = ats_alloc(sizeof(*w));
	w->name = ats_strdup(name);
	TAILQ_INIT(&w->wait);
	LIST_INSERT_HEAD(&waitlist_list, w, link);

	return (w);
}

/*
 * create a holdlist
 */
struct holdlist *
holdlist_create(const char *name, struct waitlist *wl)
{
	struct holdlist *h;

	LIST_FOREACH(h, &holdlist_list, link)
		if (strcmp(h->name, name) == 0)
			err_file("holdlist %s already exists", name);
	h = ats_alloc(sizeof(*h));
	h->name = ats_strdup(name);
	h->waitlist = wl;
	TAILQ_INIT(&h->hold);
	LIST_INSERT_HEAD(&holdlist_list, h, link);

	return (h);
}

/*
 * Put the rest of the current line on a hold list
 */
void
holdlist_put(struct holdlist *h, struct line *l)
{
	char *word;
	struct string *ss;

	if ((word = skip_blanks(l)) == NULL)
		err_file("%s: missing line for wait list", h->name);

	compress_blanks(word);

	/* save */
	ss = ats_alloc(sizeof(*ss));
	ss->string = ats_strdup(word);
	TAILQ_INSERT_TAIL(&h->hold, ss, link);
}

/*
 * create a eventlist
 */
struct eventlist *
eventlist_create(const char *name, struct waitlist *wl)
{
	struct eventlist *h;

	LIST_FOREACH(h, &eventlist_list, link)
		if (strcmp(h->name, name) == 0)
			err_file("eventlist %s already exists", name);
	h = ats_alloc(sizeof(*h));
	h->name = ats_strdup(name);
	h->waitlist = wl;
	TAILQ_INIT(&h->events);
	LIST_INSERT_HEAD(&eventlist_list, h, link);

	return (h);
}

/*
 * Put the rest of the current line on a event list
 */
void
eventlist_put(struct eventlist *h, const char *s)
{
	struct string *ss;

	/* save */
	ss = ats_alloc(sizeof(*ss));
	ss->string = ats_strdup(s);
	TAILQ_INSERT_TAIL(&h->events, ss, link);
}


/*
 * Put the rest of the current line on a wait list
 */
void
waitlist_put(struct waitlist *w, struct line *l)
{
	char *word;
	struct string *ss;

	if ((word = skip_blanks(l)) == NULL)
		err_file("%s: missing line for wait list", w->name);

	compress_blanks(word);

	/* save */
	ss = ats_alloc(sizeof(*ss));
	ss->string = ats_strdup(word);
	TAILQ_INSERT_TAIL(&w->wait, ss, link);
}

/*
 * Try to match the given string with a waitlist
 */
int
waitlist_match_ign(struct waitlist *w, const char *str)
{
	struct string *s;

	trace(3, "%s: '%s'", w->name, str);

	TAILQ_FOREACH(s, &w->wait, link)
		if (match(s->string, str) == 0)
			break;

	if (s == NULL)
		return (0);

	if (s != TAILQ_FIRST(&w->wait))
		report_problem("out of order %s data '%s'", w->name, str);

	trace(4, "removing from wait list");

	TAILQ_REMOVE(&w->wait, s, link);
	free(s->string);
	free(s);

	return (1);
}

int
waitlist_match(struct waitlist *w, const char *str)
{
	if (!waitlist_match_ign(w, str)) {
		report_problem("unexpected %s data '%s'", w->name, str);
		return (0);
	}
	return (1);
}

/*
 * create a ignlist
 */
struct ignlist *
ignlist_create(const char *name)
{
	struct ignlist *w;

	w = ats_alloc(sizeof(*w));
	w->name = ats_strdup(name);
	w->gen = 0;
	TAILQ_INIT(&w->ignore);

	return (w);
}

/*
 * Put the rest of the current line on a ign list
 */
void
ignlist_push(struct ignlist *w, int mult, struct line *l)
{
	char *word;
	struct string *ss;

	if ((word = skip_blanks(l)) == NULL)
		err_file("%s: missing line for ignore list", w->name);

	compress_blanks(word);

	/* save */
	ss = ats_alloc(sizeof(*ss));
	ss->string = ats_strdup(word);
	ss->lno = w->gen | (mult != 0);
	w->gen += 2;
	TAILQ_INSERT_TAIL(&w->ignore, ss, link);
}

/*
 * Try to match the given string with a ignlist
 */
int
ignlist_match(struct ignlist *w, const char *str)
{
	struct string *s;

	trace(3, "try %s: '%s'", w->name, str);

	TAILQ_FOREACH(s, &w->ignore, link)
		if (match(s->string, str) == 0)
			break;

	if (s == NULL)
		return (0);

	if ((s->lno & 1) == 0) {
		trace(4, "removing from ignore list");

		TAILQ_REMOVE(&w->ignore, s, link);
		free(s->string);
		free(s);
	} else
		trace(4, "match on ignore list");

	return (1);
}

/* pop ignore */
void
ignlist_pop(struct ignlist *w)
{
	struct string *ss, *ss1;

	if (w->gen == 0)
		err_file("%s -- nothing to pop", w->name);
	w->gen -= 2;

	TAILQ_FOREACH_SAFE(ss, &w->ignore, link, ss1)
		if ((ss->lno & ~1) == w->gen) {
			TAILQ_REMOVE(&w->ignore, ss, link);
			free(ss->string);
			free(ss);
		}
}

/*
 * malloc with fatal exit
 */
void *
ats_alloc(size_t s)
{
	void *p;

	p = malloc(s);
	if (p == NULL) {
		perror("libats");
		exit(1);
	}
	memset(p, 0, s);
	return (p);
}

void *
ats_realloc(void *p, size_t s)
{

	p = realloc(p, s);
	if (p == NULL) {
		perror("libats");
		exit(1);
	}
	return (p);
}

/*
 * strdup with error check
 */
char *
ats_strdup(const char *s)
{
	char *c;

	c = ats_alloc(strlen(s) + 1);
	return (strcpy(c, s));
}


struct tdata {
	void	*arg;
	void	(*func)(void *);
	int	id;
};

static void
tfunc(int tid __unused, void *uap)
{
	struct tdata *t = uap;

	t->func(t->arg);
	free(t);
}


/*
 * start a one-shot timer
 */
void *
ats_start_timer(u_int msec, void (*func)(void *), void *uarg)
{
	struct tdata *t;

	t = ats_alloc(sizeof(*t));
	t->arg = uarg;
	t->func = func;
	t->id = -1;

	t->id = poll_start_timer(msec, 0, tfunc, t);

	return (t);
}

void
ats_stop_timer(void *a)
{
	struct tdata *t = a;

	poll_stop_timer(t->id);
	free(t);
}

/* file, sockets */
struct fdata {
	int	fd;
	void	*arg;
	void	(*func)(int, void *);
	int	id;
};

static void
ffunc(int fd __unused, int mask __unused, void *uap)
{
	struct fdata *f = uap;

	(*f->func)(f->fd, f->arg);
}

void *
ats_register_fd(int fd, void (*func)(int, void *), void *uarg)
{
	struct fdata *f;

	f = ats_alloc(sizeof(*f));
	f->fd = fd;
	f->arg = uarg;
	f->func = func;
	f->id = poll_register(fd, ffunc, f, POLL_IN);
	return (f);
}

void
ats_unregister_fd(void *p)
{
	struct fdata *f = p;

	poll_unregister(f->id);
	free(f);
}
