/*
 * Copyright (c) 2003-2003
 *	Fraunhofer Institute for Open Communication Systems (FhG Fokus).
 * 	All rights reserved.
 *
 * Author: Hartmut Brandt <harti@freebsd.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Begemot: atmsupport/libats/ats_var.c,v 1.9 2004/07/08 13:33:36 brandt Exp $
 */

#include <sys/time.h>

#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <errno.h>
#include <limits.h>
#include <inttypes.h>

#include "ats.h"
#include "atspriv.h"

/* current definition level */
u_int var_level;

/* list of variables */
static struct var_list vars = LIST_HEAD_INITIALIZER(vars);

/*
 * Find a variable. The name begins at 's' and consists of alnums.
 * endp is set to the next address behind the variable's name.
 */
struct var *
var_find(const char *s, const char **endp)
{
	const char *end;
	struct var *var;

	end = s;
	while (*end != '\0' && isascii(*end) && (isalnum(*end) || *end == '_'))
		end++;

	if (!isascii(*end) || end == s)
		err_file("bad variable name '%.*s'", (int)(end - s), s);

	LIST_FOREACH(var, &vars, link)
		if (strlen(var->name) == (size_t)(end - s) &&
		    strncmp(var->name, s, end - s) == 0)
			break;

	if (var == NULL)
		err_file("unknown variable '%.*s'", (int)(end - s), s);

	if (endp != NULL)
		*endp = end;

	return (var);
}

/*
 * Set the variable's value to the given string.
 */
void
var_setn(struct var *var, const char *s, size_t len)
{
	struct value *v = var->v;

	while (v->ref != NULL)
		v = v->ref;

	if (v->set_val != NULL) {
		v->set_val(v, s, len);
		return;
	}
	free(v->val);
	v->val = ats_alloc(len + 1);
	strncpy(v->val, s, len);
	v->val[len] = '\0';

	if (ISVERB('v'))
		printf("VERB: '%s' := '%s' (%u)\n", var->name,
		    v->val, v->def_level);
}

/*
 * Define a variable and push earlier definitions
 * This is called for the '.' statement and function arguments.
 */
struct var *
var_define(const char *name, const char *value, int ref,
    struct var *ref_var)
{
	struct var *var;
	struct value *val;
	struct value *ref_value;

	LIST_FOREACH(var, &vars, link)
		if (strcmp(name, var->name) == 0)
			break;

	if (var == NULL) {
		/* new variable */
		var = ats_alloc(sizeof(*var));
		var->name = ats_strdup(name);
		var->v = NULL;

		LIST_INSERT_HEAD(&vars, var, link);

	} else {
		/* known variable */
		if (var->v->set_val != NULL) {
			/* don't allow redefinition of special variables */
			err_file("%s: cannot redefine special variable", name);
		}
		if (var->v->def_level == var_level)
			err_file("%s: cannot redefine", name);
	}

	ref_value = NULL;	/* gcc */
	if (ref)
		/* take this before stacking below */
		ref_value = ref_var->v;

	/*
	 * Allocate new value
	 */
	val = ats_alloc(sizeof(*val));
	val->def_level = var_level;
	val->stack = var->v;
	var->v = val;

	if (ref) {
		/* this is a reference. Take the value of the reference
		 * before stacking above in the case a variable references
		 * to an outer instance of itself.
		 */
		val->ref = ref_value;
		if (ISVERB('v'))
			printf("VERB: '%s' := &%s (%u)\n", var->name,
			    ref_var->name, val->def_level);
	} else {
		/* real variable */
		if (value != NULL)
			val->val = ats_strdup(value);
		if (ISVERB('v'))
			printf("VERB: '%s' := '%s' (%u)\n", var->name,
			    val->val ? val->val : "{NULL}",
			    val->def_level);
	}

	return (var);
}

/*
 * Define a global variable. This is possible only if no variable of
 * this name exists yet at local levels
 */
struct var *
var_define_global(const char *name, const char *value)
{
	struct var *var;
	struct value *val;

	LIST_FOREACH(var, &vars, link)
		if (strcmp(name, var->name) == 0)
			break;

	if (var == NULL) {
		/* new variable */
		var = ats_alloc(sizeof(*var));
		var->name = ats_strdup(name);

		LIST_INSERT_HEAD(&vars, var, link);

	} else if (var->v->def_level != 0) {
		err_file("%s: redefinining as global", name);

	} else {
		free(var->v->val);
		free(var->v);
	}

	/*
	 * Allocate value
	 */
	var->v = val = ats_alloc(sizeof(*val));
	val->def_level = 0;
	var->v = val;

	if (value != NULL)
		val->val = ats_strdup(value);
	if (ISVERB('v'))
		printf("VERB: '%s' := '%s' (%u)\n", var->name,
		    val->val ? val->val : "{NULL}",
		    val->def_level);

	return (var);
}

/*
 * Define a global, resetable variable. This variable must not exist.
 */
struct var *
var_define_resetable(const char *name, const char *value)
{
	struct var *var;
	struct value *val;

	LIST_FOREACH(var, &vars, link)
		if (strcmp(name, var->name) == 0)
			break;

	if (var != NULL)
		err_file("%s: redefinining resetable", name);

	var = ats_alloc(sizeof(*var));
	var->name = ats_strdup(name);

	LIST_INSERT_HEAD(&vars, var, link);

	/*
	 * Allocate value
	 */
	var->v = val = ats_alloc(sizeof(*val));
	val->def_level = 0;

	val->val = ats_strdup(value);
	val->reset = ats_strdup(value);
	if (ISVERB('v'))
		printf("VERB: '%s' := '%s' (%u)\n", var->name,
		    val->val, val->def_level);

	return (var);
}

/*
 * Pop all variables that have been defined above this level.
 */
void
var_pop(u_int level)
{
	struct var *var, *next;
	struct value *pop;

	var = LIST_FIRST(&vars);
	while (var != NULL) {
		next = LIST_NEXT(var, link);

		while (var->v != NULL && var->v->def_level >= level) {
			pop = var->v->stack;
			if (ISVERB('v'))
				printf("VERB: pop '%s' (%u)\n",
				    var->name, var->v->def_level);
			if (var->v->val != NULL)
				free(var->v->val);
			free(var->v);
			var->v = pop;
		}
		if (var->v == NULL) {
			if (ISVERB('v'))
				printf("VERB: delete '%s'\n", var->name);
			LIST_REMOVE(var, link);
			free(var->name);
			free(var);
		}

		var = next;
	}
}

/*
 * Make substitution on the given variable
 */
const char *
var_subst(const char *ptr, const char **end)
{
	struct var *var;
	const char *repl;

	var = var_find(ptr, end);
	repl = var_getval(var);

	if (ISVERB('v'))
		printf("VERB: replace '%s' -> '%s'\n", var->name, repl);

	return (repl);
}

/*
 * Random special variable
 */
static char *
random_get(struct value *val __unused)
{
	static char sbuf[100];

	sprintf(sbuf, "%ld", random());
	return (sbuf);
}
static void
random_set(struct value *val __unused, const char *str, size_t len)
{
	char *buf, *end;
	uintmax_t n;

	buf = ats_alloc(len + 1);
	strncpy(buf, str, len);
	buf[len] = '\0';

	errno = 0;
	n = strtoumax(buf, &end, 10);
	if (errno != 0 || *end != '\0' || end == buf)
		err_file("bad integer value for random '%s'", buf);
	if (n > ULONG_MAX)
		err_file("random seed too large");
	srandom((u_long)n);

	free(buf);
}

/*
 * Random special variable
 */
static char *
clock_get(struct value *val __unused)
{
	struct timeval tv;
	static char sbuf[100];

	if (gettimeofday(&tv, NULL) == -1)
		err_file("gettimeofday: %s", strerror(errno));
	sprintf(sbuf, "%"FMT_INTMAX"d",
	    (intmax_t)tv.tv_sec * 1000 + tv.tv_usec / 1000);
	return (sbuf);
}
static void
clock_set(struct value *val __unused, const char *str __unused,
    size_t len __unused)
{
	err_file("cannot set 'clock'");
}

/*
 * Return a pointer to the value of a variable
 */
const char *
var_getval(const struct var *var)
{
	struct value *v = var->v;

	while (v->ref != NULL)
		v = v->ref;

	if (v->get_val != NULL)
		return (v->get_val(v));
	else
		return (v->val);
}

/*
 * Create special variables
 */
void
var_create_special(void)
{
	struct var *v;

	v = var_define("random", NULL, 0, NULL);
	v->v->get_val = random_get;
	v->v->set_val = random_set;

	v = var_define("clock", NULL, 0, NULL);
	v->v->get_val = clock_get;
	v->v->set_val = clock_set;
}

/*
 * Pop all local variable
 */
void
var_reset(void)
{
	struct var *var;

	var_pop(1);
	LIST_FOREACH(var, &vars, link) {
		if (var->v != NULL && var->v->reset != NULL) {
			free(var->v->val);
			var->v->val = ats_strdup(var->v->reset);
		}
	}
}
