/*
 * Copyright (c) 2001-2003
 *	Fraunhofer Institute for Open Communication Systems (FhG Fokus).
 * 	All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Author: Harti Brandt <harti@freebsd.org>
 *
 * $Begemot: atmsupport/libtestlib/dump.c,v 1.3 2004/12/20 10:44:55 brandt Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <err.h>
#include <string.h>
#include <errno.h>
#include <limits.h>
#include "testlib.h"

/*
 * Dump a block of data in hex to stdout.
 */
void
dump_hex(u_char *data, size_t n)
{
	int pos = 0;

	while (n--) {
		if (pos % 16 == 0)
			printf("%04x  ", pos);
		if (pos % 16 == 8)
			printf(" ");
		printf(" %02x", *data++);
		if (pos % 16 == 15)
			printf("\n");
		pos++;
	}
	if (pos % 16 != 0)
		printf("\n");
}

/*
 * Save data into the next file of the given type.
 */
void
save_data(const char *prefix, const char *type, u_int *filenum,
    const u_char *data, size_t n)
{
	char fname[NAME_MAX];
	int fd;
	ssize_t ret;

	if (prefix[0] == '\0')
		return;

	snprintf(fname, sizeof(fname), "%s.%s%03d", prefix, type, (*filenum)++);
	if ((fd = open(fname, O_WRONLY | O_CREAT, 0666)) < 0) {
		warn("%s", fname);
		return;
	}

	if ((ret = write(fd, data, n)) == -1)
		warnx("%s: %s: write failed", fname, strerror(errno));
	else if ((size_t)ret != n)
		warnx("%s: short write (%zu bytes from %zu)", fname, n,
		    (size_t)ret);
	if (fsync(fd) == -1)
		warn("%s: flush failed", fname);
	if (close(fd) == -1)
		warn("%s: close error", fname);
}

/*
 * Read a line from the user and parse it.
 */
int
read_user(char *buf, char **argv)
{
	ssize_t n;
	int argc;

	if ((n = read(0, buf, LINE_MAX)) == -1)
		err(1, "stdin");

	if (n == 0)
		return (-1);

	buf[n] = '\0';
	while (n > 0 && buf[n-1] == '\n') {
		n--;
		buf[n] = '\0';
	}

	argc = 0;
	while ((*argv = strsep(&buf, " \t\n")) != NULL) {
		if (**argv != '\0') {
			if (++argc >= ARG_MAX)
				return (argc);
			argv++;
		}
	}
	return (argc);
}

/*
 * Read a file into a buffer.
 */
ssize_t
read_file(const char *fn, u_char *buf, size_t buflen)
{
	ssize_t n;
	int fd;

	if (fn == NULL) {
		fprintf(stderr, "missing file name\n");
		return (-1);
	}
	if ((fd = open(fn, O_RDONLY, 0)) < 0) {
		warn("%s", fn);
		return (-1);
	}

	if ((n = read(fd, buf, buflen)) < 0) {
		warn("%s: reading file", fn);
		(void)close(fd);
		return (-1);
	}

	(void)close(fd);

	return (n);
}
