/*
 * Copyright (c) 2001-2003
 *	Fraunhofer Institute for Open Communication Systems (FhG Fokus).
 * 	All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Author: Harti Brandt <harti@freebsd.org>
 *
 * Netgraph support functions.
 *
 * $Begemot: atmsupport/libtestlib/ng.c,v 1.4 2004/12/20 10:44:56 brandt Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <netgraph.h>
#include <err.h>
#include "testlib.h"

/*
 * Connect two already existing nodes by issuing a connect message:
 *
 *       hook    peerhook
 *  path<---------------->cpath
 */
static void
ngconnect(int csock, const char *path, const char *cpath,
    const char *hook, const char *peerhook)
{
	struct ngm_connect conn;

	strcpy(conn.path, cpath);
	strcpy(conn.ourhook, hook);
	strcpy(conn.peerhook, peerhook);

	if (NgSendMsg(csock, path, NGM_GENERIC_COOKIE, NGM_CONNECT,
	     &conn, sizeof(conn)) < 0)
		err(1, "connect(%s,%s.%s)", path, cpath, peerhook);
}

/*
 * Issue a mkpeer messages to create a new node, name it and connect it
 * to an already existing node:
 *
 *        hook                peerhook
 *   path<---------------------------->xname(type)
 */
static void
ngmkpeer(int csock, const char *path, const char *type, const char *hook,
    const char *peerhook, const char *xname)
{
	struct ngm_mkpeer mkp;
	struct ngm_name name;
	char xpath[NG_PATHLEN + 1];

	strcpy(mkp.type, type);
	strcpy(mkp.ourhook, hook);
	strcpy(mkp.peerhook, peerhook);

	if (NgSendMsg(csock, path, NGM_GENERIC_COOKIE, NGM_MKPEER,
	    &mkp, sizeof(mkp)) < 0)
		err(1, "mkpeer(%s,%s)", path, type);

	strcpy(name.name, xname);

	snprintf(xpath, sizeof(xpath), "%s%s", path, hook);
	if (NgSendMsg(csock, xpath, NGM_GENERIC_COOKIE, NGM_NAME,
	    &name, sizeof(name)) < 0)
		err(1, "name(%s,%s)", xpath, xname);
}

/*
 * Shutdown a node.
 */
static void
ngshutdown(int csock, const char *path)
{

	(void)NgSendMsg(csock, path, NGM_GENERIC_COOKIE,
	    NGM_SHUTDOWN, NULL, 0);
}

void
shutdown_stack(int csock, const struct stack *sp)
{
	char sock[NG_NODELEN+1];

	while (sp->op != STACK_END) {
		if (sp->op == STACK_MKPEER) {
			sprintf(sock, "%s:", sp->name);
			ngshutdown(csock, sock);
		}
		sp++;
	}
}

/*
 * Build a netgraph stack.
 */
void
build_stack(int *csock, int *dsock, const struct stack *sp)
{
	char sock[NG_NODELEN+1];

	sprintf(sock, "test%u", getpid());
	if (NgMkSockNode(sock, csock, dsock))
		err(1, "cannot make netgraph sockets");

	while (sp->op != STACK_END) {
		if (sp->op == STACK_MKPEER)
			ngmkpeer(*csock, sp->path, sp->cpath, sp->hook,
			    sp->peerhook, sp->name);
		else if (sp->op == STACK_CONNECT)
			ngconnect(*csock, sp->path, sp->cpath, sp->hook,
			    sp->peerhook);
		else
			abort();
		sp++;
	}
}

/*
 * Read a message from the control socket, asciify it and print it to the
 * user.
 */
void
read_cmsg(int csock)
{
	union {
		u_char buf[MAXMSG];
		struct ng_mesg m;
	} buf;
	struct ng_mesg *m = &buf.m;
	struct ng_mesg *a = (struct ng_mesg *)(void *)m->data;
	char path[NG_PATHLEN + 1];

	if (NgRecvMsg(csock, m, sizeof(buf), path) < 0)
		err(1, "NgRecvMsg");
	if (NgSendMsg(csock, path, NGM_GENERIC_COOKIE, NGM_BINARY2ASCII,
	    m, sizeof(*m) + m->header.arglen) < 0)
		err(1, "NgSendMsg: BINARY2ASCII");
	if (NgRecvMsg(csock, m, sizeof(buf), NULL) < 0)
		err(1, "NgRecvMsg: BINARY2ASCII");

	printf("received %s '%s'(%d) from '%s':\n",
	    (a->header.flags & NGF_RESP) ? "response" : "request",
	    a->header.cmdstr, a->header.cmd, path);

	if (*a->data != '\0')
		printf("Args: '%s'\n", a->data);
}
