/*
 * Copyright (c) 2004
 *	Hartmut Brandt
 * 	All rights reserved.
 *
 * Author: Hartmut Brandt <harti@freebsd.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Begemot: atmsupport/libunitcl/api.c,v 1.8 2004/07/08 13:33:40 brandt Exp $
 */
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>

#include "inctcl.h"

#include <netnatm/unimsg.h>
#include <netnatm/msg/unistruct.h>
#include <netnatm/api/unisap.h>
#include <netnatm/sig/unidef.h>
#include <netnatm/api/atmapi.h>

#include "unitcl.h"
#include "unitcl_priv.h"

static const struct {
	int32_t	resp;
	const char *name;
	const char *txt;
} responses[] = {
#define DEF(NAME, VAL, STR)	{ NAME, #NAME, STR },
ATM_DEFINE_ERRORS
#undef DEF
};

static const char *attr_names[] = {
	[ATM_ATTR_NONE] =		NULL,
	[ATM_ATTR_BLLI_SELECTOR] =	"blli-selector",
	[ATM_ATTR_BLLI] =		"blli",
	[ATM_ATTR_BEARER] =		"bearer",
	[ATM_ATTR_TRAFFIC] =		"traffic",
	[ATM_ATTR_QOS] =		"qos",
	[ATM_ATTR_EXQOS] =		"exqos",
	[ATM_ATTR_CALLED] =		"called",
	[ATM_ATTR_CALLEDSUB] =		"calledsub",
	[ATM_ATTR_CALLING] =		"calling",
	[ATM_ATTR_CALLINGSUB] =		"callingsub",
	[ATM_ATTR_AAL] =		"aal",
	[ATM_ATTR_EPREF] =		"epref",
	[ATM_ATTR_CONNED] =		"conned",
	[ATM_ATTR_CONNEDSUB] =		"connedsub",
	[ATM_ATTR_EETD] =		"eetd",
	[ATM_ATTR_ABRSETUP] =		"abrsetup",
	[ATM_ATTR_ABRADD] =		"abradd",
	[ATM_ATTR_CONNID] =		"connid",
	[ATM_ATTR_MDCR] =		"mdcr",
};

/*
 * Format an attribute
 *
 * {<name> <attr>}
 */
static int
fmt_attr(Tcl_Interp *interp, Tcl_DString *str, u_int attr,
    const u_char **ptr, size_t *len)
{
	int ret;

#define	GET_ATTR(TYPE)						\
	TYPE a;							\
								\
	if (*len < sizeof(a))					\
		return (unitcl_setres(interp, "attr too "	\
		    "short for %s", attr_names[attr]));		\
	memcpy(&a, *ptr, sizeof(a));				\
	*len -= sizeof(a);					\
	*ptr += sizeof(a);

#define	FMT_GENERIC(IE, BIG)					\
	    {							\
		GET_ATTR(struct uni_ie_##IE);			\
		if (a.h.present == 0) {				\
			Tcl_DStringStartSublist(str);		\
			ret = unitcl_fmt_iehdr(interp, str,	\
			    UNI_IE_##BIG, &a.h);		\
			Tcl_DStringEndSublist(str);		\
			if (ret == TCL_CONTINUE)		\
				return (TCL_OK);		\
			return (ret);				\
		} else						\
			return (fmt_##IE(interp, str, &a));	\
	    }

	
	switch (attr) {
	  case ATM_ATTR_NONE:
		return (TCL_OK);

	  case ATM_ATTR_BLLI_SELECTOR:
	    {
		char buf[30];

		GET_ATTR(uint32_t);

		Tcl_DStringStartSublist(str);
		Tcl_DStringAppendElement(str, attr_names[attr]);
		sprintf(buf, "%u", a);
		Tcl_DStringAppendElement(str, buf);
		Tcl_DStringEndSublist(str);
		return (TCL_OK);
	    }

	  case ATM_ATTR_BLLI:
		FMT_GENERIC(blli, BLLI);

	  case ATM_ATTR_BEARER:
		FMT_GENERIC(bearer, BEARER);

	  case ATM_ATTR_TRAFFIC:
		FMT_GENERIC(traffic, TRAFFIC);

	  case ATM_ATTR_QOS:
		FMT_GENERIC(qos, QOS);

	  case ATM_ATTR_EXQOS:
		FMT_GENERIC(exqos, EXQOS);

	  case ATM_ATTR_CALLED:
		FMT_GENERIC(called, CALLED);

	  case ATM_ATTR_CALLEDSUB:
		FMT_GENERIC(calledsub, CALLEDSUB);

	  case ATM_ATTR_CALLING:
		FMT_GENERIC(calling, CALLING);

	  case ATM_ATTR_CALLINGSUB:
		FMT_GENERIC(callingsub, CALLINGSUB);

	  case ATM_ATTR_AAL:
		FMT_GENERIC(aal, AAL);

	  case ATM_ATTR_EPREF:
		FMT_GENERIC(epref, EPREF);

	  case ATM_ATTR_CONNED:
		FMT_GENERIC(conned, CONNED);

	  case ATM_ATTR_CONNEDSUB:
		FMT_GENERIC(connedsub, CONNEDSUB);

	  case ATM_ATTR_EETD:
		FMT_GENERIC(eetd, EETD);

	  case ATM_ATTR_ABRSETUP:
		FMT_GENERIC(abrsetup, ABRSETUP);

	  case ATM_ATTR_ABRADD:
		FMT_GENERIC(abradd, ABRADD);

	  case ATM_ATTR_CONNID:
		FMT_GENERIC(connid, CONNID);

	  case ATM_ATTR_MDCR:
		FMT_GENERIC(mdcr, MDCR);
	}

	return (unitcl_setres(interp, "bad attribute %u", attr));
}

static int
fmt_attr_name(Tcl_Interp *interp, Tcl_DString *str, u_int attr)
{

	if (attr >= sizeof(attr_names) / sizeof(attr_names[0]) ||
	    attr_names[attr] == NULL)
		return (unitcl_setres(interp,
		    "%s: bad attribute %u", __func__, attr));

	Tcl_DStringAppendElement(str, attr_names[attr]);
	return (TCL_OK);
}

static int
parse_attr(Tcl_Interp *interp, const char *str, uint32_t *attr,
    u_char **buf, size_t *len, struct unicx *cx __unused)
{
	int ac;
	const char **av;
	u_int n;
	enum uni_ietype type;
	union uni_ieall ie;

	/* split to find out about special attrs */
	if (Tcl_SplitList(interp, str, &ac, &av) != TCL_OK)
		return (TCL_ERROR);

	if (ac == 0) {
		Tcl_Free((char *)av);
		return (unitcl_setres(interp, "empty attribute"));
	}

	if (strcmp(av[0], attr_names[ATM_ATTR_BLLI_SELECTOR]) == 0) {
		uint32_t sel;

		if (ac != 2) {
			Tcl_Free((char *)av);
			return (unitcl_setres(interp,
			    "bad number of args for blli-selector"));
		}
		if (unitcl_parse_num(interp, av[1], &n) != TCL_OK)
			return (TCL_ERROR);
		sel = n;

		memcpy(*buf, &sel, sizeof(uint32_t));
		*len += sizeof(uint32_t);
		*buf += sizeof(uint32_t);
		*attr = ATM_ATTR_BLLI_SELECTOR;

		Tcl_Free((char *)av);
		return (TCL_OK);
	}
	Tcl_Free((char *)av);

	/* normal IE */
	if (parse_ie(interp, str, &ie, &type) != TCL_OK)
		return (TCL_ERROR);

	switch (type) {

	  default:
		return (unitcl_setres(interp, "bad attribute IE %u",
		    (u_int)type));

#define	PUT_ATTR(NAME, Name)						\
	  case UNI_IE_##NAME:						\
		memcpy(*buf, &ie.Name, sizeof(ie.Name));		\
		*len += sizeof(ie.Name);				\
		*buf += sizeof(ie.Name);				\
		*attr = ATM_ATTR_##NAME;				\
		break

	  PUT_ATTR(BLLI, blli);
	  PUT_ATTR(BEARER, bearer);
	  PUT_ATTR(TRAFFIC, traffic);
	  PUT_ATTR(QOS, qos);
	  PUT_ATTR(EXQOS, exqos);
	  PUT_ATTR(CALLED, called);
	  PUT_ATTR(CALLEDSUB, calledsub);
	  PUT_ATTR(CALLING, calling);
	  PUT_ATTR(CALLINGSUB, callingsub);
	  PUT_ATTR(AAL, aal);
	  PUT_ATTR(EPREF, epref);
	  PUT_ATTR(CONNED, conned);
	  PUT_ATTR(CONNEDSUB, connedsub);
	  PUT_ATTR(EETD, eetd);
	  PUT_ATTR(ABRSETUP, abrsetup);
	  PUT_ATTR(ABRADD, abradd);
	  PUT_ATTR(CONNID, connid);
	  PUT_ATTR(MDCR, mdcr);
	}
	return (TCL_OK);
}

static int
parse_attr_name(Tcl_Interp *interp, const char *name, uint32_t *attr)
{
	u_int i;

	for (i = 0; i < sizeof(attr_names) / sizeof(attr_names[0]); i++)
		if (attr_names[i] != NULL && strcmp(name, attr_names[i]) == 0) {
			*attr = i;
			return (TCL_OK);
		}

	return (unitcl_setres(interp,
	    "%s: bad attribute '%s'", __func__, name));
}

/*
 * Format a port
 */
int
fmt_port_info(Tcl_DString *str, const struct atm_port_info *port)
{
	char buf[100];

	sprintf(buf, "%u", port->port);
	Tcl_DStringAppendElement(str, buf);

	sprintf(buf, "%u", port->pcr);
	Tcl_DStringAppendElement(str, buf);

	sprintf(buf, "%u", port->max_vpi_bits);
	Tcl_DStringAppendElement(str, buf);

	sprintf(buf, "%u", port->max_vci_bits);
	Tcl_DStringAppendElement(str, buf);

	sprintf(buf, "%u", port->max_svpc_vpi);
	Tcl_DStringAppendElement(str, buf);

	sprintf(buf, "%u", port->max_svcc_vpi);
	Tcl_DStringAppendElement(str, buf);

	sprintf(buf, "%u", port->min_svcc_vci);
	Tcl_DStringAppendElement(str, buf);

	sprintf(buf, "%02x:%02x:%02x:%02x:%02x:%02x", port->esi[0],
	    port->esi[1], port->esi[2], port->esi[3], port->esi[4],
	    port->esi[5]);
	Tcl_DStringAppendElement(str, buf);

	sprintf(buf, "%u", port->num_addrs);
	Tcl_DStringAppendElement(str, buf);

	return (TCL_OK);
}

static int
fmt_port(Tcl_Interp *interp __unused, Tcl_DString *str,
    const struct atm_port_info *port)
{

	Tcl_DStringStartSublist(str);
	Tcl_DStringAppendElement(str, "port");

	if (fmt_port_info(str, port) != TCL_OK)
		return (TCL_ERROR);

	Tcl_DStringEndSublist(str);

	return (TCL_OK);
}

int
parse_port_info(Tcl_Interp *interp, int argc, const char **argv,
    struct atm_port_info *port)
{
	u_int n;
	u_int esi[6];
	int q, i;

	if (argc != 9)
		return (unitcl_setres(interp, "bad number of args for port"));

	if (unitcl_parse_num(interp, argv[0], &n) != TCL_OK)
		return (TCL_ERROR);
	port->port = n;

	if (unitcl_parse_num(interp, argv[1], &n) != TCL_OK)
		return (TCL_ERROR);
	port->pcr = n;

	if (unitcl_parse_num(interp, argv[2], &n) != TCL_OK)
		return (TCL_ERROR);
	port->max_vpi_bits = n;

	if (unitcl_parse_num(interp, argv[3], &n) != TCL_OK)
		return (TCL_ERROR);
	port->max_vci_bits = n;

	if (unitcl_parse_num(interp, argv[4], &n) != TCL_OK)
		return (TCL_ERROR);
	port->max_svpc_vpi = n;

	if (unitcl_parse_num(interp, argv[5], &n) != TCL_OK)
		return (TCL_ERROR);
	port->max_svcc_vpi = n;

	if (unitcl_parse_num(interp, argv[6], &n) != TCL_OK)
		return (TCL_ERROR);
	port->min_svcc_vci = n;

	if ((i = sscanf(argv[7], "%x:%x:%x:%x:%x:%x%n",
	    &esi[0], &esi[1], &esi[2], &esi[3], &esi[4], &esi[5], &q)) != 6 ||
	    (size_t)q != strlen(argv[7]))
		return (unitcl_setres(interp, "bad ESI '%s' (%d, %d, %zu)",
		    argv[7], i, q, strlen(argv[7])));
	port->esi[0] = esi[0];
	port->esi[1] = esi[1];
	port->esi[2] = esi[2];
	port->esi[3] = esi[3];
	port->esi[4] = esi[4];
	port->esi[5] = esi[5];

	if (unitcl_parse_num(interp, argv[8], &n) != TCL_OK)
		return (TCL_ERROR);
	port->num_addrs = n;

	return (TCL_OK);
}

static int
parse_port(Tcl_Interp *interp, int argc, const char **argv, u_char *buf)
{

	return (parse_port_info(interp, argc, argv,
	    (struct atm_port_info *)(void *)buf));
}

/*
 * Format an address
 *
 * {addr <type> <addr>}
 */
static int
fmt_apiaddr(Tcl_Interp *interp __unused, Tcl_DString *str,
    const struct uni_addr *addr)
{

	Tcl_DStringStartSublist(str);
	Tcl_DStringAppendElement(str, "addr");

	if (fmt_addr(interp, str, addr) != TCL_OK)
		return (TCL_ERROR);

	Tcl_DStringEndSublist(str);

	return (TCL_OK);
}

static int
parse_apiaddr(Tcl_Interp *interp, int argc, const char **argv, u_char *buf)
{
	struct uni_addr *addr = (struct uni_addr *)(void *)buf;

	if (argc != 2)
		return (unitcl_setres(interp, "bad number of args for addr"));

	if (parse_addr(interp, argv, addr) != TCL_OK)
		return (TCL_ERROR);

	return (TCL_OK);
}

/*
 * Format an ATMAPI response.
 *
 * {atm-resp <code> {<message>}
 *   ({ATTRS <attr>*} |
 *    {PORTS <port>* <addr>*} |
 *    {STATE <name> <state>})}
 */
static int
fmt_atm_resp(Tcl_Interp *interp, Tcl_DString *str,
    const u_char *ptr, size_t ilen)
{
	const struct atm_resp *resp =
	    (const struct atm_resp *)(const void *)ptr;
	int err;
	u_int i;
	char sbuf[100];
	size_t len = ilen;

	for (i = 0; i < sizeof(responses) / sizeof(responses[0]); i++)
		if (responses[i].resp == resp->resp)
			break;

	if (i == sizeof(responses) / sizeof(responses[0])) {
		sprintf(sbuf, "%d", resp->resp);
		Tcl_DStringAppendElement(str, sbuf);
	} else {
		Tcl_DStringAppendElement(str, responses[i].name);
	}
	len -= sizeof(*resp);
	ptr += sizeof(*resp);

	switch (resp->data) {

	  case ATMRESP_NONE:
		if (len != 0)
			return (unitcl_setres(interp,
			    "bad length for atm-resp (%zu, need %zu)",
			    ilen, sizeof(*resp)));
		break;

	  case ATMRESP_ATTRS:
	    {
		const uint32_t *attr = (const uint32_t *)(const void *)ptr;
		uint32_t count;

		if (len < sizeof(uint32_t))
			return (unitcl_setres(interp,
			    "short length for atm-resp (%zu, no count)"
			    " with RESP_ATTR", ilen));
		count = *attr++;
		len -= sizeof(uint32_t);
		ptr += sizeof(uint32_t);

		if (len < count * sizeof(uint32_t))
			return (unitcl_setres(interp,
			    "short length for atm-resp (%zu, %u attrs)"
			    " with RESP_ATTR", ilen, count));
		len -= count * sizeof(uint32_t);
		ptr += count * sizeof(uint32_t);

		Tcl_DStringStartSublist(str);
		Tcl_DStringAppendElement(str, "ATTRS");
		for (i = 0; i < count; i++) {
			err = fmt_attr(interp, str, attr[i], &ptr, &len);
			if (err != TCL_OK)
				return (err);
		}
		Tcl_DStringEndSublist(str);
		if (len != 0)
			return (unitcl_setres(interp,
			    "long length for atm-resp (%zu, rest %zu)"
			    " with RESP_ATTR", ilen, len));
		break;
	    }


	  case ATMRESP_PORTS:
	    {
		const struct atm_port_list *arg;
		const struct atm_port_info *p;
		const struct uni_addr *a;

		if (len < sizeof(*arg))
			return (unitcl_setres(interp, "response too short for "
			    "LOCAL_PORT_LIST (%zu, need %zu)", ilen,
			    sizeof(*resp) + sizeof(*arg)));

		len -= sizeof(*arg);
		arg = (const struct atm_port_list *)(const void *)ptr;
		ptr += sizeof(*arg);

		if (len != arg->num_ports * sizeof(*p) +
		    arg->num_addrs * sizeof(*a))
			return (unitcl_setres(interp,
			    "response bad length for LOCAL_PORT_LIST (%zu, "
			    "%u ports, %u addrs)", ilen, arg->num_ports,
			    arg->num_addrs));

		p = (const struct atm_port_info *)(const void *)ptr;
		a = (const struct uni_addr *)(const void *)(ptr +
		    arg->num_ports * sizeof(*p));

		Tcl_DStringStartSublist(str);
		Tcl_DStringAppendElement(str, "PORTS");
		for (i = 0; i < arg->num_ports; i++) {
			err = fmt_port(interp, str, &p[i]);
			if (err != TCL_OK)
				return (err);
		}
		for (i = 0; i < arg->num_addrs; i++) {
			err = fmt_apiaddr(interp, str, &a[i]);
			if (err != TCL_OK)
				return (err);
		}
		Tcl_DStringEndSublist(str);
		break;
	    }

	  case ATMRESP_STATE:
	    {
		const struct atm_epstate *arg;
		char buf[100];

		if (len < sizeof(*arg))
			return (unitcl_setres(interp, "response too short for "
			    "ATMRESP_STATE (%zu, need %zu)", ilen,
			    sizeof(*arg)));

		arg = (const struct atm_epstate *)ptr;

		Tcl_DStringStartSublist(str);
		Tcl_DStringAppendElement(str, "STATE");

		Tcl_DStringAppendElement(str, arg->name);

		sprintf(buf, "%u", arg->state);
		Tcl_DStringAppendElement(str, buf);
		Tcl_DStringEndSublist(str);
		break;
	    }

	  default:
		return (unitcl_setres(interp, "bad response data %u",
		    resp->data));
	}
	return (TCL_OK);
}

/*
 * Parse an ATM response
 */
static int
parse_atm_resp(Tcl_Interp *interp, int argc, const char **argv,
    u_char *buf, size_t *len, struct unicx *cx)
{
	u_int i;
	struct atm_resp *resp = (struct atm_resp *)(void *)buf;
	int ac, n;
	const char **av;
	int err;

	if (argc != 2 && argc != 3)
		return (unitcl_setres(interp,
		    "%s: wrong number of args", __func__));

	/* error code or name */
	for (i = 0; i < sizeof(responses) / sizeof(responses[0]); i++)
		if (strcmp(argv[0], responses[i].name) == 0) {
			resp->resp = responses[i].resp;
			break;
		}

	if (i == sizeof(responses) / sizeof(responses[0])) {
		if (unitcl_parse_num(interp, argv[0], &i) != TCL_OK)
			return (TCL_ERROR);
		resp->resp = i;
	}

	*len += sizeof(*resp);
	buf += sizeof(*resp);

	if (argc == 2) {
		resp->data = ATMRESP_NONE;
		return (TCL_OK);
	}

	if (Tcl_SplitList(interp, argv[1], &ac, &av) != TCL_OK)
		return (TCL_ERROR);

	if (ac == 0) {
		Tcl_Free((char *)av);
		return (unitcl_setres(interp, "no type for response data"));
	}

	if (strcmp(av[0], "ATTRS") == 0) {
		/*
		 * attribute list
		 */
		uint32_t *attr = (uint32_t *)(void *)buf;

		resp->data = ATMRESP_ATTRS;
		attr[0] = ac - 1;

		buf += ac * sizeof(uint32_t);
		*len += ac * sizeof(uint32_t);

		for (n = 1; n < ac; n++) {
			err = parse_attr(interp, av[n], &attr[n],
			    &buf, len, cx);
			if (err != TCL_OK) {
				Tcl_Free((char *)av);
				return (err);
			}
		}

	} else if (strcmp(av[0], "PORTS") == 0) {
		/*
		 * Ports and address list
		 */
		int in_addr = 0;
		int ac1;
		const char **av1;
		struct atm_port_list *arg;

		resp->data = ATMRESP_PORTS;

		arg = (struct atm_port_list *)(void *)buf;
		buf += sizeof(*arg);
		*len += sizeof(*arg);

		arg->num_ports = arg->num_addrs = 0;
		for (n = 1; n < ac; n++) {
			if (Tcl_SplitList(interp, av[n], &ac1, &av1) != TCL_OK)
				return (TCL_ERROR);
			if (ac1 == 0) {
				Tcl_Free((char *)av1);
				Tcl_Free((char *)av);
				return (unitcl_setres(interp,
				    "'port' or 'addr' expected"));
			}
			if (strcmp(av1[0], "port") == 0) {
				if (in_addr) {
					Tcl_Free((char *)av1);
					Tcl_Free((char *)av);
					return (unitcl_setres(interp,
					    "port in address list"));
				}
				err = parse_port(interp, ac1 - 1, av1 + 1, buf);
				if (err != TCL_OK) {
					Tcl_Free((char *)av1);
					Tcl_Free((char *)av);
					return (err);
				}
				arg->num_ports++;
				buf += sizeof(struct atm_port_info);
				*len += sizeof(struct atm_port_info);

			} else if (strcmp(av1[0], "addr") == 0) {
				in_addr = 1;
				err = parse_apiaddr(interp, ac1 - 1,
				    av1 + 1, buf);
				if (err != TCL_OK) {
					Tcl_Free((char *)av1);
					Tcl_Free((char *)av);
					return (err);
				}
				arg->num_addrs++;
				buf += sizeof(struct uni_addr);
				*len += sizeof(struct uni_addr);

			} else {
				Tcl_Free((char *)av1);
				Tcl_Free((char *)av);
				return (unitcl_setres(interp,
				    "'port' or 'addr' expected"));
			}
		}

	} else {
		Tcl_Free((char *)av);
		return (unitcl_setres(interp, "bad type for response data"));
	}

	Tcl_Free((char *)av);
	return (TCL_OK);
}

/*
 * Format an atm-abort-connection
 *
 * {atm-abort-connection <leaf-ident> <ie-cause>}
 */
static int
fmt_atm_abort_connection(Tcl_Interp *interp, Tcl_DString *str,
    const u_char *ptr, size_t ilen __unused)
{
	const struct atm_abort_connection *arg =
	    (const struct atm_abort_connection *)(const void *)ptr;

	return (fmt_cause(interp, str, &arg->cause));
}

static int
parse_atm_abort_connection(Tcl_Interp *interp, int argc, const char **argv,
    u_char *buf, size_t *len, struct unicx *cx __unused)
{
	struct atm_abort_connection *arg =
	    (struct atm_abort_connection *)(void *)buf;
	int error;
	union uni_ieall ie;
	u_int ietype;

	if (argc != 1)
		return (unitcl_setres(interp,
		    "%s: wrong number of args", __func__));

	if ((error = parse_ie(interp, argv[0], &ie, &ietype)) != TCL_OK)
		return (error);

	if (ietype != UNI_IE_CAUSE)
		return (unitcl_setres(interp,
		    "%s: expecting {cause ...}", __func__));

	arg->cause = ie.cause;
	*len += sizeof(*arg);

	return (TCL_OK);
}

/*
 * Format an atm-accept-incoming-call
 *
 * {atm-accept-incoming-call
 */
static int
fmt_atm_accept_incoming_call(Tcl_Interp *interp __unused, Tcl_DString *str,
    const u_char *ptr, size_t ilen __unused)
{
	const struct atm_accept_incoming_call *arg =
	    (const struct atm_accept_incoming_call *)ptr;

	Tcl_DStringAppendElement(str, arg->newep);
	return (TCL_OK);
}

static int
parse_atm_accept_incoming_call(Tcl_Interp *interp, int argc, const char **argv,
    u_char *buf, size_t *len, struct unicx *cx __unused)
{
	struct atm_accept_incoming_call *arg =
	    (struct atm_accept_incoming_call *)buf;

	if (argc != 1)
		return (unitcl_setres(interp,
		    "%s: wrong number of args", __func__));

	if (strlen(argv[0]) >= ATM_EPNAMSIZ)
		return (unitcl_setres(interp,
		    "%s: endpoint name too long", __func__));

	strcpy(arg->newep, argv[0]);
	*len += sizeof(*arg);

	return (TCL_OK);
}

/*
 * Format an atm-add-party
 *
 * {atm-add-party <leaf-ident> <ie-called>}
 */
static int
fmt_atm_add_party(Tcl_Interp *interp, Tcl_DString *str,
    const u_char *ptr, size_t ilen __unused)
{
	const struct atm_add_party *arg =
	    (const struct atm_add_party *)(const void *)ptr;
	char buf[100];

	sprintf(buf, "%u", arg->leaf_ident);
	Tcl_DStringAppendElement(str, buf);

	return (fmt_called(interp, str, &arg->called));
}

static int
parse_atm_add_party(Tcl_Interp *interp, int argc, const char **argv,
    u_char *buf, size_t *len, struct unicx *cx __unused)
{
	struct atm_add_party *arg =
	    (struct atm_add_party *)(void *)buf;
	int error;
	union uni_ieall ie;
	u_int ietype, num;

	if (argc != 2)
		return (unitcl_setres(interp,
		    "%s: wrong number of args", __func__));

	if ((error = unitcl_parse_num(interp, argv[0], &num)) != TCL_OK)
		return (error);
	if (num > 32767)
		return (unitcl_setres(interp,
		    "%s: leaf identifier too large", __func__));
	arg->leaf_ident = num;
	
	if ((error = parse_ie(interp, argv[1], &ie, &ietype)) != TCL_OK)
		return (error);

	if (ietype != UNI_IE_CALLED)
		return (unitcl_setres(interp,
		    "%s: expecting {called ...}", __func__));

	arg->called = ie.called;
	*len += sizeof(*arg);

	return (TCL_OK);
}

/*
 * Format an atm-add-party-reject
 *
 * {atm-add-party-reject <leaf-ident> <ie-cause>}
 */
static int
fmt_atm_add_party_reject(Tcl_Interp *interp, Tcl_DString *str,
    const u_char *ptr, size_t ilen __unused)
{
	const struct atm_add_party_reject *arg =
	    (const struct atm_add_party_reject *)(const void *)ptr;
	char buf[100];

	sprintf(buf, "%u", arg->leaf_ident);
	Tcl_DStringAppendElement(str, buf);

	return (fmt_cause(interp, str, &arg->cause));
}

static int
parse_atm_add_party_reject(Tcl_Interp *interp, int argc, const char **argv,
    u_char *buf, size_t *len, struct unicx *cx __unused)
{
	struct atm_add_party_reject *arg =
	    (struct atm_add_party_reject *)(void *)buf;
	int error;
	union uni_ieall ie;
	u_int ietype, num;

	if (argc != 2)
		return (unitcl_setres(interp,
		    "%s: wrong number of args", __func__));

	if ((error = unitcl_parse_num(interp, argv[0], &num)) != TCL_OK)
		return (error);
	if (num > 32767)
		return (unitcl_setres(interp,
		    "%s: leaf identifier too large", __func__));
	arg->leaf_ident = num;
	
	if ((error = parse_ie(interp, argv[1], &ie, &ietype)) != TCL_OK)
		return (error);

	if (ietype != UNI_IE_CAUSE)
		return (unitcl_setres(interp,
		    "%s: expecting {cause ...}", __func__));

	arg->cause = ie.cause;
	*len += sizeof(*arg);

	return (TCL_OK);
}

/*
 * Format an atm-add-party-success
 *
 * {atm-add-party-success <leaf-ident> <ie-cause>}
 */
static int
fmt_atm_add_party_success(Tcl_Interp *interp __unused, Tcl_DString *str,
    const u_char *ptr, size_t ilen __unused)
{
	const struct atm_add_party_success *arg =
	    (const struct atm_add_party_success *)(const void *)ptr;
	char buf[100];

	sprintf(buf, "%u", arg->leaf_ident);
	Tcl_DStringAppendElement(str, buf);

	return (TCL_OK);
}

static int
parse_atm_add_party_success(Tcl_Interp *interp, int argc, const char **argv,
    u_char *buf, size_t *len, struct unicx *cx __unused)
{
	struct atm_add_party_success *arg =
	    (struct atm_add_party_success *)(void *)buf;
	int error;
	u_int num;

	if (argc != 1)
		return (unitcl_setres(interp,
		    "%s: wrong number of args", __func__));

	if ((error = unitcl_parse_num(interp, argv[0], &num)) != TCL_OK)
		return (error);
	if (num > 32767)
		return (unitcl_setres(interp,
		    "%s: leaf identifier too large", __func__));
	arg->leaf_ident = num;
	
	*len += sizeof(*arg);

	return (TCL_OK);
}

/*
 * Arrival of incoming call
 *
 *	{atm-arrival-of-incoming-call}
 */
static int
fmt_atm_arrival_of_incoming_call(Tcl_Interp *interp __unused,
    Tcl_DString *str __unused, const u_char *ptr __unused, size_t ilen __unused)
{
	return (TCL_OK);
}

static int
parse_atm_arrival_of_incoming_call(Tcl_Interp *interp, int argc,
    const char **argv __unused,
    u_char *buf __unused, size_t *len __unused, struct unicx *cx __unused)
{

	if (argc != 0)
		return (unitcl_setres(interp,
		    "%s: wrong number of args", __func__));

	return (TCL_OK);
}


/*
 * Format an atm-call-release
 *
 * {atm-call-release <cause1> [<cause2>]}
 */
static int
fmt_atm_call_release(Tcl_Interp *interp, Tcl_DString *str,
    const u_char *ptr, size_t ilen __unused)
{
	const struct atm_call_release *arg =
	    (const struct atm_call_release *)(const void *)ptr;
	int error;

	if ((error = fmt_cause(interp, str, &arg->cause[0])) != TCL_OK)
		return (error);
	if (IE_ISPRESENT(arg->cause[1])) {
		if ((error = fmt_cause(interp, str, &arg->cause[1])) != TCL_OK)
			return (error);
	}
	return (TCL_OK);
}

static int
parse_atm_call_release(Tcl_Interp *interp, int argc, const char **argv,
    u_char *buf, size_t *len, struct unicx *cx __unused)
{
	struct atm_call_release *arg =
	    (struct atm_call_release *)(void *)buf;
	int error;
	union uni_ieall ie;
	u_int ietype;

	if (argc != 1 && argc != 2)
		return (unitcl_setres(interp,
		    "%s: wrong number of args", __func__));

	*len += sizeof(*arg);

	if ((error = parse_ie(interp, argv[0], &ie, &ietype)) != TCL_OK)
		return (error);

	if (ietype != UNI_IE_CAUSE)
		return (unitcl_setres(interp,
		    "%s: expecting {cause ...}", __func__));

	arg->cause[0] = ie.cause;

	if (argc == 1) {
		arg->cause[1].h.present = 0;
		return (TCL_OK);
	}

	if ((error = parse_ie(interp, argv[1], &ie, &ietype)) != TCL_OK)
		return (error);

	if (ietype != UNI_IE_CAUSE)
		return (unitcl_setres(interp,
		    "%s: expecting {cause ...}", __func__));

	arg->cause[1] = ie.cause;

	return (TCL_OK);
}

/*
 * Format an atm-connect-outgoing-call
 *
 * {atm-connect-outgoing-call <ie-called>}
 */
static int
fmt_atm_connect_outgoing_call(Tcl_Interp *interp, Tcl_DString *str,
    const u_char *ptr, size_t ilen __unused)
{
	const struct atm_connect_outgoing_call *arg =
	    (const struct atm_connect_outgoing_call *)(const void *)ptr;

	return (fmt_called(interp, str, &arg->called));
}

static int
parse_atm_connect_outgoing_call(Tcl_Interp *interp, int argc, const char **argv,
    u_char *buf, size_t *len, struct unicx *cx __unused)
{
	struct atm_connect_outgoing_call *arg =
	    (struct atm_connect_outgoing_call *)(void *)buf;
	int error;
	union uni_ieall ie;
	u_int ietype;

	if (argc != 1)
		return (unitcl_setres(interp,
		    "%s: wrong number of args", __func__));

	if ((error = parse_ie(interp, argv[0], &ie, &ietype)) != TCL_OK)
		return (error);

	if (ietype != UNI_IE_CALLED)
		return (unitcl_setres(interp,
		    "%s: expecting {called ...}", __func__));

	arg->called = ie.called;
	*len += sizeof(*arg);

	return (TCL_OK);
}

/*
 * Format an atm-drop-party
 *
 * {atm-drop-party <leaf-ident> <ie-cause>}
 */
static int
fmt_atm_drop_party(Tcl_Interp *interp, Tcl_DString *str,
    const u_char *ptr, size_t ilen __unused)
{
	const struct atm_drop_party *arg =
	    (const struct atm_drop_party *)(const void *)ptr;
	char buf[100];

	sprintf(buf, "%u", arg->leaf_ident);
	Tcl_DStringAppendElement(str, buf);

	return (fmt_cause(interp, str, &arg->cause));
}

static int
parse_atm_drop_party(Tcl_Interp *interp, int argc, const char **argv,
    u_char *buf, size_t *len, struct unicx *cx __unused)
{
	struct atm_drop_party *arg =
	    (struct atm_drop_party *)(void *)buf;
	int error;
	union uni_ieall ie;
	u_int ietype, num;

	if (argc != 2)
		return (unitcl_setres(interp,
		    "%s: wrong number of args", __func__));

	if ((error = unitcl_parse_num(interp, argv[0], &num)) != TCL_OK)
		return (error);
	if (num > 32767)
		return (unitcl_setres(interp,
		    "%s: leaf identifier too large", __func__));
	arg->leaf_ident = num;
	
	if ((error = parse_ie(interp, argv[1], &ie, &ietype)) != TCL_OK)
		return (error);

	if (ietype != UNI_IE_CAUSE)
		return (unitcl_setres(interp,
		    "%s: expecting {cause ...}", __func__));

	arg->cause = ie.cause;
	*len += sizeof(*arg);

	return (TCL_OK);
}

/*
 * Format an atm-get-local-port-info
 *
 * {atm-get-local-port-info <portno>}
 */
static int
fmt_atm_get_local_port_info(Tcl_Interp *interp __unused, Tcl_DString *str,
    const u_char *ptr, size_t ilen __unused)
{
	const struct atm_get_local_port_info *arg =
	    (const struct atm_get_local_port_info *)(const void *)ptr;
	char buf[100];

	sprintf(buf, "%u", arg->port);
	Tcl_DStringAppendElement(str, buf);

	return (TCL_OK);
}

static int
parse_atm_get_local_port_info(Tcl_Interp *interp, int argc, const char **argv,
    u_char *buf, size_t *len, struct unicx *cx __unused)
{
	struct atm_get_local_port_info *arg =
	    (struct atm_get_local_port_info *)(void *)buf;
	int error;
	u_int num;

	if (argc != 1)
		return (unitcl_setres(interp,
		    "%s: wrong number of args", __func__));

	if ((error = unitcl_parse_num(interp, argv[0], &num)) != TCL_OK)
		return (error);
	if (num > 0xffffffff)
		return (unitcl_setres(interp,
		    "%s: portno too large", __func__));
	arg->port = num;
	*len += sizeof(*arg);

	return (TCL_OK);
}

/*
 * Format an atm-p2mp-call-active
 *
 * {atm-p2mp-call-active <connid>}
 */
static int
fmt_atm_p2mp_call_active(Tcl_Interp *interp, Tcl_DString *str,
    const u_char *ptr, size_t ilen __unused)
{
	const struct atm_p2mp_call_active *arg =
	    (const struct atm_p2mp_call_active *)(const void *)ptr;

	return (fmt_connid(interp, str, &arg->connid));
}

static int
parse_atm_p2mp_call_active(Tcl_Interp *interp, int argc, const char **argv,
    u_char *buf, size_t *len, struct unicx *cx __unused)
{
	struct atm_p2mp_call_active *arg =
	    (struct atm_p2mp_call_active *)(void *)buf;
	int error;
	union uni_ieall ie;
	u_int ietype;

	if (argc != 1)
		return (unitcl_setres(interp,
		    "%s: wrong number of args", __func__));

	if ((error = parse_ie(interp, argv[0], &ie, &ietype)) != TCL_OK)
		return (error);

	if (ietype != UNI_IE_CONNID)
		return (unitcl_setres(interp,
		    "%s: expecting {connid ...}", __func__));

	arg->connid = ie.connid;
	*len += sizeof(*arg);

	return (TCL_OK);
}

/*
 * Format an atm-p2p-call-active
 *
 * {atm-p2p-call-active <connid>}
 */
static int
fmt_atm_p2p_call_active(Tcl_Interp *interp, Tcl_DString *str,
    const u_char *ptr, size_t ilen __unused)
{
	const struct atm_p2p_call_active *arg =
	    (const struct atm_p2p_call_active *)(const void *)ptr;

	return (fmt_connid(interp, str, &arg->connid));
}

static int
parse_atm_p2p_call_active(Tcl_Interp *interp, int argc, const char **argv,
    u_char *buf, size_t *len, struct unicx *cx __unused)
{
	struct atm_p2p_call_active *arg =
	    (struct atm_p2p_call_active *)(void *)buf;
	int error;
	union uni_ieall ie;
	u_int ietype;

	if (argc != 1)
		return (unitcl_setres(interp,
		    "%s: wrong number of args", __func__));

	if ((error = parse_ie(interp, argv[0], &ie, &ietype)) != TCL_OK)
		return (error);

	if (ietype != UNI_IE_CONNID)
		return (unitcl_setres(interp,
		    "%s: expecting {connid ...}", __func__));

	arg->connid = ie.connid;
	*len += sizeof(*arg);

	return (TCL_OK);
}

/*
 * Format an atm-prepare-incoming-call
 *
 * {atm-prepare-incoming-call <queue-len> <sap>}
 */
static int
fmt_atm_prepare_incoming_call(Tcl_Interp *interp, Tcl_DString *str,
    const u_char *ptr, size_t ilen __unused)
{
	const struct atm_prepare_incoming_call *arg =
	    (const struct atm_prepare_incoming_call *)(const void *)ptr;
	char buf[100];

	sprintf(buf, "%u", arg->queue_size);
	Tcl_DStringAppendElement(str, buf);

	return (fmt_sap(interp, str, &arg->sap));
}

static int
parse_atm_prepare_incoming_call(Tcl_Interp *interp, int argc, const char **argv,
    u_char *buf, size_t *len, struct unicx *cx __unused)
{
	struct atm_prepare_incoming_call *arg =
	    (struct atm_prepare_incoming_call *)(void *)buf;
	u_int n;

	if (argc < 2)
		return (unitcl_setres(interp,
		    "%s: wrong number of args", __func__));

	if (unitcl_parse_num(interp, argv[0], &n) != TCL_OK)
		return (TCL_ERROR);
	arg->queue_size = n;

	if (parse_sap(interp, argc - 1, argv + 1, &arg->sap) != TCL_OK)
		return (TCL_ERROR);

	*len += sizeof(*arg);

	return (TCL_OK);
}

/*
 * Prepare outgoing call
 *
 *	{atm-prepare-outgoing-call}
 */
static int
fmt_atm_prepare_outgoing_call(Tcl_Interp *interp __unused,
    Tcl_DString *str __unused, const u_char *ptr __unused, size_t ilen __unused)
{
	return (TCL_OK);
}

static int
parse_atm_prepare_outgoing_call(Tcl_Interp *interp, int argc,
    const char **argv __unused,
    u_char *buf __unused, size_t *len __unused, struct unicx *cx __unused)
{

	if (argc != 0)
		return (unitcl_setres(interp,
		    "%s: wrong number of args", __func__));

	return (TCL_OK);
}


/*
 * Format an atm-query-connection-attributes
 *
 * {atm-query-connection-attributes <attr>}
 */
static int
fmt_atm_query_connection_attributes(Tcl_Interp *interp, Tcl_DString *str,
    const u_char *ptr, size_t ilen __unused)
{
	const struct atm_query_connection_attributes *arg =
	    (const struct atm_query_connection_attributes *)(const void *)ptr;

	return (fmt_attr_name(interp, str, arg->attr));
}
static int
parse_atm_query_connection_attributes(Tcl_Interp *interp, int argc,
    const char **argv, u_char *buf, size_t *len, struct unicx *cx __unused)
{
	struct atm_query_connection_attributes *arg =
	    (struct atm_query_connection_attributes *)(void *)buf;
	int e;

	if (argc != 1)
		return (unitcl_setres(interp,
		    "%s: wrong number of args", __func__));

	if ((e = parse_attr_name(interp, argv[0], &arg->attr)) != TCL_OK)
		return (e);

	*len += sizeof(*arg);
	return (TCL_OK);
}

/*
 * Format an atm-reject-incoming-call
 *
 * {atm-reject-incoming-call <cause>}
 */
static int
fmt_atm_reject_incoming_call(Tcl_Interp *interp, Tcl_DString *str,
    const u_char *ptr, size_t ilen __unused)
{
	const struct atm_reject_incoming_call *arg =
	    (const struct atm_reject_incoming_call *)(const void *)ptr;

	return (fmt_cause(interp, str, &arg->cause));
}

static int
parse_atm_reject_incoming_call(Tcl_Interp *interp, int argc, const char **argv,
    u_char *buf, size_t *len, struct unicx *cx __unused)
{
	struct atm_reject_incoming_call *arg =
	    (struct atm_reject_incoming_call *)(void *)buf;
	int error;
	union uni_ieall ie;
	u_int ietype;

	if (argc != 1)
		return (unitcl_setres(interp,
		    "%s: wrong number of args", __func__));

	if ((error = parse_ie(interp, argv[0], &ie, &ietype)) != TCL_OK)
		return (error);

	if (ietype != UNI_IE_CAUSE)
		return (unitcl_setres(interp,
		    "%s: expecting {cause ...}", __func__));

	arg->cause = ie.cause;
	*len += sizeof(*arg);

	return (TCL_OK);
}

/*
 * Format an atm-set-connection-attributes
 *
 * {atm-set-connection-attributes <attr>}
 */
static int
fmt_atm_set_connection_attributes(Tcl_Interp *interp, Tcl_DString *str,
    const u_char *ptr, size_t ilen)
{
	const struct atm_set_connection_attributes *arg =
	    (const struct atm_set_connection_attributes *)(const void *)ptr;
	const u_char *val = ptr + sizeof(*arg);
	size_t len = ilen - sizeof(*arg);
	int e;

	if ((e = fmt_attr(interp, str, arg->attr, &val, &len)) != TCL_OK)
		return (e);
	if (len != 0)
		return (unitcl_setres(interp,
		    "%s: attribute too long", __func__));
	return (TCL_OK);
}

static int
parse_atm_set_connection_attributes(Tcl_Interp *interp, int argc,
    const char **argv, u_char *buf, size_t *len, struct unicx *cx)
{
	struct atm_set_connection_attributes *arg =
	    (struct atm_set_connection_attributes *)(void *)buf;
	int e;

	if (argc != 1)
		return (unitcl_setres(interp,
		    "%s: wrong number of args", __func__));

	buf += sizeof(*arg);
	e = parse_attr(interp, argv[0], &arg->attr, &buf, len, cx);
	if (e != TCL_OK)
		return (e);

	*len += sizeof(*arg);
	return (TCL_OK);
}

/*
 * Wait on incoming call
 *
 *	{atm-wait-on-incoming-call}
 */
static int
fmt_atm_wait_on_incoming_call(Tcl_Interp *interp __unused,
    Tcl_DString *str __unused, const u_char *ptr __unused, size_t ilen __unused)
{
	return (TCL_OK);
}

static int
parse_atm_wait_on_incoming_call(Tcl_Interp *interp, int argc,
    const char **argv __unused,
    u_char *buf __unused, size_t *len __unused, struct unicx *cx __unused)
{

	if (argc != 0)
		return (unitcl_setres(interp,
		    "%s: wrong number of args", __func__));

	return (TCL_OK);
}

/*
 * Format an atm-set-connection-attributes-x
 *
 * {atm-set-connection-attributes-x <attr>*}
 */
static int
fmt_atm_set_connection_attributes_x(Tcl_Interp *interp, Tcl_DString *str,
    const u_char *ptr, size_t ilen)
{
	const struct atm_set_connection_attributes_x *arg =
	    (const struct atm_set_connection_attributes_x *)(const void *)ptr;

	const u_char *val = ptr + sizeof(*arg);
	size_t len = ilen - sizeof(*arg);
	u_int i;
	int e;

	if (len < arg->count * sizeof(arg->attr[0]))
		return (unitcl_setres(interp,
		    "%s: arg too short for count=%u: %zu", __func__,
		    arg->count, ilen));

	for (i = 0; i < arg->count; i++) {
		e = fmt_attr(interp, str, arg->attr[i], &val, &len);
		if (e != TCL_OK)
			return (e);
	}
	if (len != 0)
		return (unitcl_setres(interp,
		    "%s: attributes too long", __func__));
	return (TCL_OK);
}

static int
parse_atm_set_connection_attributes_x(Tcl_Interp *interp, int argc,
    const char **argv, u_char *buf, size_t *len, struct unicx *cx)
{
	struct atm_set_connection_attributes_x *arg =
	    (struct atm_set_connection_attributes_x *)(void *)buf;
	int e, a;
	u_char *ptr;

	ptr = buf + sizeof(*arg) + argc * sizeof(arg->attr[0]);
	for (a = 0; a < argc; a++) {
		e = parse_attr(interp, argv[a], &arg->attr[a], &ptr, len, cx);
		if (e != TCL_OK)
			return (e);
	}

	*len += sizeof(*arg) + argc * sizeof(arg->attr[0]);
	arg->count = argc;
	return (TCL_OK);
}

/*
 * Format an atm-query-connection-attributes-x
 *
 * {atm-query-connection-attributes-x <attr>*}
 */
static int
fmt_atm_query_connection_attributes_x(Tcl_Interp *interp, Tcl_DString *str,
    const u_char *ptr, size_t ilen)
{
	const struct atm_query_connection_attributes_x *arg =
	    (const struct atm_query_connection_attributes_x *)(const void *)ptr;
	u_int i;
	int e;

	if (ilen != arg->count * sizeof(uint32_t) + sizeof(*arg))
		return (unitcl_setres(interp,
		    "%s: bad length (count=%u, ilen=%zu)", __func__,
		    arg->count, ilen));

	for (i = 0; i < arg->count; i++)
		if ((e = fmt_attr_name(interp, str, arg->attr[i])) != TCL_OK)
			return (e);

	return (TCL_OK);
}
static int
parse_atm_query_connection_attributes_x(Tcl_Interp *interp, int argc,
    const char **argv, u_char *buf, size_t *len, struct unicx *cx __unused)
{
	struct atm_query_connection_attributes_x *arg =
	    (struct atm_query_connection_attributes_x *)(void *)buf;
	int a, e;

	for (a = 0; a < argc; a++) {
		e = parse_attr_name(interp, argv[a], &arg->attr[a]);
		if (e != TCL_OK)
			return (e);
	}

	arg->count = argc;
	*len += sizeof(*arg) + argc * sizeof(arg->attr[0]);

	return (TCL_OK);
}

/*
 * Query state request
 *
 *	{atm-query-state}
 */
static int
fmt_atm_query_state(Tcl_Interp *interp __unused, Tcl_DString *str __unused,
    const u_char *ptr __unused, size_t ilen __unused)
{
	return (TCL_OK);
}

static int
parse_atm_query_state(Tcl_Interp *interp, int argc, const char **argv __unused,
    u_char *buf __unused, size_t *len __unused, struct unicx *cx __unused)
{

	if (argc != 0)
		return (unitcl_setres(interp,
		    "%s: wrong number of args", __func__));

	return (TCL_OK);
}

/*
 * List of all known signals
 */
static const struct {
	const char	*name;
	int		exact;
	size_t		minlen;
	int		(*fmt)(Tcl_Interp *, Tcl_DString *,
			    const u_char *ptr, size_t);
	int		(*parse)(Tcl_Interp *, int, const char **,
			    u_char *, size_t *, struct unicx *);
} signals[] = {
	[ATMOP_RESP] = {
		"atm-resp",
		0, sizeof(struct atm_resp),
		fmt_atm_resp,
		parse_atm_resp,
	},
	[ATMOP_ABORT_CONNECTION] = {
		"atm-abort-connection",
		1, sizeof(struct atm_abort_connection),
		fmt_atm_abort_connection,
		parse_atm_abort_connection,
	},
	[ATMOP_ACCEPT_INCOMING_CALL] = {
		"atm-accept-incoming-call",
		1, sizeof(struct atm_accept_incoming_call),
		fmt_atm_accept_incoming_call,
		parse_atm_accept_incoming_call,
	},
	[ATMOP_ADD_PARTY] = {
		"atm-add-party",
		1, sizeof(struct atm_add_party),
		fmt_atm_add_party,
		parse_atm_add_party,
	},
	[ATMOP_ADD_PARTY_REJECT] = {
		"atm-add-party-reject",
		1, sizeof(struct atm_add_party_reject),
		fmt_atm_add_party_reject,
		parse_atm_add_party_reject,
	},
	[ATMOP_ADD_PARTY_SUCCESS] = {		/* 5 */
		"atm-add-party-success",
		1, sizeof(struct atm_add_party_success),
		fmt_atm_add_party_success,
		parse_atm_add_party_success,
	},
	[ATMOP_ARRIVAL_OF_INCOMING_CALL] = {
		"atm-arrival-of-incoming-call",
		1, 0,
		fmt_atm_arrival_of_incoming_call,
		parse_atm_arrival_of_incoming_call,
	},
	[ATMOP_CALL_RELEASE] = {
		"atm-call-release",
		1, sizeof(struct atm_call_release),
		fmt_atm_call_release,
		parse_atm_call_release,
	},
	[ATMOP_CONNECT_OUTGOING_CALL] = {
		"atm-connect-outgoing-call",
		1, sizeof(struct atm_connect_outgoing_call),
		fmt_atm_connect_outgoing_call,
		parse_atm_connect_outgoing_call,
	},
	[ATMOP_DROP_PARTY] = {
		"atm-drop-party",
		1, sizeof(struct atm_drop_party),
		fmt_atm_drop_party,
		parse_atm_drop_party,
	},
	[ATMOP_GET_LOCAL_PORT_INFO] = {
		"atm-get-local-port-info",
		1, sizeof(struct atm_get_local_port_info),
		fmt_atm_get_local_port_info,
		parse_atm_get_local_port_info,
	},
	[ATMOP_P2MP_CALL_ACTIVE] = {
		"atm-p2mp-call-active",
		1, sizeof(struct atm_p2mp_call_active),
		fmt_atm_p2mp_call_active,
		parse_atm_p2mp_call_active,
	},
	[ATMOP_P2P_CALL_ACTIVE] = {
		"atm-p2p-call-active",
		1, sizeof(struct atm_p2p_call_active),
		fmt_atm_p2p_call_active,
		parse_atm_p2p_call_active,
	},
	[ATMOP_PREPARE_INCOMING_CALL] = {
		"atm-prepare-incoming-call",
		1, sizeof(struct atm_prepare_incoming_call),
		fmt_atm_prepare_incoming_call,
		parse_atm_prepare_incoming_call,
	},
	[ATMOP_PREPARE_OUTGOING_CALL] = {
		"atm-prepare-outgoing-call",
		1, 0,
		fmt_atm_prepare_outgoing_call,
		parse_atm_prepare_outgoing_call,
	},
	[ATMOP_QUERY_CONNECTION_ATTRIBUTES] = {
		"atm-query-connection-attributes",
		1, sizeof(struct atm_query_connection_attributes),
		fmt_atm_query_connection_attributes,
		parse_atm_query_connection_attributes,
	},
	[ATMOP_REJECT_INCOMING_CALL] = {
		"atm-reject-incoming-call",
		1, sizeof(struct atm_reject_incoming_call),
		fmt_atm_reject_incoming_call,
		parse_atm_reject_incoming_call,
	},
	[ATMOP_SET_CONNECTION_ATTRIBUTES] = {
		"atm-set-connection-attributes",
		0, sizeof(struct atm_set_connection_attributes),
		fmt_atm_set_connection_attributes,
		parse_atm_set_connection_attributes,
	},
	[ATMOP_WAIT_ON_INCOMING_CALL] = {
		"atm-wait-on-incoming-call",
		1, 0,
		fmt_atm_wait_on_incoming_call,
		parse_atm_wait_on_incoming_call,
	},
	[ATMOP_SET_CONNECTION_ATTRIBUTES_X] = {
		"atm-set-connection-attributes-x",
		0, sizeof(struct atm_set_connection_attributes_x),
		fmt_atm_set_connection_attributes_x,
		parse_atm_set_connection_attributes_x,
	},
	[ATMOP_QUERY_CONNECTION_ATTRIBUTES_X] = {
		"atm-query-connection-attributes-x",
		0, sizeof(struct atm_query_connection_attributes_x),
		fmt_atm_query_connection_attributes_x,
		parse_atm_query_connection_attributes_x,
	},
	[ATMOP_QUERY_STATE] = {
		"atm-query-state",
		1, 0,
		fmt_atm_query_state,
		parse_atm_query_state,
	},
};

/*
 * Format an API signal
 */
int
fmt_atmapi(Tcl_Interp *interp, Tcl_DString *str, u_int sig, const u_char *ptr,
    size_t len)
{
	if (sig >= sizeof(signals) / sizeof(signals[0]) ||
	    signals[sig].name == NULL)
		return (unitcl_setres(interp, "bad ATMAPI signal %u", sig));

	if (signals[sig].exact) {
		if (len != signals[sig].minlen)
			return (unitcl_setres(interp,
			    "bad length for %s (%zu, need %zu)",
			    signals[sig].name, len, signals[sig].minlen));
	} else {
		if (len < signals[sig].minlen)
			return (unitcl_setres(interp,
			    "small length for %s (%zu, need %zu)",
			    signals[sig].name, len, signals[sig].minlen));
	}
	return (signals[sig].fmt(interp, str, ptr, len));
}

/*
 * Get the name of an API signal
 */
int
fmt_atmapi_sig(Tcl_Interp *interp, Tcl_DString *str, u_int sig)
{

	if (sig >= sizeof(signals) / sizeof(signals[0]) ||
	    signals[sig].name == NULL)
		return (unitcl_setres(interp, "bad ATMAPI signal %u", sig));
	Tcl_DStringAppendElement(str, signals[sig].name);
	return (TCL_OK);
}

/*
 * Parse API signal name
 */
int
parse_atmapi_sig(Tcl_Interp *interp, const char *arg, u_int *sig)
{

	for (*sig = 0; *sig < sizeof(signals) / sizeof(signals[0]); (*sig)++)
		if (signals[*sig].name != NULL &&
		    strcmp(signals[*sig].name, arg) == 0)
			return (TCL_OK);
	return (unitcl_setres(interp, "bad ATMAPI signal name '%s'", arg));
}

/*
 * Parse ATMAPI signal
 */
int
parse_atmapi(Tcl_Interp *interp, int argc, const char **argv, u_int sig,
    u_char *u, size_t *lenp, struct unicx *cx)
{

	if (sig >= sizeof(signals) / sizeof(signals[0]) ||
	    signals[sig].name == NULL)
		return (unitcl_setres(interp, "bad ATMAPI signal %u", sig));
	*lenp = 0;
	return (signals[sig].parse(interp, argc, argv, u, lenp, cx));
}

/*
 * Saps
 */
int
parse_sap(Tcl_Interp *interp, int argc, const char **argv, struct uni_sap *sap)
{

	if (argc != 5)
		return (unitcl_setres(interp, "bad number of args for sap"));

	if (parse_addr_sve(interp, argv[0], &sap->addr) != TCL_OK ||
	    parse_selector_sve(interp, argv[1], &sap->selector) != TCL_OK ||
	    parse_blli_id2_sve(interp, argv[2], &sap->blli_id2) != TCL_OK ||
	    parse_blli_id3_sve(interp, argv[3], &sap->blli_id3) != TCL_OK ||
	    parse_bhli_sve(interp, argv[4], &sap->bhli) != TCL_OK)
		return (TCL_ERROR);

	return (TCL_OK);
}

int
fmt_sap(Tcl_Interp *interp, Tcl_DString *str, const struct uni_sap *sap)
{

	if (fmt_addr_sve(interp, str, &sap->addr) != TCL_OK ||
	    fmt_selector_sve(interp, str, &sap->selector) != TCL_OK ||
	    fmt_blli_id2_sve(interp, str, &sap->blli_id2) != TCL_OK ||
	    fmt_blli_id3_sve(interp, str, &sap->blli_id3) != TCL_OK ||
	    fmt_bhli_sve(interp, str, &sap->bhli) != TCL_OK)
		return (TCL_ERROR);

	return (TCL_OK);
}

static int
cmp_ep(const void *p1, const void *p2)
{
	const struct atm_exstatus_ep *e1 = p1;
	const struct atm_exstatus_ep *e2 = p2;

	return (strcmp(e1->name, e2->name));
}

static int
cmp_port(const void *p1, const void *p2)
{
	const struct atm_exstatus_port *e1 = p1;
	const struct atm_exstatus_port *e2 = p2;

	if (e1->portno < e2->portno)
		return (-1);
	if (e1->portno > e2->portno)
		return (+1);
	abort();
}

static int
cmp_conn(const void *p1, const void *p2)
{
	const struct atm_exstatus_conn *e1 = p1;
	const struct atm_exstatus_conn *e2 = p2;

	if (e1->cref < e2->cref)
		return (-1);
	if (e1->cref > e2->cref)
		return (+1);
	return (strcmp(e1->ep, e2->ep));
}
	
static int
cmp_party(const void *p1, const void *p2)
{
	const struct atm_exstatus_party *e1 = p1;
	const struct atm_exstatus_party *e2 = p2;

	if (e1->epref < e2->epref)
		return (-1);
	if (e1->epref > e2->epref)
		return (+1);
	return (0);
}

static void
exstat_fmt_conn(Tcl_DString *str,
    const struct atm_exstatus_conn *conn, u_int nparties,
    const struct atm_exstatus_party *party)
{
	char nbuf[100];
	u_int p;

	Tcl_DStringStartSublist(str);
	sprintf(nbuf, "%u", conn->cref & 0x7fffff);
	Tcl_DStringAppendElement(str, nbuf);
	sprintf(nbuf, "%u", (conn->cref >> 23) & 1);
	Tcl_DStringAppendElement(str, nbuf);
	Tcl_DStringAppendElement(str, conn->ep);
	sprintf(nbuf, "%u", conn->state);
	Tcl_DStringAppendElement(str, nbuf);

	for (p = 0; p < nparties; p++) {
		if (party[p].connid == conn->id) {
			Tcl_DStringStartSublist(str);
			sprintf(nbuf, "%u", party[p].epref);
			Tcl_DStringAppendElement(str, nbuf);
			sprintf(nbuf, "%u", party[p].state);
			Tcl_DStringAppendElement(str, nbuf);
			Tcl_DStringEndSublist(str);
		}
	}

	Tcl_DStringEndSublist(str);
}

/*
 * exstatus:	{<conn>*} {<ep>*} <port>*
 * conn:	{<cref> <flag> <epname> <state> <party>*}
 * ep:		{<epname> <state>}
 * port: 	{<portno> <status> <conn>*}
 * party:	{<epref> <state>}
 */
int
atmapi_fmt_exstatus(Tcl_Interp *interp __unused, Tcl_DString *str,
    const struct atm_exstatus *status, struct atm_exstatus_ep *ep,
    struct atm_exstatus_port *port, struct atm_exstatus_conn *conn,
    struct atm_exstatus_party *party)
{
	u_int i, p;
	char nbuf[100];

	qsort(ep, status->neps, sizeof(*ep), cmp_ep);
	qsort(port, status->nports, sizeof(*port), cmp_port);
	qsort(conn, status->nconns, sizeof(*conn), cmp_conn);
	qsort(party, status->nparties, sizeof(*party), cmp_party);

	/* orphaned connections */
	Tcl_DStringStartSublist(str);
	for (i = 0; i < status->nconns; i++)
		if (conn[i].port == 0)
			exstat_fmt_conn(str, &conn[i], status->nparties, party);
	Tcl_DStringEndSublist(str);

	/* endpoints */
	Tcl_DStringStartSublist(str);
	for (i = 0; i < status->neps; i++) {
		Tcl_DStringStartSublist(str);
		Tcl_DStringAppendElement(str, ep[i].name);
		sprintf(nbuf, "%u", ep[i].state);
		Tcl_DStringAppendElement(str, nbuf);
		Tcl_DStringEndSublist(str);
	}
	Tcl_DStringEndSublist(str);

	/* ports */
	for (p = 0; p < status->nports; p++) {
		Tcl_DStringStartSublist(str);

		sprintf(nbuf, "%u", port[p].portno);
		Tcl_DStringAppendElement(str, nbuf);
		sprintf(nbuf, "%u", port[p].state);
		Tcl_DStringAppendElement(str, nbuf);

		for (i = 0; i < status->nconns; i++) {
			if (conn[i].port == port[p].portno)
				exstat_fmt_conn(str, &conn[i],
				    status->nparties, party);
		}

		Tcl_DStringEndSublist(str);
	}

	return (TCL_OK);
}
