/*
 * Copyright (c) 2001-2003
 *	Fraunhofer Institute for Open Communication Systems (FhG Fokus).
 * 	All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Author: Hartmut Brandt <harti@freebsd.org>
 *
 * $Begemot: atmsupport/libunitcl/ie_abr.c,v 1.9 2004/05/12 11:50:04 novo Exp $
 */
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>

#include "inctcl.h"

#include <netnatm/unimsg.h>
#include <netnatm/msg/unistruct.h>

#include "unitcl.h"
#include "unitcl_priv.h"

REGISTER_IE(UNI_IE_ABRADD, abradd, 0)
REGISTER_IE(UNI_IE_ABRSETUP, abrsetup, 0)


/*
 * {abradd {{<type> <val>}*} {{<type> <val>}*}}
 * type := nrm | trm | cdf | adtf
 */
static int
parse_abr_rec(Tcl_Interp *interp, const char *arg, struct uni_abr_rec *rec)
{
	int ac;
	const char **av, **aptr;

	rec->present = 0;
	if (Tcl_SplitList(interp, arg, &ac, &av) != TCL_OK)
		return (TCL_ERROR);

	aptr = av;
	while (ac--) {
		int xc;
		const char **xv;
		u_int num;

		if (Tcl_SplitList(interp, *aptr, &xc, &xv) != TCL_OK) {
			Tcl_Free((char *)av);
			return (TCL_ERROR);
		}
		if (xc != 2) {
			Tcl_Free((char *)av);
			Tcl_Free((char *)xv);
			return (unitcl_setres(interp,
			    "bad # of args in abradd subfield"));
		}
		if (unitcl_parse_num(interp, xv[1], &num) != TCL_OK) {
			Tcl_Free((char *)av);
			Tcl_Free((char *)xv);
			return (TCL_ERROR);
		}

		if (strcmp(xv[0], "nrm") == 0) {
			rec->nrm = num;
			rec->present |= UNI_ABR_REC_NRM_P;
		} else if (strcmp(xv[0], "trm") == 0) {
			rec->trm = num;
			rec->present |= UNI_ABR_REC_TRM_P;
		} else if (strcmp(xv[0], "cdf") == 0) {
			rec->cdf = num;
			rec->present |= UNI_ABR_REC_CDF_P;
		} else if (strcmp(xv[0], "adtf") == 0) {
			rec->adtf = num;
			rec->present |= UNI_ABR_REC_ADTF_P;
		} else {
			Tcl_Free((char *)av);
			Tcl_Free((char *)xv);
			return (unitcl_setres(interp,
			    "bad subfield in abradd"));
		}
		Tcl_Free((char *)xv);
		aptr++;
	}

	Tcl_Free((char *)av);
	return (TCL_OK);
}

static int
fmt_abr_rec(Tcl_Interp *interp, Tcl_DString *str, const struct uni_abr_rec *rec)
{
	char buf[100];

	interp = interp;
#define DOFMT(BIG,SMALL)						\
	if(rec->present & UNI_ABR_REC_##BIG##_P) {			\
		Tcl_DStringStartSublist(str);				\
		Tcl_DStringAppendElement(str, #SMALL);			\
		sprintf(buf, "%d", rec->SMALL);				\
		Tcl_DStringAppendElement(str, buf);			\
		Tcl_DStringEndSublist(str);				\
	}

	DOFMT(NRM, nrm);
	DOFMT(TRM, trm);
	DOFMT(CDF, cdf);
	DOFMT(ADTF, adtf);
#undef DOFMT

	return TCL_OK;
}

int
parse_abradd(Tcl_Interp *interp, int argc, const char **argv,
    union uni_ieall *ie)
{
	if(argc != 2) {
		unitcl_setres(interp, "bad # of args for abradd");
		return TCL_ERROR;
	}

	if(parse_abr_rec(interp, argv[0], &ie->abradd.fwd) != TCL_OK)
		return TCL_ERROR;
	if(parse_abr_rec(interp, argv[1], &ie->abradd.bwd) != TCL_OK)
		return TCL_ERROR;

	IE_ADDPRESENT(ie->abradd);

	return TCL_OK;
}

FMT_IE_START(UNI_IE_ABRADD, abradd)
{
	Tcl_DStringStartSublist(str);
	ret = fmt_abr_rec(interp, str, &ie->fwd);
	Tcl_DStringEndSublist(str);
	if(ret == TCL_OK) {
		Tcl_DStringStartSublist(str);
		ret = fmt_abr_rec(interp, str, &ie->bwd);
		Tcl_DStringEndSublist(str);
	}
}
FMT_IE_END()

/********************************************************************/

static const struct toff {
	const char	*name;
	size_t		off;
	u_int		flag;
} offtab[] = {
  {"ficr", offsetof(struct uni_ie_abrsetup, ficr), UNI_ABRSETUP_FICR_P },
  {"bicr", offsetof(struct uni_ie_abrsetup, bicr), UNI_ABRSETUP_BICR_P },
  {"ftbe", offsetof(struct uni_ie_abrsetup, ftbe), UNI_ABRSETUP_FTBE_P },
  {"btbe", offsetof(struct uni_ie_abrsetup, btbe), UNI_ABRSETUP_BTBE_P },
  {"rmfrt",offsetof(struct uni_ie_abrsetup, rmfrt),UNI_ABRSETUP_RMFRT_P },
  {"frif", offsetof(struct uni_ie_abrsetup, frif), UNI_ABRSETUP_FRIF_P },
  {"brif", offsetof(struct uni_ie_abrsetup, brif), UNI_ABRSETUP_BRIF_P },
  {"frdf", offsetof(struct uni_ie_abrsetup, frdf), UNI_ABRSETUP_FRDF_P },
  {"brdf", offsetof(struct uni_ie_abrsetup, brdf), UNI_ABRSETUP_BRDF_P },
  { NULL, 0, 0 }
};

/*
 * {abrsetup {<param> <val>}*}
 * param := ficr | bicr | ftbe | btbe | rmfrt | frif | brif | frdf | brdf
 */
int
parse_abrsetup(Tcl_Interp *interp, int argc, const char **argv,
    union uni_ieall *ie)
{
	int ac;
	const char **av;
	u_int i;
	u_int val;

	if(argc == 0)
		return TCL_OK;

	while(argc--) {
		if(Tcl_SplitList(interp, *argv, &ac, &av) != TCL_OK)
			return TCL_ERROR;

		if(ac != 2) {
			unitcl_setres(interp, "bad # of fields in abrsetup subfield");
			Tcl_Free((char *)av);
			return TCL_ERROR;
		}

		if(unitcl_parse_num(interp, av[1], &val) != TCL_OK) {
			Tcl_Free((char *)av);
			return TCL_ERROR;
		}
		for(i = 0; offtab[i].name; i++)
			if(strcmp(offtab[i].name, av[0]) == 0) {
				*(u_int *)(void *)((u_char *)&ie->abrsetup +
				    offtab[i].off) = val;
				ie->abrsetup.h.present |= offtab[i].flag;
				break;
			}
		if(!offtab[i].name) {
			unitcl_setres(interp, "bad abrsetup subfield");
			Tcl_Free((char *)av);
			return TCL_ERROR;
		}
		Tcl_Free((char *)av);
		argv++;
	}

	IE_ADDPRESENT(ie->abrsetup);

	return TCL_OK;
}

FMT_IE_START(UNI_IE_ABRSETUP, abrsetup)
{
	char buf[100];
#define DOFMT(BIG,SMALL)						\
	if(ie->h.present & UNI_ABRSETUP_##BIG##_P) {			\
		Tcl_DStringStartSublist(str);				\
		Tcl_DStringAppendElement(str, #SMALL);			\
		sprintf(buf, "%d", ie->SMALL);				\
		Tcl_DStringAppendElement(str, buf);			\
		Tcl_DStringEndSublist(str);				\
	}
	DOFMT(FICR, ficr);
	DOFMT(BICR, bicr);
	DOFMT(FTBE, ftbe);
	DOFMT(BTBE, btbe);
	DOFMT(RMFRT, rmfrt);
	DOFMT(FRIF, frif);
	DOFMT(BRIF, brif);
	DOFMT(FRDF, frdf);
	DOFMT(BRDF, brdf);
#undef DOFMT
}
FMT_IE_END()
