/*
 * Copyright (c) 2004
 *	Hartmut Brandt
 * Copyright (c) 1996-2003
 *	Fraunhofer Institute for Open Communication Systems (FhG Fokus).
 * 	All rights reserved.
 *
 * Author: Hartmut Brandt <harti@freebsd.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Begemot: atmsupport/libunitcl/ie_addr.c,v 1.10 2004/07/08 13:33:42 brandt Exp $
 */
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>

#include "inctcl.h"

#include <netnatm/unimsg.h>
#include <netnatm/addr.h>
#include <netnatm/msg/unistruct.h>
#include <netnatm/api/unisap.h>

#include "unitcl.h"
#include "unitcl_priv.h"

REGISTER_IE(UNI_IE_CALLED, called, 0)
REGISTER_IE(UNI_IE_CALLEDSUB, calledsub, 0)
REGISTER_IE(UNI_IE_CALLING, calling, 0)
REGISTER_IE(UNI_IE_CALLINGSUB, callingsub, 0)
REGISTER_IE(UNI_IE_CONNED, conned, 0)
REGISTER_IE(UNI_IE_CONNEDSUB, connedsub, 0)

/*
 * Parse an ATM address
 *
 *	atme <nsap>
 *    | e164 <e164>
 */
int
parse_addr(Tcl_Interp *interp, const char **argv, struct uni_addr *addr)
{

	if (strcmp(argv[0], "atme") == 0) {
		addr->plan = UNI_ADDR_ATME;
		addr->type = UNI_ADDR_UNKNOWN;
		if (uni_str2nsap(addr->addr, argv[1]))
			return (unitcl_setres(interp, "bad ATME address"));
		addr->len = 20;

	} else if (strcmp(argv[0], "e164") == 0) {
		addr->plan = UNI_ADDR_E164;
		addr->type = UNI_ADDR_INTERNATIONAL;
		if (strspn(argv[1], "0123456789") != strlen(argv[1]))
			return (unitcl_setres(interp,
			    "bad E.164 address"));
		if (strlen(argv[1]) > 16)
			return (unitcl_setres(interp,
			    "E.164 address too long"));
		(void)strcpy((char *)addr->addr, argv[1]);
		addr->len = strlen(argv[1]);

	} else
		return (unitcl_setres(interp, "bad number type"));

	return (TCL_OK);
}

int
fmt_addr(Tcl_Interp *interp, Tcl_DString *str, const struct uni_addr *addr)
{
	char buf[100];

	if (addr->plan == UNI_ADDR_E164) {
		(void)strncpy(buf, (const char *)addr->addr, addr->len);
		buf[addr->len] = 0;
		Tcl_DStringAppendElement(str, "e164");
		Tcl_DStringAppendElement(str, buf);

	} else if (addr->plan == UNI_ADDR_ATME) {
		uni_nsap2str(buf, addr->addr, 1);
		Tcl_DStringAppendElement(str, "atme");
		Tcl_DStringAppendElement(str, buf);
	} else
		return (unitcl_setres(interp, "bad address type to format"));

	return (TCL_OK);
}

/*
 * Parse a subaddress
 *
 *	nsap <nsap>
 *    | atme <nsap>
 */
int
parse_subaddr(Tcl_Interp *interp, const char **argv, struct uni_subaddr *addr)
{

	if (strcmp(argv[0], "nsap") == 0)
		addr->type = UNI_SUBADDR_NSAP;
	else if (strcmp(argv[0], "atme") == 0)
		addr->type = UNI_SUBADDR_ATME;
	else
		return (unitcl_setres(interp, "bad subaddress type"));
	
	if (uni_str2nsap(addr->addr, argv[1]))
		return (unitcl_setres(interp, "bad subaddress"));
	addr->len = 20;

	return (TCL_OK);
}

int
fmt_subaddr(Tcl_Interp *interp, Tcl_DString *str,
    const struct uni_subaddr *addr)
{
	char buf[100];

	if (addr->type == UNI_SUBADDR_NSAP)
		Tcl_DStringAppendElement(str, "nsap");
	else if (addr->type == UNI_SUBADDR_ATME)
		Tcl_DStringAppendElement(str, "atme");
	else
		return (unitcl_setres(interp, "bad subaddress type to format"));

	uni_nsap2str(buf, addr->addr, 1);
	Tcl_DStringAppendElement(str, buf);

	return (TCL_OK);
}

/*
 * Parse the presentation and screening indicator
 */
static int
parse_pres(Tcl_Interp *interp, const char **argv, enum uni_addr_pres *pres,
    enum uni_addr_screen *screen)
{

	if (strcmp(argv[0], "pres") == 0)
		*pres = UNI_ADDR_PRES;
	else if (strcmp(argv[0], "restrict") == 0)
		*pres = UNI_ADDR_RESTRICT;
	else if (strcmp(argv[0], "nonumber") == 0)
		*pres = UNI_ADDR_NONUMBER;
	else
		return (unitcl_setres(interp, "bad presentation indicator"));

	if (strcmp(argv[1], "not") == 0)
		*screen = UNI_ADDR_SCREEN_NOT;
	else if (strcmp(argv[1], "passed") == 0)
		*screen = UNI_ADDR_SCREEN_PASSED;
	else if (strcmp(argv[1], "failed") == 0)
		*screen = UNI_ADDR_SCREEN_FAILED;
	else if (strcmp(argv[1], "net") == 0)
		*screen = UNI_ADDR_SCREEN_NET;
	else
		return (unitcl_setres(interp, "bad screening indicator"));

	return (TCL_OK);
}

static int
fmt_pres(Tcl_Interp *interp, Tcl_DString *str, enum uni_addr_pres pres,
    enum uni_addr_screen screen)
{

	if (pres == UNI_ADDR_PRES)
		Tcl_DStringAppendElement(str, "pres");
	else if (pres == UNI_ADDR_RESTRICT)
		Tcl_DStringAppendElement(str, "restrict");
	else if (pres == UNI_ADDR_NONUMBER)
		Tcl_DStringAppendElement(str, "nonumber");
	else
		return (unitcl_setres(interp,
		    "bad representation indicator to fmt"));

	if (screen == UNI_ADDR_SCREEN_NOT)
		Tcl_DStringAppendElement(str, "not");
	else if (screen == UNI_ADDR_SCREEN_PASSED)
		Tcl_DStringAppendElement(str, "passed");
	else if (screen == UNI_ADDR_SCREEN_FAILED)
		Tcl_DStringAppendElement(str, "failed");
	else if (screen == UNI_ADDR_SCREEN_NET)
		Tcl_DStringAppendElement(str, "net");
	else
		return (unitcl_setres(interp,
		    "bad screening indicator to fmt"));

	return (TCL_OK);
}
/*************************************************************************/
/*
 * {called atme/e164 addr}
 */
int
parse_called(Tcl_Interp *interp, int argc, const char **argv,
    union uni_ieall *ie)
{

	if (argc != 2)
		return (unitcl_setres(interp, "bad # of args for called"));

	if (parse_addr(interp, argv, &ie->called.addr) != TCL_OK)
		return (TCL_ERROR);
	IE_ADDPRESENT(ie->called);

	return (TCL_OK);
}

FMT_IE_START(UNI_IE_CALLED, called)
{
	ret = fmt_addr(interp, str, &ie->addr);
}
FMT_IE_END()

/*************************************************************************/
/*
 * {calledsub nsap/atme addr}
 */
int
parse_calledsub(Tcl_Interp *interp, int argc, const char **argv,
    union uni_ieall *ie)
{

	if (argc != 2)
		return (unitcl_setres(interp, "bad # of args for calledsub"));

	if (parse_subaddr(interp, argv, &ie->calledsub.addr) != TCL_OK)
		return (TCL_ERROR);
	IE_ADDPRESENT(ie->calledsub);

	return (TCL_OK);
}

FMT_IE_START(UNI_IE_CALLEDSUB, calledsub)
{
	ret = fmt_subaddr(interp, str, &ie->addr);
}
FMT_IE_END()

/*************************************************************************/
/*
 * {calling nsap/e164 addr [pres screen]}
 */
int
parse_calling(Tcl_Interp *interp, int argc, const char **argv,
    union uni_ieall *ie)
{

	if (argc != 2 && argc != 4)
		return (unitcl_setres(interp, "bad # of args for calling"));

	if (parse_addr(interp, argv, &ie->calling.addr) != TCL_OK)
		return (TCL_ERROR);

	if (argc == 4) {
		ie->calling.h.present |= UNI_CALLING_SCREEN_P;
		if (parse_pres(interp, argv + 2, &ie->calling.pres,
		    &ie->calling.screen) != TCL_OK)
			return (TCL_ERROR);
	}
	IE_ADDPRESENT(ie->calling);

	return (TCL_OK);
}

FMT_IE_START(UNI_IE_CALLING, calling)
{
	if ((ret = fmt_addr(interp, str, &ie->addr)) == TCL_OK)
		if (ie->h.present & UNI_CALLING_SCREEN_P)
			ret = fmt_pres(interp, str, ie->pres, ie->screen);
}
FMT_IE_END()

/*************************************************************************/
/*
 * {callingsub nsap/atme addr}
 */
int
parse_callingsub(Tcl_Interp *interp, int argc, const char **argv,
    union uni_ieall *ie)
{

	if (argc != 2)
		return (unitcl_setres(interp, "bad # of args for callingsub"));

	if (parse_subaddr(interp, argv, &ie->callingsub.addr) != TCL_OK)
		return (TCL_ERROR);
	IE_ADDPRESENT(ie->callingsub);

	return (TCL_OK);
}

FMT_IE_START(UNI_IE_CALLINGSUB, callingsub)
{
	ret = fmt_subaddr(interp, str, &ie->addr);
}
FMT_IE_END()

/*************************************************************************/
/*
 * {conned nsap/e164 addr [pres screen]}
 */
int
parse_conned(Tcl_Interp *interp, int argc, const char **argv,
    union uni_ieall *ie)
{

	if (argc != 2 && argc != 4)
		return (unitcl_setres(interp, "bad # of args for conned"));

	if (parse_addr(interp, argv, &ie->conned.addr) != TCL_OK)
		return (TCL_ERROR);

	if (argc == 4) {
		ie->conned.h.present |= UNI_CONNED_SCREEN_P;
		if (parse_pres(interp, argv + 2, &ie->conned.pres,
		    &ie->conned.screen) != TCL_OK)
			return (TCL_ERROR);
	}
	IE_ADDPRESENT(ie->conned);

	return (TCL_OK);
}

FMT_IE_START(UNI_IE_CONNED, conned)
{
	if ((ret = fmt_addr(interp, str, &ie->addr)) == TCL_OK)
		if (ie->h.present & UNI_CONNED_SCREEN_P)
			ret = fmt_pres(interp, str, ie->pres, ie->screen);
}
FMT_IE_END()

/*************************************************************************/
/*
 * {connedsub nsap/atme addr}
 */
int
parse_connedsub(Tcl_Interp *interp, int argc, const char **argv,
    union uni_ieall *ie)
{

	if (argc != 2)
		return (unitcl_setres(interp, "bad # of args for connedsub"));

	if (parse_subaddr(interp, argv, &ie->connedsub.addr) != TCL_OK)
		return (TCL_ERROR);
	IE_ADDPRESENT(ie->connedsub);

	return (TCL_OK);
}

FMT_IE_START(UNI_IE_CONNEDSUB, connedsub)
{
	ret = fmt_subaddr(interp, str, &ie->addr);
}
FMT_IE_END()


/*************************************************************************/
/*
 * {addr nsap/atme <addr>}
 */
int
parse_addr_sve(Tcl_Interp *interp, const char *arg, struct unisve_addr *sve)
{
	const char **argv, **av;
	int argc;

	if (Tcl_SplitList(interp, arg, &argc, &argv) != TCL_OK)
		return (TCL_ERROR);
	av = argv;

	if (argc < 2) {
		Tcl_Free((char *)av);
		return (unitcl_setres(interp,
		    "addr SVE: {addr <tag> ...}"));
	}
	if (strcmp(argv[0], "addr") != 0) {
		Tcl_Free((char *)av);
		return (unitcl_setres(interp,
		    "SVE type not 'addr': %s", argv[0]));
	}

	if (parse_svetag(interp, argv[1], &sve->tag) != TCL_OK) {
		Tcl_Free((char *)av);
		return (TCL_ERROR);
	}

	argc -= 2;
	argv += 2;

	if (sve->tag == UNISVE_PRESENT) {
		if (argc < 2) {
			Tcl_Free((char *)av);
			return (unitcl_setres(interp,
			    "addr SVE: {addr present <type> <addr>}"));
		}
		if (strcmp(argv[0], "e164") == 0) {
			if (strlen(argv[1]) == 0 || strlen(argv[1]) > 15) {
				Tcl_Free((char *)av);
				return (unitcl_setres(interp,
				    "E.164 address to long"));
			}
			strcpy((char *)sve->addr, argv[1]);
			sve->len = strlen((char *)sve->addr);
			sve->type = UNI_ADDR_INTERNATIONAL;
			sve->plan = UNI_ADDR_E164;

		} else if (strcmp(argv[0], "atme") == 0) {
			char tmp[100];

			if (strlen(argv[1]) > 90) {
				Tcl_Free((char *)av);
				return (unitcl_setres(interp,
				    "bad ATME address"));
			}
			strcpy(tmp, argv[1]);
			strcat(tmp, "00");
			if (uni_str2nsap(sve->addr, tmp)) {
				Tcl_Free((char *)av);
				return (unitcl_setres(interp,
				    "bad ATME address"));
			}
			sve->type = UNI_ADDR_UNKNOWN;
			sve->plan = UNI_ADDR_ATME;
			sve->len = 19;

		} else {
			Tcl_Free((char *)av);
			return (unitcl_setres(interp,
			    "bad address type '%s'", argv[0]));
		}
		argc -= 2;
		argv += 2;
	}

	if (argc != 0) {
		Tcl_Free((char *)av);
		return (unitcl_setres(interp, "excess args to address SVE"));
	}

	Tcl_Free((char *)av);
	return (TCL_OK);
}

int
fmt_addr_sve(Tcl_Interp *interp, Tcl_DString *str,
    const struct unisve_addr *sve)
{
	char buf[100];

	Tcl_DStringStartSublist(str);

	Tcl_DStringAppendElement(str, "addr");
	if (fmt_svetag(interp, str, sve->tag) != TCL_OK)
		return (TCL_ERROR);

	if (sve->tag == UNISVE_PRESENT) {
		if (sve->plan == UNI_ADDR_E164) {
			Tcl_DStringAppendElement(str, "e164");

			sprintf(buf, "%.*s", (int)sve->len, sve->addr);
			Tcl_DStringAppendElement(str, buf);

		} else if (sve->plan == UNI_ADDR_ATME) {
			Tcl_DStringAppendElement(str, "atme");

			uni_prefix2str(buf, sve->addr, sve->len, 0);
			Tcl_DStringAppendElement(str, buf);

		} else
			return (unitcl_setres(interp,
			    "bad address plan %u", sve->plan));
	}

	Tcl_DStringEndSublist(str);
	return (TCL_OK);
}

int
parse_selector_sve(Tcl_Interp *interp, const char *arg,
    struct unisve_selector *sve)
{
	const char **argv, **av;
	char *end;
	int argc;
	u_long sel;

	if (Tcl_SplitList(interp, arg, &argc, &argv) != TCL_OK)
		return (TCL_ERROR);
	av = argv;

	if (argc < 2) {
		Tcl_Free((char *)av);
		return (unitcl_setres(interp,
		    "selector SVE: {selector <tag> [<sel>]}"));
	}
	if (strcmp(argv[0], "selector") != 0) {
		Tcl_Free((char *)av);
		return (unitcl_setres(interp,
		    "SVE type not 'selector': %s", argv[0]));
	}
	if (parse_svetag(interp, argv[1], &sve->tag) != TCL_OK) {
		Tcl_Free((char *)av);
		return (TCL_ERROR);
	}

	argc -= 2;
	argv += 2;

	if (sve->tag == UNISVE_PRESENT) {
		if (argc < 1) {
			Tcl_Free((char *)av);
			return (unitcl_setres(interp,
			    "selector SVE: {selector present <sel>}"));
		}
		sel = strtoul(argv[0], &end, 0);
		if (*end != '\0') {
			Tcl_Free((char *)av);
			return (unitcl_setres(interp,
			    "bad selector '%s'", argv[0]));
		}
		if (sel > 255) {
			Tcl_Free((char *)av);
			return (unitcl_setres(interp,
			    "selector too large '%s'", argv[0]));
		}
		sve->selector = (uint8_t)sel;
		argc--;
		argv++;
	}

	if (argc != 0) {
		Tcl_Free((char *)av);
		return (unitcl_setres(interp, "excess args to selector SVE"));
	}

	Tcl_Free((char *)av);
	return (TCL_OK);
}

int
fmt_selector_sve(Tcl_Interp *interp, Tcl_DString *str,
    const struct unisve_selector *sve)
{
	char buf[100];

	Tcl_DStringStartSublist(str);

	Tcl_DStringAppendElement(str, "selector");
	if (fmt_svetag(interp, str, sve->tag) != TCL_OK)
		return (TCL_ERROR);

	if (sve->tag == UNISVE_PRESENT) {
		sprintf(buf, "%u", sve->selector);
		Tcl_DStringAppendElement(str, buf);
	}

	Tcl_DStringEndSublist(str);
	return (TCL_OK);

}
