/*
 * Copyright (c) 1996-2003
 *	Fraunhofer Institute for Open Communication Systems (FhG Fokus).
 * 	All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Author: Hartmut Brandt <harti@freebsd.org>
 *
 * $Begemot: atmsupport/libunitcl/ie_bearer.c,v 1.9 2004/01/02 15:15:54 hbb Exp $
 */
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>

#include "inctcl.h"

#include <netnatm/unimsg.h>
#include <netnatm/msg/unistruct.h>

#include "unitcl.h"
#include "unitcl_priv.h"

REGISTER_IE(UNI_IE_BEARER, bearer, 0)

/*
 * {bearer <class> [<atc>] <clip> <conf>}
 * class := a | c | x | p
 * atc := cbr | cbr1 | vbr | vbr1 | nvbr | nvbr1 | abr | <num>
 * clip := clip | noclip
 * conf := p2p | mp
 */
int
parse_bearer(Tcl_Interp *interp, int argc, const char **argv,
    union uni_ieall *ie)
{
	u_int num;

	if (argc != 3 && argc != 4) {
		unitcl_setres(interp, "bad # of args for bearer cap (%d)",
		    argc);
		return (TCL_ERROR);
	}
	if (strcmp(argv[0], "a") == 0)
		ie->bearer.bclass = UNI_BEARER_A;
	else if (strcmp(argv[0], "c") == 0)
		ie->bearer.bclass = UNI_BEARER_C;
	else if (strcmp(argv[0], "x") == 0)
		ie->bearer.bclass = UNI_BEARER_X;
	else if (strcmp(argv[0], "p") == 0)
		ie->bearer.bclass = UNI_BEARER_TVP;
	else {
		unitcl_setres(interp, "bad bearer class");
		return (TCL_ERROR);
	}
	argv++;
	argc--;

	if (argc == 3) {
		ie->bearer.h.present |= UNI_BEARER_ATC_P;
		if (strcmp(argv[0], "cbr") == 0)
			ie->bearer.atc = UNI_BEARER_ATC_CBR;
		else if (strcmp(argv[0], "cbr1") == 0)
			ie->bearer.atc = UNI_BEARER_ATC_CBR1;
		else if (strcmp(argv[0], "vbr") == 0)
			ie->bearer.atc = UNI_BEARER_ATC_VBR;
		else if (strcmp(argv[0], "vbr1") == 0)
			ie->bearer.atc = UNI_BEARER_ATC_VBR1;
		else if (strcmp(argv[0], "nvbr") == 0)
			ie->bearer.atc = UNI_BEARER_ATC_NVBR;
		else if (strcmp(argv[0], "nvbr1") == 0)
			ie->bearer.atc = UNI_BEARER_ATC_NVBR1;
		else if (strcmp(argv[0], "abr") == 0)
			ie->bearer.atc = UNI_BEARER_ATC_ABR;
		else {
			if (unitcl_parse_num(interp, argv[0], &num) != TCL_OK)
				return (TCL_ERROR);
			ie->bearer.atc = num;
		}
		argv++;
		argc--;
	}

	if (strcmp(argv[0], "clip") == 0)
		ie->bearer.clip = UNI_BEARER_CLIP;
	else if (strcmp(argv[0], "noclip") == 0)
		ie->bearer.clip = UNI_BEARER_NOCLIP;
	else {
		unitcl_setres(interp, "bad clipping '%s'", argv[0]);
		return (TCL_ERROR);
	}
	argc--;
	argv++;

	if (strcmp(argv[0], "p2p") == 0)
		ie->bearer.cfg = UNI_BEARER_P2P;
	else if (strcmp(argv[0], "mp") == 0)
		ie->bearer.cfg = UNI_BEARER_MP;
	else {
		unitcl_setres(interp, "bad bearer configuration '%s'", argv[0]);
		return (TCL_ERROR);
	}

	IE_ADDPRESENT(ie->bearer);

	return (TCL_OK);
}

FMT_IE_START(UNI_IE_BEARER, bearer)
{
	char buf[100];

	switch (ie->bclass) {

	  default:
		unitcl_setres(interp, "bad bearer class to format");
		return (TCL_ERROR);

	  case UNI_BEARER_A:
		Tcl_DStringAppendElement(str, "a");
		break;

	  case UNI_BEARER_C:
		Tcl_DStringAppendElement(str, "c");
		break;

	  case UNI_BEARER_X:
		Tcl_DStringAppendElement(str, "x");
		break;

	  case UNI_BEARER_TVP:
		Tcl_DStringAppendElement(str, "p");
		break;
	}

	if (ie->h.present & UNI_BEARER_ATC_P) {
		switch (ie->atc) {

		  case UNI_BEARER_ATC_CBR:
			Tcl_DStringAppendElement(str, "cbr");
			break;

		  case UNI_BEARER_ATC_CBR1:
			Tcl_DStringAppendElement(str, "cbr1");
			break;

		  case UNI_BEARER_ATC_VBR:
			Tcl_DStringAppendElement(str, "vbr");
			break;

		  case UNI_BEARER_ATC_VBR1:
			Tcl_DStringAppendElement(str, "vbr1");
			break;

		  case UNI_BEARER_ATC_NVBR:
			Tcl_DStringAppendElement(str, "nvbr");
			break;

		  case UNI_BEARER_ATC_NVBR1:
			Tcl_DStringAppendElement(str, "nvbr1");
			break;

		  case UNI_BEARER_ATC_ABR:
			Tcl_DStringAppendElement(str, "abr");
			break;

		  default:
			sprintf(buf, "%u", ie->atc);
			Tcl_DStringAppendElement(str, buf);
			break;
		}
	}
	if (ie->clip == UNI_BEARER_CLIP)
		Tcl_DStringAppendElement(str, "clip");
	else if (ie->clip == UNI_BEARER_NOCLIP)
		Tcl_DStringAppendElement(str, "noclip");
	else {
		unitcl_setres(interp, "bad clipping to format");
		return (TCL_ERROR);
	}

	if (ie->cfg == UNI_BEARER_P2P)
		Tcl_DStringAppendElement(str, "p2p");
	else if (ie->cfg == UNI_BEARER_MP)
		Tcl_DStringAppendElement(str, "mp");
	else {
		unitcl_setres(interp, "bad bearer config to format");
		return (TCL_ERROR);
	}
}
FMT_IE_END()
