/*
 * Copyright (c) 2004
 *	Hartmut Brandt
 * Copyright (c) 1996-2003
 *	Fraunhofer Institute for Open Communication Systems (FhG Fokus).
 * 	All rights reserved.
 *
 * Author: Hartmut Brandt <harti@freebsd.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Begemot: atmsupport/libunitcl/ie_blli.c,v 1.10 2004/05/12 11:50:04 novo Exp $
 */
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>

#include "inctcl.h"

#include <netnatm/unimsg.h>
#include <netnatm/msg/unistruct.h>
#include <netnatm/api/unisap.h>

#include "unitcl.h"
#include "unitcl_priv.h"

REGISTER_IE(UNI_IE_BLLI, blli, 0)

static int parse_blli_l1(Tcl_Interp *, int, const char **,
    struct uni_ie_blli *);
static int parse_blli_l2(Tcl_Interp *, int, const char **,
    struct uni_ie_blli *);
static int parse_blli_l3(Tcl_Interp *, int, const char **,
    struct uni_ie_blli *);

/*
 * {blli [{l1 <proto>}]
 *	 [{l2 <proto> [{q933 <q933> <mode> [<wsize>]}]
 *		      [{user <proto>}]}]
 *	 [{l3 <proto> [{mode <mode> [<psiz> [<wsiz>]]}]
 *		      [{user <proto>}]
 *		      [{ipi <ipi> [<oui> <pid>]}]
 *		      [{ttype <ttype> [<fmux> <bmux>]}]}]}
 */
int
parse_blli(Tcl_Interp *interp, int argc, const char **argv, union uni_ieall *ie)
{
	int i, ret;

	if (argc == 0)
		return (TCL_OK);

	for (i = 0; i < argc; i++) {
		const char **av;
		int ac;

		if (Tcl_SplitList(interp, argv[i], &ac, &av) != TCL_OK)
			return (TCL_ERROR);

		if (ac < 2) {
			Tcl_Free((char *)av);
			return (unitcl_setres(interp,
			    "not enough args in blli IE"));
		}

		if (strcmp(av[0], "l1") == 0)
			ret = parse_blli_l1(interp, ac - 1, av + 1, &ie->blli);
		else if (strcmp(av[0], "l2") == 0)
			ret = parse_blli_l2(interp, ac - 1, av + 1, &ie->blli);
		else if (strcmp(av[0], "l3") == 0)
			ret = parse_blli_l3(interp, ac - 1, av + 1, &ie->blli);
		else {
			Tcl_Free((char *)av);
			return (unitcl_setres(interp,
			    "bad blli element ident"));
		}

		Tcl_Free((char *)av);
		if (ret != TCL_OK)
			return (TCL_ERROR);
	}

	IE_ADDPRESENT(ie->blli);

	return (TCL_OK);
}

static int
parse_blli_l1(Tcl_Interp *interp, int argc, const char **argv,
    struct uni_ie_blli *ie)
{
	u_int n;

	if (argc != 1)
		return (unitcl_setres(interp, "bad # of args for blli l1"));
	if (unitcl_parse_num(interp, argv[0], &n))
		return (TCL_ERROR);
	ie->h.present |= UNI_BLLI_L1_P;
	ie->l1 = n;

	return (TCL_OK);
}

/*
 * {l2 <proto> [{q933 <q933> <mode> [<wsize>]}] [{user <proto>}]}
 */
static int
parse_blli_l2(Tcl_Interp *interp, int argc, const char **argv,
    struct uni_ie_blli *ie)
{
	int i, ac;
	const char **av;
	u_int n;

	if (argc < 1)
		return (unitcl_setres(interp, "bad # of args for blli l2"));
	if (unitcl_parse_num(interp, argv[0], &n))
		return (TCL_ERROR);

	ie->h.present |= UNI_BLLI_L2_P;
	ie->l2 = n;

	for (i = 1; i < argc; i++) {
		if (Tcl_SplitList(interp, argv[i], &ac, &av) != TCL_OK)
			return (TCL_ERROR);

		if (ac < 2) {
			Tcl_Free((char *)av);
			return (unitcl_setres(interp,
			    "not enough args in blli l2 IE"));
		}

		if (strcmp(av[0], "q933") == 0) {
			if (ac != 3 && ac != 4) {
				Tcl_Free((char *)av);
				return (unitcl_setres(interp,
				    "bad # of args for blli l2 q933"));
			}
			if (unitcl_parse_num(interp, av[1], &n)) {
				Tcl_Free((char *)av);
				return (TCL_ERROR);
			}
			ie->l2_q933 = n;
			if (unitcl_parse_num(interp, av[2], &n)) {
				Tcl_Free((char *)av);
				return (TCL_ERROR);
			}
			ie->l2_mode = n;
			if (ac == 4) {
				if (unitcl_parse_num(interp, av[3], &n)) {
					Tcl_Free((char *)av);
					return (TCL_ERROR);
				}
				ie->l2_wsiz = n;
				ie->h.present |= UNI_BLLI_L2_WSIZ_P;
			}
			ie->h.present |= UNI_BLLI_L2_Q933_P;

		} else if (strcmp(av[0], "user") == 0) {
			if (ac != 2) {
				Tcl_Free((char *)av);
				return (unitcl_setres(interp,
				    "bad # of args for blli l2 user"));
			}
			if (unitcl_parse_num(interp, av[1], &n)) {
				Tcl_Free((char *)av);
				return (TCL_ERROR);
			}
			ie->h.present |= UNI_BLLI_L2_USER_P;
			ie->l2_user = n;
		} else {
			Tcl_Free((char *)av);
			return (unitcl_setres(interp, "bad blli l2 element"));
		}
		Tcl_Free((char *)av);
	}

	return (TCL_OK);
}

/*
 * {l3 <proto> [{mode <mode> [<psiz> [<wsiz>]]}]
 *	       [{user <proto>}]
 * 	       [{ipi <ipi> [<oui> <pid>]}]
 *	       [{ttype <ttype> [<fmux> <bmux>]}]}
 */
static int
parse_blli_l3(Tcl_Interp *interp, int argc, const char **argv,
    struct uni_ie_blli *ie)
{
	int i, ac;
	const char **av;
	u_int n;

	if (argc < 1)
		return (unitcl_setres(interp, "bad # of args for blli l3"));
	if (unitcl_parse_num(interp, argv[0], &n))
		return (TCL_ERROR);

	ie->h.present |= UNI_BLLI_L3_P;
	ie->l3 = n;

	for (i = 1; i < argc; i++) {
		if (Tcl_SplitList(interp, argv[i], &ac, &av) != TCL_OK)
			return (TCL_ERROR);

		if (ac < 2) {
			Tcl_Free((char *)av);
			return (unitcl_setres(interp,
			    "not enough args in blli l3"));
		}

		if (strcmp(av[0], "mode") == 0) {
			if (ac != 2 && ac != 3 && ac != 4) {
				Tcl_Free((char *)av);
				return (unitcl_setres(interp,
				    "bad # of args for blli l3 mode"));
			}
			if (unitcl_parse_num(interp, av[1], &n)) {
				Tcl_Free((char *)av);
				return (TCL_ERROR);
			}
			ie->l3_mode = n;
			ie->h.present |= UNI_BLLI_L3_MODE_P;
			if (ac > 2) {
				if (unitcl_parse_num(interp, av[2], &n)) {
					Tcl_Free((char *)av);
					return (TCL_ERROR);
				}
				ie->l3_psiz = n;
				ie->h.present |= UNI_BLLI_L3_PSIZ_P;
				if (ac > 3) {
					if (unitcl_parse_num(interp,av[3],&n)) {
						Tcl_Free((char *)av);
						return (TCL_ERROR);
					}
					ie->l3_wsiz = n;
					ie->h.present |= UNI_BLLI_L3_WSIZ_P;
				}
			}

		} else if (strcmp(av[0], "ipi") == 0) {
			if (ac != 2 && ac != 4) {
				Tcl_Free((char *)av);
				return (unitcl_setres(interp,
				    "bad # of args for blli l3 ipi"));
			}
			if (unitcl_parse_num(interp, av[1], &n)) {
				Tcl_Free((char *)av);
				return (TCL_ERROR);
			}
			ie->l3_ipi = n;
			ie->h.present |= UNI_BLLI_L3_IPI_P;
			if (ac == 4) {
				if (unitcl_parse_num(interp, av[2], &n)) {
					Tcl_Free((char *)av);
					return (TCL_ERROR);
				}
				ie->oui = n;
				ie->h.present |= UNI_BLLI_L3_SNAP_P;
				if (unitcl_parse_num(interp, av[3], &n)) {
					Tcl_Free((char *)av);
					return (TCL_ERROR);
				}
				ie->pid = n;
			}

		} else if (strcmp(av[0], "user") == 0) {
			if (ac != 2) {
				Tcl_Free((char *)av);
				return (unitcl_setres(interp,
				    "bad # of args for blli l3 user"));
			}
			if (unitcl_parse_num(interp, av[1], &n)) {
				Tcl_Free((char *)av);
				return (TCL_ERROR);
			}
			ie->h.present |= UNI_BLLI_L3_USER_P;
			ie->l3_user = n;

		} else if (strcmp(av[0], "ttype") == 0) {
			if (ac != 2 && ac != 4) {
				Tcl_Free((char *)av);
				return (unitcl_setres(interp,
				    "bad # of args for blli l3 ttype"));
			}
			if (strcmp(av[1], "recv") == 0)
				ie->l3_ttype = UNI_BLLI_L3_TTYPE_RECV;
			else if (strcmp(av[1], "send") == 0)
				ie->l3_ttype = UNI_BLLI_L3_TTYPE_SEND;
			else if (strcmp(av[1], "both") == 0)
				ie->l3_ttype = UNI_BLLI_L3_TTYPE_BOTH;
			else {
				Tcl_Free((char *)av);
				return (unitcl_setres(interp,
				    "bad blli l3 ttype"));
			}
			if (ac == 4) {
				u_int dir, mux[2];

				for (dir = 0; dir < 2; dir++) {
					if (strcmp(av[2 + dir], "nomux") == 0)
						mux[dir] = UNI_BLLI_L3_MUX_NOMUX;
					else if (strcmp(av[2 + dir], "ts") == 0)
						mux[dir] = UNI_BLLI_L3_MUX_TS;
					else if (strcmp(av[2 + dir], "tsfec") == 0)
						mux[dir] = UNI_BLLI_L3_MUX_TSFEC;
					else if (strcmp(av[2 + dir], "ps") == 0)
						mux[dir] = UNI_BLLI_L3_MUX_PS;
					else if (strcmp(av[2 + dir], "psfec") == 0)
						mux[dir] = UNI_BLLI_L3_MUX_PSFEC;
					else if (strcmp(av[2 + dir], "h221") == 0)
						mux[dir] = UNI_BLLI_L3_MUX_H221;
					else {
						Tcl_Free((char *)av);
						return (unitcl_setres(interp,
						    "bad blli l3 mux"));
					}
				}
				ie->l3_fmux = mux[0];
				ie->l3_bmux = mux[1];
				ie->h.present |= UNI_BLLI_L3_MUX_P;
			}
			ie->h.present |= UNI_BLLI_L3_TTYPE_P;

		} else {
			Tcl_Free((char *)av);
			return (unitcl_setres(interp, "bad blli l3 element"));
		}
		Tcl_Free((char *)av);
	}

	return (TCL_OK);
}


FMT_IE_START(UNI_IE_BLLI, blli)
{
	char buf[100];

	if (ie->h.present & UNI_BLLI_L1_P) {
		Tcl_DStringStartSublist(str);
		Tcl_DStringAppendElement(str, "l1");

		sprintf(buf, "0x%02x", ie->l1);
		Tcl_DStringAppendElement(str, buf);

		Tcl_DStringEndSublist(str);
	}

	if (ie->h.present & UNI_BLLI_L2_P) {
		Tcl_DStringStartSublist(str);
		Tcl_DStringAppendElement(str, "l2");

		sprintf(buf, "0x%02x", ie->l2);
		Tcl_DStringAppendElement(str, buf);

		if (ie->h.present & UNI_BLLI_L2_Q933_P) {
			Tcl_DStringStartSublist(str);
			Tcl_DStringAppendElement(str, "q933");

			sprintf(buf, "0x%02x", ie->l2_q933);
			Tcl_DStringAppendElement(str, buf);

			sprintf(buf, "0x%02x", ie->l2_mode);
			Tcl_DStringAppendElement(str, buf);

			if (ie->h.present & UNI_BLLI_L2_WSIZ_P) {
				sprintf(buf, "0x%02x", ie->l2_wsiz);
				Tcl_DStringAppendElement(str, buf);
			}
			Tcl_DStringEndSublist(str);
		}

		if (ie->h.present & UNI_BLLI_L2_USER_P) {
			Tcl_DStringStartSublist(str);
			Tcl_DStringAppendElement(str, "user");

			sprintf(buf, "0x%02x", ie->l2_user);
			Tcl_DStringAppendElement(str, buf);

			Tcl_DStringEndSublist(str);
		}

		Tcl_DStringEndSublist(str);
	}

	if (ie->h.present & UNI_BLLI_L3_P) {
		Tcl_DStringStartSublist(str);
		Tcl_DStringAppendElement(str, "l3");

		sprintf(buf, "0x%02x", ie->l3);
		Tcl_DStringAppendElement(str, buf);

		if (ie->h.present & UNI_BLLI_L3_MODE_P) {
			Tcl_DStringStartSublist(str);
			Tcl_DStringAppendElement(str, "mode");

			sprintf(buf, "0x%02x", ie->l3_mode);
			Tcl_DStringAppendElement(str, buf);

			if (ie->h.present & UNI_BLLI_L3_PSIZ_P) {
				sprintf(buf, "0x%02x", ie->l3_psiz);
				Tcl_DStringAppendElement(str, buf);

				if (ie->h.present & UNI_BLLI_L3_WSIZ_P) {
					sprintf(buf, "0x%02x", ie->l3_wsiz);
					Tcl_DStringAppendElement(str, buf);
				}
			}
			Tcl_DStringEndSublist(str);
		}

		if (ie->h.present & UNI_BLLI_L3_USER_P) {
			Tcl_DStringStartSublist(str);
			Tcl_DStringAppendElement(str, "user");

			sprintf(buf, "0x%02x", ie->l3_user);
			Tcl_DStringAppendElement(str, buf);

			Tcl_DStringEndSublist(str);
		}

		if (ie->h.present & UNI_BLLI_L3_IPI_P) {
			Tcl_DStringStartSublist(str);
			Tcl_DStringAppendElement(str, "ipi");

			sprintf(buf, "0x%02x", ie->l3_ipi);
			Tcl_DStringAppendElement(str, buf);

			if (ie->h.present & UNI_BLLI_L3_SNAP_P) {
				sprintf(buf, "0x%06x", ie->oui);
				Tcl_DStringAppendElement(str, buf);
				sprintf(buf, "0x%04x", ie->pid);
				Tcl_DStringAppendElement(str, buf);
			}
			Tcl_DStringEndSublist(str);
		}

		if (ie->h.present & UNI_BLLI_L3_TTYPE_P) {
			Tcl_DStringStartSublist(str);
			Tcl_DStringAppendElement(str, "ttype");

			switch (ie->l3_ttype) {
			  default:
				sprintf(buf, "%u", ie->l3_ttype);
				Tcl_DStringAppendElement(str, buf);
				break;

			  case UNI_BLLI_L3_TTYPE_RECV:
				Tcl_DStringAppendElement(str, "recv");
				break;

			  case UNI_BLLI_L3_TTYPE_SEND:
				Tcl_DStringAppendElement(str, "send");
				break;

			  case UNI_BLLI_L3_TTYPE_BOTH:
				Tcl_DStringAppendElement(str, "both");
				break;
			}
			if (ie->h.present & UNI_BLLI_L3_MUX_P) {
				int mux;

				for (mux = 0; mux < 2; mux++) {
				    switch (mux ? ie->l3_bmux : ie->l3_fmux) {

				      default:
					sprintf(buf, "%u", mux ? ie->l3_bmux
						: ie->l3_fmux);
					Tcl_DStringAppendElement(str, buf);
					break;

				      case UNI_BLLI_L3_MUX_NOMUX:
					Tcl_DStringAppendElement(str, "nomux");
					break;

				      case UNI_BLLI_L3_MUX_TS:
					Tcl_DStringAppendElement(str, "ts");
					break;

				      case UNI_BLLI_L3_MUX_TSFEC:
					Tcl_DStringAppendElement(str, "tsfec");
					break;

				      case UNI_BLLI_L3_MUX_PS:
					Tcl_DStringAppendElement(str, "ps");
					break;

				      case UNI_BLLI_L3_MUX_PSFEC:
					Tcl_DStringAppendElement(str, "psfec");
					break;

				      case UNI_BLLI_L3_MUX_H221:
					Tcl_DStringAppendElement(str, "h221");
					break;
				    }
				}
			}
			Tcl_DStringEndSublist(str);
		}
		Tcl_DStringEndSublist(str);
	}
}
FMT_IE_END()

/************************************************************/
/*
 * {blli_id2 <tag> [<proto> [<user]]}
 */
int
parse_blli_id2_sve(Tcl_Interp *interp, const char *arg,
    struct unisve_blli_id2 *sve)
{
	char *end;
	const char **argv, **av;
	int argc;
	u_long proto;

	if (Tcl_SplitList(interp, arg, &argc, &argv) != TCL_OK)
		return (TCL_ERROR);
	av = argv;

	if (argc < 2) {
		Tcl_Free((char *)av);
		return (unitcl_setres(interp,
		    "blli_id2 SVE: {blli_id2 <tag> ...}"));
	}
	if (strcmp(argv[0], "blli_id2") != 0) {
		Tcl_Free((char *)av);
		return (unitcl_setres(interp,
		    "SVE type not 'blli_id2': %s", argv[0]));
	}
	if (parse_svetag(interp, argv[1], &sve->tag) != TCL_OK) {
		Tcl_Free((char *)av);
		return (TCL_ERROR);
	}
	argc -= 2;
	argv += 2;

	if (sve->tag == UNISVE_PRESENT) {
		if (argc < 1) {
			Tcl_Free((char *)av);
			return (unitcl_setres(interp,
			    "need protocol to blli_id2"));
		}
		proto = strtoul(argv[0], &end, 0);
		if (*end != '\0') {
			Tcl_Free((char *)av);
			return (unitcl_setres(interp,
			    "bad L2 proto '%s'", argv[0]));
		}
		if (proto > 0x1f) {
			Tcl_Free((char *)av);
			return (unitcl_setres(interp,
			    "L2 proto too large '%s'", argv[0]));
		}
		sve->proto = proto;
		argv++;
		argc--;
		if (sve->proto == UNI_BLLI_L2_USER) {
			if (argc < 1) {
				Tcl_Free((char *)av);
				return (unitcl_setres(interp,
				    "need L2 user proto"));
			}
			proto = strtoul(argv[0], &end, 0);
			if (*end != '\0') {
				Tcl_Free((char *)av);
				return (unitcl_setres(interp,
				    "bad L2 user proto '%s", argv[0]));
			}
			if (proto > 0x7f) {
				Tcl_Free((char *)av);
				return (unitcl_setres(interp,
				    "L2 user proto too large '%s'", argv[0]));
			}
			sve->user = proto;
			argv++;
			argc--;
		}
	}
	if (argc != 0) {
		Tcl_Free((char *)av);
		return (unitcl_setres(interp, "excess args to blli_id2 SVE"));
	}
	Tcl_Free((char *)av);
	return (TCL_OK);
}

int
fmt_blli_id2_sve(Tcl_Interp *interp, Tcl_DString *str,
    const struct unisve_blli_id2 *sve)
{
	char buf[100];

	Tcl_DStringStartSublist(str);

	Tcl_DStringAppendElement(str, "blli_id2");
	if (fmt_svetag(interp, str, sve->tag) != TCL_OK)
		return (TCL_ERROR);

	if (sve->tag == UNISVE_PRESENT) {
		sprintf(buf, "%u", sve->proto);
		Tcl_DStringAppendElement(str, buf);

		if (sve->proto == UNI_BLLI_L2_USER) {
			sprintf(buf, "%u", sve->user);
			Tcl_DStringAppendElement(str, buf);
		}
	}

	Tcl_DStringEndSublist(str);
	return (TCL_OK);
}

/*
 * {blli_id3 <tag> [<proto>]}
 * {blli_id3 present <proto> <user>}
 * {blli_id3 present <proto> <ipi> [<oui> <pid>]}
 */
int
parse_blli_id3_sve(Tcl_Interp *interp, const char *arg,
    struct unisve_blli_id3 *sve)
{
	const char **argv, **av;
	char *end;
	int argc;
	u_long proto;

	if (Tcl_SplitList(interp, arg, &argc, &argv) != TCL_OK)
		return (TCL_ERROR);
	av = argv;

	if (argc < 2) {
		Tcl_Free((char *)av);
		return (unitcl_setres(interp,
		    "blli_id3 SVE: {blli_id3 <tag> ...}"));
	}
	if (strcmp(argv[0], "blli_id3") != 0) {
		Tcl_Free((char *)av);
		return (unitcl_setres(interp,
		    "SVE type not 'blli_id3': %s", argv[0]));
	}
	if (parse_svetag(interp, argv[1], &sve->tag) != TCL_OK) {
		Tcl_Free((char *)av);
		return (TCL_ERROR);
	}
	argc -= 2;
	argv += 2;

	if (sve->tag == UNISVE_PRESENT) {
		if (argc < 1) {
			Tcl_Free((char *)av);
			return (unitcl_setres(interp,
			    "need protocol to blli_id3"));
		}
		proto = strtoul(argv[0], &end, 0);
		if (*end != '\0') {
			Tcl_Free((char *)av);
			return (unitcl_setres(interp,
			    "bad L3 proto '%s'", argv[2]));
		}
		if (proto > 0x1f) {
			Tcl_Free((char *)av);
			return (unitcl_setres(interp,
			    "L3 proto too large '%s'", argv[2]));
		}
		sve->proto = proto;
		argv++;
		argc--;

		if (sve->proto == UNI_BLLI_L3_USER) {
			if (argc < 1) {
				Tcl_Free((char *)av);
				return (unitcl_setres(interp,
				    "need L3 user proto"));
			}
			proto = strtoul(argv[0], &end, 0);
			if (*end != '\0') {
				Tcl_Free((char *)av);
				return (unitcl_setres(interp,
				    "bad L3 user proto '%s", argv[0]));
			}
			if (proto > 0x7f) {
				Tcl_Free((char *)av);
				return (unitcl_setres(interp,
				    "L3 user proto too large '%s'", argv[0]));
			}
			sve->user = proto;
			argv++;
			argc--;

		} else if (sve->proto == UNI_BLLI_L3_TR9577) {
			if (argc == 0) {
				sve->noipi = 1;

			} else {
				sve->noipi = 0;
				proto = strtoul(argv[0], &end, 0);
				if (*end != '\0') {
					Tcl_Free((char *)av);
					return (unitcl_setres(interp,
					    "bad L3 IPI '%s", argv[0]));
				}
				if (proto > 0xff) {
					Tcl_Free((char *)av);
					return (unitcl_setres(interp,
					    "L3 IPI too large '%s'", argv[0]));
				}
				sve->ipi = proto;
				argv++;
				argc--;
				if (sve->ipi == UNI_BLLI_L3_SNAP) {
					if (argc < 2) {
						Tcl_Free((char *)av);
						return (unitcl_setres(interp,
						    "need L3 OUI and PID"));
					}
					proto = strtoul(argv[0], &end, 0);
					if (*end != '\0') {
						Tcl_Free((char *)av);
						return (unitcl_setres(interp,
						    "bad L3 OUI '%s", argv[0]));
					}
					if (proto > 0xffffff) {
						Tcl_Free((char *)av);
						return (unitcl_setres(interp,
						    "L3 OUI too large '%s'",
						    argv[0]));
					}
					sve->oui = proto;

					proto = strtoul(argv[1], &end, 0);
					if (*end != '\0') {
						Tcl_Free((char *)av);
						return (unitcl_setres(interp,
						    "bad L3 PID '%s", argv[1]));
					}
					if (proto > 0xffff) {
						Tcl_Free((char *)av);
						return (unitcl_setres(interp,
						    "L3 PID too large '%s'",
						    argv[1]));
					}
					sve->pid = proto;

					argc -= 2;
					argv += 2;
				}
			}
		}
	}
	if (argc != 0) {
		Tcl_Free((char *)av);
		return (unitcl_setres(interp, "excess args to blli_id3 SVE"));
	}

	Tcl_Free((char *)av);
	return (TCL_OK);
}

int
fmt_blli_id3_sve(Tcl_Interp *interp, Tcl_DString *str,
    const struct unisve_blli_id3 *sve)
{
	char buf[100];

	Tcl_DStringStartSublist(str);

	Tcl_DStringAppendElement(str, "blli_id3");
	if (fmt_svetag(interp, str, sve->tag) != TCL_OK)
		return (TCL_ERROR);

	if (sve->tag == UNISVE_PRESENT) {
		sprintf(buf, "%u", sve->proto);
		Tcl_DStringAppendElement(str, buf);

		if (sve->proto == UNI_BLLI_L3_USER) {
			sprintf(buf, "%u", sve->user);
			Tcl_DStringAppendElement(str, buf);

		} else if (sve->proto == UNI_BLLI_L3_TR9577) {
			if (!sve->noipi) {
				sprintf(buf, "%u", sve->ipi);
				Tcl_DStringAppendElement(str, buf);

				if (sve->ipi == UNI_BLLI_L3_SNAP) {
					sprintf(buf, "%u", sve->oui);
					Tcl_DStringAppendElement(str, buf);

					sprintf(buf, "%u", sve->pid);
					Tcl_DStringAppendElement(str, buf);
				}
			}
		}
	}

	Tcl_DStringEndSublist(str);
	return (TCL_OK);
}

