/*
 * Copyright (c) 1996-2003
 *	Fraunhofer Institute for Open Communication Systems (FhG Fokus).
 * 	All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Author: Hartmut Brandt <harti@freebsd.org>
 *
 * $Begemot: atmsupport/libunitcl/ie_cause.c,v 1.9 2004/01/02 15:15:55 hbb Exp $
 */
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>

#include "inctcl.h"

#include <netnatm/unimsg.h>
#include <netnatm/msg/unistruct.h>

#include "unitcl.h"
#include "unitcl_priv.h"

REGISTER_IE(UNI_IE_CAUSE, cause, 0)

/*
 * {cause <loc> <cause> ( {cond <pu> <na> <cond>}
 *			| {tns <IE>}
 *			| {rej <reason> <cond> [<user>|<ie>]}
 *			| {number <IE>}
 *			| {traffic <key> [...]}
 *			| {attr {NUM [NUM [NUM]]}+}
 *			| {param NUM}
 *			| {timer <timer>} ) }
 *			| {mtype <mtype>}
 *			| {ie <ie> [...]}
 *			| {vpci <vpci> <vci>}
 */
static int
parse_diag_cond(Tcl_Interp *interp, struct uni_ie_cause *ie,
    int argc, const char **argv)
{
	u_int n;

 	/* {cond <pu> <na> <cond>} */
	if (argc != 3)
		return (unitcl_setres(interp,
		    "need 4 args for cond diagnostic"));

	if (unitcl_parse_num(interp, argv[0], &n))
		return (TCL_ERROR);
	ie->u.cond.pu = n;

	if (unitcl_parse_num(interp, argv[1], &n))
		return (TCL_ERROR);
	ie->u.cond.na = n;

	if (unitcl_parse_num(interp, argv[2], &n))
		return (TCL_ERROR);
	ie->u.cond.cond = n;

	ie->h.present |= UNI_CAUSE_COND_P;

	return (TCL_OK);
}

static int
parse_diag_rej(Tcl_Interp *interp, struct uni_ie_cause *ie,
    int argc, const char **argv)
{
	u_int n;

	/* {rej <reason> <cond> [<user>|<ie>]} */
	if (argc != 2 && argc != 3)
		return (unitcl_setres(interp,
		    "bad # of args for reject diagnostic"));

	if (unitcl_parse_num(interp, argv[0], &n))
		return (TCL_ERROR);
	ie->u.rej.reason = n;

	if (unitcl_parse_num(interp, argv[1], &n))
		return (TCL_ERROR);
	ie->u.rej.cond = n;

	if (argc == 3) {
		if (ie->u.rej.reason == UNI_CAUSE_REASON_USER) {
			ie->h.present |= UNI_CAUSE_REJ_USER_P;
			if (unitcl_parse_num(interp, argv[2], &n))
				return (TCL_ERROR);
			ie->u.rej.user = n;
		} else {
			ie->h.present |= UNI_CAUSE_REJ_IE_P;
			if (unitcl_parse_num(interp, argv[2], &n))
				return (TCL_ERROR);
			ie->u.rej.ie = n;
		}
	}

	ie->h.present |= UNI_CAUSE_REJ_P;

	return (TCL_OK);
}

static int
parse_diag_ie(Tcl_Interp *interp, struct uni_ie_cause *ie,
    int argc, const char **argv)
{
	u_int n;
	int i;

	/* {ie <ie> [...]} */
	if (argc < 1)
		return (unitcl_setres(interp,
		    "not enough args for ie diagnostic"));
	if (argc > UNI_CAUSE_IE_N)
		return (unitcl_setres(interp,
		    "too many ie's in ie diagnostic"));
	for (i = 0; i < argc; i++) {
		if (unitcl_parse_num(interp, argv[i], &n))
			return (TCL_ERROR);
		ie->u.ie.ie[ie->u.ie.len++] = n;
	}
	ie->h.present |= UNI_CAUSE_IE_P;

	return (TCL_OK);
}

static int
parse_diag_traffic(Tcl_Interp *interp, struct uni_ie_cause *ie,
    int argc, const char **argv)
{
	u_int n;
	int i;

	/* {traffic <key> [...]} */
	if (argc < 1)
		return (unitcl_setres(interp,
		    "not enough args for traffic diagnostic"));
	if (argc > UNI_CAUSE_TRAFFIC_N)
		return (unitcl_setres(interp,
		    "too many traffic's in traffic diagnostic"));
	for (i = 0; i < argc; i++) {
		if (unitcl_parse_num(interp, argv[i], &n))
			return (TCL_ERROR);
		ie->u.traffic.traffic[ie->u.traffic.len++] = n;
	}
	ie->h.present |= UNI_CAUSE_TRAFFIC_P;

	return (TCL_OK);
}

static int
parse_diag_vpci(Tcl_Interp *interp, struct uni_ie_cause *ie,
    int argc, const char **argv)
{
	u_int n;

	/* {vpci <vpci> <vci>} */
	if (argc != 2)
		return (unitcl_setres(interp, "need 2 args for vpci diag"));

	if (unitcl_parse_num(interp, argv[0], &n))
		return (TCL_ERROR);
	ie->u.vpci.vpci = n;

	if (unitcl_parse_num(interp, argv[1], &n))
		return (TCL_ERROR);
	ie->u.vpci.vci = n;

	ie->h.present |= UNI_CAUSE_VPCI_P;

	return (TCL_OK);
}

static int
parse_diag_mtype(Tcl_Interp *interp, struct uni_ie_cause *ie,
    int argc, const char **argv)
{
	u_int n;

	/* {mtype <mtype>} */
	if (argc != 1)
		return (unitcl_setres(interp, "need 1 arg for mtype diag"));

	if (unitcl_parse_num(interp, argv[0], &n))
		return (TCL_ERROR);
	ie->u.mtype = n;

	ie->h.present |= UNI_CAUSE_MTYPE_P;

	return (TCL_OK);
}

static int
parse_diag_timer(Tcl_Interp *interp, struct uni_ie_cause *ie,
    int argc, const char **argv)
{
	/* {timer <timer>} */
	if (argc != 1)
		return (unitcl_setres(interp,
		    "need 1 arg for timer diagnostic"));

	if (strlen(argv[0]) != 5 || argv[0][0] != '"' || argv[0][4] != '"')
		return (unitcl_setres(interp, "bad timer string"));

	if (!isascii(argv[0][1]) || !isascii(argv[0][2]) ||
	    !isascii(argv[0][3]) || !isalnum(argv[0][1]) ||
	    !isalnum(argv[0][2]) || !isalnum(argv[0][3]))
		return (unitcl_setres(interp, "bad syntax in timer name"));

	ie->u.timer[0] = argv[0][1];
	ie->u.timer[1] = argv[0][2];
	ie->u.timer[2] = argv[0][3];

	ie->h.present |= UNI_CAUSE_TIMER_P;

	return (TCL_OK);
}

static int
parse_diag_param(Tcl_Interp *interp, struct uni_ie_cause *ie,
    int argc, const char **argv)
{
	u_int n;

	/* {param <param>} */
	if (argc != 1)
		return (unitcl_setres(interp, "need 1 arg for param diag"));

	if (unitcl_parse_num(interp, argv[0], &n))
		return (TCL_ERROR);
	ie->u.param = n;

	ie->h.present |= UNI_CAUSE_PARAM_P;

	return (TCL_OK);
}

static int
parse_diag_attr(Tcl_Interp *interp, struct uni_ie_cause *ie,
    int argc, const char **argv)
{
	int ac, i, j;
	const char **av;
	u_int n;

	/* {attr {NUM [NUM [NUM]]}+} */
	if (argc < 1)
		return (unitcl_setres(interp,
		    "need at last 1 arg for attr diagnostic"));
	if (argc > UNI_CAUSE_ATTR_N)
		return (unitcl_setres(interp,
		    "too many attributes in attr diagnostic"));

	for (i = 0; i < argc; i++) {
		if (Tcl_SplitList(interp, argv[i], &ac, &av) != TCL_OK)
			return (TCL_ERROR);
		if (ac < 1 || ac > 3) {
			Tcl_Free((char *)av);
			return (unitcl_setres(interp,
			    "bad number of args in attribute"));
		}
		for (j = 0; j < ac; j++) {
			if (unitcl_parse_num(interp, av[j], &n)) {
				Tcl_Free((char *)av);
				return (TCL_ERROR);
			}
			ie->u.attr.attr[i][j] = n;
		}
		Tcl_Free((char *)av);
	}
	ie->u.attr.nattr = i;

	ie->h.present |= UNI_CAUSE_ATTR_P;

	return (TCL_OK);
}

static int
parse_diag_tns(Tcl_Interp *interp, struct uni_ie_cause *ie,
    int argc, const char **argv)
{
	enum uni_ietype ietype;
	union uni_ieall tns;

	if (argc != 1)
		return (unitcl_setres(interp, "need 1 arg in tns diagnostic"));

	if (parse_ie(interp, argv[0], &tns, &ietype) != TCL_OK)
		return (TCL_ERROR);

	if (ietype != UNI_IE_TNS)
		return (unitcl_setres(interp,
		    "expecting TNS IE in tns diagnostic"));

	ie->u.tns = tns.tns;
	ie->h.present |= UNI_CAUSE_TNS_P;

	return (TCL_OK);
}

static int
parse_diag_number(Tcl_Interp *interp, struct uni_ie_cause *ie,
    int argc, const char **argv)
{
	enum uni_ietype ietype;
	union uni_ieall called;

	if (argc != 1)
		return (unitcl_setres(interp,
		    "need 1 arg in number diagnostic"));

	if (parse_ie(interp, argv[0], &called, &ietype) != TCL_OK)
		return (TCL_ERROR);

	if (ietype != UNI_IE_CALLED)
		return (unitcl_setres(interp,
		    "expecting CALLED IE in number diagnostic"));

	ie->u.number = called.called;
	ie->h.present |= UNI_CAUSE_NUMBER_P;

	return (TCL_OK);
}

int
parse_cause(Tcl_Interp *interp, int argc, const char **argv,
    union uni_ieall *ie)
{
	u_int n;
	int ac;
	const char **av;
	int ret;

	if (argc < 2 || argc > 3)
		return (unitcl_setres(interp, "bad # of args for cause"));

	if (unitcl_parse_num(interp, argv[0], &n))
		return (TCL_ERROR);
	ie->cause.loc = n;
	if (unitcl_parse_num(interp, argv[1], &n))
		return (TCL_ERROR);
	ie->cause.cause = n;

	if (argc == 3) {
		if (Tcl_SplitList(interp, argv[2], &ac, &av) != TCL_OK)
			return (TCL_ERROR);
		if (ac < 1) {
			Tcl_Free((char *)av);
			return (unitcl_setres(interp,
			    "not enough args in diagnostic"));
		}
		if (strcmp(av[0], "cond") == 0)
			ret = parse_diag_cond(interp, &ie->cause,
			    ac - 1, av + 1);
		else if (strcmp(av[0], "tns") == 0)
			ret = parse_diag_tns(interp, &ie->cause,
			    ac - 1, av + 1);
		else if (strcmp(av[0], "rej") == 0)
			ret = parse_diag_rej(interp, &ie->cause,
			    ac - 1, av + 1);
		else if (strcmp(av[0], "number") == 0)
			ret = parse_diag_number(interp, &ie->cause,
			    ac - 1, av + 1);
		else if (strcmp(av[0], "traffic") == 0)
			ret = parse_diag_traffic(interp, &ie->cause,
			    ac - 1, av + 1);
		else if (strcmp(av[0], "attr") == 0)
			ret = parse_diag_attr(interp, &ie->cause,
			    ac - 1, av + 1);
		else if (strcmp(av[0], "param") == 0)
			ret = parse_diag_param(interp, &ie->cause,
			    ac - 1, av + 1);
		else if (strcmp(av[0], "timer") == 0)
			ret = parse_diag_timer(interp, &ie->cause,
			    ac - 1, av + 1);
		else if (strcmp(av[0], "mtype") == 0)
			ret = parse_diag_mtype(interp, &ie->cause,
			    ac - 1, av + 1);
		else if (strcmp(av[0], "ie") == 0)
			ret = parse_diag_ie(interp, &ie->cause,
			    ac - 1, av + 1);
		else if (strcmp(av[0], "vpci") == 0)
			ret = parse_diag_vpci(interp, &ie->cause,
			    ac - 1, av + 1);
		else  {
			ret = TCL_ERROR;
			unitcl_setres(interp, "bad diagnostic '%s'", av[0]);
		}
		Tcl_Free((char *)av);
		if (ret != TCL_OK)
			return (TCL_ERROR);
	}

	IE_ADDPRESENT(ie->cause);

	return (TCL_OK);
}

FMT_IE_START(UNI_IE_CAUSE, cause)
{
	char buf[100];
	u_int i;

	sprintf(buf, "%u", ie->loc);
	Tcl_DStringAppendElement(str, buf);
	sprintf(buf, "%u", ie->cause);
	Tcl_DStringAppendElement(str, buf);

	if (ie->h.present & UNI_CAUSE_COND_P) {
 		/* {cond <pu> <na> <cond>} */
		Tcl_DStringStartSublist(str);
		Tcl_DStringAppendElement(str, "cond");
		sprintf(buf, "%u", ie->u.cond.pu);
		Tcl_DStringAppendElement(str, buf);
		sprintf(buf, "%u", ie->u.cond.na);
		Tcl_DStringAppendElement(str, buf);
		sprintf(buf, "%u", ie->u.cond.cond);
		Tcl_DStringAppendElement(str, buf);
		Tcl_DStringEndSublist(str);

	} else if (ie->h.present & UNI_CAUSE_REJ_P) {
 		/* {rej <reason> <cond> [<user>|<ie>]} */
		Tcl_DStringStartSublist(str);
		Tcl_DStringAppendElement(str, "rej");
		sprintf(buf, "%u", ie->u.rej.reason);
		Tcl_DStringAppendElement(str, buf);
		sprintf(buf, "%u", ie->u.rej.cond);
		Tcl_DStringAppendElement(str, buf);
		if (ie->h.present & UNI_CAUSE_REJ_USER_P) {
			sprintf(buf, "%u", ie->u.rej.user);
			Tcl_DStringAppendElement(str, buf);
		}
		if (ie->h.present & UNI_CAUSE_REJ_IE_P) {
			sprintf(buf, "%u", ie->u.rej.ie);
			Tcl_DStringAppendElement(str, buf);
		}
		Tcl_DStringEndSublist(str);

	} else if (ie->h.present & UNI_CAUSE_IE_P) {
 		/* {ie <ie> [...]} */
		Tcl_DStringStartSublist(str);
		Tcl_DStringAppendElement(str, "ie");
		for (i = 0; i < ie->u.ie.len; i++) {
			sprintf(buf, "%u", ie->u.ie.ie[i]);
			Tcl_DStringAppendElement(str, buf);
		}
		Tcl_DStringEndSublist(str);

	} else if (ie->h.present & UNI_CAUSE_TRAFFIC_P) {
 		/* {traffic <key> [...]} */
		Tcl_DStringStartSublist(str);
		Tcl_DStringAppendElement(str, "traffic");
		for (i = 0; i < ie->u.traffic.len; i++) {
			sprintf(buf, "%u", ie->u.traffic.traffic[i]);
			Tcl_DStringAppendElement(str, buf);
		}
		Tcl_DStringEndSublist(str);

	} else if (ie->h.present & UNI_CAUSE_VPCI_P) {
 		/* {vpci <vpci> <vci>} */
		Tcl_DStringStartSublist(str);
		Tcl_DStringAppendElement(str, "vpci");
		sprintf(buf, "%u", ie->u.vpci.vpci);
		Tcl_DStringAppendElement(str, buf);
		sprintf(buf, "%u", ie->u.vpci.vci);
		Tcl_DStringAppendElement(str, buf);
		Tcl_DStringEndSublist(str);

	} else if (ie->h.present & UNI_CAUSE_MTYPE_P) {
 		/* {mtype <mtype>} */
		Tcl_DStringStartSublist(str);
		Tcl_DStringAppendElement(str, "mtype");
		sprintf(buf, "%u", ie->u.mtype);
		Tcl_DStringAppendElement(str, buf);
		Tcl_DStringEndSublist(str);

	} else if (ie->h.present & UNI_CAUSE_TIMER_P) {
		/* {timer <timer>} */
		Tcl_DStringStartSublist(str);
		Tcl_DStringAppendElement(str, "timer");
		sprintf(buf, "\"%.3s\"", ie->u.timer);
		Tcl_DStringAppendElement(str, buf);
		Tcl_DStringEndSublist(str);

	} else if (ie->h.present & UNI_CAUSE_TNS_P) {
 		/* {tns <IE>} */
		Tcl_DStringStartSublist(str);
		Tcl_DStringAppendElement(str, "tns");
		if (fmt_tns(interp, str, &ie->u.tns) != TCL_OK)
			return (TCL_ERROR);
		Tcl_DStringEndSublist(str);

	} else if (ie->h.present & UNI_CAUSE_NUMBER_P) {
 		/* {number <IE>} */
		Tcl_DStringStartSublist(str);
		Tcl_DStringAppendElement(str, "number");
		if (fmt_called(interp, str, &ie->u.number) != TCL_OK)
			return (TCL_ERROR);
		Tcl_DStringEndSublist(str);

	} else if (ie->h.present & UNI_CAUSE_ATTR_P) {
 		/* {attr {NUM [NUM [NUM]]}+} */
		u_int j;

		Tcl_DStringStartSublist(str);
		Tcl_DStringAppendElement(str, "attr");
		for (i = 0; i < ie->u.attr.nattr; i++) {
			Tcl_DStringStartSublist(str);
			for (j = 0; j < 3; j++) {
				sprintf(buf, "0x%02x", ie->u.attr.attr[i][j]);
				Tcl_DStringAppendElement(str, buf);
				if (ie->u.attr.attr[i][j] & 0x80)
					break;
			}
			Tcl_DStringEndSublist(str);
		}
		Tcl_DStringEndSublist(str);

	} else if (ie->h.present & UNI_CAUSE_PARAM_P) {
		/* {param NUM} */
		Tcl_DStringStartSublist(str);
		Tcl_DStringAppendElement(str, "param");
		sprintf(buf, "0x%02x", ie->u.param);
		Tcl_DStringAppendElement(str, buf);
		Tcl_DStringEndSublist(str);
	}
}
FMT_IE_END()
