/*
 * Copyright (c) 2001-2003
 *	Fraunhofer Institute for Open Communication Systems (FhG Fokus).
 * 	All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Author: Hartmut Brandt <harti@freebsd.org>
 *
 * $Begemot: atmsupport/libunitcl/ie_crankback.c,v 1.9 2004/01/02 15:15:55 hbb Exp $
 */
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>

#include "inctcl.h"

#include <netnatm/unimsg.h>
#include <netnatm/msg/unistruct.h>

#include "unitcl.h"
#include "unitcl_priv.h"

REGISTER_IE(UNI_IE_CRANKBACK, crankback, 0)

/*
 * {crankback <level> <type>
 *	[{<level> <node-id>} | {<prec-level> <prec-id> <port> <succ-level> <succ-id>}]
 *	<cause>
 *	[{<dir> <port> <AvCR> [<CRM> <VF>]} | {<qos>}}]
 * type := if | node | link
 * qos := ctd | cdv | clr | other
 * dir := f | b
 */
int
parse_crankback(Tcl_Interp *interp, int argc, const char **argv,
    union uni_ieall *ie)
{
	u_int n;
	int ac, i;
	const char **av;

	if(argc < 2) {
		unitcl_setres(interp, "not enough args for crankback");
		return TCL_ERROR;
	}
	if(unitcl_parse_num(interp, argv[0], &n) != TCL_OK)
		return TCL_ERROR;
	ie->crankback.level = n;
	argc--;
	argv++;

	if(strcmp(argv[0], "if") == 0) {
		ie->crankback.type = UNI_CRANKBACK_IF;
		argc--;
		argv++;
	} else if(strcmp(argv[0], "node") == 0) {
		ie->crankback.type = UNI_CRANKBACK_NODE;
		argc--;
		argv++;
		if(argc == 0) {
			unitcl_setres(interp, "missing node id for crankback");
			return TCL_ERROR;
		}
		if(Tcl_SplitList(interp, argv[0], &ac, &av) != TCL_OK)
			return TCL_ERROR;
		if(ac != 2) {
			unitcl_setres(interp, "bad no of args for crankback node id");
			Tcl_Free((char *)av);
			return TCL_ERROR;
		}
		if(unitcl_parse_nodeid(interp, (const char **)av,
			&ie->crankback.id.node.level,
			ie->crankback.id.node.id) != TCL_OK) {
			Tcl_Free((char *)av);
			return TCL_ERROR;
		}
		Tcl_Free((char *)av);
		argc--;
		argv++;
	} else if(strcmp(argv[0], "link") == 0) {
		ie->crankback.type = UNI_CRANKBACK_LINK;
		argc--;
		argv++;
		if(argc == 0) {
			unitcl_setres(interp, "missing link id for crankback");
			return TCL_ERROR;
		}
		if(Tcl_SplitList(interp, argv[0], &ac, &av) != TCL_OK)
			return TCL_ERROR;
		if(ac != 5) {
			unitcl_setres(interp, "bad no of args for crankback link id");
			Tcl_Free((char *)av);
			return TCL_ERROR;
		}
		if(unitcl_parse_nodeid(interp, (const char **)av,
			&ie->crankback.id.link.plevel,
			ie->crankback.id.link.pid) != TCL_OK) {
			Tcl_Free((char *)av);
			return TCL_ERROR;
		}
		if(unitcl_parse_num(interp, av[2], &n)) {
			Tcl_Free((char *)av);
			return TCL_ERROR;
		}
		ie->crankback.id.link.port = n;
		if(unitcl_parse_nodeid(interp, (const char **)(av + 3),
			&ie->crankback.id.link.slevel,
			ie->crankback.id.link.sid) != TCL_OK) {
			Tcl_Free((char *)av);
			return TCL_ERROR;
		}
		Tcl_Free((char *)av);
		argc--;
		argv++;
	} else {
		unitcl_setres(interp, "back crankback type: %s", argv[0]);
		return TCL_ERROR;
	}
	if(argc == 0) {
		unitcl_setres(interp, "missing cause for crankback");
		return TCL_ERROR;
	}
	if(unitcl_parse_num(interp, argv[0], &n))
		return TCL_ERROR;
	ie->crankback.cause = n;
	argc--;
	argv++;


	switch(ie->crankback.cause) {

	  case UNI_CAUSE_CRATE_NAVL:
		if(argc > 0) {
			if(Tcl_SplitList(interp, argv[0], &ac, &av) != TCL_OK)
				return TCL_ERROR;
			if(ac != 3 && ac != 5) {
				unitcl_setres(interp, "bad # of args for crankback diag");
				Tcl_Free((char *)av);
				return TCL_ERROR;
			}
			if(strcmp(av[0], "f") == 0)
				ie->crankback.diag.top.dir = 0;
			else if(strcmp(av[0], "b") == 0)
				ie->crankback.diag.top.dir = 1;
			else {
				unitcl_setres(interp, "bad direction in cranback diag");
				Tcl_Free((char *)av);
				return TCL_ERROR;
			}
			if(unitcl_parse_num(interp, av[1], &n)) {
				Tcl_Free((char *)av);
				return TCL_ERROR;
			}
			ie->crankback.diag.top.port = n;

			if(unitcl_parse_num(interp, av[2], &n)) {
				Tcl_Free((char *)av);
				return TCL_ERROR;
			}
			ie->crankback.diag.top.avcr = n;
			ie->crankback.h.present |= UNI_CRANKBACK_TOP_P;

			if(ac == 5) {
				if(unitcl_parse_num(interp, av[3], &n)) {
					Tcl_Free((char *)av);
					return TCL_ERROR;
				}
				ie->crankback.diag.top.crm = n;

				if(unitcl_parse_num(interp, av[4], &n)) {
					Tcl_Free((char *)av);
					return TCL_ERROR;
				}
				ie->crankback.diag.top.vf = n;

				ie->crankback.h.present |= UNI_CRANKBACK_TOPX_P;
			}

			Tcl_Free((char *)av);
			argc--;
			argv++;
		}

		break;

	  case UNI_CAUSE_QOS_NAVL:
		if(argc > 0) {
			if(Tcl_SplitList(interp, argv[0], &ac, &av) != TCL_OK)
				return TCL_ERROR;

			for(i = 0; i < ac; i++) {
				if(strcmp(av[i], "ctd") == 0)
					ie->crankback.diag.qos.ctd = 1;
				else if(strcmp(av[i], "cdv") == 0)
					ie->crankback.diag.qos.cdv = 1;
				else if(strcmp(av[i], "clr") == 0)
					ie->crankback.diag.qos.clr = 1;
				else if(strcmp(av[i], "other") == 0)
					ie->crankback.diag.qos.other = 1;
				else {
					unitcl_setres(interp, "bad crankback QoS diag '%s'", av[i]);
					Tcl_Free((char *)av);
					return TCL_ERROR;
				}
			}
			ie->crankback.h.present |= UNI_CRANKBACK_QOS_P;

			Tcl_Free((char *)av);
			argc--;
			argv++;
		}

		break;
	}
	if(argc != 0) {
		unitcl_setres(interp, "excess args in crankback");
		return TCL_ERROR;
	}

	IE_ADDPRESENT(ie->crankback);
	return TCL_OK;
}

FMT_IE_START(UNI_IE_CRANKBACK, crankback)
{
	char buf[100];
	int j;

	sprintf(buf, "%d", ie->level);
	Tcl_DStringAppendElement(str, buf);

	switch(ie->type) {

	  case UNI_CRANKBACK_IF:
		Tcl_DStringAppendElement(str, "if");
		break;

	  case UNI_CRANKBACK_NODE:
		Tcl_DStringAppendElement(str, "node");
		Tcl_DStringStartSublist(str);

		sprintf(buf, "%d", ie->id.node.level);
		Tcl_DStringAppendElement(str, buf);

		for(j = 0; j < 21; j++)
			sprintf(buf+2*j, "%02x", ie->id.node.id[j]);
		Tcl_DStringAppendElement(str, buf);

		Tcl_DStringEndSublist(str);
		break;

	  case UNI_CRANKBACK_LINK:
		Tcl_DStringAppendElement(str, "link");
		Tcl_DStringStartSublist(str);

		sprintf(buf, "%d", ie->id.link.plevel);
		Tcl_DStringAppendElement(str, buf);

		for(j = 0; j < 21; j++)
			sprintf(buf+2*j, "%02x", ie->id.link.pid[j]);
		Tcl_DStringAppendElement(str, buf);

		sprintf(buf, "0x%d", ie->id.link.port);
		Tcl_DStringAppendElement(str, buf);

		sprintf(buf, "%d", ie->id.link.slevel);
		Tcl_DStringAppendElement(str, buf);

		for(j = 0; j < 21; j++)
			sprintf(buf+2*j, "%02x", ie->id.link.sid[j]);
		Tcl_DStringAppendElement(str, buf);

		Tcl_DStringEndSublist(str);
		break;

	  default:
		unitcl_setres(interp, "bad crankback type: %d", ie->type);
		return TCL_ERROR;
	}

	sprintf(buf, "%d", ie->cause);
	Tcl_DStringAppendElement(str, buf);

	if(ie->h.present & UNI_CRANKBACK_TOP_P) {
		Tcl_DStringStartSublist(str);

		switch(ie->diag.top.dir) {

		  case 0:
			Tcl_DStringAppendElement(str, "f");
			break;

		  case 1:
			Tcl_DStringAppendElement(str, "b");
			break;

		  default:
			unitcl_setres(interp, "bad crankback diag direction: %d", ie->diag.top.dir);
			return TCL_ERROR;
		}

		sprintf(buf, "0x%x", ie->diag.top.port);
		Tcl_DStringAppendElement(str, buf);

		sprintf(buf, "%u", ie->diag.top.avcr);
		Tcl_DStringAppendElement(str, buf);

		if(ie->h.present & UNI_CRANKBACK_TOPX_P) {
			sprintf(buf, "%u", ie->diag.top.crm);
			Tcl_DStringAppendElement(str, buf);

			sprintf(buf, "%u", ie->diag.top.vf);
			Tcl_DStringAppendElement(str, buf);
		}
		Tcl_DStringEndSublist(str);
	}
	if(ie->h.present & UNI_CRANKBACK_QOS_P) {
		Tcl_DStringStartSublist(str);

		if(ie->diag.qos.ctd)
			Tcl_DStringAppendElement(str, "ctd");
		if(ie->diag.qos.cdv)
			Tcl_DStringAppendElement(str, "cdv");
		if(ie->diag.qos.clr)
			Tcl_DStringAppendElement(str, "clr");
		if(ie->diag.qos.other)
			Tcl_DStringAppendElement(str, "other");

		Tcl_DStringEndSublist(str);
	}
}
FMT_IE_END()
