/*
 * Copyright (c) 2001-2003
 *	Fraunhofer Institute for Open Communication Systems (FhG Fokus).
 * 	All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Author: Hartmut Brandt <harti@freebsd.org>
 *
 * $Begemot: atmsupport/libunitcl/ie_git.c,v 1.8 2004/01/02 15:15:55 hbb Exp $
 */
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>

#include "inctcl.h"

#include <netnatm/unimsg.h>
#include <netnatm/msg/unistruct.h>

#include "unitcl.h"
#include "unitcl_priv.h"

REGISTER_IE(UNI_IE_GIT, git, 0)

/*
 * {git <std> [{<type> [<byte> ...]}]*}
 *
 * std := dsmcc | h254 | N
 * type := session | resource
 */
int
parse_git(Tcl_Interp *interp, int argc, const char **argv, union uni_ieall *ie)
{
	u_int n;
	int ac, i;
	const char **av;

	if(argc < 1) {
		unitcl_setres(interp, "bad # of args to git");
		return TCL_ERROR;
	}

	if(strcmp(argv[0], "dsmcc") == 0)
		ie->git.std = UNI_GIT_STD_DSMCC;
	else if(strcmp(argv[0], "h245") == 0)
		ie->git.std = UNI_GIT_STD_H245;
	else if(unitcl_parse_num(interp, argv[0], &n) != TCL_OK) {
		unitcl_setres(interp, "bad git standard value");
		return TCL_ERROR;
	}

	argc--; argv++;
	while(argc--) {
		if(ie->git.numsub >= UNI_GIT_MAXSUB) {
			unitcl_setres(interp, "too many git idents");
			return TCL_ERROR;
		}
		if(Tcl_SplitList(interp, *argv, &ac, &av) != TCL_OK)
			return TCL_ERROR;

		if(ac < 2 || ac > UNI_GIT_MAXVAL + 1) {
			unitcl_setres(interp, "bad # of args in git ident");
			Tcl_Free((char *)av);
			return TCL_ERROR;
		}
		if(strcmp(av[0], "session") == 0)
			ie->git.sub[ie->git.numsub].type = UNI_GIT_TYPE_SESS;
		else if(strcmp(av[0], "resource") == 0)
			ie->git.sub[ie->git.numsub].type = UNI_GIT_TYPE_RES;
		else {
			unitcl_setres(interp, "bad git type identifier");
			Tcl_Free((char *)av);
			return TCL_ERROR;
		}
		for(i = 1; i < ac; i++) {
			if(unitcl_parse_num(interp, av[i], &n) != TCL_OK) {
				Tcl_Free((char *)av);
				return TCL_ERROR;
			}
			ie->git.sub[ie->git.numsub].val[i-1] = n;
		}
		ie->git.sub[ie->git.numsub].len = ac - 1;
		ie->git.numsub++;
		Tcl_Free((char *)av);
	}

	IE_ADDPRESENT(ie->git);

	return TCL_OK;
}

FMT_IE_START(UNI_IE_GIT, git)
{
	char buf[100];
	u_int i, n;

	if(ie->std == UNI_GIT_STD_DSMCC)
		Tcl_DStringAppendElement(str, "dsmcc");
	else if(ie->std == UNI_GIT_STD_H245)
		Tcl_DStringAppendElement(str, "h245");
	else {
		sprintf(buf, "0x%02x", ie->std);
		Tcl_DStringAppendElement(str, buf);
	}

	for(n = 0; n < ie->numsub; n++) {
		Tcl_DStringStartSublist(str);
		if(ie->sub[n].type == UNI_GIT_TYPE_SESS)
			Tcl_DStringAppendElement(str, "session");
		else if(ie->sub[n].type == UNI_GIT_TYPE_RES)
			Tcl_DStringAppendElement(str, "resource");
		else {
			unitcl_setres(interp, "bad git type to format");
			return TCL_ERROR;
		}
		for(i = 0; i < ie->sub[n].len; i++) {
			sprintf(buf, "0x%02x", ie->sub[n].val[i]);
			Tcl_DStringAppendElement(str, buf);
		}
		Tcl_DStringEndSublist(str);
	}
}
FMT_IE_END()
