/*
 * Copyright (c) 1996-2003
 *	Fraunhofer Institute for Open Communication Systems (FhG Fokus).
 * 	All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Author: Hartmut Brandt <harti@freebsd.org>
 *
 * $Begemot: atmsupport/libunitcl/ie_traffic.c,v 1.8 2004/01/02 15:15:55 hbb Exp $
 */
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>

#include "inctcl.h"

#include <netnatm/unimsg.h>
#include <netnatm/msg/unistruct.h>

#include "unitcl.h"
#include "unitcl_priv.h"

REGISTER_IE(UNI_IE_TRAFFIC, traffic, 0)
REGISTER_IE(UNI_IE_MINTRAFFIC, mintraffic, 0)
REGISTER_IE(UNI_IE_ATRAFFIC, atraffic, 0)
REGISTER_IE(UNI_IE_MDCR, mdcr, 0)


static const struct toff {
	const char	*dir;
	const char	*name;
	const char	*clp;
	size_t		off;
	u_int		flag;
} traffic_off[] = {
  {"f", "pcr", "0", offsetof(struct uni_xtraffic, fpcr0), UNI_TRAFFIC_FPCR0_P },
  {"b", "pcr", "0", offsetof(struct uni_xtraffic, bpcr0), UNI_TRAFFIC_BPCR0_P },
  {"f", "pcr", "1", offsetof(struct uni_xtraffic, fpcr1), UNI_TRAFFIC_FPCR1_P },
  {"b", "pcr", "1", offsetof(struct uni_xtraffic, bpcr1), UNI_TRAFFIC_BPCR1_P },
  {"f", "scr", "0", offsetof(struct uni_xtraffic, fscr0), UNI_TRAFFIC_FSCR0_P },
  {"b", "scr", "0", offsetof(struct uni_xtraffic, bscr0), UNI_TRAFFIC_BSCR0_P },
  {"f", "scr", "1", offsetof(struct uni_xtraffic, fscr1), UNI_TRAFFIC_FSCR1_P },
  {"b", "scr", "1", offsetof(struct uni_xtraffic, bscr1), UNI_TRAFFIC_BSCR1_P },
  {"f", "mbs", "0", offsetof(struct uni_xtraffic, fmbs0), UNI_TRAFFIC_FMBS0_P },
  {"b", "mbs", "0", offsetof(struct uni_xtraffic, bmbs0), UNI_TRAFFIC_BMBS0_P },
  {"f", "mbs", "1", offsetof(struct uni_xtraffic, fmbs1), UNI_TRAFFIC_FMBS1_P },
  {"b", "mbs", "1", offsetof(struct uni_xtraffic, bmbs1), UNI_TRAFFIC_BMBS1_P },
  {"f", "abr", "1", offsetof(struct uni_xtraffic, fabr1), UNI_TRAFFIC_FABR1_P },
  {"b", "abr", "1", offsetof(struct uni_xtraffic, babr1), UNI_TRAFFIC_BABR1_P },
  { NULL, NULL, NULL, 0, 0 }
},
mintraffic_off[] = {
  {"f", "pcr", "0", offsetof(struct uni_ie_mintraffic,fpcr0),UNI_MINTRAFFIC_FPCR0_P},
  {"b", "pcr", "0", offsetof(struct uni_ie_mintraffic,bpcr0),UNI_MINTRAFFIC_BPCR0_P},
  {"f", "pcr", "1", offsetof(struct uni_ie_mintraffic,fpcr1),UNI_MINTRAFFIC_FPCR1_P},
  {"b", "pcr", "1", offsetof(struct uni_ie_mintraffic,bpcr1),UNI_MINTRAFFIC_BPCR1_P},
  {"f", "abr", "1", offsetof(struct uni_ie_mintraffic,fabr1),UNI_MINTRAFFIC_FABR1_P},
  {"b", "abr", "1", offsetof(struct uni_ie_mintraffic,babr1),UNI_MINTRAFFIC_BABR1_P},
  { NULL, NULL, NULL, 0, 0 }
};

/*
 * {traffic {<direction> {<param> <clp> <val>}*}* [be] [tf] [tb] [df] [db]
 * where:
 *	direction=f|b
 *	param=pcr|scr|mbs|abr
 *	clp=0|1
 */
static int
parse_cr(Tcl_Interp *interp, int argc, const char **argv, void *out,
	const struct toff *tab, u_int *pres)
{
	int ac;
	const char **av;
	const char *dir, *clp;
	u_int val, i;

	dir = *argv++;
	argc--;
	while(argc--) {
		if(Tcl_SplitList(interp, argv[0], &ac, &av) != TCL_OK)
			return TCL_ERROR;
		if(ac != 3) {
			unitcl_setres(interp, "bad number of fields in "
				"traffic subfield: %s", argv[0]);
			Tcl_Free((char *)av);
			return TCL_ERROR;
		}
		clp = av[1];
		if(unitcl_parse_num(interp, av[2], &val) != TCL_OK) {
			Tcl_Free((char *)av);
			return TCL_ERROR;
		}
		for(i = 0; tab[i].dir; i++) {
			if(strcmp(dir, tab[i].dir) == 0 &&
			   strcmp(clp, tab[i].clp) == 0 &&
			   strcmp(av[0], tab[i].name) == 0) {
				*(u_int *)(void *)((u_char *)out + tab[i].off) =
				    val;
				*pres |= tab[i].flag;
				break;
			}
		}
		if(tab[i].dir == NULL) {
			unitcl_setres(interp, "bad traffic subfield %s %s %s",
				dir, av[0], clp);
			Tcl_Free((char *)av);
			return TCL_ERROR;
		}
		Tcl_Free((char *)av);
		argv++;
	}
	return TCL_OK;
}

static int
parse_traffic_common(Tcl_Interp *interp, int argc, const char **argv,
	struct uni_xtraffic *t, u_int *present)
{
	int ac;
	const char **av;

	while(argc--) {
		if(strcmp(argv[0], "be") == 0) {
			*present |= UNI_TRAFFIC_BEST_P;
		} else if(strcmp(argv[0], "tf") == 0) {
			*present |= UNI_TRAFFIC_MOPT_P;
			t->ftag = 1;
		} else if(strcmp(argv[0], "tb") == 0) {
			*present |= UNI_TRAFFIC_MOPT_P;
			t->btag = 1;
		} else if(strcmp(argv[0], "df") == 0) {
			*present |= UNI_TRAFFIC_MOPT_P;
			t->fdisc = 1;
		} else if(strcmp(argv[0], "db") == 0) {
			*present |= UNI_TRAFFIC_MOPT_P;
			t->bdisc = 1;
		} else {
			if(Tcl_SplitList(interp, argv[0], &ac, &av) != TCL_OK)
				return TCL_ERROR;

			if(ac < 2) {
				unitcl_setres(interp, "bad # of parameters in "
					"traffic descriptor");
				Tcl_Free((char *)av);
				return TCL_ERROR;
			}

			if(parse_cr(interp, ac, av, t, traffic_off, present)) {
				Tcl_Free((char *)av);
				return TCL_ERROR;
			}
			Tcl_Free((char *)av);
		}
		argv++;
	}

	return TCL_OK;
}


int
parse_traffic(Tcl_Interp *interp, int argc, const char **argv,
    union uni_ieall *ie)
{
	int ret;

	ret = parse_traffic_common(interp, argc, argv, &ie->traffic.t,
	    &ie->traffic.h.present);
	if (ret == TCL_OK)
		IE_ADDPRESENT(ie->traffic);
	return (ret);
}
int
parse_atraffic(Tcl_Interp *interp, int argc, const char **argv,
    union uni_ieall *ie)
{
	int ret;

	ret = parse_traffic_common(interp, argc, argv, &ie->atraffic.t,
	    &ie->atraffic.h.present);
	if (ret == TCL_OK)
		IE_ADDPRESENT(ie->atraffic);
	return (ret);
}

/*
 * {mintraffic {<kind> <pcr>}+}
 * where: kind=f0|f1|b0|b1
 */
int
parse_mintraffic(Tcl_Interp *interp, int argc, const char **argv,
    union uni_ieall *ie)
{
	int ac;
	const char **av;

	while (argc--) {
		if (Tcl_SplitList(interp, argv[0], &ac, &av) != TCL_OK)
			return TCL_ERROR;

		if (ac != 2) {
			unitcl_setres(interp, "bad # of parameters in "
				"minimum traffic descriptor");
			Tcl_Free((char *)av);
			return (TCL_ERROR);
		}
		if (parse_cr(interp, ac, av, &ie->mintraffic, mintraffic_off,
		    &ie->mintraffic.h.present)) {
			Tcl_Free((char *)av);
			return (TCL_ERROR);
		}
		Tcl_Free((char *)av);
		argv++;
	}

	IE_ADDPRESENT(ie->mintraffic);

	return (TCL_OK);
}


static void
fmt_traffic_common(Tcl_DString *str, const struct uni_xtraffic *ie,
    u_int present)
{
	char buf[100];

#define FMT_SUBFIELD(L,S,field,FIELD,CLP)				\
	if (present & UNI_TRAFFIC_##L##FIELD##CLP##_P) {		\
		Tcl_DStringStartSublist(str);				\
		Tcl_DStringAppendElement(str, #field);			\
		Tcl_DStringAppendElement(str, #CLP);			\
		sprintf(buf, "%u", ie->S##field##CLP);			\
		Tcl_DStringAppendElement(str, buf);			\
		Tcl_DStringEndSublist(str);				\
	}								\

#define FMT_CR(L,S)							\
	Tcl_DStringStartSublist(str);					\
	Tcl_DStringAppendElement(str, #S);				\
	FMT_SUBFIELD(L,S,pcr,PCR, 0);					\
	FMT_SUBFIELD(L,S,scr,SCR, 0);					\
	FMT_SUBFIELD(L,S,mbs,MBS, 0);					\
	FMT_SUBFIELD(L,S,pcr,PCR, 1);					\
	FMT_SUBFIELD(L,S,scr,SCR, 1);					\
	FMT_SUBFIELD(L,S,mbs,MBS, 1);					\
	FMT_SUBFIELD(L,S,abr,ABR, 1);					\
	Tcl_DStringEndSublist(str);

	FMT_CR(F,f);
	FMT_CR(B,b);
#undef FMT_CR
#undef FMT_SUBFIELD

	if (present & UNI_TRAFFIC_BEST_P)
		Tcl_DStringAppendElement(str, "be");
	if (present & UNI_TRAFFIC_MOPT_P) {
		if (ie->ftag)
			Tcl_DStringAppendElement(str, "tf");
		if (ie->btag)
			Tcl_DStringAppendElement(str, "tb");
		if (ie->fdisc)
			Tcl_DStringAppendElement(str, "df");
		if (ie->bdisc)
			Tcl_DStringAppendElement(str, "db");
	}
}

FMT_IE_START(UNI_IE_TRAFFIC, traffic)
{
	fmt_traffic_common(str, &ie->t, ie->h.present);
}
FMT_IE_END()

FMT_IE_START(UNI_IE_ATRAFFIC, atraffic)
{
	fmt_traffic_common(str, &ie->t, ie->h.present);
}
FMT_IE_END()

FMT_IE_START(UNI_IE_MINTRAFFIC, mintraffic)
{
	char buf[100];

#define FMT_CR(L,S,N,TYPE,type)					\
	if (ie->h.present & UNI_MINTRAFFIC_##L##TYPE##N##_P) {		\
		Tcl_DStringStartSublist(str);				\
		Tcl_DStringAppendElement(str, #S #N);			\
		sprintf(buf, "%u", ie->S##type##N);			\
		Tcl_DStringAppendElement(str, buf);			\
		Tcl_DStringEndSublist(str);				\
	}
	FMT_CR(F,f,1,PCR,pcr);
	FMT_CR(B,b,1,PCR,pcr);
	FMT_CR(F,f,0,PCR,pcr);
	FMT_CR(B,b,0,PCR,pcr);
	FMT_CR(F,f,1,ABR,abr);
	FMT_CR(B,b,1,ABR,abr);
#undef FMT_CR
}
FMT_IE_END()

/*
 * {mdcr [user|net] <fmdcr> <bmdcr>}
 */
int
parse_mdcr(Tcl_Interp *interp, int argc, const char **argv,
   union uni_ieall *ie)
{
	u_int n;

	if (argc != 3) {
		unitcl_setres(interp, "mdcr needs three arguments");
		return TCL_ERROR;
	}
	if (strcmp(argv[0], "user") == 0)
		ie->mdcr.origin = UNI_MDCR_ORIGIN_USER;
	else if (strcmp(argv[0], "net") == 0)
		ie->mdcr.origin = UNI_MDCR_ORIGIN_NET;
	else {
		unitcl_setres(interp, "bad origin for mdcr");
		return (TCL_ERROR);
	}
	if (unitcl_parse_num(interp, argv[1], &n))
		return (TCL_ERROR);
	ie->mdcr.fmdcr = n;
	if (unitcl_parse_num(interp, argv[2], &n))
		return (TCL_ERROR);
	ie->mdcr.bmdcr = n;

	IE_ADDPRESENT(ie->mdcr);

	return (TCL_OK);
}

FMT_IE_START(UNI_IE_MDCR, mdcr)
{
	char buf[100];

	if (ie->origin == UNI_MDCR_ORIGIN_USER)
		Tcl_DStringAppendElement(str, "user");
	else if (ie->origin == UNI_MDCR_ORIGIN_NET)
		Tcl_DStringAppendElement(str, "net");
	else
		Tcl_DStringAppendElement(str, "error");
	sprintf(buf, "%u", ie->fmdcr);
	Tcl_DStringAppendElement(str, buf);
	sprintf(buf, "%u", ie->bmdcr);
	Tcl_DStringAppendElement(str, buf);
}
FMT_IE_END()
