/*
 * Copyright (c) 2003-2003
 *	Fraunhofer Institute for Open Communication Systems (FhG Fokus).
 * 	All rights reserved.
 *
 * Author: Hartmut Brandt <harti@freebsd.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Begemot: atmsupport/libunitcl/saal.c,v 1.5 2004/01/02 15:15:55 hbb Exp $
 */
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>

#include "inctcl.h"

#include <netnatm/unimsg.h>
#include <netnatm/saal/sscopdef.h>
#include <netnatm/saal/sscfudef.h>

#include "unitcl.h"
#include "unitcl_priv.h"

#define	PDU_BGN		1
#define	PDU_BGAK	2
#define	PDU_BGREJ	7
#define	PDU_END		3
#define	PDU_ENDAK	4
#define	PDU_RS		5
#define	PDU_RSAK	6
#define	PDU_ER		9
#define	PDU_ERAK	15
#define	PDU_SD		8
#define	PDU_POLL	10
#define	PDU_STAT	11
#define	PDU_USTAT	12
#define	PDU_UD		13
#define	PDU_MD		14

static const struct {
	const char *const name;
	int	data;
} sigs[] = {
	[SAAL_ESTABLISH_request] =	{ "saal-establish-request", 1 },
	[SAAL_ESTABLISH_indication] =	{ "saal-establish-indication", 1 },
	[SAAL_ESTABLISH_confirm] =	{ "saal-establish-confirm", 1 },
	[SAAL_RELEASE_request] =	{ "saal-release-request", 1 },
	[SAAL_RELEASE_confirm] =	{ "saal-release-confirm", 0 },
	[SAAL_RELEASE_indication] =	{ "saal-release-indication", 1 },
	[SAAL_DATA_request] =		{ "saal-data-request", 1 },
	[SAAL_DATA_indication] =	{ "saal-data-indication", 1 },
	[SAAL_UDATA_request] =		{ "saal-udata-request", 1 },
	[SAAL_UDATA_indication] =	{ "saal-udata-indication", 1 },
};

/*
 * {<elem>*}
 *
 * where elem is either:
 *  n:
 * or
 *  v
 */
static int
fmt_sscop_data(Tcl_DString *str, const struct uni_msg *msg,
    size_t len)
{
	u_int a;
	int z;
	char sbuf[100];

	Tcl_DStringStartSublist(str);

	z = 0;
	for (a = 0; a < len; a++) {
		if (msg->b_rptr[a] == 0) {
			z = 1;
		} else {
			if (z != 0) {
				sprintf(sbuf, "%u:", a);
				Tcl_DStringAppendElement(str, sbuf);
				z = 0;
			}
			sprintf(sbuf, "0x%02x", msg->b_rptr[a]);
			Tcl_DStringAppendElement(str, sbuf);
		}
	}
	if (z != 0) {
		sprintf(sbuf, "%u:", a);
		Tcl_DStringAppendElement(str, sbuf);
	}

	Tcl_DStringEndSublist(str);

	return (TCL_OK);
}

static int
parse_sscop_data1(Tcl_Interp *interp, struct uni_msg *msg,
    int argc, const char **argv)
{
	int i;
	u_long v;
	char *end;

	for (i = 0; i < argc; i++) {
		errno = 0;
		v = strtoul(argv[i], &end, 0);
		if (errno != 0)
			return (unitcl_setres(interp,
			    "bad data element '%s': %s", argv[i],
			    strerror(errno)));
		if (*end == '\0') {
			if (v > 0xff)
				return (unitcl_setres(interp,
				    "data element too large '%s'", argv[i]));
			if (uni_msg_ensure(msg, 1))
				return (unitcl_setres(interp,
				    "out of memory for sscop message"));
			*msg->b_wptr++ = v;
		} else if (end[0] == ':' && end[1] == '\0') {
			if (v > 0xffff)
				return (unitcl_setres(interp,
				    "data index too large '%s'", argv[i]));
			if (v < uni_msg_len(msg))
				return (unitcl_setres(interp,
				    "data address backward '%s'", argv[i]));
			if (uni_msg_ensure(msg, v - uni_msg_len(msg)))
				return (unitcl_setres(interp,
				    "out of memory for sscop message"));
			memset(msg->b_wptr, 0, v - uni_msg_len(msg));
			msg->b_wptr += v - uni_msg_len(msg);

		} else if (*end != '\0')
			return (unitcl_setres(interp,
			    "bad data element '%s'", argv[i]));
	}
	return (TCL_OK);
}

static int
parse_sscop_data(Tcl_Interp *interp, struct uni_msg *msg, const char *arg)
{
	int ac;
	int ret;
	const char **av;

	if (Tcl_SplitList(interp, arg, &ac, &av) != TCL_OK)
		return (TCL_ERROR);
	ret = parse_sscop_data1(interp, msg, ac, av);
	Tcl_Free((char *)av);
	return (ret);
}

static int
parse_sscop_data_opt(Tcl_Interp *interp, struct uni_msg **msg, const char *arg)
{

	if ((*msg = uni_msg_alloc(100)) == NULL)
		return (unitcl_setres(interp, "out of memory"));
	if (parse_sscop_data(interp, *msg, arg) != TCL_OK) {
		uni_msg_destroy(*msg);
		*msg = NULL;
		return (TCL_ERROR);
	}
	if (uni_msg_len(*msg) == 0) {
		uni_msg_destroy(*msg);
		*msg = NULL;
	}
	return (TCL_OK);
}

/*
 * Format an SAAL signal
 */
int
fmt_saal_sig(Tcl_Interp *interp, Tcl_DString *str, u_int sig,
    const struct uni_msg *msg)
{
	if (sig >= sizeof(sigs) / sizeof(sigs[0]) || sigs[sig].name == NULL)
		return (unitcl_setres(interp, "unknown signal %u", (u_int)sig));

	Tcl_DStringAppendElement(str, sigs[sig].name);
	if (sigs[sig].data) {
		if (msg == NULL) {
			Tcl_DStringAppendElement(str, "");
			return (TCL_OK);
		}
		return (fmt_sscop_data(str, msg, uni_msg_len(msg)));
	} else
		return (TCL_OK);
}

int
parse_saal_sig(Tcl_Interp *interp, u_int *sigp, int argc, const char *argv[],
    struct uni_msg *msg)
{

	if (argc < 1)
		return (unitcl_setres(interp, "%s: need arg", __func__));

	for (*sigp = 0; *sigp < sizeof(sigs) / sizeof(sigs[0]); (*sigp)++)
		if (sigs[*sigp].name != NULL &&
		    strcmp(sigs[*sigp].name, argv[0]) == 0)
			break;

	if (*sigp == sizeof(sigs) / sizeof(sigs[0]))
		return (unitcl_setres(interp, "unknown signal %s", argv[0]));

	if (sigs[*sigp].data) {
		if (argc != 2)
			return (unitcl_setres(interp, "%s: need data",
			    argv[0]));
		return (parse_sscop_data(interp, msg, argv[1]));
	} else {
		if (argc != 1)
			return (unitcl_setres(interp, "%s: excess args",
			    argv[0]));
		return (TCL_OK);
	}
}

/*
 * {bgn <nmr> <nsq> {<data>}}
 * {bgak <nmr> {<data>}}
 * {end <source> {<data>}}
 * {endak}
 * {rs <nmr> <nsq> {<data>}}
 * {rsak <nmr>}
 * {bgrej <data>}
 * {sd <ns> {<data>}}
 * {er <nmr> <nsq>}
 * {poll <ns> <nps>}
 * {stat <nr> <nmr> <nps> <le>*}
 * {ustat <nr> <nmr> <le1> <le2>}
 * {ud {<data>}}
 * {md {<data>}}
 * {erak <nmr>}
 */
int
fmt_sscop(Tcl_Interp *interp, Tcl_DString *str, const struct uni_msg *msg)
{
	size_t len;
	uint32_t w;
	u_int pad;
	char sbuf[100];
	u_int i;

	len = uni_msg_len(msg);
	if (len < 4)
		return (unitcl_setres(interp, "sscop message too short (%zu)",
		    len));
	if (len % 4 != 0)
		return (unitcl_setres(interp, "sscop message not aligned (%zu)",
		    len));

	w = uni_msg_trail32(msg, -1);
	pad = (w >> 30) & 0x3;

	switch ((w >> 24) & 0x0f) {

	  case 0:	/* illegal */
	  default:
		return (unitcl_setres(interp, "bad sscop message type ",
		    (w >> 24) & 0xf));

	  case PDU_BGN:
		if (len < 8)
			return (unitcl_setres(interp, "BGN too short %zu",
			    len));
		if ((w & 0x30000000) != 0)
			return (unitcl_setres(interp, "BGN Rsvd1 not 0"));
		Tcl_DStringAppendElement(str, "bgn");
		sprintf(sbuf, "%u", (u_int)(w & 0xffffff));
		Tcl_DStringAppendElement(str, sbuf);
		w = uni_msg_trail32(msg, -2);
		if ((w & 0xffffff00) != 0)
			return (unitcl_setres(interp, "BGN Rsvd2 not 0"));
		sprintf(sbuf, "%u", (u_int)(w & 0xff));
		Tcl_DStringAppendElement(str, sbuf);
		return (fmt_sscop_data(str, msg, len - 8 - pad));

	  case PDU_BGAK:
		if (len < 8)
			return (unitcl_setres(interp, "BGAK too short %zu",
			    len));
		if ((w & 0x30000000) != 0)
			return (unitcl_setres(interp, "BGAK Rsvd1 not 0"));
		Tcl_DStringAppendElement(str, "bgak");
		sprintf(sbuf, "%u", (u_int)(w & 0xffffff));
		Tcl_DStringAppendElement(str, sbuf);
		w = uni_msg_trail32(msg, -2);
		if (w != 0)
			return (unitcl_setres(interp, "BGAK Rsvd2 not 0"));
		return (fmt_sscop_data(str, msg, len - 8 - pad));

	  case PDU_END:
		if (len < 8)
			return (unitcl_setres(interp, "END too short %zu",
			    len));
		if ((w & 0x20000000) != 0)
			return (unitcl_setres(interp, "END Rsvd1 not 0"));
		if ((w & 0x00ffffff) != 0)
			return (unitcl_setres(interp, "END Rsvd2 not 0"));
		Tcl_DStringAppendElement(str, "end");
		if (w & 0x10000000)
			Tcl_DStringAppendElement(str, "sscop");
		else
			Tcl_DStringAppendElement(str, "user");
		w = uni_msg_trail32(msg, -2);
		if (w != 0)
			return (unitcl_setres(interp, "END Rsvd3 not 0"));
		return (fmt_sscop_data(str, msg, len - 8 - pad));

	  case PDU_ENDAK:
		if (len != 8)
			return (unitcl_setres(interp, "ENDAK bad length %zu",
			    len));
		if ((w & 0xf0000000) != 0)
			return (unitcl_setres(interp, "ENDAK Rsvd1 not 0"));
		if ((w & 0x00ffffff) != 0)
			return (unitcl_setres(interp, "ENDAK Rsvd2 not 0"));
		Tcl_DStringAppendElement(str, "endak");
		w = uni_msg_trail32(msg, -2);
		if (w != 0)
			return (unitcl_setres(interp, "ENDAK Rsvd3 not 0"));
		return (TCL_OK);

	  case PDU_RS:
		if (len < 8)
			return (unitcl_setres(interp, "RS too short %zu",
			    len));
		if ((w & 0x30000000) != 0)
			return (unitcl_setres(interp, "RS Rsvd1 not 0"));
		Tcl_DStringAppendElement(str, "rs");
		sprintf(sbuf, "%u", (u_int)(w & 0xffffff));
		Tcl_DStringAppendElement(str, sbuf);
		w = uni_msg_trail32(msg, -2);
		if ((w & 0xffffff00) != 0)
			return (unitcl_setres(interp, "RS Rsvd2 not 0"));
		sprintf(sbuf, "%u", (u_int)(w & 0xff));
		Tcl_DStringAppendElement(str, sbuf);
		return (fmt_sscop_data(str, msg, len - 8 - pad));

	  case PDU_RSAK:
		if (len != 8)
			return (unitcl_setres(interp, "RSAK bad length %zu",
			    len));
		if ((w & 0x30000000) != 0)
			return (unitcl_setres(interp, "RSAK Rsvd1 not 0"));
		Tcl_DStringAppendElement(str, "rsak");
		sprintf(sbuf, "%u", (u_int)(w & 0xffffff));
		Tcl_DStringAppendElement(str, sbuf);
		w = uni_msg_trail32(msg, -2);
		if (w != 0)
			return (unitcl_setres(interp, "RSAK Rsvd2 not 0"));
		return (TCL_OK);

	  case PDU_BGREJ:
		if (len < 8)
			return (unitcl_setres(interp, "BGREJ too short %zu",
			    len));
		if ((w & 0x30000000) != 0)
			return (unitcl_setres(interp, "BGREJ Rsvd1 not 0"));
		if ((w & 0x00ffffff) != 0)
			return (unitcl_setres(interp, "BGREJ Rsvd2 not 0"));
		Tcl_DStringAppendElement(str, "bgrej");
		w = uni_msg_trail32(msg, -2);
		if (w != 0)
			return (unitcl_setres(interp, "BGREJ Rsvd3 not 0"));
		return (fmt_sscop_data(str, msg, len - 8 - pad));

	  case PDU_SD:
		if ((w & 0x30000000) != 0)
			return (unitcl_setres(interp, "SD Rsvd1 not 0"));
		Tcl_DStringAppendElement(str, "sd");
		sprintf(sbuf, "%u", (u_int)(w & 0xffffff));
		Tcl_DStringAppendElement(str, sbuf);
		return (fmt_sscop_data(str, msg, len - 4 - pad));

	  case PDU_ER:
		if (len != 8)
			return (unitcl_setres(interp, "ER bad length %zu",
			    len));
		if ((w & 0xf0000000) != 0)
			return (unitcl_setres(interp, "ER Rsvd1 not 0"));
		Tcl_DStringAppendElement(str, "er");
		sprintf(sbuf, "%u", (u_int)(w & 0xffffff));
		Tcl_DStringAppendElement(str, sbuf);
		w = uni_msg_trail32(msg, -2);
		if ((w & 0xffffff00) != 0)
			return (unitcl_setres(interp, "ER Rsvd2 not 0"));
		sprintf(sbuf, "%u", (u_int)(w & 0xff));
		Tcl_DStringAppendElement(str, sbuf);
		return (TCL_OK);

	  case PDU_POLL:
		if (len != 8)
			return (unitcl_setres(interp, "POLL bad length %zu",
			    len));
		if ((w & 0xf0000000) != 0)
			return (unitcl_setres(interp, "POLL Rsvd1 not 0"));
		Tcl_DStringAppendElement(str, "poll");
		sprintf(sbuf, "%u", (u_int)(w & 0xffffff));
		Tcl_DStringAppendElement(str, sbuf);
		w = uni_msg_trail32(msg, -2);
		if ((w & 0xff000000) != 0)
			return (unitcl_setres(interp, "POLL Rsvd2 not 0"));
		sprintf(sbuf, "%u", (u_int)(w & 0xffffff));
		Tcl_DStringAppendElement(str, sbuf);
		return (TCL_OK);

	  case PDU_STAT:
		if (len < 12)
			return (unitcl_setres(interp, "STAT too short %zu",
			    len));

		if ((w & 0xf0000000) != 0)
			return (unitcl_setres(interp, "STAT Rsvd1 not 0"));
		Tcl_DStringAppendElement(str, "stat");
		sprintf(sbuf, "%u", (u_int)(w & 0xffffff));
		Tcl_DStringAppendElement(str, sbuf);

		w = uni_msg_trail32(msg, -2);
		if ((w & 0xf0000000) != 0)
			return (unitcl_setres(interp, "STAT Rsvd2 not 0"));
		sprintf(sbuf, "%u", (u_int)(w & 0xffffff));
		Tcl_DStringAppendElement(str, sbuf);

		w = uni_msg_trail32(msg, -3);
		if ((w & 0xf0000000) != 0)
			return (unitcl_setres(interp, "STAT Rsvd3 not 0"));
		sprintf(sbuf, "%u", (u_int)(w & 0xffffff));
		Tcl_DStringAppendElement(str, sbuf);

		for (i = 0; i < (len - 12) / 4; i++) {
			w = uni_msg_trail32(msg, - (len / 4) + i);
			sprintf(sbuf, "%u", (u_int)(w & 0xffffff));
			Tcl_DStringAppendElement(str, sbuf);
		}
		return (TCL_OK);

	  case PDU_USTAT:
		if (len != 16)
			return (unitcl_setres(interp, "USTAT bad length %zu",
			    len));

		if ((w & 0xf0000000) != 0)
			return (unitcl_setres(interp, "USTAT Rsvd1 not 0"));
		Tcl_DStringAppendElement(str, "ustat");
		sprintf(sbuf, "%u", (u_int)(w & 0xffffff));
		Tcl_DStringAppendElement(str, sbuf);

		w = uni_msg_trail32(msg, -2);
		if ((w & 0xf0000000) != 0)
			return (unitcl_setres(interp, "USTAT Rsvd2 not 0"));
		sprintf(sbuf, "%u", (u_int)(w & 0xffffff));
		Tcl_DStringAppendElement(str, sbuf);

		w = uni_msg_trail32(msg, -4);
		sprintf(sbuf, "%u", (u_int)(w & 0xffffff));
		Tcl_DStringAppendElement(str, sbuf);

		w = uni_msg_trail32(msg, -3);
		sprintf(sbuf, "%u", (u_int)(w & 0xffffff));
		Tcl_DStringAppendElement(str, sbuf);
		return (TCL_OK);

	  case PDU_UD:
		if ((w & 0x30000000) != 0)
			return (unitcl_setres(interp, "UD Rsvd1 not 0"));
		if ((w & 0x00ffffff) != 0)
			return (unitcl_setres(interp, "UD Rsvd2 not 0"));
		Tcl_DStringAppendElement(str, "ud");
		return (fmt_sscop_data(str, msg, len - 4 - pad));

	  case PDU_MD:
		if ((w & 0x30000000) != 0)
			return (unitcl_setres(interp, "MD Rsvd1 not 0"));
		if ((w & 0x00ffffff) != 0)
			return (unitcl_setres(interp, "MD Rsvd2 not 0"));
		Tcl_DStringAppendElement(str, "md");
		return (fmt_sscop_data(str, msg, len - 4 - pad));

	  case PDU_ERAK:
		if (len != 8)
			return (unitcl_setres(interp, "ERAK bad length %zu",
			    len));
		if ((w & 0x30000000) != 0)
			return (unitcl_setres(interp, "ERAK Rsvd1 not 0"));
		Tcl_DStringAppendElement(str, "erak");
		sprintf(sbuf, "%u", (u_int)(w & 0xffffff));
		Tcl_DStringAppendElement(str, sbuf);
		w = uni_msg_trail32(msg, -2);
		if (w != 0)
			return (unitcl_setres(interp, "ERAK Rsvd2 not 0"));
		return (TCL_OK);
	}
}

static int
parse_sscop_num(Tcl_Interp *interp, const char *arg, u_int *n,
    u_int maxval, const char *w)
{
	if (unitcl_parse_num(interp, arg, n) != TCL_OK)
		return (TCL_ERROR);
	if (*n > maxval)
		return (unitcl_setres(interp, "%s too large: '%s'", w, arg));
	return (TCL_OK);
}

int
parse_sscop(Tcl_Interp *interp, struct uni_msg *msg,
    int argc, const char **argv)
{
	u_int n1, n2, n3, n4, pad, i;
	int a;

	if (argc < 1)
		return (unitcl_setres(interp, "no args for sscop message"));

	if (strcmp(argv[0], "bgn") == 0) {
		if (argc != 4)
			return (unitcl_setres(interp, "need 3 args for bgn"));
		if (parse_sscop_num(interp, argv[1], &n1, 0xffffff,
		    "bng N(MR)") != TCL_OK)
			return (TCL_ERROR);
		if (parse_sscop_num(interp, argv[2], &n2, 0xff,
		    "bng N(SQ)") != TCL_OK)
			return (TCL_ERROR);

		if (parse_sscop_data(interp, msg, argv[3]) != TCL_OK)
			return (TCL_ERROR);
		if ((pad = 4 - uni_msg_len(msg) % 4) == 4)
			pad = 0;
		if (uni_msg_ensure(msg, pad + 8))
			return (unitcl_setres(interp, "out of memory"));

		for (i = 0; i < pad; i++)
			uni_msg_append8(msg, 0);
		uni_msg_append32(msg, n2);
		uni_msg_append32(msg, (pad << 30) | (PDU_BGN << 24) | n1);

	} else if (strcmp(argv[0], "bgak") == 0) {
		if (argc != 3)
			return (unitcl_setres(interp, "need 2 args for bgak"));
		if (parse_sscop_num(interp, argv[1], &n1, 0xffffff,
		    "bgak N(MR)") != TCL_OK)
			return (TCL_ERROR);

		if (parse_sscop_data(interp, msg, argv[2]) != TCL_OK)
			return (TCL_ERROR);
		if ((pad = 4 - uni_msg_len(msg) % 4) == 4)
			pad = 0;
		if (uni_msg_ensure(msg, pad + 8))
			return (unitcl_setres(interp, "out of memory"));

		for (i = 0; i < pad; i++)
			uni_msg_append8(msg, 0);
		uni_msg_append32(msg, 0);
		uni_msg_append32(msg, (pad << 30) | (PDU_BGAK << 24) | n1);

	} else if (strcmp(argv[0], "end") == 0) {
		if (argc != 3)
			return (unitcl_setres(interp, "need 2 args for end"));
		if (strcmp(argv[1], "sscop") == 0)
			n1 = 1;
		else if (strcmp(argv[1], "user") == 0)
			n1 = 0;
		else
			return (unitcl_setres(interp, "bad END source '%s'",
			    argv[1]));

		if (parse_sscop_data(interp, msg, argv[2]) != TCL_OK)
			return (TCL_ERROR);
		if ((pad = 4 - uni_msg_len(msg) % 4) == 4)
			pad = 0;
		if (uni_msg_ensure(msg, pad + 8))
			return (unitcl_setres(interp, "out of memory"));

		for (i = 0; i < pad; i++)
			uni_msg_append8(msg, 0);
		uni_msg_append32(msg, 0);
		uni_msg_append32(msg, (pad << 30) | (PDU_END << 24) | (n1 << 28));

	} else if (strcmp(argv[0], "endak") == 0) {
		if (argc != 1)
			return (unitcl_setres(interp, "need no args for endak"));

		if (uni_msg_ensure(msg, 8))
			return (unitcl_setres(interp, "out of memory"));
		uni_msg_append32(msg, 0);
		uni_msg_append32(msg, (PDU_ENDAK << 24));

	} else if (strcmp(argv[0], "rs") == 0) {
		if (argc != 4)
			return (unitcl_setres(interp, "need 3 args for rs"));
		if (parse_sscop_num(interp, argv[1], &n1, 0xffffff,
		    "rs N(MR)") != TCL_OK)
			return (TCL_ERROR);
		if (parse_sscop_num(interp, argv[2], &n2, 0xff,
		    "rs N(SQ)") != TCL_OK)
			return (TCL_ERROR);

		if (parse_sscop_data(interp, msg, argv[3]) != TCL_OK)
			return (TCL_ERROR);
		if ((pad = 4 - uni_msg_len(msg) % 4) == 4)
			pad = 0;
		if (uni_msg_ensure(msg, pad + 8))
			return (unitcl_setres(interp, "out of memory"));

		for (i = 0; i < pad; i++)
			uni_msg_append8(msg, 0);
		uni_msg_append32(msg, n2);
		uni_msg_append32(msg, (pad << 30) | (PDU_RS << 24) | n1);

	} else if (strcmp(argv[0], "rsak") == 0) {
		if (argc != 2)
			return (unitcl_setres(interp, "need 1 arg for rsak"));
		if (parse_sscop_num(interp, argv[1], &n1, 0xffffff,
		    "rsak N(MR)") != TCL_OK)
			return (TCL_ERROR);

		if (uni_msg_ensure(msg, 8))
			return (unitcl_setres(interp, "out of memory"));

		uni_msg_append32(msg, 0);
		uni_msg_append32(msg, (PDU_RSAK << 24) | n1);

	} else if (strcmp(argv[0], "bgrej") == 0) {
		if (argc != 2)
			return (unitcl_setres(interp, "need 1 arg for bgrej"));

		if (parse_sscop_data(interp, msg, argv[1]) != TCL_OK)
			return (TCL_ERROR);
		if ((pad = 4 - uni_msg_len(msg) % 4) == 4)
			pad = 0;
		if (uni_msg_ensure(msg, pad + 8))
			return (unitcl_setres(interp, "out of memory"));

		for (i = 0; i < pad; i++)
			uni_msg_append8(msg, 0);
		uni_msg_append32(msg, 0);
		uni_msg_append32(msg, (pad << 30) | (PDU_BGREJ << 24));

	} else if (strcmp(argv[0], "sd") == 0) {
		if (argc != 3)
			return (unitcl_setres(interp, "need 2 args for sd"));
		if (parse_sscop_num(interp, argv[1], &n1, 0xffffff,
		    "sd N(S)") != TCL_OK)
			return (TCL_ERROR);

		if (parse_sscop_data(interp, msg, argv[2]) != TCL_OK)
			return (TCL_ERROR);
		if ((pad = 4 - uni_msg_len(msg) % 4) == 4)
			pad = 0;
		if (uni_msg_ensure(msg, pad + 8))
			return (unitcl_setres(interp, "out of memory"));

		for (i = 0; i < pad; i++)
			uni_msg_append8(msg, 0);
		uni_msg_append32(msg, (pad << 30) | (PDU_SD << 24) | n1);

	} else if (strcmp(argv[0], "er") == 0) {
		if (argc != 3)
			return (unitcl_setres(interp, "need 2 args for er"));
		if (parse_sscop_num(interp, argv[1], &n1, 0xffffff,
		    "er N(MR)") != TCL_OK)
			return (TCL_ERROR);
		if (parse_sscop_num(interp, argv[2], &n2, 0xff,
		    "er N(SQ)") != TCL_OK)
			return (TCL_ERROR);

		if (uni_msg_ensure(msg, 8))
			return (unitcl_setres(interp, "out of memory"));

		uni_msg_append32(msg, n2);
		uni_msg_append32(msg, (PDU_ER << 24) | n1);

	} else if (strcmp(argv[0], "poll") == 0) {
		if (argc != 3)
			return (unitcl_setres(interp, "need 2 args for poll"));
		if (parse_sscop_num(interp, argv[1], &n1, 0xffffff,
		    "poll N(S)") != TCL_OK)
			return (TCL_ERROR);
		if (parse_sscop_num(interp, argv[2], &n2, 0xffffff,
		    "poll N(PS)") != TCL_OK)
			return (TCL_ERROR);

		if (uni_msg_ensure(msg, 8))
			return (unitcl_setres(interp, "out of memory"));

		uni_msg_append32(msg, n2);
		uni_msg_append32(msg, (PDU_POLL << 24) | n1);

	} else if (strcmp(argv[0], "stat") == 0) {
		if (argc < 4)
			return (unitcl_setres(interp, "need >2 args for stat"));
		if (parse_sscop_num(interp, argv[1], &n1, 0xffffff,
		    "stat N(R)") != TCL_OK)
			return (TCL_ERROR);
		if (parse_sscop_num(interp, argv[2], &n2, 0xffffff,
		    "stat N(MR)") != TCL_OK)
			return (TCL_ERROR);
		if (parse_sscop_num(interp, argv[3], &n3, 0xffffff,
		    "stat N(PS)") != TCL_OK)
			return (TCL_ERROR);

		if (uni_msg_ensure(msg, 12U + (argc - 4) * 4))
			return (unitcl_setres(interp, "out of memory"));

		for (a = 4; a < argc; a++) {
			if (parse_sscop_num(interp, argv[a], &n4, 0xffffff,
			    "stat LE") != TCL_OK)
				return (TCL_ERROR);
			uni_msg_append32(msg, n4);
		}
		uni_msg_append32(msg, n3);
		uni_msg_append32(msg, n2);
		uni_msg_append32(msg, (PDU_STAT << 24) | n1);

	} else if (strcmp(argv[0], "ustat") == 0) {
		if (argc != 5)
			return (unitcl_setres(interp, "need 4 args for ustat"));
		if (parse_sscop_num(interp, argv[1], &n1, 0xffffff,
		    "ustat N(R)") != TCL_OK)
			return (TCL_ERROR);
		if (parse_sscop_num(interp, argv[2], &n2, 0xffffff,
		    "ustat N(MR)") != TCL_OK)
			return (TCL_ERROR);
		if (parse_sscop_num(interp, argv[3], &n3, 0xffffff,
		    "ustat LE1") != TCL_OK)
			return (TCL_ERROR);
		if (parse_sscop_num(interp, argv[4], &n4, 0xffffff,
		    "ustat LE2") != TCL_OK)
			return (TCL_ERROR);

		if (uni_msg_ensure(msg, 16))
			return (unitcl_setres(interp, "out of memory"));

		uni_msg_append32(msg, n3);
		uni_msg_append32(msg, n4);
		uni_msg_append32(msg, n2);
		uni_msg_append32(msg, (PDU_USTAT << 24) | n1);

	} else if (strcmp(argv[0], "ud") == 0) {
		if (argc != 2)
			return (unitcl_setres(interp, "need 1 arg for ud"));

		if (parse_sscop_data(interp, msg, argv[1]) != TCL_OK)
			return (TCL_ERROR);
		if ((pad = 4 - uni_msg_len(msg) % 4) == 4)
			pad = 0;
		if (uni_msg_ensure(msg, pad + 4))
			return (unitcl_setres(interp, "out of memory"));

		for (i = 0; i < pad; i++)
			uni_msg_append8(msg, 0);
		uni_msg_append32(msg, (pad << 30) | (PDU_UD << 24));

	} else if (strcmp(argv[0], "md") == 0) {
		if (argc != 2)
			return (unitcl_setres(interp, "need 1 arg for md"));

		if (parse_sscop_data(interp, msg, argv[1]) != TCL_OK)
			return (TCL_ERROR);
		if ((pad = 4 - uni_msg_len(msg) % 4) == 4)
			pad = 0;
		if (uni_msg_ensure(msg, pad + 4))
			return (unitcl_setres(interp, "out of memory"));

		for (i = 0; i < pad; i++)
			uni_msg_append8(msg, 0);
		uni_msg_append32(msg, (pad << 30) | (PDU_MD << 24));

	} else if (strcmp(argv[0], "erak") == 0) {
		if (argc != 2)
			return (unitcl_setres(interp, "need 1 arg for bgn"));
		if (parse_sscop_num(interp, argv[1], &n1, 0xffffff,
		    "erak N(MR)") != TCL_OK)
			return (TCL_ERROR);

		if (uni_msg_ensure(msg, 8))
			return (unitcl_setres(interp, "out of memory"));

		uni_msg_append32(msg, 0);
		uni_msg_append32(msg, (PDU_ERAK << 24) | n1);

	} else if (strcmp(argv[0], "invalid") == 0) {
		if (argc != 2)
			return (unitcl_setres(interp, "need 1 arg for invalid"));
		return (parse_sscop_data(interp, msg, argv[1]) != TCL_OK);

	} else
		return (unitcl_setres(interp, "bad sscop type '%s'", argv[0]));

	return (TCL_OK);
}

/*
 *
 */
int
fmt_sscop_sig(Tcl_Interp *interp, Tcl_DString *str, u_int sig, u_int arg,
    const struct uni_msg *msg)
{
	char sbuf[100];

	switch (sig) {

	  case SSCOP_ESTABLISH_request:		/* <- UU, BR */
		Tcl_DStringAppendElement(str, "sscop-establish-request");
		sprintf(sbuf, "%u", arg != 0 ? 1 : 0);
		Tcl_DStringAppendElement(str, sbuf);
		return (fmt_sscop_data(str, msg,
		    msg != NULL ? uni_msg_len(msg) : 0));

	  case SSCOP_ESTABLISH_indication:	/* -> UU */
		Tcl_DStringAppendElement(str, "sscop-establish-indication");
		return (fmt_sscop_data(str, msg,
		    msg != NULL ? uni_msg_len(msg) : 0));

	  case SSCOP_ESTABLISH_response:	/* <- UU, BR */
		Tcl_DStringAppendElement(str, "sscop-establish-response");
		sprintf(sbuf, "%u", arg != 0 ? 1 : 0);
		Tcl_DStringAppendElement(str, sbuf);
		return (fmt_sscop_data(str, msg,
		    msg != NULL ? uni_msg_len(msg) : 0));

	  case SSCOP_ESTABLISH_confirm:		/* -> UU */
		Tcl_DStringAppendElement(str, "sscop-establish-confirm");
		return (fmt_sscop_data(str, msg,
		    msg != NULL ? uni_msg_len(msg) : 0));

	  case SSCOP_RELEASE_request:		/* <- UU */
		Tcl_DStringAppendElement(str, "sscop-release-request");
		return (fmt_sscop_data(str, msg,
		    msg != NULL ? uni_msg_len(msg) : 0));

	  case SSCOP_RELEASE_indication:	/* -> UU, SRC */
		Tcl_DStringAppendElement(str, "sscop-release-indication");
		if (arg != 0)
			Tcl_DStringAppendElement(str, "sscop");
		else
			Tcl_DStringAppendElement(str, "user");
		return (fmt_sscop_data(str, msg,
		    msg != NULL ? uni_msg_len(msg) : 0));

	  case SSCOP_RELEASE_confirm:		/* -> */
		Tcl_DStringAppendElement(str, "sscop-release-confirm");
		return (TCL_OK);

	  case SSCOP_DATA_request:		/* <- MU */
		Tcl_DStringAppendElement(str, "sscop-data-request");
		return (fmt_sscop_data(str, msg, uni_msg_len(msg)));

	  case SSCOP_DATA_indication:		/* -> MU, SN */
		Tcl_DStringAppendElement(str, "sscop-data-indication");
		sprintf(sbuf, "%u", arg);
		Tcl_DStringAppendElement(str, sbuf);
		return (fmt_sscop_data(str, msg, uni_msg_len(msg)));

	  case SSCOP_UDATA_request:		/* <- MU */
		Tcl_DStringAppendElement(str, "sscop-udata-request");
		return (fmt_sscop_data(str, msg, uni_msg_len(msg)));

	  case SSCOP_UDATA_indication:		/* -> MU */
		Tcl_DStringAppendElement(str, "sscop-udata-indication");
		return (fmt_sscop_data(str, msg, uni_msg_len(msg)));

	  case SSCOP_RECOVER_indication:	/* -> */
		Tcl_DStringAppendElement(str, "sscop-recover-indication");
		return (TCL_OK);

	  case SSCOP_RECOVER_response:		/* <- */
		Tcl_DStringAppendElement(str, "sscop-recover-response");
		return (TCL_OK);

	  case SSCOP_RESYNC_request:		/* <- UU */
		Tcl_DStringAppendElement(str, "sscop-resync-request");
		return (fmt_sscop_data(str, msg,
		    msg != NULL ? uni_msg_len(msg) : 0));

	  case SSCOP_RESYNC_indication:		/* -> UU */
		Tcl_DStringAppendElement(str, "sscop-resync-indication");
		return (fmt_sscop_data(str, msg,
		    msg != NULL ? uni_msg_len(msg) : 0));

	  case SSCOP_RESYNC_response:		/* <- */
		Tcl_DStringAppendElement(str, "sscop-resync-response");
		return (TCL_OK);

	  case SSCOP_RESYNC_confirm:		/* -> */
		Tcl_DStringAppendElement(str, "sscop-resync-confirm");
		return (TCL_OK);

	  case SSCOP_RETRIEVE_request:		/* <- RN */
		Tcl_DStringAppendElement(str, "sscop-retrieve-request");
		if (arg == SSCOP_RETRIEVE_UNKNOWN)
			Tcl_DStringAppendElement(str, "unknown");
		else if (arg == SSCOP_RETRIEVE_TOTAL)
			Tcl_DStringAppendElement(str, "total");
		else if (arg > SSCOP_MAXSEQNO)
			return (unitcl_setres(interp, "bad seqno 0x%x", arg));
		else {
			sprintf(sbuf, "%u", arg);
			Tcl_DStringAppendElement(str, sbuf);
		}
		return (TCL_OK);

	  case SSCOP_RETRIEVE_indication:	/* -> MU */
		Tcl_DStringAppendElement(str, "sscop-retrieve-indication");
		return (fmt_sscop_data(str, msg, uni_msg_len(msg)));

	  case SSCOP_RETRIEVE_COMPL_indication:/* -> */
		Tcl_DStringAppendElement(str, "sscop-retrieve-complete-indication");
		return (TCL_OK);

	  default:
		return (unitcl_setres(interp, "bad sscop signal %u", sig));
	}
	return (TCL_OK);
}

int
parse_sscop_sig(Tcl_Interp *interp, u_int *sig, u_int *arg,
    struct uni_msg **msg, int argc, const char **argv)
{
	u_int n;

	if (argc < 1)
		return (unitcl_setres(interp, "no args for sscop-sig"));

	if (strcmp(argv[0], "sscop-establish-request") == 0) {
		if (argc != 3)
			return (unitcl_setres(interp, "%s needs 2 args",
			    argv[0]));
		*sig = SSCOP_ESTABLISH_request;
		if (unitcl_parse_num(interp, argv[1], &n) != TCL_OK)
			return (TCL_ERROR);
		*arg = (n != 0);
		return (parse_sscop_data_opt(interp, msg, argv[2]));

	} else if (strcmp(argv[0], "sscop-establish-indication") == 0) {
		if (argc != 2)
			return (unitcl_setres(interp, "%s needs 1 arg",
			    argv[0]));
		*sig = SSCOP_ESTABLISH_indication;
		return (parse_sscop_data_opt(interp, msg, argv[1]));

	} else if (strcmp(argv[0], "sscop-establish-response") == 0) {
		if (argc != 3)
			return (unitcl_setres(interp, "%s needs 2 args",
			    argv[0]));
		*sig = SSCOP_ESTABLISH_response;
		if (unitcl_parse_num(interp, argv[1], &n) != TCL_OK)
			return (TCL_ERROR);
		*arg = (n != 0);
		return (parse_sscop_data_opt(interp, msg, argv[2]));

	} else if (strcmp(argv[0], "sscop-establish-confirm") == 0) {
		if (argc != 2)
			return (unitcl_setres(interp, "%s needs 1 arg",
			    argv[0]));
		*sig = SSCOP_ESTABLISH_confirm;
		return (parse_sscop_data_opt(interp, msg, argv[1]));

	} else if (strcmp(argv[0], "sscop-release-request") == 0) {
		if (argc != 2)
			return (unitcl_setres(interp, "%s needs 1 arg",
			    argv[0]));
		*sig = SSCOP_RELEASE_request;
		return (parse_sscop_data_opt(interp, msg, argv[1]));

	} else if (strcmp(argv[0], "sscop-release-indication") == 0) {
		if (argc != 3)
			return (unitcl_setres(interp, "%s needs 2 args",
			    argv[0]));
		*sig = SSCOP_RELEASE_indication;
		if (strcmp(argv[1], "sscop") == 0)
			*arg = 1;
		else if (strcmp(argv[1], "user") == 0)
			*arg = 0;
		else
			return (unitcl_setres(interp, "bad release source %s",
			    argv[1]));
		return (parse_sscop_data_opt(interp, msg, argv[2]));

	} else if (strcmp(argv[0], "sscop-release-confirm") == 0) {
		if (argc != 1)
			return (unitcl_setres(interp, "%s needs no args",
			    argv[0]));
		*sig = SSCOP_RELEASE_confirm;
		*msg = NULL;
		return (TCL_OK);

	} else if (strcmp(argv[0], "sscop-data-request") == 0) {
		*sig = SSCOP_DATA_request;

  pdata:
		if (argc != 2)
			return (unitcl_setres(interp, "%s needs 1 arg",
			    argv[0]));
		if ((*msg = uni_msg_alloc(100)) == NULL)
			return (unitcl_setres(interp, "out of memory"));
		if (parse_sscop_data(interp, *msg, argv[1]) != TCL_OK) {
			uni_msg_destroy(*msg);
			*msg = NULL;
			return (TCL_ERROR);
		}
		return (TCL_OK);

	} else if (strcmp(argv[0], "sscop-data-indication") == 0) {
		if (argc != 3)
			return (unitcl_setres(interp, "%s needs 2 args",
			    argv[0]));
		*sig = SSCOP_DATA_indication;
		if (unitcl_parse_num(interp, argv[1], arg) != TCL_OK)
			return (TCL_ERROR);

		if ((*msg = uni_msg_alloc(100)) == NULL)
			return (unitcl_setres(interp, "out of memory"));
		if (parse_sscop_data(interp, *msg, argv[2]) != TCL_OK) {
			uni_msg_destroy(*msg);
			*msg = NULL;
			return (TCL_ERROR);
		}
		return (TCL_OK);

	} else if (strcmp(argv[0], "sscop-udata-request") == 0) {
		*sig = SSCOP_UDATA_request;
		goto pdata;

	} else if (strcmp(argv[0], "sscop-udata-indication") == 0) {
		*sig = SSCOP_UDATA_indication;
		goto pdata;

	} else if (strcmp(argv[0], "sscop-recover-indication") == 0) {
		if (argc != 1)
			return (unitcl_setres(interp, "%s needs no args",
			    argv[0]));
		*sig = SSCOP_RECOVER_indication;
		*msg = NULL;
		return (TCL_OK);

	} else if (strcmp(argv[0], "sscop-recover-response") == 0) {
		if (argc != 1)
			return (unitcl_setres(interp, "%s needs no args",
			    argv[0]));
		*sig = SSCOP_RECOVER_response;
		*msg = NULL;
		return (TCL_OK);

	} else if (strcmp(argv[0], "sscop-resync-request") == 0) {
		if (argc != 2)
			return (unitcl_setres(interp, "%s needs 1 arg",
			    argv[0]));
		*sig = SSCOP_RESYNC_request;
		return (parse_sscop_data_opt(interp, msg, argv[1]));

	} else if (strcmp(argv[0], "sscop-resync-indication") == 0) {
		if (argc != 2)
			return (unitcl_setres(interp, "%s needs 1 arg",
			    argv[0]));
		*sig = SSCOP_RESYNC_indication;
		return (parse_sscop_data_opt(interp, msg, argv[1]));

	} else if (strcmp(argv[0], "sscop-resync-response") == 0) {
		if (argc != 1)
			return (unitcl_setres(interp, "%s needs no args",
			    argv[0]));
		*sig = SSCOP_RESYNC_response;
		*msg = NULL;
		return (TCL_OK);

	} else if (strcmp(argv[0], "sscop-resync-confirm") == 0) {
		if (argc != 1)
			return (unitcl_setres(interp, "%s needs no args",
			    argv[0]));
		*sig = SSCOP_RESYNC_confirm;
		*msg = NULL;
		return (TCL_OK);

	} else if (strcmp(argv[0], "sscop-retrieve-request") == 0) {
		if (argc != 2)
			return (unitcl_setres(interp, "%s needs 1 arg",
			    argv[0]));
		*sig = SSCOP_RETRIEVE_request;
		*msg = NULL;
		if (strcmp(argv[1], "unknown") == 0) {
			*arg = SSCOP_RETRIEVE_UNKNOWN;
			return (TCL_OK);
		} else if (strcmp(argv[1], "total") == 0) {
			*arg = SSCOP_RETRIEVE_TOTAL;
			return (TCL_OK);
		} else
			return (unitcl_parse_num(interp, argv[1], arg));

	} else if (strcmp(argv[0], "sscop-retrieve-indication") == 0) {
		*sig = SSCOP_RETRIEVE_indication;
		goto pdata;

	} else if (strcmp(argv[0], "sscop-retrieve-complete-indication") == 0) {
		if (argc != 1)
			return (unitcl_setres(interp, "%s needs no args",
			    argv[0]));
		*sig = SSCOP_RETRIEVE_COMPL_indication;
		*msg = NULL;
		return (TCL_OK);

	} else
		return (unitcl_setres(interp,
		    "bad sscop signal '%s'", argv[0]));
}

/*
 *
 */
int
fmt_sscop_msig(Tcl_Interp *interp, Tcl_DString *str, u_int sig, u_int arg1,
    u_int arg2, const struct uni_msg *msg)
{
	char sbuf[100];

	switch (sig) {

	  case SSCOP_MDATA_request:		/* <- MU */
		Tcl_DStringAppendElement(str, "sscop-mdata-request");
		return (fmt_sscop_data(str, msg, uni_msg_len(msg)));

	  case SSCOP_MDATA_indication:		/* -> MU */
		Tcl_DStringAppendElement(str, "sscop-mdata-indication");
		return (fmt_sscop_data(str, msg, uni_msg_len(msg)));

	  case SSCOP_MERROR_indication:		/* CODE, CNT */
		if (arg1 < 'A' && arg1 > 'Z')
			return (unitcl_setres(interp, "bad error code 0x%02x",
			    arg1));
		Tcl_DStringAppendElement(str, "sscop-merror-indication");
		sprintf(sbuf, "%c", arg1);
		Tcl_DStringAppendElement(str, sbuf);
		sprintf(sbuf, "%u", arg2);
		Tcl_DStringAppendElement(str, sbuf);
		return (TCL_OK);

	  default:
		return (unitcl_setres(interp, "bad sscop msignal %u", sig));
	}
}

int
parse_sscop_msig(Tcl_Interp *interp, u_int *sig, u_int *arg1, u_int *arg2,
    struct uni_msg **msg, int argc, const char **argv)
{

	if (argc < 1)
		return (unitcl_setres(interp, "no args for sscop-msig"));

	if (strcmp(argv[0], "sscop-mdata-request") == 0) {
		*sig = SSCOP_MDATA_request;

  pdata:
		if (argc != 2)
			return (unitcl_setres(interp, "%s needs 1 arg",
			    argv[0]));
		if ((*msg = uni_msg_alloc(100)) == NULL)
			return (unitcl_setres(interp, "out of memory"));
		if (parse_sscop_data(interp, *msg, argv[1]) != TCL_OK) {
			uni_msg_destroy(*msg);
			*msg = NULL;
			return (TCL_ERROR);
		}
		return (TCL_OK);

	} else if (strcmp(argv[0], "sscop-mdata-indication") == 0) {
		*sig = SSCOP_MDATA_indication;
		goto pdata;

	} else if (strcmp(argv[0], "sscop-merror-indication") == 0) {
		if (argc != 3)
			return (unitcl_setres(interp, "%s needs 2 args",
			    argv[0]));
		*sig = SSCOP_MERROR_indication;

		if (strlen(argv[1]) != 1 || argv[1][0] < 'A' ||
		    argv[1][0] > 'Z')
			return (unitcl_setres(interp, "bad error code '%s'",
			    argv[1]));
		*arg1 = argv[1][0];

		*msg = NULL;

		return (unitcl_parse_num(interp, argv[2], arg2));

	} else
		return (unitcl_setres(interp,
		    "bad sscop msignal '%s'", argv[0]));
}
