/*
 * Copyright (c) 1996-2003
 *	Fraunhofer Institute for Open Communication Systems (FhG Fokus).
 * 	All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Author: Hartmut Brandt <harti@freebsd.org>
 *
 * $Begemot: atmsupport/libunitcl/support.c,v 1.15 2004/05/14 09:26:53 novo Exp $
 */
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>

#include "inctcl.h"

#include <netnatm/unimsg.h>
#include <netnatm/msg/unistruct.h>
#include <netnatm/msg/unimsglib.h>
#include <netnatm/api/unisap.h>

#include "unitcl.h"
#include "unitcl_priv.h"

#define MAX_TCLERROR 1024

static const struct iedecl *ietable[256];
static const struct msgdecl *msgtable[257];


static const struct valtab {
	u_int	val;
	const char *name;
}
coding_tab[] = {
	{ UNI_CODING_ITU,	"itut" },
	{ UNI_CODING_NET,	"atmf" },
	{ 0, NULL }
},
ieact_tab[] = {
	{ UNI_IEACT_CLEAR,	"clear" },
	{ UNI_IEACT_IGNORE,	"ignore" },
	{ UNI_IEACT_REPORT,	"report" },
	{ UNI_IEACT_MSG_IGNORE,	"msg-ignore" },
	{ UNI_IEACT_MSG_REPORT,	"msg-report" },
	{ UNI_IEACT_DEFAULT,	"default" },
	{ 0, NULL }
},
msgact_tab[] = {
	{ UNI_MSGACT_CLEAR,	"clear" },
	{ UNI_MSGACT_IGNORE,	"ignore" },
	{ UNI_MSGACT_REPORT,	"report" },
	{ UNI_MSGACT_DEFAULT,	"default" },
	{ 0, NULL }
};


static __inline int
app_tab(Tcl_DString *str, const struct valtab *tab, u_int val)
{
	while(tab->name) {
		if(tab->val == val) {
			Tcl_DStringAppendElement(str, tab->name);
			return 0;
		}
		tab++;
	}
	return -1;
}

void
unitcl_register_ie(u_int ietype, const struct iedecl *decl)
{
	ietable[ietype] = decl;
}

void
unitcl_register_msg(u_int type, const struct msgdecl *decl)
{
	msgtable[type] = decl;
}

/*
 * <type>
 * {<type> [<coding> [<action> [error | absent]]]}
 */
static int
parse_iehdr(Tcl_Interp *interp, const char *arg, struct uni_iehdr *hdr,
    u_int *type)
{
	u_int i;
	int ac;
	const char **av;

	if (Tcl_SplitList(interp, arg, &ac, &av) != TCL_OK)
		return (TCL_ERROR);

	if (ac < 1 || ac > 4) {
		Tcl_Free((char *)av);
		return (unitcl_setres(interp,
		    "bad number of elements in IE hdr (%d)", ac));
	}
	for (i = 0; i < 256; i++)
		if (ietable[i] && strcmp(av[0], ietable[i]->name) == 0)
			break;

	if (i == 256) {
		unitcl_setres(interp, "bad IE type '%s'", av[0]);
		Tcl_Free((char *)av);
		return (TCL_ERROR);
	}

	*type = i;
	hdr->coding = UNI_CODING_ITU;
	hdr->act = UNI_IEACT_DEFAULT;

	if (ac == 1) {
		Tcl_Free((char *)av);
		return (TCL_OK);
	}

	for (i = 0; coding_tab[i].name; i++)
		if (strcmp(coding_tab[i].name, av[1]) == 0)
			break;
	if (coding_tab[i].name == NULL) {
		unitcl_setres(interp, "bad IE coding '%s'", av[1]);
		Tcl_Free((char *)av);
		return (TCL_ERROR);
	}
	hdr->coding = coding_tab[i].val;

	if (ac == 2) {
		Tcl_Free((char *)av);
		return (TCL_OK);
	}

	for (i = 0; ieact_tab[i].name; i++)
		if (strcmp(ieact_tab[i].name, av[2]) == 0)
			break;
	if (ieact_tab[i].name == NULL) {
		unitcl_setres(interp, "bad IE action '%s'", av[2]);
		Tcl_Free((char *)av);
		return (TCL_ERROR);
	}
	hdr->act =ieact_tab[i].val;

	if (ac == 3) {
		Tcl_Free((char *)av);
		return (TCL_OK);
	}

	if (strcmp(av[3], "error") == 0) {
		hdr->present = UNI_IE_ERROR | UNI_IE_PRESENT;
		Tcl_Free((char *)av);
		return (TCL_OK);
	}
	if (strcmp(av[3], "absent") == 0) {
		hdr->present = 0;
		Tcl_Free((char *)av);
		return (TCL_OK);
	}
	unitcl_setres(interp, "4th element in IE hdr must be error or absent");
	Tcl_Free((char *)av);
	return (TCL_ERROR);
}

int
parse_ie(Tcl_Interp *interp, const char *arg, union uni_ieall *ie,
    u_int *type)
{
	int argc, ret;
	const char **argv;

	memset(ie, 0, sizeof(union uni_ieall));

	if (Tcl_SplitList(interp, arg, &argc, &argv) != TCL_OK)
		return (TCL_ERROR);

	if (argc < 1) {
		Tcl_Free((char *)argv);
		return (unitcl_setres(interp,
		    "bad number of elements in IE (%d)", argc));
	}

	IE_SETEMPTY(ie->aal);

	if (parse_iehdr(interp, argv[0], &ie->h, type) != TCL_OK) {
		Tcl_Free((char *)argv);
		return (TCL_ERROR);
	}

	if (argc == 1) {
		Tcl_Free((char *)argv);
		return (TCL_OK);
	}
	ie->aal.h.present &= ~UNI_IE_EMPTY;

	ret = (*ietable[*type]->parse)(interp, argc - 1, argv + 1, ie);

	Tcl_Free((char *)argv);

	return (ret);
}

int
unitcl_parse_num(Tcl_Interp *interp, const char *arg, u_int *num)
{
	char *end;

	errno = 0;
	*num = strtoul(arg, &end, 0);
	if (errno != 0 || *end != '\0' || end == arg) {
		unitcl_setres(interp, "bad number: '%s'", arg);
		return (TCL_ERROR);
	}
	return (TCL_OK);
}

int
unitcl_parse_nodeid(Tcl_Interp *interp, const char **args, u_char *level, u_char *id)
{
	u_int n;
	int j;

	if (unitcl_parse_num(interp, *args, &n))
		return (TCL_ERROR);
	*level = n;

	args++;

	if (strlen(*args) != 42 ||
	    strspn(*args, "0123456789abcdefABCDEF") != 42) {
		unitcl_setres(interp, "need 42 hex digits for node id");
		return (TCL_ERROR);
	}
	for (j = 0; j < 21; j++) {
		(void)sscanf(*args + 2 * j, "%2x", &n);
		*id++ = n;
	}

	return (TCL_OK);
}

int
unitcl_fmt_iehdr(Tcl_Interp *interp, Tcl_DString *str, u_int type,
    const struct uni_iehdr *hdr)
{

	if (type >= 256 || ietable[type] == NULL)
		return (unitcl_setres(interp, "Bad IE type '%u'", type));

	/*
	 * Optimize output for the common case.
	 */
	if (IE_ISGOOD(*(const struct uni_ie_aal *)hdr) &&
	    hdr->coding == UNI_CODING_ITU &&
	    hdr->act == UNI_IEACT_DEFAULT) {
		Tcl_DStringAppendElement(str, ietable[type]->name);
		return (TCL_OK);
	}

	Tcl_DStringStartSublist(str);
	Tcl_DStringAppendElement(str, ietable[type]->name);

	if (app_tab(str, coding_tab, hdr->coding))
		return (unitcl_setres(interp,
		    "bad IE coding 0x%x", hdr->coding));
	if (app_tab(str, ieact_tab, hdr->act))
		return (unitcl_setres(interp,
		    "bad IE action 0x%x", hdr->act));

	if (IE_ISEMPTY(*(const struct uni_ie_aal *)hdr)) {
		Tcl_DStringEndSublist(str);
		return (TCL_CONTINUE);
	} else if (IE_ISERROR(*(const struct uni_ie_aal *)hdr)) {
		Tcl_DStringAppendElement(str, "error");
		Tcl_DStringEndSublist(str);
		return (TCL_CONTINUE);
	} else if (IE_ISPRESENT(*(const struct uni_ie_aal *)hdr)) {
		Tcl_DStringEndSublist(str);
		return (TCL_OK);
	} else if (hdr->present == 0) {
		Tcl_DStringAppendElement(str, "absent");
		Tcl_DStringEndSublist(str);
		return (TCL_CONTINUE);
	}
	return (unitcl_setres(interp, "bad IE flags: 0x%x", hdr->present));
}

#define MAXREC 20
static int
substitute_vars1(Tcl_Interp *interp, const char *arg, Tcl_DString *str)
{
	Tcl_DString inp;
	const char *ptr;
	const char *end;
	const char *p;
	int didit, rec;

	Tcl_DStringInit(&inp);
	Tcl_DStringAppend(&inp, arg, -1);
	rec = 0;
	while(1) {
		ptr = Tcl_DStringValue(&inp);
		Tcl_DStringInit(str);
		didit = 0;
		while(*ptr != '\0') {
			end = ptr;
			while(*end != '\0') {
				if(*end == '\\') {
					end++;
					if(*end != '\0')
						end++;
					continue;
				} else if(*end == '$')
					break;
				end++;
			}
			if(end > ptr) {
				Tcl_DStringAppend(str, ptr, end - ptr);
				ptr = end;
			}
			if(*ptr == '\0')
				break;

			didit = 1;
			if((p = Tcl_ParseVar(interp, ptr, &end)) == NULL) {
				Tcl_DStringFree(&inp);
				Tcl_DStringFree(str);
				return TCL_ERROR;
			}

			Tcl_DStringAppend(str, p, -1);

			ptr = end;
		}
		Tcl_DStringFree(&inp);

		if(didit == 0)
			break;

		if(rec >= MAXREC) {
			unitcl_setres(interp, "too many recursions");
			Tcl_DStringFree(str);
			return TCL_ERROR;
		}

		/* make inp a copy of str and free str */
		Tcl_DStringInit(&inp);
		Tcl_DStringAppend(&inp, Tcl_DStringValue(str), -1);
		Tcl_DStringFree(str);
	}

	return TCL_OK;
}

char *
substitute_vars(Tcl_Interp *interp, const char *arg)
{
	const char *sub;
	char *res;
	Tcl_DString str;

	if ((sub = Tcl_GetVar(interp, "uni_substitute", 0)) == NULL)
		sub = Tcl_GetVar(interp, "uni_substitute", TCL_GLOBAL_ONLY);

	if (sub == NULL || *sub == '\0') {
		if((res = malloc(strlen(arg) + 1)) == NULL)
			unitclbug("out of memory");
		return (strcpy(res, arg));
	}

	if (substitute_vars1(interp, arg, &str) != TCL_OK)
		return (NULL);

	if ((res = malloc(strlen(Tcl_DStringValue(&str)) + 1)) == NULL)
		unitclbug("out of memory");
	strcpy(res, Tcl_DStringValue(&str));
	Tcl_DStringFree(&str);

	return (res);
}

/*
 * Format a message. This does not format the header part.
 */
int
fmt_msg(Tcl_Interp *interp, Tcl_DString *str, const struct uni_all *m)
{
	if(m->mtype >= 256 || msgtable[m->mtype] == NULL) {
		unitcl_setres(interp, "bad message to format %x", m->mtype);
		return TCL_ERROR;
	}

	return (*msgtable[m->mtype]->fmt)(interp, str, &m->u);
}

/*
 * Parse a message. Each argument is assumed to be an information element.
 * This does not parse the header.
 */
int
parse_msg(Tcl_Interp *interp, int argc, const char **argv, struct uni_all *msg,
    struct unicx *cx)
{

	if (msg->mtype >= 256 || msgtable[msg->mtype] == NULL) {
		unitcl_setres(interp, "%s: bad message type %x", __func__,
		    msg->mtype);
		return (TCL_ERROR);
	}

	return (*msgtable[msg->mtype]->parse)(interp, argc, argv, &msg->u, cx);
}

/*
 * Format the message type.
 */
int
fmt_mtype(Tcl_Interp *interp, Tcl_DString *str, u_int mtype)
{

	if (mtype >= 256 || msgtable[mtype] == NULL) {
		unitcl_setres(interp, "bad message type to format %x", mtype);
		return (TCL_ERROR);
	}

	Tcl_DStringAppendElement(str, msgtable[mtype]->name);

	return (TCL_OK);
}

/*
 * Parse the message type
 */
int
parse_mtype(Tcl_Interp *interp, const char *arg, u_int *mtype)
{
	u_int i;

	for (i = 0; i < 256; i++)
		if (msgtable[i] && strcmp(arg, msgtable[i]->name) == 0) {
			*mtype = i;
			return (TCL_OK);
		}

	unitcl_setres(interp, "bad MSG type '%s'", arg);
	return (TCL_ERROR);
}

/*
 * Format the message action indicator
 */
int
fmt_msgact(Tcl_Interp *interp, Tcl_DString *str, u_int act)
{

	if (app_tab(str, msgact_tab, act)) {
		unitcl_setres(interp, "bad msg act %u", (u_int)act);
		return (TCL_ERROR);
	}
	return (TCL_OK);
}

/*
 * Parse message action indicator
 */
int
parse_msgact(Tcl_Interp *interp, const char *arg, struct uni_msghdr *hdr)
{
	u_int i;

	for (i = 0; msgact_tab[i].name; i++)
		if (strcmp(msgact_tab[i].name, arg) == 0) {
			hdr->act = msgact_tab[i].val;
			return (TCL_OK);
		}
	unitcl_setres(interp, "bad message action indicator '%s'", arg);
	return (TCL_ERROR);
}



/*
 * Format a message header. The format is either:
 *
 *	<mtype> <cref> <cref-flag>
 * or
 *
 *	{<mtype> <act> <pass>} <cref> <cref-flag>
 *
 * where <pass> is either "pass" or "nopass". The first form is used when
 * the action indicator is default and pass is "nopass".
 */
int
fmt_msghdr(Tcl_Interp *interp, Tcl_DString *str, u_int mtype,
    const struct uni_msghdr *h)
{

	if (msgtable[mtype] == NULL) {
		unitcl_setres(interp, "bad message type %d", mtype);
		return (TCL_ERROR);
	}
	if (h->act == UNI_MSGACT_DEFAULT && !h->pass) {
		/* short cut for default/nopass */
		Tcl_DStringAppendElement(str, msgtable[mtype]->name);
	} else {
		Tcl_DStringStartSublist(str);
		Tcl_DStringAppendElement(str, msgtable[mtype]->name);
		if (fmt_msgact(interp, str, h->act) != TCL_OK)
			return (TCL_ERROR);
		if (h->pass)
			Tcl_DStringAppendElement(str, "pass");
		else
			Tcl_DStringAppendElement(str, "nopass");
		Tcl_DStringEndSublist(str);
	}
	fmt_cref(str, &h->cref);

	return (TCL_OK);
}

/*
 * Parse a message header. The recognized formats are:
 *
 *	<mtype> <cref> <cref-flag>
 *	{<mtype> [<act> [<pass>]]} <cref> <cref-flag>
 *
 * In any case this uses three arguments.
 */
int
parse_msghdr(Tcl_Interp *interp, int argc, const char **argv,
    u_int *mtype, struct uni_msghdr *h)
{
	int ac;
	const char **av;

	if (argc < 3) {
		unitcl_setres(interp, "not enough args for msghdr %d", argc);
		return (TCL_ERROR);
	}
	if (Tcl_SplitList(interp, argv[0], &ac, &av) != TCL_OK)
		return (TCL_ERROR);

	if (ac < 1 || ac > 3) {
		unitcl_setres(interp, "bad # of elements in MSG hdr (%d)", ac);
		goto err;
	}
	if (parse_mtype(interp, av[0], mtype) != TCL_OK)
		goto err;

	h->act = UNI_MSGACT_DEFAULT;
	h->pass = 0;

	if (ac > 1) {
		if (parse_msgact(interp, av[1], h) != TCL_OK)
			goto err;
		if (ac > 2) {
			if (strcmp(av[2], "pass") == 0)
				h->pass = 1;
			else if (strcmp(av[2], "nopass") == 0)
				h->pass = 0;
			else {
				unitcl_setres(interp, "bad pass flag '%s'",
				    av[2]);
				goto err;
			}
		}
	}
	Tcl_Free((char *)av);

	return (parse_cref(interp, argc - 1, argv + 1, &h->cref));

  err:
	Tcl_Free((char *)av);
	return (TCL_ERROR);
}

/*
 * Format a call reference
 */
void
fmt_cref(Tcl_DString *str, const struct uni_cref *cref)
{
	char buf[100];

	if (cref->cref == CREF_GLOBAL)
		Tcl_DStringAppendElement(str, "global");
	else if (cref->cref == CREF_DUMMY)
		Tcl_DStringAppendElement(str, "dummy");
	else {
		sprintf(buf, "%u", cref->cref);
		Tcl_DStringAppendElement(str, buf);
	}

	if (cref->flag)
		Tcl_DStringAppendElement(str, "1");
	else
		Tcl_DStringAppendElement(str, "0");
}

/*
 * Parse a call reference
 */
int
parse_cref(Tcl_Interp *interp, int argc, const char **argv,
    struct uni_cref *cref)
{

	if (argc < 2) {
		unitcl_setres(interp, "need 2 args for call reference");
		return (TCL_ERROR);
	}
	if (strcmp(argv[0], "global") == 0)
		cref->cref = CREF_GLOBAL;
	else if (strcmp(argv[0], "dummy") == 0)
		cref->cref = CREF_DUMMY;
	else if (unitcl_parse_num(interp, argv[0], &cref->cref) != TCL_OK)
		return (TCL_ERROR);
	if (unitcl_parse_num(interp, argv[1], &cref->flag) != TCL_OK)
		return (TCL_ERROR);
	if (cref->flag != 0 && cref->flag != 1) {
		unitcl_setres(interp, "cref flag must be 0 or 1");
		return (TCL_ERROR);
	}
	if (cref->cref >= (1 << 23)) {
		unitcl_setres(interp, "cref too large");
		return (TCL_ERROR);
	}
	return (TCL_OK);
}

int
unitcl_setres(Tcl_Interp *interp, const char *fmt, ...)
{
	va_list ap;
	char tcl_errbuf[MAX_TCLERROR];

	va_start(ap, fmt);
#ifdef HAVE_VSNPRINTF
	vsnprintf(tcl_errbuf, MAX_TCLERROR, fmt, ap);
#else
	vsprintf(tcl_errbuf, fmt, ap);
#endif
	va_end(ap);
	Tcl_SetResult(interp, tcl_errbuf, TCL_VOLATILE);

	return (TCL_ERROR);
}

/*
 * Parse/format the tag field of an SAP
 */
int
parse_svetag(Tcl_Interp *interp, const char *arg, u_int *tag)
{

	if (strcmp(arg, "any") == 0)
		*tag = UNISVE_ANY;
	else if (strcmp(arg, "absent") == 0)
		*tag = UNISVE_ABSENT;
	else if (strcmp(arg, "present") == 0)
		*tag = UNISVE_PRESENT;
	else
		return (unitcl_setres(interp, "bad SVE tag: '%s'", arg));

	return (TCL_OK);
}
int
fmt_svetag(Tcl_Interp *interp, Tcl_DString *str, u_int tag)
{

	switch (tag) {

	  case UNISVE_ANY:
		Tcl_DStringAppendElement(str, "any");
		return (TCL_OK);

	  case UNISVE_PRESENT:
		Tcl_DStringAppendElement(str, "present");
		return (TCL_OK);

	  case UNISVE_ABSENT:
		Tcl_DStringAppendElement(str, "absent");
		return (TCL_OK);

	}
	return (unitcl_setres(interp, "bad SVE tag: %u", tag));
}

/*
 * A weak version of bug
 */
void
unitclbug(const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	fprintf(stderr, "libunitcl: ");
	vfprintf(stderr, fmt, ap);
	fprintf(stderr, "\n");
	va_end(ap);
	abort();
}
