/*
 * Copyright (c) 2003-2003
 *	Fraunhofer Institute for Open Communication Systems (FhG Fokus).
 * 	All rights reserved.
 *
 * Author: Hartmut Brandt <harti@freebsd.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Begemot: atmsupport/libunitcl/uni.c,v 1.19 2004/07/08 13:33:43 brandt Exp $
 */
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>

#include "inctcl.h"

#include <netnatm/unimsg.h>
#include <netnatm/msg/unistruct.h>
#include <netnatm/sig/unidef.h>

#include "unitcl.h"
#include "unitcl_priv.h"

static const char *const reasons[] = {
#define DEF(NAME, VAL, STR) [UNIAPI_##NAME] = STR,
UNIAPI_DEF_ERRORS(DEF)
#undef DEF
};

#define	PARSE_START_IE(MSG, IDX)					\
		const char *_msg = (MSG);				\
		if (parse_ie(interp, argv[(IDX)], &ie, &type) != TCL_OK)\
			return (TCL_ERROR);				\
		switch (type) {						\
									\
		  default:						\
			return (unitcl_setres(interp,			\
			    "%s: illegal IE %u", _msg, (u_int)type));	\

#define	PARSE_HANDLE_IE(NAME, IEL, IEP)					\
		  case UNI_IE_##NAME:					\
			if (IE_ISPRESENT(IEP))				\
				return (unitcl_setres(interp,		\
				    "%s."#IEL": already present",	\
				    _msg));				\
			IEP = ie.IEL;					\
			break

#define	PARSE_END_IE()							\
		}
static int
fmt_uniapi_error(Tcl_Interp *interp __unused, Tcl_DString *str,
    const void *ptr, size_t len __unused)
{
	const struct uniapi_error *e = ptr;
	char sbuf[100];

	sprintf(sbuf, "%u", e->state);
	Tcl_DStringAppendElement(str, sbuf);
	sprintf(sbuf, "%u", e->reason);
	Tcl_DStringAppendElement(str, sbuf);
	Tcl_DStringStartSublist(str);
	if (e->reason == UNIAPI_OK)
		; /* nothing */
	else if (e->reason >= sizeof(reasons) / sizeof(reasons[0]) ||
	    reasons[e->reason] == NULL)
		Tcl_DStringAppend(str, "unknown reason", -1);
	else
		Tcl_DStringAppend(str, reasons[e->reason], -1);
	Tcl_DStringEndSublist(str);

	return (TCL_OK);
}

static int
parse_uniapi_error(Tcl_Interp *interp, int argc, const char **argv,
    union uniapi_all *u, struct unicx *cx __unused)
{
	if (argc != 2 && argc != 3)
		return (unitcl_setres(interp,
		    "%s: wrong number of args", __func__));
	if (unitcl_parse_num(interp, argv[0], &u->error.state) != TCL_OK ||
	    unitcl_parse_num(interp, argv[1], &u->error.reason) != TCL_OK)
		return (TCL_ERROR);
	return (TCL_OK);
}

static int
fmt_uniapi_call_created(Tcl_Interp *interp __unused, Tcl_DString *str,
    const void *ptr, size_t len __unused)
{
	const struct uniapi_call_created *e = ptr;

	fmt_cref(str, &e->cref);
	return (TCL_OK);
}

static int
parse_uniapi_call_created(Tcl_Interp *interp, int argc, const char **argv,
    union uniapi_all *u, struct unicx *cx __unused)
{

	if (parse_cref(interp, argc, argv, &u->call_created.cref) != TCL_OK)
		return (TCL_ERROR);
	argc -= 2;
	argv += 2;

	if (argc != 0)
		return (unitcl_setres(interp,
		    "%s: wrong number of args", __func__));
	return (TCL_OK);
}

static int
fmt_uniapi_call_destroyed(Tcl_Interp *interp __unused, Tcl_DString *str,
    const void *ptr, size_t len __unused)
{
	const struct uniapi_call_destroyed *e = ptr;

	fmt_cref(str, &e->cref);
	return (TCL_OK);
}

static int
parse_uniapi_call_destroyed(Tcl_Interp *interp, int argc, const char **argv,
    union uniapi_all *u, struct unicx *cx __unused)
{

	if (parse_cref(interp, argc, argv, &u->call_destroyed.cref) != TCL_OK)
		return (TCL_ERROR);
	argc -= 2;
	argv += 2;

	if (argc != 0)
		return (unitcl_setres(interp,
		    "%s: wrong number of args", __func__));
	return (TCL_OK);
}

static int
fmt_uniapi_party_created(Tcl_Interp *interp, Tcl_DString *str,
    const void *ptr, size_t len __unused)
{
	const struct uniapi_party_created *e = ptr;
	int ret;

	fmt_cref(str, &e->cref);
	if ((ret = fmt_epref(interp, str, &e->epref)) != TCL_OK)
		return (ret);
	return (TCL_OK);
}

static int
parse_uniapi_party_created(Tcl_Interp *interp, int argc, const char **argv,
    union uniapi_all *u, struct unicx *cx __unused)
{
	int i;
	enum uni_ietype type;
	union uni_ieall ie;
	struct uniapi_party_created *e = &u->party_created;

	if (parse_cref(interp, argc, argv, &e->cref) != TCL_OK)
		return (TCL_ERROR);

	for (i = 2; i < argc; i++) {
		PARSE_START_IE("party-created", i);
		  PARSE_HANDLE_IE(EPREF, epref, e->epref);
		PARSE_END_IE();
	}
	return (TCL_OK);
}

static int
fmt_uniapi_party_destroyed(Tcl_Interp *interp __unused, Tcl_DString *str,
    const void *ptr, size_t len __unused)
{
	const struct uniapi_party_destroyed *e = ptr;
	int ret;

	fmt_cref(str, &e->cref);
	if ((ret = fmt_epref(interp, str, &e->epref)) != TCL_OK)
		return (ret);
	return (TCL_OK);
}

static int
parse_uniapi_party_destroyed(Tcl_Interp *interp, int argc, const char **argv,
    union uniapi_all *u, struct unicx *cx __unused)
{
	int i;
	enum uni_ietype type;
	union uni_ieall ie;
	struct uniapi_party_destroyed *e = &u->party_destroyed;

	if (parse_cref(interp, argc, argv, &e->cref) != TCL_OK)
		return (TCL_ERROR);

	for (i = 2; i < argc; i++) {
		PARSE_START_IE("party-destroyed", i);
		  PARSE_HANDLE_IE(EPREF, epref, e->epref);
		PARSE_END_IE();
	}
	return (TCL_OK);
}

static int
fmt_uniapi_abort_call_request(Tcl_Interp *interp __unused, Tcl_DString *str,
    const void *ptr, size_t len __unused)
{
	const struct uniapi_abort_call_request *e = ptr;

	fmt_cref(str, &e->cref);
	return (TCL_OK);
} 

static int
parse_uniapi_abort_call_request(Tcl_Interp *interp, int argc,
    const char **argv, union uniapi_all *u, struct unicx *cx __unused)
{

	if (parse_cref(interp, argc, argv, &u->abort_call_request.cref) !=
	    TCL_OK)
		return (TCL_ERROR);
	argc -= 2;
	argv += 2;

	if (argc != 0)
		return (unitcl_setres(interp,
		    "%s: wrong number of args", __func__));
	return (TCL_OK);
}

/************************************************************************/
/*
 * RESET
 */

static int
fmt_reset_common(Tcl_Interp *interp, Tcl_DString *str,
    const struct uni_ie_restart *restart, const struct uni_ie_connid *connid)
{
	int ret;

	if ((ret = fmt_restart(interp, str, restart)) != TCL_OK ||
	    (ret = fmt_connid(interp, str, connid)) != TCL_OK)
		return (ret);
	return (TCL_OK);
}

static int
parse_reset_common(Tcl_Interp *interp, const char *msg, int argc,
    const char **argv, struct uni_ie_restart *restart,
    struct uni_ie_connid *connid)
{
	int i;
	enum uni_ietype type;
	union uni_ieall ie;

	for (i = 0; i < argc; i++) {
		PARSE_START_IE(msg, i);
		  PARSE_HANDLE_IE(RESTART, restart, *restart);
		  PARSE_HANDLE_IE(CONNID, connid, *connid);
		PARSE_END_IE();
	}
	return (TCL_OK);
}

static int
fmt_uniapi_reset_indication(Tcl_Interp *interp, Tcl_DString *str,
    const void *ptr, size_t len __unused)
{
	const struct uniapi_reset_indication *e = ptr;

	return (fmt_reset_common(interp, str, &e->restart, &e->connid));
}

static int
parse_uniapi_reset_indication(Tcl_Interp *interp, int argc, const char **argv,
    union uniapi_all *u, struct unicx *cx __unused)
{

	return (parse_reset_common(interp, "reset_indication", argc, argv,
	    &u->reset_indication.restart, &u->reset_indication.connid));
}

static int
fmt_uniapi_reset_response(Tcl_Interp *interp, Tcl_DString *str,
    const void *ptr, size_t len __unused)
{
	const struct uniapi_reset_response *e = ptr;

	return (fmt_reset_common(interp, str, &e->restart, &e->connid));
}

static int
parse_uniapi_reset_response(Tcl_Interp *interp, int argc, const char **argv,
    union uniapi_all *u, struct unicx *cx __unused)
{

	return (parse_reset_common(interp, "reset_response", argc, argv,
	    &u->reset_response.restart, &u->reset_response.connid));
}

static int
fmt_uniapi_reset_request(Tcl_Interp *interp, Tcl_DString *str,
    const void *ptr, size_t len __unused)
{
	const struct uniapi_reset_request *e = ptr;

	return (fmt_reset_common(interp, str, &e->restart, &e->connid));
}

static int
parse_uniapi_reset_request(Tcl_Interp *interp, int argc, const char **argv,
    union uniapi_all *u, struct unicx *cx __unused)
{
	return (parse_reset_common(interp, "reset_request", argc, argv,
	    &u->reset_request.restart, &u->reset_request.connid));
}

static int
fmt_uniapi_reset_confirm(Tcl_Interp *interp, Tcl_DString *str,
    const void *ptr, size_t len __unused)
{
	const struct uniapi_reset_confirm *e = ptr;

	return (fmt_reset_common(interp, str, &e->restart, &e->connid));
}

static int
parse_uniapi_reset_confirm(Tcl_Interp *interp, int argc, const char **argv,
    union uniapi_all *u, struct unicx *cx __unused)
{
	return (parse_reset_common(interp, "reset_confirm", argc, argv,
	    &u->reset_confirm.restart, &u->reset_confirm.connid));
}


static int
fmt_uniapi_reset_error_indication(Tcl_Interp *interp, Tcl_DString *str,
    const void *ptr, size_t len __unused)
{
	const struct uniapi_reset_error_indication *e = ptr;
	char sbuf[20];

	if (e->source == 0)
		Tcl_DStringAppendElement(str, "start");
	else if (e->source == 1)
		Tcl_DStringAppendElement(str, "respond");
	else
		return (unitcl_setres(interp,
		    "%s: bad source %u", __func__, e->source));

	sprintf(sbuf, "%u", e->reason);
	Tcl_DStringAppendElement(str, sbuf);

	return (TCL_OK);
}

static int
parse_uniapi_reset_error_indication(Tcl_Interp *interp, int argc,
    const char **argv, union uniapi_all *u, struct unicx *cx __unused)
{
	char *end;

	if (argc != 2)
		return (unitcl_setres(interp, "%s: bad # of args", __func__));
	if (strcmp(argv[0], "start") == 0)
		u->reset_error_indication.source = 0;
	else if (strcmp(argv[0], "respond") == 0)
		u->reset_error_indication.source = 1;
	else
		return (unitcl_setres(interp, "%s: bad source", __func__));

	errno = 0;
	u->reset_error_indication.reason = strtoul(argv[1], &end, 10);
	if (*end != '\0' || errno != 0 || end == argv[1])
		return (unitcl_setres(interp,
		    "%s: bad reason '%s'", __func__, argv[1]));
	return (TCL_OK);
}

static int
fmt_uniapi_reset_status_indication(Tcl_Interp *interp, Tcl_DString *str,
    const void *ptr, size_t len __unused)
{
	const struct uniapi_reset_status_indication *e = ptr;
	int ret;

	fmt_cref(str, &e->cref);
	if ((ret = fmt_callstate(interp, str, &e->callstate)) != TCL_OK ||
	    (ret = fmt_cause(interp, str, &e->cause)) != TCL_OK)
		return (ret);
	return (TCL_OK);
}

static int
parse_uniapi_reset_status_indication(Tcl_Interp *interp,
    int argc, const char **argv, union uniapi_all *u, struct unicx *cx __unused)
{
	int i;
	enum uni_ietype type;
	union uni_ieall ie;
	struct uniapi_reset_status_indication *e = &u->reset_status_indication;
	

	if (parse_cref(interp, argc, argv, &e->cref) != TCL_OK)
		return (TCL_ERROR);

	for (i = 2; i < argc; i++) {
		PARSE_START_IE("reset-status-indication", i);
		  PARSE_HANDLE_IE(CALLSTATE, callstate, e->callstate);
		  PARSE_HANDLE_IE(CAUSE, cause, e->cause);
		PARSE_END_IE();
	}
	return (TCL_OK);
}

/************************************************************************/
/*
 * SETUP
 */
static int
fmt_uniapi_setup_indication(Tcl_Interp *interp __unused, Tcl_DString *str __unused,
    const void *ptr, size_t len __unused)
{
	const struct uniapi_setup_indication *e = ptr;

	if (fmt_msghdr(interp, str, UNI_SETUP, &e->setup.hdr) != TCL_OK ||
	    fmt_msg_setup(interp, str, &e->setup) != TCL_OK)
		return (TCL_ERROR);
	return (TCL_OK);
}

static int
parse_uniapi_setup_indication(Tcl_Interp *interp __unused, int argc __unused,
    const char **argv __unused, union uniapi_all *u, struct unicx *cx __unused)
{
	enum uni_msgtype mtype;

	if (parse_msghdr(interp, argc, argv, &mtype,
	    &u->setup_indication.setup.hdr) != TCL_OK)
		return (TCL_ERROR);
	if (mtype != UNI_SETUP)
		return (unitcl_setres(interp, "%s: message type not 'setup'",
		    __func__));
	if (parse_msg_setup(interp, argc - 3, argv + 3,
	    &u->setup_indication.setup, cx) != TCL_OK)
		return (TCL_ERROR);
	return (TCL_OK);
}

static int
fmt_uniapi_setup_response(Tcl_Interp *interp, Tcl_DString *str,
    const void *ptr, size_t len __unused)
{
	const struct uniapi_setup_response *e = ptr;

	if (fmt_msghdr(interp, str, UNI_CONNECT, &e->connect.hdr) != TCL_OK ||
	    fmt_msg_connect(interp, str, &e->connect) != TCL_OK)
		return (TCL_ERROR);
	return (TCL_OK);
}

static int
parse_uniapi_setup_response(Tcl_Interp *interp, int argc,
    const char **argv, union uniapi_all *u, struct unicx *cx __unused)
{
	enum uni_msgtype mtype;

	if (parse_msghdr(interp, argc, argv, &mtype,
	    &u->setup_response.connect.hdr) != TCL_OK)
		return (TCL_ERROR);
	if (mtype != UNI_CONNECT)
		return (unitcl_setres(interp, "%s: message type not 'connect'",
		    __func__));
	if (parse_msg_connect(interp, argc - 3, argv + 3,
	    &u->setup_response.connect, cx) != TCL_OK)
		return (TCL_ERROR);
	return (TCL_OK);
}

static int
fmt_uniapi_setup_complete_indication(Tcl_Interp *interp, Tcl_DString *str,
    const void *ptr, size_t len __unused)
{
	const struct uniapi_setup_complete_indication *e = ptr;

	if (fmt_msghdr(interp, str, UNI_CONNECT_ACK,
	    &e->connect_ack.hdr) != TCL_OK ||
	    fmt_msg_connect_ack(interp, str, &e->connect_ack) != TCL_OK)
		return (TCL_ERROR);
	return (TCL_OK);
}

static int
parse_uniapi_setup_complete_indication(Tcl_Interp *interp, int argc,
    const char **argv, union uniapi_all *u, struct unicx *cx __unused)
{
	enum uni_msgtype mtype;

	if (parse_msghdr(interp, argc, argv, &mtype,
	    &u->setup_complete_indication.connect_ack.hdr) != TCL_OK)
		return (TCL_ERROR);
	if (mtype != UNI_CONNECT_ACK)
		return (unitcl_setres(interp,
		    "%s: message type not 'connect_ack'", __func__));
	if (parse_msg_connect_ack(interp, argc - 3, argv + 3,
	    &u->setup_complete_indication.connect_ack, cx) != TCL_OK)
		return (TCL_ERROR);
	return (TCL_OK);
}

static int
fmt_uniapi_setup_complete_request(Tcl_Interp *interp, Tcl_DString *str,
    const void *ptr, size_t len __unused)
{
	const struct uniapi_setup_complete_request *e = ptr;

	if (fmt_msghdr(interp, str, UNI_CONNECT_ACK,
	    &e->connect_ack.hdr) != TCL_OK ||
	    fmt_msg_connect_ack(interp, str, &e->connect_ack) != TCL_OK)
		return (TCL_ERROR);
	return (TCL_OK);
}

static int
parse_uniapi_setup_complete_request(Tcl_Interp *interp, int argc,
    const char **argv, union uniapi_all *u, struct unicx *cx __unused)
{
	enum uni_msgtype mtype;

	if (parse_msghdr(interp, argc, argv, &mtype,
	    &u->setup_complete_request.connect_ack.hdr) != TCL_OK)
		return (TCL_ERROR);
	if (mtype != UNI_CONNECT_ACK)
		return (unitcl_setres(interp,
		    "%s: message type not 'connect_ack'", __func__));
	if (parse_msg_connect_ack(interp, argc - 3, argv + 3,
	    &u->setup_complete_request.connect_ack, cx) != TCL_OK)
		return (TCL_ERROR);
	return (TCL_OK);
}

static int
fmt_uniapi_setup_request(Tcl_Interp *interp, Tcl_DString *str,
    const void *ptr, size_t len __unused)
{
	const struct uniapi_setup_request *e = ptr;

	if (fmt_msghdr(interp, str, UNI_SETUP, &e->setup.hdr) != TCL_OK ||
	    fmt_msg_setup(interp, str, &e->setup) != TCL_OK)
		return (TCL_ERROR);
	return (TCL_OK);
}

static int
parse_uniapi_setup_request(Tcl_Interp *interp, int argc,
    const char **argv, union uniapi_all *u, struct unicx *cx)
{
	enum uni_msgtype mtype;

	if (parse_msghdr(interp, argc, argv, &mtype,
	    &u->setup_request.setup.hdr) != TCL_OK)
		return (TCL_ERROR);
	if (mtype != UNI_SETUP)
		return (unitcl_setres(interp, "%s: message type not 'setup'",
		    __func__));
	if (parse_msg_setup(interp, argc - 3, argv + 3,
	    &u->setup_request.setup, cx) != TCL_OK)
		return (TCL_ERROR);
	return (TCL_OK);
}

static int
fmt_uniapi_setup_confirm(Tcl_Interp *interp, Tcl_DString *str,
    const void *ptr, size_t len __unused)
{
	const struct uniapi_setup_confirm *e = ptr;

	if (fmt_msghdr(interp, str, UNI_CONNECT, &e->connect.hdr) != TCL_OK ||
	    fmt_msg_connect(interp, str, &e->connect) != TCL_OK)
		return (TCL_ERROR);
	return (TCL_OK);
}

static int
parse_uniapi_setup_confirm(Tcl_Interp *interp, int argc,
    const char **argv, union uniapi_all *u, struct unicx *cx)
{
	enum uni_msgtype mtype;

	if (parse_msghdr(interp, argc, argv, &mtype,
	    &u->setup_confirm.connect.hdr) != TCL_OK)
		return (TCL_ERROR);
	if (mtype != UNI_CONNECT)
		return (unitcl_setres(interp, "%s: message type not 'connect'",
		    __func__));
	if (parse_msg_connect(interp, argc - 3, argv + 3,
	    &u->setup_confirm.connect, cx) != TCL_OK)
		return (TCL_ERROR);
	return (TCL_OK);
}

/************************************************************************/
/*
 * PROCEEDING
 */
static int
fmt_uniapi_proceeding_request(Tcl_Interp *interp, Tcl_DString *str,
    const void *ptr, size_t len __unused)
{
	const struct uniapi_proceeding_request *e = ptr;

	if (fmt_msghdr(interp, str, UNI_CALL_PROC, &e->call_proc.hdr) != TCL_OK ||
	    fmt_msg_call_proc(interp, str, &e->call_proc) != TCL_OK)
		return (TCL_ERROR);
	return (TCL_OK);
}

static int
parse_uniapi_proceeding_request(Tcl_Interp *interp, int argc,
    const char **argv, union uniapi_all *u, struct unicx *cx __unused)
{
	enum uni_msgtype mtype;

	if (parse_msghdr(interp, argc, argv, &mtype,
	    &u->proceeding_request.call_proc.hdr) != TCL_OK)
		return (TCL_ERROR);
	if (mtype != UNI_CALL_PROC)
		return (unitcl_setres(interp,
		    "%s: message type not 'call_proc'", __func__));
	if (parse_msg_call_proc(interp, argc - 3, argv + 3,
	    &u->proceeding_request.call_proc, cx) != TCL_OK)
		return (TCL_ERROR);
	return (TCL_OK);
}

static int
fmt_uniapi_proceeding_indication(Tcl_Interp *interp, Tcl_DString *str,
    const void *ptr, size_t len __unused)
{
	const struct uniapi_proceeding_indication *e = ptr;

	if (fmt_msghdr(interp, str, UNI_CALL_PROC, &e->call_proc.hdr) != TCL_OK ||
	    fmt_msg_call_proc(interp, str, &e->call_proc) != TCL_OK)
		return (TCL_ERROR);
	return (TCL_OK);
}

static int
parse_uniapi_proceeding_indication(Tcl_Interp *interp, int argc,
    const char **argv, union uniapi_all *u, struct unicx *cx __unused)
{
	enum uni_msgtype mtype;

	if (parse_msghdr(interp, argc, argv, &mtype,
	    &u->proceeding_indication.call_proc.hdr) != TCL_OK)
		return (TCL_ERROR);
	if (mtype != UNI_CALL_PROC)
		return (unitcl_setres(interp,
		    "%s: message type not 'call_proc'", __func__));
	if (parse_msg_call_proc(interp, argc - 3, argv + 3,
	    &u->proceeding_indication.call_proc, cx) != TCL_OK)
		return (TCL_ERROR);
	return (TCL_OK);
}

/************************************************************************/
/*
 * RELEASE
 */
static int
fmt_uniapi_release_response(Tcl_Interp *interp, Tcl_DString *str,
    const void *ptr, size_t len __unused)
{
	const struct uniapi_release_response *e = ptr;

	if (fmt_msghdr(interp, str, UNI_RELEASE_COMPL,
	    &e->release_compl.hdr) != TCL_OK ||
	    fmt_msg_release_compl(interp, str, &e->release_compl) != TCL_OK)
		return (TCL_ERROR);
	return (TCL_OK);
}

static int
parse_uniapi_release_response(Tcl_Interp *interp, int argc,
    const char **argv, union uniapi_all *u, struct unicx *cx __unused)
{
	enum uni_msgtype mtype;

	if (parse_msghdr(interp, argc, argv, &mtype,
	    &u->release_response.release_compl.hdr) != TCL_OK)
		return (TCL_ERROR);
	if (mtype != UNI_RELEASE_COMPL)
		return (unitcl_setres(interp,
		    "%s: message type not 'release_compl'", __func__));
	if (parse_msg_release_compl(interp, argc - 3, argv + 3,
	    &u->release_response.release_compl, cx) != TCL_OK)
		return (TCL_ERROR);
	return (TCL_OK);
}

static int
fmt_release_common(Tcl_Interp *interp, Tcl_DString *str,
    const struct uni_release *e)
{
	if (fmt_msghdr(interp, str, UNI_RELEASE, &e->hdr) != TCL_OK ||
	    fmt_msg_release(interp, str, e) != TCL_OK)
		return (TCL_ERROR);
	return (TCL_OK);
}

static int
parse_release_common(Tcl_Interp *interp, int argc, const char **argv,
    struct uni_release *e, struct unicx *cx)
{
	enum uni_msgtype mtype;

	if (parse_msghdr(interp, argc, argv, &mtype, &e->hdr) != TCL_OK)
		return (TCL_ERROR);
	if (mtype != UNI_RELEASE)
		return (unitcl_setres(interp,
		    "%s: message type not 'release'", __func__));
	if (parse_msg_release(interp, argc - 3, argv + 3, e, cx) != TCL_OK)
		return (TCL_ERROR);
	return (TCL_OK);
}

static int
fmt_uniapi_release_indication(Tcl_Interp *interp, Tcl_DString *str,
    const void *ptr, size_t len __unused)
{
	return (fmt_release_common(interp, str,
	    &((const struct uniapi_release_indication *)ptr)->release));
}

static int
parse_uniapi_release_indication(Tcl_Interp *interp, int argc,
    const char **argv, union uniapi_all *u, struct unicx *cx __unused)
{
	return (parse_release_common(interp, argc, argv,
	    &u->release_indication.release, cx));
}

static int
fmt_uniapi_release_confirm(Tcl_Interp *interp, Tcl_DString *str,
    const void *ptr, size_t len __unused)
{
	return (fmt_release_common(interp, str,
	    &((const struct uniapi_release_confirm *)ptr)->release));
}

static int
parse_uniapi_release_confirm(Tcl_Interp *interp, int argc,
    const char **argv, union uniapi_all *u, struct unicx *cx __unused)
{
	return (parse_release_common(interp, argc, argv,
	    &u->release_confirm.release, cx));
}

static int
fmt_uniapi_release_request(Tcl_Interp *interp, Tcl_DString *str,
    const void *ptr, size_t len __unused)
{
	return (fmt_release_common(interp, str,
	    &((const struct uniapi_release_request *)ptr)->release));
}

static int
parse_uniapi_release_request(Tcl_Interp *interp, int argc,
    const char **argv, union uniapi_all *u, struct unicx *cx __unused)
{
	return (parse_release_common(interp, argc, argv,
	    &u->release_request.release, cx));
}

/************************************************************************/
/*
 * STATUS
 */
static int
fmt_uniapi_status_enquiry_request(Tcl_Interp *interp, Tcl_DString *str,
    const void *ptr, size_t len __unused)
{
	const struct uniapi_status_enquiry_request *e = ptr;
	int ret;

	fmt_cref(str, &e->cref);
	if ((ret = fmt_epref(interp, str, &e->epref)) != TCL_OK)
		return (ret);
	return (TCL_OK);
}

static int
parse_uniapi_status_enquiry_request(Tcl_Interp *interp, int argc,
    const char **argv, union uniapi_all *u, struct unicx *cx __unused)
{
	int i;
	enum uni_ietype type;
	union uni_ieall ie;
	struct uniapi_status_enquiry_request *e = &u->status_enquiry_request;

	if (parse_cref(interp, argc, argv, &e->cref) != TCL_OK)
		return (TCL_ERROR);

	for (i = 2; i < argc; i++) {
		PARSE_START_IE("status-enquiry-request", i);
		  PARSE_HANDLE_IE(EPREF, epref, e->epref);
		PARSE_END_IE();
	}
	return (TCL_OK);
}

static int
fmt_uniapi_status_indication(Tcl_Interp *interp, Tcl_DString *str,
    const void *ptr, size_t len __unused)
{
	const struct uniapi_status_indication *e = ptr;
	char sbuf[100];
	int ret;

	fmt_cref(str, &e->cref);

	sprintf(sbuf, "%u", e->my_state);
	Tcl_DStringAppendElement(str, sbuf);

	sprintf(sbuf, "%u", e->my_cause);
	Tcl_DStringAppendElement(str, sbuf);

	if ((ret = fmt_callstate(interp, str, &e->his_state)) != TCL_OK ||
	    (ret = fmt_cause(interp, str, &e->his_cause)) != TCL_OK ||
	    (ret = fmt_epref(interp, str, &e->epref)) != TCL_OK ||
	    (ret = fmt_epstate(interp, str, &e->epstate)) != TCL_OK)
		return (ret);
	return (TCL_OK);
}

static int
parse_uniapi_status_indication(Tcl_Interp *interp, int argc,
    const char **argv, union uniapi_all *u, struct unicx *cx __unused)
{
	int i;
	enum uni_ietype type;
	union uni_ieall ie;
	char *end;
	struct uniapi_status_indication *e = &u->status_indication;

	if (parse_cref(interp, argc, argv, &e->cref) != TCL_OK)
		return (TCL_ERROR);

	if (argc < 4)
		return (unitcl_setres(interp, "%s: wrong # of args", argc));

	errno = 0;
	e->my_state = strtoul(argv[2], &end, 10);
	if (*end != '\0' || errno != 0 || end == argv[2])
		return (unitcl_setres(interp,
		    "%s: bad callstate '%s'", __func__, argv[2]));

	errno = 0;
	e->my_cause = strtoul(argv[3], &end, 10);
	if (*end != '\0' || errno != 0 || end == argv[3])
		return (unitcl_setres(interp,
		    "%s: bad cause '%s'", __func__, argv[3]));

	for (i = 4; i < argc; i++) {
		PARSE_START_IE("status-enquiry-request", i);
		  PARSE_HANDLE_IE(CALLSTATE, callstate, e->his_state);
		  PARSE_HANDLE_IE(CAUSE, cause, e->his_cause);
		  PARSE_HANDLE_IE(EPREF, epref, e->epref);
		  PARSE_HANDLE_IE(EPSTATE, epstate, e->epstate);
		PARSE_END_IE();
	}
	return (TCL_OK);
}

/************************************************************************/
/*
 * ADD PARTY
 */
static int
fmt_uniapi_add_party_request(Tcl_Interp *interp, Tcl_DString *str,
    const void *ptr, size_t len __unused)
{
	const struct uniapi_add_party_request *e = ptr;

	if (fmt_msghdr(interp, str, UNI_ADD_PARTY, &e->add.hdr) != TCL_OK ||
	    fmt_msg_add_party(interp, str, &e->add) != TCL_OK)
		return (TCL_ERROR);
	return (TCL_OK);
}

static int
parse_uniapi_add_party_request(Tcl_Interp *interp, int argc,
    const char **argv, union uniapi_all *u, struct unicx *cx __unused)
{
	enum uni_msgtype mtype;
	struct uni_add_party *e = &u->add_party_request.add;

	if (parse_msghdr(interp, argc, argv, &mtype, &e->hdr) != TCL_OK)
		return (TCL_ERROR);
	if (mtype != UNI_ADD_PARTY)
		return (unitcl_setres(interp,
		    "%s: message type not 'add_party'", __func__));
	if (parse_msg_add_party(interp, argc - 3, argv + 3, e, cx) != TCL_OK)
		return (TCL_ERROR);
	return (TCL_OK);
}

static int
fmt_uniapi_add_party_indication(Tcl_Interp *interp, Tcl_DString *str,
    const void *ptr, size_t len __unused)
{
	const struct uniapi_add_party_indication *e = ptr;

	if (fmt_msghdr(interp, str, UNI_ADD_PARTY, &e->add.hdr) != TCL_OK ||
	    fmt_msg_add_party(interp, str, &e->add) != TCL_OK)
		return (TCL_ERROR);
	return (TCL_OK);
}

static int
parse_uniapi_add_party_indication(Tcl_Interp *interp, int argc,
    const char **argv, union uniapi_all *u, struct unicx *cx __unused)
{
	enum uni_msgtype mtype;
	struct uni_add_party *e = &u->add_party_indication.add;

	if (parse_msghdr(interp, argc, argv, &mtype, &e->hdr) != TCL_OK)
		return (TCL_ERROR);
	if (mtype != UNI_ADD_PARTY)
		return (unitcl_setres(interp,
		    "%s: message type not 'add_party'", __func__));
	if (parse_msg_add_party(interp, argc - 3, argv + 3, e, cx) != TCL_OK)
		return (TCL_ERROR);
	return (TCL_OK);
}

static int
fmt_uniapi_add_party_ack_request(Tcl_Interp *interp, Tcl_DString *str,
    const void *ptr, size_t len __unused)
{
	const struct uniapi_add_party_ack_request *e = ptr;

	if (fmt_msghdr(interp, str, UNI_ADD_PARTY_ACK, &e->ack.hdr) != TCL_OK ||
	    fmt_msg_add_party_ack(interp, str, &e->ack) != TCL_OK)
		return (TCL_ERROR);
	return (TCL_OK);
}

static int
parse_uniapi_add_party_ack_request(Tcl_Interp *interp, int argc,
    const char **argv, union uniapi_all *u, struct unicx *cx __unused)
{
	enum uni_msgtype mtype;
	struct uni_add_party_ack *e = &u->add_party_ack_request.ack;

	if (parse_msghdr(interp, argc, argv, &mtype, &e->hdr) != TCL_OK)
		return (TCL_ERROR);
	if (mtype != UNI_ADD_PARTY_ACK)
		return (unitcl_setres(interp,
		    "%s: message type not 'add_party_ack'", __func__));
	if (parse_msg_add_party_ack(interp, argc - 3, argv + 3, e, cx) != TCL_OK)
		return (TCL_ERROR);
	return (TCL_OK);
}


static int
fmt_uniapi_add_party_ack_indication(Tcl_Interp *interp, Tcl_DString *str,
    const void *ptr, size_t len __unused)
{
	const struct uniapi_add_party_ack_indication *e = ptr;

	if (fmt_msghdr(interp, str, UNI_ADD_PARTY_ACK, &e->ack.hdr) != TCL_OK ||
	    fmt_msg_add_party_ack(interp, str, &e->ack) != TCL_OK)
		return (TCL_ERROR);
	return (TCL_OK);
}

static int
parse_uniapi_add_party_ack_indication(Tcl_Interp *interp, int argc,
    const char **argv, union uniapi_all *u, struct unicx *cx __unused)
{
	enum uni_msgtype mtype;
	struct uni_add_party_ack *e = &u->add_party_ack_indication.ack;

	if (parse_msghdr(interp, argc, argv, &mtype, &e->hdr) != TCL_OK)
		return (TCL_ERROR);
	if (mtype != UNI_ADD_PARTY_ACK)
		return (unitcl_setres(interp,
		    "%s: message type not 'add_party_ack'", __func__));
	if (parse_msg_add_party_ack(interp, argc - 3, argv + 3, e, cx) != TCL_OK)
		return (TCL_ERROR);
	return (TCL_OK);
}

static int
fmt_uniapi_add_party_rej_request(Tcl_Interp *interp, Tcl_DString *str,
    const void *ptr, size_t len __unused)
{
	const struct uniapi_add_party_rej_request *e = ptr;

	if (fmt_msghdr(interp, str, UNI_ADD_PARTY_REJ, &e->rej.hdr) != TCL_OK ||
	    fmt_msg_add_party_rej(interp, str, &e->rej) != TCL_OK)
		return (TCL_ERROR);
	return (TCL_OK);
}

static int
parse_uniapi_add_party_rej_request(Tcl_Interp *interp, int argc,
    const char **argv, union uniapi_all *u, struct unicx *cx __unused)
{
	enum uni_msgtype mtype;
	struct uni_add_party_rej *e = &u->add_party_rej_request.rej;

	if (parse_msghdr(interp, argc, argv, &mtype, &e->hdr) != TCL_OK)
		return (TCL_ERROR);
	if (mtype != UNI_ADD_PARTY_REJ)
		return (unitcl_setres(interp,
		    "%s: message type not 'add_party_rej'", __func__));
	if (parse_msg_add_party_rej(interp, argc - 3, argv + 3, e, cx) != TCL_OK)
		return (TCL_ERROR);
	return (TCL_OK);
}

static int
fmt_uniapi_add_party_rej_indication(Tcl_Interp *interp, Tcl_DString *str,
    const void *ptr, size_t len __unused)
{
	const struct uniapi_add_party_rej_indication *e = ptr;

	if (fmt_msghdr(interp, str, UNI_ADD_PARTY_REJ, &e->rej.hdr) != TCL_OK ||
	    fmt_msg_add_party_rej(interp, str, &e->rej) != TCL_OK)
		return (TCL_ERROR);
	return (TCL_OK);
}

static int
parse_uniapi_add_party_rej_indication(Tcl_Interp *interp, int argc,
    const char **argv, union uniapi_all *u, struct unicx *cx __unused)
{
	enum uni_msgtype mtype;
	struct uni_add_party_rej *e = &u->add_party_rej_indication.rej;

	if (parse_msghdr(interp, argc, argv, &mtype, &e->hdr) != TCL_OK)
		return (TCL_ERROR);
	if (mtype != UNI_ADD_PARTY_REJ)
		return (unitcl_setres(interp,
		    "%s: message type not 'add_party_rej'", __func__));
	if (parse_msg_add_party_rej(interp, argc - 3, argv + 3, e, cx) != TCL_OK)
		return (TCL_ERROR);
	return (TCL_OK);
}

/************************************************************************/
/*
 * DROP PARTY
 */
static int
fmt_drop_party_common(Tcl_Interp *interp, Tcl_DString *str,
    const struct uni_drop_party *drop)
{

	if (fmt_msghdr(interp, str, UNI_DROP_PARTY, &drop->hdr) != TCL_OK ||
	    fmt_msg_drop_party(interp, str, drop) != TCL_OK)
		return (TCL_ERROR);
	return (TCL_OK);
}

static int
parse_drop_party_common(Tcl_Interp *interp, int argc,
    const char **argv, struct uni_drop_party *e, struct unicx *cx)
{
	enum uni_msgtype mtype;

	if (parse_msghdr(interp, argc, argv, &mtype, &e->hdr) != TCL_OK)
		return (TCL_ERROR);
	if (mtype != UNI_DROP_PARTY)
		return (unitcl_setres(interp,
		    "%s: message type not 'drop_party'", __func__));
	if (parse_msg_drop_party(interp, argc - 3, argv + 3, e, cx) != TCL_OK)
		return (TCL_ERROR);
	return (TCL_OK);
}

static int
fmt_uniapi_drop_party_request(Tcl_Interp *interp, Tcl_DString *str,
    const void *ptr, size_t len __unused)
{
	const struct uniapi_drop_party_request *e = ptr;

	return (fmt_drop_party_common(interp, str, &e->drop));
}

static int
parse_uniapi_drop_party_request(Tcl_Interp *interp, int argc,
    const char **argv, union uniapi_all *u, struct unicx *cx)
{

	return (parse_drop_party_common(interp, argc, argv,
	    &u->drop_party_request.drop, cx));
}

static int
fmt_uniapi_drop_party_indication(Tcl_Interp *interp, Tcl_DString *str,
    const void *ptr, size_t len __unused)
{
	const struct uniapi_drop_party_indication *e = ptr;

	Tcl_DStringStartSublist(str);
	if (fmt_msghdr(interp, str, UNI_DROP_PARTY, &e->drop.hdr) != TCL_OK ||
	    fmt_msg_drop_party(interp, str, &e->drop) != TCL_OK)
		return (TCL_ERROR);
	Tcl_DStringEndSublist(str);

	if (fmt_cause(interp, str, &e->my_cause) != TCL_OK)
		return (TCL_ERROR);

	return (TCL_OK);
}

static int
parse_uniapi_drop_party_indication(Tcl_Interp *interp, int argc,
    const char **argv, union uniapi_all *u, struct unicx *cx)
{
	struct uniapi_drop_party_indication *e = &u->drop_party_indication;
	enum uni_msgtype mtype;
	int ac, i;
	const char **av;
	union uni_ieall ie;
	enum uni_ietype type;

	if (argc < 1)
		return (unitcl_setres(interp, "%s: not enough args", __func__));

	if (Tcl_SplitList(interp, argv[0], &ac, &av) != TCL_OK)
		return (TCL_ERROR);

	if (parse_msghdr(interp, ac, av, &mtype, &e->drop.hdr) != TCL_OK)
		return (TCL_ERROR);
	if (mtype != UNI_DROP_PARTY)
		return (unitcl_setres(interp,
		    "%s: message type not 'drop_party'", __func__));
	if (parse_msg_drop_party(interp, ac - 3, av + 3,
	    &e->drop, cx) != TCL_OK)
		return (TCL_ERROR);

	Tcl_Free((char *)av);

	for (i = 1; i < argc; i++) {
		PARSE_START_IE("drop-party-indication", i);
		  PARSE_HANDLE_IE(CAUSE, cause, e->my_cause);
		PARSE_END_IE();
	}

	return (TCL_OK);
}

static int
fmt_uniapi_drop_party_ack_indication(Tcl_Interp *interp, Tcl_DString *str,
    const void *ptr, size_t len __unused)
{
	const struct uniapi_drop_party_ack_indication *e = ptr;

	Tcl_DStringStartSublist(str);
	if (fmt_msghdr(interp, str, UNI_DROP_PARTY, &e->drop.hdr) != TCL_OK)
		return (TCL_ERROR);
	if (fmt_msg_drop_party(interp, str, &e->drop) != TCL_OK)
		return (TCL_ERROR);
	Tcl_DStringEndSublist(str);

	if (fmt_crankback(interp, str, &e->crankback) != TCL_OK)
		return (TCL_ERROR);

	return (TCL_OK);
}

static int
parse_uniapi_drop_party_ack_indication(Tcl_Interp *interp, int argc,
    const char **argv, union uniapi_all *u, struct unicx *cx)
{
	struct uniapi_drop_party_ack_indication *e =
	    &u->drop_party_ack_indication;
	enum uni_msgtype mtype;
	int ac, i;
	const char **av;
	enum uni_ietype type;
	union uni_ieall ie;

	if (argc < 1)
		return (unitcl_setres(interp, "%s: not enough args", __func__));

	if (Tcl_SplitList(interp, argv[0], &ac, &av) != TCL_OK)
		return (TCL_ERROR);
	if (parse_msghdr(interp, ac, av, &mtype, &e->drop.hdr) != TCL_OK) {
		Tcl_Free((char *)av);
		return (TCL_ERROR);
	}
	if (mtype != UNI_DROP_PARTY) {
		Tcl_Free((char *)av);
		return (unitcl_setres(interp,
		    "%s: message type not 'drop_party'", __func__));
	}
	if (parse_msg_drop_party(interp, ac - 3, av + 3, &e->drop,
	    cx) != TCL_OK) {
		Tcl_Free((char *)av);
		return (TCL_ERROR);
	}
	Tcl_Free((char *)av);

	for (i = 1; i < argc; i++) {
		PARSE_START_IE("drop-party-ack-indication", i);
		  PARSE_HANDLE_IE(CRANKBACK, crankback, e->crankback);
		PARSE_END_IE();
	}
	return (TCL_OK);
}

static int
fmt_uniapi_drop_party_ack_request(Tcl_Interp *interp, Tcl_DString *str,
    const void *ptr, size_t len __unused)
{
	const struct uniapi_drop_party_ack_request *e = ptr;

	if (fmt_msghdr(interp, str, UNI_DROP_PARTY_ACK, &e->ack.hdr) != TCL_OK ||
	    fmt_msg_drop_party_ack(interp, str, &e->ack) != TCL_OK)
		return (TCL_ERROR);
	return (TCL_OK);
}

static int
parse_uniapi_drop_party_ack_request(Tcl_Interp *interp, int argc,
    const char **argv, union uniapi_all *u, struct unicx *cx)
{
	enum uni_msgtype mtype;
	struct uni_drop_party_ack *e = &u->drop_party_ack_request.ack;

	if (parse_msghdr(interp, argc, argv, &mtype, &e->hdr) != TCL_OK)
		return (TCL_ERROR);
	if (mtype != UNI_DROP_PARTY_ACK)
		return (unitcl_setres(interp,
		    "%s: message type not 'drop_party_ack'", __func__));
	if (parse_msg_drop_party_ack(interp, argc - 3, argv + 3, e, cx) != TCL_OK)
		return (TCL_ERROR);
	return (TCL_OK);
}

/************************************************************************/
/*
 * ALERTING
 */
static int
fmt_uniapi_alerting_request(Tcl_Interp *interp, Tcl_DString *str,
    const void *ptr, size_t len __unused)
{
	const struct uniapi_alerting_request *e = ptr;

	if (fmt_msghdr(interp, str, UNI_ALERTING, &e->alerting.hdr) != TCL_OK ||
	    fmt_msg_alerting(interp, str, &e->alerting) != TCL_OK)
		return (TCL_ERROR);
	return (TCL_OK);
}
static int
parse_uniapi_alerting_request(Tcl_Interp *interp, int argc, const char **argv,
    union uniapi_all *u, struct unicx *cx __unused)
{
	enum uni_msgtype mtype;
	struct uni_alerting *e = &u->alerting_request.alerting;

	if (parse_msghdr(interp, argc, argv, &mtype, &e->hdr) != TCL_OK)
		return (TCL_ERROR);
	if (mtype != UNI_ALERTING)
		return (unitcl_setres(interp,
		    "%s: message type not 'alerting'", __func__));
	return (parse_msg_alerting(interp, argc - 3, argv + 3, e, cx));
}

static int
fmt_uniapi_alerting_indication(Tcl_Interp *interp, Tcl_DString *str,
    const void *ptr, size_t len __unused)
{
	const struct uniapi_alerting_indication *e = ptr;

	if (fmt_msghdr(interp, str, UNI_ALERTING, &e->alerting.hdr) != TCL_OK ||
	    fmt_msg_alerting(interp, str, &e->alerting) != TCL_OK)
		return (TCL_ERROR);
	return (TCL_OK);
}
static int
parse_uniapi_alerting_indication(Tcl_Interp *interp, int argc,
    const char **argv, union uniapi_all *u, struct unicx *cx __unused)
{
	enum uni_msgtype mtype;
	struct uni_alerting *e = &u->alerting_indication.alerting;

	if (parse_msghdr(interp, argc, argv, &mtype, &e->hdr) != TCL_OK)
		return (TCL_ERROR);
	if (mtype != UNI_ALERTING)
		return (unitcl_setres(interp,
		    "%s: message type not 'alerting'", __func__));
	return (parse_msg_alerting(interp, argc - 3, argv + 3, e, cx));
}

static int
fmt_uniapi_party_alerting_request(Tcl_Interp *interp, Tcl_DString *str,
    const void *ptr, size_t len __unused)
{
	const struct uniapi_party_alerting_request *e = ptr;

	if (fmt_msghdr(interp, str, UNI_PARTY_ALERTING, &e->alert.hdr) != TCL_OK ||
	    fmt_msg_party_alerting(interp, str, &e->alert) != TCL_OK)
		return (TCL_ERROR);
	return (TCL_OK);
}
static int
parse_uniapi_party_alerting_request(Tcl_Interp *interp, int argc,
    const char **argv, union uniapi_all *u, struct unicx *cx __unused)
{
	enum uni_msgtype mtype;
	struct uni_party_alerting *e = &u->party_alerting_request.alert;

	if (parse_msghdr(interp, argc, argv, &mtype, &e->hdr) != TCL_OK)
		return (TCL_ERROR);
	if (mtype != UNI_PARTY_ALERTING)
		return (unitcl_setres(interp,
		    "%s: message type not 'party alerting'", __func__));
	return (parse_msg_party_alerting(interp, argc - 3, argv + 3, e, cx));
}

static int
fmt_uniapi_party_alerting_indication(Tcl_Interp *interp, Tcl_DString *str,
    const void *ptr, size_t len __unused)
{
	const struct uniapi_party_alerting_indication *e = ptr;

	if (fmt_msghdr(interp, str, UNI_PARTY_ALERTING, &e->alert.hdr) != TCL_OK ||
	    fmt_msg_party_alerting(interp, str, &e->alert) != TCL_OK)
		return (TCL_ERROR);
	return (TCL_OK);
}
static int
parse_uniapi_party_alerting_indication(Tcl_Interp *interp, int argc,
    const char **argv, union uniapi_all *u, struct unicx *cx __unused)
{
	enum uni_msgtype mtype;
	struct uni_party_alerting *e = &u->party_alerting_indication.alert;

	if (parse_msghdr(interp, argc, argv, &mtype, &e->hdr) != TCL_OK)
		return (TCL_ERROR);
	if (mtype != UNI_PARTY_ALERTING)
		return (unitcl_setres(interp,
		    "%s: message type not 'party alerting'", __func__));
	return (parse_msg_party_alerting(interp, argc - 3, argv + 3, e, cx));
}

/************************************************************************/
/*
 * NOTIFY
 */

static int
fmt_notify_common(Tcl_Interp *interp, Tcl_DString *str,
    const struct uni_notify *notify)
{

	if (fmt_msghdr(interp, str, UNI_NOTIFY, &notify->hdr) != TCL_OK ||
	    fmt_msg_notify(interp, str, notify) != TCL_OK)
		return (TCL_ERROR);
	return (TCL_OK);
}

static int
parse_notify_common(Tcl_Interp *interp, int argc,
    const char **argv, struct uni_notify *e, struct unicx *cx)
{
	enum uni_msgtype mtype;

	if (parse_msghdr(interp, argc, argv, &mtype, &e->hdr) != TCL_OK)
		return (TCL_ERROR);
	if (mtype != UNI_NOTIFY)
		return (unitcl_setres(interp,
		    "%s: message type not 'notify'", __func__));
	if (parse_msg_notify(interp, argc - 3, argv + 3, e, cx) != TCL_OK)
		return (TCL_ERROR);
	return (TCL_OK);
}

static int
fmt_uniapi_notify_request(Tcl_Interp *interp, Tcl_DString *str,
    const void *ptr, size_t len __unused)
{
	const struct uniapi_notify_request *e = ptr;

	return (fmt_notify_common(interp, str, &e->notify));
}
static int
fmt_uniapi_notify_indication(Tcl_Interp *interp, Tcl_DString *str,
    const void *ptr, size_t len __unused)
{
	const struct uniapi_notify_indication *e = ptr;

	return (fmt_notify_common(interp, str, &e->notify));

}

static int
parse_uniapi_notify_request(Tcl_Interp *interp, int argc, const char **argv,
    union uniapi_all *u, struct unicx *cx)
{

	return (parse_notify_common(interp, argc, argv,
	    &u->notify_request.notify, cx));
}
static int
parse_uniapi_notify_indication(Tcl_Interp *interp, int argc, const char **argv,
    union uniapi_all *u, struct unicx *cx)
{

	return (parse_notify_common(interp, argc, argv,
	    &u->notify_indication.notify, cx));
}

static int
fmt_uniapi_link_establish_confirm(Tcl_Interp *interp __unused,
    Tcl_DString *str __unused, const void *ptr __unused, size_t len __unused)
{
	return (TCL_OK);
}
static int
parse_uniapi_link_establish_confirm(Tcl_Interp *interp, int argc,
    const char **argv __unused, union uniapi_all *u __unused,
    struct unicx *cx __unused)
{
	if (argc != 0)
		return (unitcl_setres(interp, "bad number of args"));
	return (TCL_OK);
}

static int
fmt_uniapi_link_establish_request(Tcl_Interp *interp __unused,
    Tcl_DString *str __unused, const void *ptr __unused, size_t len __unused)
{
	return (TCL_OK);
}
static int
parse_uniapi_link_establish_request(Tcl_Interp *interp, int argc,
    const char **argv __unused, union uniapi_all *u __unused,
    struct unicx *cx __unused)
{
	if (argc != 0)
		return (unitcl_setres(interp, "bad number of args"));
	return (TCL_OK);
}

static int
fmt_uniapi_link_release_confirm(Tcl_Interp *interp __unused,
    Tcl_DString *str __unused, const void *ptr __unused, size_t len __unused)
{
	return (TCL_OK);
}
static int
parse_uniapi_link_release_confirm(Tcl_Interp *interp, int argc,
    const char **argv __unused, union uniapi_all *u __unused,
    struct unicx *cx __unused)
{
	if (argc != 0)
		return (unitcl_setres(interp, "bad number of args"));
	return (TCL_OK);
}

static int
fmt_uniapi_link_release_request(Tcl_Interp *interp __unused,
    Tcl_DString *str __unused, const void *ptr __unused, size_t len __unused)
{
	return (TCL_OK);
}
static int
parse_uniapi_link_release_request(Tcl_Interp *interp, int argc,
    const char **argv __unused, union uniapi_all *u __unused,
    struct unicx *cx __unused)
{
	if (argc != 0)
		return (unitcl_setres(interp, "bad number of args"));
	return (TCL_OK);
}

static int fmt_uniapi_reset_error_response(Tcl_Interp *interp __unused, Tcl_DString *str __unused, const void *ptr __unused, size_t len __unused) {  unitcl_setres(interp, "%s: implement me", __func__); return (TCL_ERROR); }
static int parse_uniapi_reset_error_response(Tcl_Interp *interp __unused, int argc __unused, const char **argv __unused, union uniapi_all *u __unused, struct unicx *cx __unused) {  unitcl_setres(interp, "%s: implement me", __func__); return (TCL_ERROR); }

static const struct {
	const char	*name;
	size_t		len;
	int		(*fmt)(Tcl_Interp *, Tcl_DString *,
			    const void *ptr, size_t);
	int		(*parse)(Tcl_Interp *, int, const char **,
			    union uniapi_all *, struct unicx *);
} signals[] = {
	[UNIAPI_ERROR] = {
		"uni-error",
		sizeof(struct uniapi_error),
		fmt_uniapi_error,
		parse_uniapi_error,
	},
	[UNIAPI_CALL_CREATED] = {
		"uni-call-created",
		sizeof(struct uniapi_call_created),
		fmt_uniapi_call_created,
		parse_uniapi_call_created,
	},
	[UNIAPI_CALL_DESTROYED] = {
		"uni-call-destroyed",
		sizeof(struct uniapi_call_destroyed),
		fmt_uniapi_call_destroyed,
		parse_uniapi_call_destroyed,
	},
	[UNIAPI_PARTY_CREATED] = {
		"uni-party-created",
		sizeof(struct uniapi_party_created),
		fmt_uniapi_party_created,
		parse_uniapi_party_created,
	},
	[UNIAPI_PARTY_DESTROYED] = {
		"uni-party-destroyed",
		sizeof(struct uniapi_party_destroyed),
		fmt_uniapi_party_destroyed,
		parse_uniapi_party_destroyed,
	},
	[UNIAPI_LINK_ESTABLISH_request] = {
		"uni-link-establish-request",
		0,
		fmt_uniapi_link_establish_request,
		parse_uniapi_link_establish_request,
	},
	[UNIAPI_LINK_ESTABLISH_confirm] = {
		"uni-link-establish-confirm",
		0,
		fmt_uniapi_link_establish_confirm,
		parse_uniapi_link_establish_confirm,
	},
	[UNIAPI_LINK_RELEASE_request] = {
		"uni-link-release-request",
		0,
		fmt_uniapi_link_release_request,
		parse_uniapi_link_release_request,
	},
	[UNIAPI_LINK_RELEASE_confirm] = {
		"uni-link-release-confirm",
		0,
		fmt_uniapi_link_release_confirm,
		parse_uniapi_link_release_confirm,
	},
	[UNIAPI_RESET_request] = {
		"uni-reset-request",
		sizeof(struct uniapi_reset_request),
		fmt_uniapi_reset_request,
		parse_uniapi_reset_request,
	},
	[UNIAPI_RESET_confirm] = {
		"uni-reset-confirm",
		sizeof(struct uniapi_reset_confirm),
		fmt_uniapi_reset_confirm,
		parse_uniapi_reset_confirm,
	},
	[UNIAPI_RESET_indication] = {
		"uni-reset-indication",
		sizeof(struct uniapi_reset_indication),
		fmt_uniapi_reset_indication,
		parse_uniapi_reset_indication,
	},
	[UNIAPI_RESET_ERROR_indication] = {
		"uni-reset-error-indication",
		sizeof(struct uniapi_reset_error_indication),
		fmt_uniapi_reset_error_indication,
		parse_uniapi_reset_error_indication,
	},
	[UNIAPI_RESET_response] = {
		"uni-reset-response",
		sizeof(struct uniapi_reset_response),
		fmt_uniapi_reset_response,
		parse_uniapi_reset_response,
	},
	[UNIAPI_RESET_ERROR_response] = {
		"uni-reset-error-response",
		(size_t)-1,
		fmt_uniapi_reset_error_response,
		parse_uniapi_reset_error_response,
	},
	[UNIAPI_RESET_STATUS_indication] = {
		"uni-reset-status-indication",
		sizeof(struct uniapi_reset_status_indication),
		fmt_uniapi_reset_status_indication,
		parse_uniapi_reset_status_indication,
	},
	[UNIAPI_SETUP_request] = {
		"uni-setup-request",
		sizeof(struct uniapi_setup_request),
		fmt_uniapi_setup_request,
		parse_uniapi_setup_request,
	},
	[UNIAPI_SETUP_indication] = {
		"uni-setup-indication",
		sizeof(struct uniapi_setup_indication),
		fmt_uniapi_setup_indication,
		parse_uniapi_setup_indication,
	},
	[UNIAPI_SETUP_response] = {
		"uni-setup-response",
		sizeof(struct uniapi_setup_response),
		fmt_uniapi_setup_response,
		parse_uniapi_setup_response,
	},
	[UNIAPI_SETUP_confirm] = {
		"uni-setup-confirm",
		sizeof(struct uniapi_setup_confirm),
		fmt_uniapi_setup_confirm,
		parse_uniapi_setup_confirm,
	},
	[UNIAPI_SETUP_COMPLETE_indication] = {
		"uni-setup-complete-indication",
		sizeof(struct uniapi_setup_complete_indication),
		fmt_uniapi_setup_complete_indication,
		parse_uniapi_setup_complete_indication,
	},
	[UNIAPI_SETUP_COMPLETE_request] = {
		"uni-setup-complete-request",
		sizeof(struct uniapi_setup_complete_request),
		fmt_uniapi_setup_complete_request,
		parse_uniapi_setup_complete_request,
	},
	[UNIAPI_ALERTING_request] = {
		"uni-alerting-request",
		sizeof(struct uniapi_alerting_request),
		fmt_uniapi_alerting_request,
		parse_uniapi_alerting_request,
	},
	[UNIAPI_ALERTING_indication] = {
		"uni-alerting-indication",
		sizeof(struct uniapi_alerting_indication),
		fmt_uniapi_alerting_indication,
		parse_uniapi_alerting_indication,
	},
	[UNIAPI_PROCEEDING_request] = {
		"uni-proceeding-request",
		sizeof(struct uniapi_proceeding_request),
		fmt_uniapi_proceeding_request,
		parse_uniapi_proceeding_request,
	},
	[UNIAPI_PROCEEDING_indication] = {
		"uni-proceeding-indication",
		sizeof(struct uniapi_proceeding_indication),
		fmt_uniapi_proceeding_indication,
		parse_uniapi_proceeding_indication,
	},
	[UNIAPI_RELEASE_request] = {
		"uni-release-request",
		sizeof(struct uniapi_release_request),
		fmt_uniapi_release_request,
		parse_uniapi_release_request,
	},
	[UNIAPI_RELEASE_indication] = {
		"uni-release-indication",
		sizeof(struct uniapi_release_indication),
		fmt_uniapi_release_indication,
		parse_uniapi_release_indication,
	},
	[UNIAPI_RELEASE_response] = {
		"uni-release-response",
		sizeof(struct uniapi_release_response),
		fmt_uniapi_release_response,
		parse_uniapi_release_response,
	},
	[UNIAPI_RELEASE_confirm] = {
		"uni-release-confirm",
		sizeof(struct uniapi_release_confirm),
		fmt_uniapi_release_confirm,
		parse_uniapi_release_confirm,
	},
	[UNIAPI_NOTIFY_request] = {
		"uni-notify-request",
		sizeof(struct uniapi_notify_request),
		fmt_uniapi_notify_request,
		parse_uniapi_notify_request,
	},
	[UNIAPI_NOTIFY_indication] = {
		"uni-notify-indication",
		sizeof(struct uniapi_notify_indication),
		fmt_uniapi_notify_indication,
		parse_uniapi_notify_indication,
	},
	[UNIAPI_STATUS_indication] = {
		"uni-status-indication",
		sizeof(struct uniapi_status_indication),
		fmt_uniapi_status_indication,
		parse_uniapi_status_indication,
	},
	[UNIAPI_STATUS_ENQUIRY_request] = {
		"uni-status-enquiry-request",
		sizeof(struct uniapi_status_enquiry_request),
		fmt_uniapi_status_enquiry_request,
		parse_uniapi_status_enquiry_request,
	},
	[UNIAPI_ADD_PARTY_request] = {
		"uni-add-party-request",
		sizeof(struct uniapi_add_party_request),
		fmt_uniapi_add_party_request,
		parse_uniapi_add_party_request,
	},
	[UNIAPI_ADD_PARTY_indication] = {
		"uni-add-party-indication",
		sizeof(struct uniapi_add_party_indication),
		fmt_uniapi_add_party_indication,
		parse_uniapi_add_party_indication,
	},
	[UNIAPI_PARTY_ALERTING_request] = {
		"uni-party-alerting-request",
		sizeof(struct uniapi_party_alerting_request),
		fmt_uniapi_party_alerting_request,
		parse_uniapi_party_alerting_request,
	},
	[UNIAPI_PARTY_ALERTING_indication] = {
		"uni-party-alerting-indication",
		sizeof(struct uniapi_party_alerting_indication),
		fmt_uniapi_party_alerting_indication,
		parse_uniapi_party_alerting_indication,
	},
	[UNIAPI_ADD_PARTY_ACK_request] = {
		"uni-add-party-ack-request",
		sizeof(struct uniapi_add_party_ack_request),
		fmt_uniapi_add_party_ack_request,
		parse_uniapi_add_party_ack_request,
	},
	[UNIAPI_ADD_PARTY_ACK_indication] = {
		"uni-add-party-ack-indication",
		sizeof(struct uniapi_add_party_ack_indication),
		fmt_uniapi_add_party_ack_indication,
		parse_uniapi_add_party_ack_indication,
	},
	[UNIAPI_ADD_PARTY_REJ_request] = {
		"uni-add-party-rej-request",
		sizeof(struct uniapi_add_party_rej_request),
		fmt_uniapi_add_party_rej_request,
		parse_uniapi_add_party_rej_request,
	},
	[UNIAPI_ADD_PARTY_REJ_indication] = {
		"uni-add-party-rej-indication",
		sizeof(struct uniapi_add_party_rej_indication),
		fmt_uniapi_add_party_rej_indication,
		parse_uniapi_add_party_rej_indication,
	},
	[UNIAPI_DROP_PARTY_request] = {
		"uni-drop-party-request",
		sizeof(struct uniapi_drop_party_request),
		fmt_uniapi_drop_party_request,
		parse_uniapi_drop_party_request,
	},
	[UNIAPI_DROP_PARTY_indication] = {
		"uni-drop-party-indication",
		sizeof(struct uniapi_drop_party_indication),
		fmt_uniapi_drop_party_indication,
		parse_uniapi_drop_party_indication,
	},
	[UNIAPI_DROP_PARTY_ACK_request] = {
		"uni-drop-party-ack-request",
		sizeof(struct uniapi_drop_party_ack_request),
		fmt_uniapi_drop_party_ack_request,
		parse_uniapi_drop_party_ack_request,
	},
	[UNIAPI_DROP_PARTY_ACK_indication] = {
		"uni-drop-party-ack-indication",
		sizeof(struct uniapi_drop_party_ack_indication),
		fmt_uniapi_drop_party_ack_indication,
		parse_uniapi_drop_party_ack_indication,
	},
	[UNIAPI_ABORT_CALL_request] = {
		"uni-abort-call-request",
		sizeof(struct uniapi_abort_call_request),
		fmt_uniapi_abort_call_request,
		parse_uniapi_abort_call_request,
	},
};

int
fmt_uniapi(Tcl_Interp *interp, Tcl_DString *str, u_int sig, const void *ptr,
    size_t len)
{

	if (sig >= sizeof(signals) / sizeof(signals[0]) ||
	    signals[sig].name == NULL)
		return (unitcl_setres(interp, "bad UNI signal %u", sig));
	if (len != signals[sig].len)
		return (unitcl_setres(interp,
		    "bad length for %s (%zu, need %zu)",
		    signals[sig].name, len, signals[sig].len));
	return (signals[sig].fmt(interp, str, ptr, len));
}

int
fmt_uniapi_sig(Tcl_Interp *interp, Tcl_DString *str, u_int sig)
{
	if (sig >= sizeof(signals) / sizeof(signals[0]) ||
	    signals[sig].name == NULL)
		return (unitcl_setres(interp, "bad UNI signal %u", sig));
	Tcl_DStringAppendElement(str, signals[sig].name);
	return (TCL_OK);
}

int
parse_uniapi_sig(Tcl_Interp *interp, const char *arg, u_int *sig)
{
	for (*sig = 0; *sig < sizeof(signals) / sizeof(signals[0]); (*sig)++)
		if (signals[*sig].name != NULL &&
		    strcmp(signals[*sig].name, arg) == 0)
			return (TCL_OK);
	return (unitcl_setres(interp, "bad uniapi signal name '%s'", arg));
}

int
parse_uniapi(Tcl_Interp *interp, int argc, const char **argv, u_int sig,
    union uniapi_all *u, size_t *lenp, struct unicx *cx)
{
	if (sig >= sizeof(signals) / sizeof(signals[0]) ||
	    signals[sig].name == NULL)
		return (unitcl_setres(interp, "bad UNI signal %u", sig));
	*lenp = signals[sig].len;
	return (signals[sig].parse(interp, argc, argv, u, cx));
}
