/*
 * Copyright (c) 2001-2003
 *	Fraunhofer Institute for Open Communication Systems (FhG Fokus).
 * 	All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Author: Harti Brandt <harti@freebsd.org>
 *
 * ng_sscop test program
 *
 * $Begemot: atmsupport/test_sscfu/sscfutest.c,v 1.5 2004/12/20 10:44:58 brandt Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <err.h>
#include <vis.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <limits.h>
#include <netgraph.h>
#include <rpoll.h>

#include <netnatm/saal/sscopdef.h>
#include <netnatm/saal/sscfudef.h>
#include <netgraph/atm/ng_sscop.h>
#include <netgraph/atm/ng_sscfu.h>
#include "testlib.h"

static int	csock, dsock;
static char	prefix[100];
static u_int	filenum;
static int	doprompt;		/* need prompt on next loop */
static const char *device = "fatm0:";
static u_int	sscop_verbose;
static u_int	sscf_verbose;

/*                [atmtest]
 *		      |(uni5)
 *                    +----------------+
 *                    |                |(upper)
 *                    |              [sscfu5]
 *                    |                |(lower)
 *                    |(man5) (manage) |(upper)
 *                    +--------------[sscop5]
 *                                     |(lower)
 *                                     |(sig5)
 *                                  [fatm]
 */
static struct stack stack[] = {
	{ STACK_MKPEER, NULL, "sscop", "sig5", "lower", "sscop5" },
	{ STACK_MKPEER, "sscop5:", "sscfu", "upper", "lower", "sscfu5" },
	{ STACK_CONNECT, ".", "sscop5:", "man5", "manage", NULL },
	{ STACK_CONNECT, ".", "sscfu5:", "uni5", "upper", NULL },
	{ STACK_END, NULL, NULL, NULL, NULL, NULL }
};

static const char help[] = "\
d [data]		data-request\n\
D file			data-request from file\n\
e [uu-data]		establish-request\n\
f N			send QUEUE FULL message\n\
F N			send QUEUE EMPTY message\n\
G H L			set high and low water mark\n\
h			help\n\
m [data]		mdata-request\n\
M file			Mdata-request from file\n\
p [name]		set/clear data save prefix\n\
r [uu-data]		release-request\n\
u [data]		udata-request\n\
U file			udata-request from file\n\
v (sscop|sscf) X	set verbosity to level X\n\
";

static const char usgtxt[] = "\
SSCF test program\n\
usage: sscfutest [-h] [-d dev:] [s proto=X] [-v]\n\
Flags:\n\
  -d dev:	use device dev instead of fatm0.\n\
  -h		print this help text\n\
  -s proto=X	set verbosity of protocol proto to X\n\
  -v		increment netgraph verbosity level\n\
'proto' is one of sscop or sscf.\n\
";

static struct ngm_queue_state qstate = {
	(u_int)-1,
	0,
	128,
	10,
	100,
	0
};

/*
 * Called on shutdown via atexit(3) to remove the stack
 */
static void
shut(void)
{	

	shutdown_stack(csock, stack);
}

/*
 * SIGINT handler
 */
static void
onintr(int sig)
{

	sig = sig;
	exit(0);
}

/*
 * Initialize the stack
 */
static void
init_stack(void)
{
	if (NgSendAsciiMsg(csock, "sscfu5:",
	    "enable") < 0)
		err(1, "failed to enable SSCF");

	if (NgSendAsciiMsg(csock, "sscop5:",
	    "setparam {mask=0x800 param={flags=0x01}}") < 0)
		err(1, "failed to set robust mode");
	read_cmsg(csock);

	if (NgSendAsciiMsg(csock, "sscop5:",
	    "enable") < 0)
		err(1, "failed to enable SSCOP");

	if (NgSendAsciiMsg(csock, device,
	    "cpcsinit {name=\"sig5\" flags=0x4 traffic=0 aal=5 vci=5}") < 0)
		err(1, "failed to init VCI");

	if (NgSendAsciiMsg(csock, "sscop5:",
	    "setdebug %#x", sscop_verbose) < 0)
		err(1, "failed to set sscop debug level");
	if (NgSendAsciiMsg(csock, "sscfu5:",
	    "setdebug %#x", sscf_verbose) < 0)
		err(1, "failed to set sscf debug level");
}

/*
 * csock is ready
 */
static void
ready_cmsg(int fd __unused, int mask __unused, void *uap __unused)
{

	read_cmsg(csock);
	doprompt = 1;
}

/*************************************************************
 *
 * Input from data socket.
 */
static void
dump_uni(u_char *buf, size_t n)
{
	struct sscfu_arg *a = (struct sscfu_arg *)(void *)buf;

	if (n < sizeof(*a)) {
		printf("uni message too short\n");
		return;
	}
	n -= sizeof(*a);

# define D(X) 								\
	  case X:							\
		printf("%s: data=%zu\n", #X, n);				\
		if (n > 0) {						\
			dump_hex(a->data, n);				\
			save_data(prefix, "d", &filenum, a->data, n);	\
		}							\
		break;

	switch (a->sig) {

	  default:
		printf("bad signal: %u", a->sig);
		break;

	  D(SAAL_ESTABLISH_request)
	  D(SAAL_ESTABLISH_indication)
	  D(SAAL_ESTABLISH_confirm)
	  D(SAAL_RELEASE_request)
	  D(SAAL_RELEASE_confirm)
	  D(SAAL_RELEASE_indication)
	  D(SAAL_DATA_request)
	  D(SAAL_DATA_indication)
	  D(SAAL_UDATA_request)
	  D(SAAL_UDATA_indication)
	}
# undef D
}

static void
dump_man(u_char *buf, size_t n)
{
	struct sscop_marg *a = (struct sscop_marg *)(void *)buf;
	struct sscop_merr *e = (struct sscop_merr *)(void *)buf;

	if (n < sizeof(a->sig)) {
		printf("management message too short\n");
		return;
	}

	switch (a->sig) {

	  case SSCOP_MDATA_indication:
		n -= sizeof(*a);
		printf("SSCOP_MDATA_indication: data=%zu\n", n);
		if (n > 0) {
			dump_hex(a->data, n);
			save_data(prefix, "m", &filenum, a->data, n);
		}
		break;

	  case SSCOP_MDATA_request:
		n -= sizeof(*a);
		printf("SSCOP_MDATA_request: data=%zu\n", n);
		if (n > 0) {
			dump_hex(a->data, n);
			save_data(prefix, "m", &filenum, a->data, n);
		}
		break;

	  case SSCOP_MERROR_indication:
		if (n != sizeof(*e)) {
			fprintf(stderr, "bad size of error: %zu\n", n);
			break;
		}
		printf("SSCOP_MERROR_indication: err='%c' cnt=%u\n",
		    e->err, e->cnt);
		break;

	  default:
		printf("bad signal: %u", a->sig);
		break;
	}
}

static void
ready_dmsg(int fd __unused, int mask __unused, void *uap __unused)
{
	u_char buf[MAXMSG];
	char hook[NG_HOOKLEN + 1];
	ssize_t n;

	if ((n = NgRecvData(dsock, buf, sizeof(buf), hook)) == -1)
		err(1, "NgRcvData");

	if (n == 0)
		errx(1, "EOF from '%s'", hook);

	if (strcmp(hook, "uni5") == 0) {
		dump_uni(buf, (size_t)n);
		doprompt = 1;

	} else if (strcmp(hook, "man5") == 0) {
		dump_man(buf, (size_t)n);
		doprompt = 1;

	} else {
		printf("%zd bytes from '%s'\n", n, hook);
		doprompt = 1;
	}
}

/************************************************************
 *
 * User interaction.
 */
static void
do_mdmsg(u_int sig, u_char *uu, size_t n)
{
	u_char buf[2 * LINE_MAX];
	struct sscop_marg *arg = (struct sscop_marg *)(void *)buf;

	memcpy(arg->data, uu, n);
	arg->sig = sig;

	if (NgSendData(dsock, "man5", (u_char *)arg, n + sizeof(*arg)) < 0)
		err(1, "NgSendData");
}

static void
do_msg(u_int sig, char *uu)
{
	u_char obuf[LINE_MAX + 1];
	struct sscfu_arg *arg = (struct sscfu_arg *)(void *)obuf;
	int n;

	if (uu != NULL) {
		if ((n = strunvis((char *)arg->data, uu)) == -1) {
			fprintf(stderr, "bad string '%s'\n", uu);
			return;
		}
	} else
		n = 0;

	arg->sig = sig;

	if (NgSendData(dsock, "uni5", (u_char *)arg, n + sizeof(*arg)) < 0)
		err(1, "NgSendData");
}

static void
do_dmsg(u_int sig, u_char *uu, size_t n)
{
	u_char buf[2 * LINE_MAX];
	struct sscfu_arg *arg = (struct sscfu_arg *)(void *)buf;

	memcpy(arg->data, uu, n);
	arg->sig = sig;

	if (NgSendData(dsock, "uni5", (u_char *)arg, n + sizeof(*arg)) < 0)
		err(1, "NgSendData");
}

static void
do_mmsg(u_int sig, char *uu)
{
	u_char obuf[LINE_MAX + 1];
	struct sscop_marg *arg = (struct sscop_marg *)(void *)obuf;
	int n;

	if (uu != NULL) {
		if ((n = strunvis((char *)arg->data, uu)) == -1) {
			fprintf(stderr, "bad string '%s'\n", uu);
			return;
		}
	} else
		n = 0;

	arg->sig = sig;

	if (NgSendData(dsock, "man5", (u_char *)arg, n + sizeof(*arg)) < 0)
		err(1, "NgSendData");
}

static void
set_flow_control(char **av)
{
	u_int h, l;
	char *end;

	if (av[1] == NULL || av[2] == NULL) {
		fprintf(stderr, "need two args\n");
		return;
	}
	l = (u_int)strtoul(av[1], &end, 10);
	if (*end != '\0') {
		fprintf(stderr, "bad low water mark\n");
		return;
	}
	h = (u_int)strtoul(av[2], &end, 10);
	if (*end != '\0') {
		fprintf(stderr, "bad high water mark\n");
		return;
	}
	if (l >= h) {
		fprintf(stderr, "low >= high water mark\n");
		return;
	}
	qstate.low_watermark = l;
	qstate.high_watermark = h;
}

static void
flow_control(char **av, int m)
{
	u_int curr;
	char *end;

	if (av[1] == NULL) {
		fprintf(stderr, "need one args\n");
		return;
	}
	curr = (u_int)strtoul(av[1], &end, 10);
	if (*end != '\0') {
		fprintf(stderr, "bad current\n");
		return;
	}
	if ((m == NGM_LOW_WATER_PASSED && curr > qstate.low_watermark) ||
	    (m == NGM_HIGH_WATER_PASSED && curr < qstate.high_watermark) ||
	    (curr > qstate.max_queuelen_packets)) {
		fprintf(stderr, "not in range %u...%u (%u)\n",
		    qstate.low_watermark, qstate.high_watermark,
		    qstate.max_queuelen_packets);
		return;
	}
	qstate.current = curr;

	if (NgSendMsg(csock, "sscfu5:lower", NGM_FLOW_COOKIE, m,
	     &qstate, sizeof(qstate)) < 0) {
		fprintf(stderr, "send msg: %s\n", strerror(errno));
	}
}


static void
ready_user(int fd __unused, int mask __unused, void *uap __unused)
{
	char line[LINE_MAX + 1];
	u_char buf[65536];
	int ac;
	int cnt, loop;
	ssize_t n;
	char *av[ARG_MAX];

	if ((ac = read_user(line, av)) == -1)
		exit(0);

	doprompt = 1;
	if (ac == 0)
		return;

	if (strlen(av[0]) != 1) {
		printf("%s", help);
		return;
	}

	switch (av[0][0]) {

	  default:
	  case 'h':
	  case '?':
		printf("%s", help);
		return;

	  case 'e':
		do_msg(SAAL_ESTABLISH_request, av[1]);
		break;

	  case 'r':
		do_msg(SAAL_RELEASE_request, av[1]);
		break;

	  case 'd':
		do_msg(SAAL_DATA_request, av[1]);
		break;

	  case 'D':
		if (ac >= 3)
			cnt = strtoul(av[2], NULL, 10);
		else
			cnt = 1;
		if ((n = read_file(av[1], buf, sizeof(buf))) >= 0) {
			for (loop = 0; loop < cnt; loop++)
				do_dmsg(SAAL_DATA_request, buf, (size_t)n);
		}
		break;

	  case 'u':
		do_msg(SAAL_UDATA_request, av[1]);
		break;

	  case 'U':
		if ((n = read_file(av[1], buf, sizeof(buf))) >= 0)
			do_mdmsg(SAAL_UDATA_request, buf, (size_t)n);
		break;

	  case 'm':
		do_mmsg(SSCOP_MDATA_request, av[1]);
		break;

	  case 'M':
		if ((n = read_file(av[1], buf, sizeof(buf))) >= 0)
			do_mdmsg(SSCOP_MDATA_request, buf, (size_t)n);
		break;

	  case 'p':
		if (av[1] == NULL)
			prefix[0] = '\0';
		else
			snprintf(prefix, sizeof(prefix), "%s", av[1]);
		break;

	  case 'f':
		flow_control(av, NGM_HIGH_WATER_PASSED);
		break;

	  case 'F':
		flow_control(av, NGM_LOW_WATER_PASSED);
		break;

	  case 'G':
		set_flow_control(av);
		break;

	  case 'v':
		if (av[1] == NULL || av[2] == NULL) {
			fprintf(stderr, "need protocol and level");
			break;
		}
		if (strcmp(av[1], "sscop") == 0) {
			sscop_verbose = strtoul(av[2], NULL, 16);
			if (NgSendAsciiMsg(csock, "sscop5:",
			    "setdebug %#x", sscop_verbose) < 0)
				err(1, "failed to set sscop debug level");
		} else if (strcmp(av[1], "sscf") == 0) {
			sscf_verbose = strtoul(av[2], NULL, 16);
			if (NgSendAsciiMsg(csock, "sscfu5:",
			    "setdebug %#x", sscf_verbose) < 0)
				err(1, "failed to set sscf debug level");
		} else
			fprintf(stderr, "bad protocol '%s'", av[1]);
		break;
	}
}

int
main(int argc, char *argv[])
{
	int opt;
	int debug = 0;
	u_int value;
	int n;

	while ((opt = getopt(argc, argv, "d:hs:v")) != -1)
		switch (opt) {

		  case 'd':
			device = optarg;
			break;

		  case 'h':
			fprintf(stderr, usgtxt);
			exit(0);

		  case 'v':
			NgSetDebug(++debug);
			break;

		  case 's':
			if (sscanf(optarg, "sscop=%u%n", &value, &n) == 2 &&
			    (int)strlen(optarg) == n)
				sscop_verbose = value;
			else if (sscanf(optarg, "sscf=%u%n", &value, &n) == 2 &&
			    (int)strlen(optarg) == n)
				sscf_verbose = value;
			else
				errx(1, "cannot parse -s proto=value");
			break;
		}

	argc -= optind;
	argv += optind;

	stack[0].path = device;

	atexit(shut);
	signal(SIGINT, onintr);

	build_stack(&csock, &dsock, stack);

	(void)poll_register(fileno(stdin), ready_user, NULL, POLL_IN);
	(void)poll_register(csock, ready_cmsg, NULL, POLL_IN);
	(void)poll_register(dsock, ready_dmsg, NULL, POLL_IN);

	init_stack();

	ev_loop(&doprompt, "sscfu");
}
