/*
 * Copyright (c) 2001-2003
 *	Fraunhofer Institute for Open Communication Systems (FhG Fokus).
 * 	All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Author: Harti Brandt <harti@freebsd.org>
 *
 * $Begemot: atmsupport/test_sscfu_user/sscfu_umain.c,v 1.5 2004/12/20 10:44:59 brandt Exp $
 *
 * User space SSCF-UNI test program.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <unistd.h>
#include <assert.h>
#include <fcntl.h>
#include <err.h>
#include <rpoll.h>

#include <netnatm/unimsg.h>
#include <netnatm/saal/sscop.h>
#include <netnatm/saal/sscfu.h>
#include "common.h"

static int end_at_eof = 1;	/* release connection at user EOF */
static int siged;		/* SIGINT catched */
static char *outf;		/* output file prefix */
static u_int sscf_vflag;	/* SSCF verbosity */

struct glue {
	struct sscfu *sscf;
	struct sscop *sscop;
} glue;

static const char usgtxt[] = "\
SSCF(U) test program\n\
Usage: sscf_u [-h] [-beFrx] [-ap=v] [-ofile] [-SX] [-tt=m] [-vX] [-VX] [-WN]\n\
Flags:\n\
  -a p=v  set parameter 'p' to 'v'\n\
  -b	  enable robustness enhancement\n\
  -e	  don't RELEASE_request on user EOF\n\
  -F	  use framing for sscop also\n\
  -h	  print this info\n\
  -o file output to file file.N\n\
  -r	  reverse user and sscop file descriptors\n\
  -S X	  set SSCF verbosity to X\n\
  -t t=m  set timer 't' to 'm' milliseconds\n\
  -v X	  set verbose flag to X\n\
  -V X	  set SSCOP verbosity to X\n\
  -W N	  set initial window to N\n\
  -x	  enable POLL after retransmission\n\
Timers are cc, poll, ka, nr or idle; parameters are j, k, cc, pd or stat.\n";

# define PROMPT(S)	write(user_fd, (S), strlen((S)))

static void sscop_send_manage(struct sscop *, void *, enum sscop_maasig,
	struct uni_msg *, u_int, u_int);
static void sscop_send_upper(struct sscop *, void *, enum sscop_aasig,
	struct uni_msg *, u_int);
static void sscop_send_lower(struct sscop *, void *, struct uni_msg *);

const struct sscop_funcs sscop_funcs = {
	sscop_send_manage,
	sscop_send_upper,
	sscop_send_lower,
	sscop_verbose,
	sscop_start_timer,
	sscop_stop_timer
};

static void sscfu_send_upper(struct sscfu *, void *, enum saal_sig,
	struct uni_msg *);
static void sscfu_send_lower(struct sscfu *, void *, enum sscop_aasig,
	struct uni_msg *, u_int);
static void sscfu_window(struct sscfu *, void *, u_int);
static void sscfu_verbose(struct sscfu *, void *, const char *, ...)
	__printflike(3, 4);

const struct sscfu_funcs sscfu_funcs = {
	sscfu_send_upper,
	sscfu_send_lower,
	sscfu_window,
	sscfu_verbose
};

static void
onsig(int sig __unused)
{

	siged++;
}

/*
 * SSCOP file descriptor is ready. Allocate and read one
 * message and dispatch a signal.
 */
static void
proto_infunc(int fd, int mask __unused, void *uap)
{
	struct uni_msg *m;

	if ((m = proto_msgin(fd)) != NULL)
		sscop_input((struct sscop *)uap, m);
}

/*
 * Got line from user. Interpret it.
 */
static void
uinput(struct sscfu *sscf, struct uni_msg *m)
{
	struct stat sb;
	int fd;

	static const char hlptxt[] =
		"h|?	- this help text\n"
		"e...	- establish request\n"
		"r...	- release request\n"
		"f...	- read data from file\n"
		"d...	- data request\n"
		"u...	- udata request\n";

	if (uni_msg_len(m) == 0) {
		fprintf(stderr, "ignoring zero-byte message\n");
		uni_msg_destroy(m);
		PROMPT("<- ");
		return;
	}

	switch (*m->b_rptr++) {

	  default:
		PROMPT("???\n");

	  case 'h':
	  case '?':
		fprintf(stderr, "%s", hlptxt);
		uni_msg_destroy(m);
		PROMPT("<- ");
		return;

	  case 'e':
		sscfu_saalsig(sscf, SAAL_ESTABLISH_request, m);
		break;

	  case 'r':
		sscfu_saalsig(sscf, SAAL_RELEASE_request, m);
		break;

	  case 'd':
		sscfu_saalsig(sscf, SAAL_DATA_request, m);
		break;

	  case 'u':
		sscfu_saalsig(sscf, SAAL_UDATA_request, m);
		break;

	  case 'f':
		if (uni_msg_len(m) < 1) {
			fprintf(stderr, "which file?\n");
			uni_msg_destroy(m);
			break;
		}
		m->b_wptr[0] = 0;
		if (stat((char *)m->b_rptr, &sb)) {
			fprintf(stderr, "stat(%s) failed: %s\n",
			    m->b_rptr, strerror(errno));
			uni_msg_destroy(m);
			break;
		}
		if (sb.st_size == 0) {
			fprintf(stderr, "%s: emtpy file", m->b_rptr);
			uni_msg_destroy(m);
			break;
		}
		if (sb.st_size > MAXUSRMSG) {
			fprintf(stderr, "%s: too large", m->b_rptr);
			uni_msg_destroy(m);
			break;
		}
		if ((fd = open((char *)m->b_rptr, O_RDONLY)) < 0) {
			fprintf(stderr, "%s: %s", m->b_rptr, strerror(errno));
			uni_msg_destroy(m);
			break;
		}
		uni_msg_destroy(m);
		if ((m = uni_msg_alloc((size_t)sb.st_size)) == NULL)
			err(1, NULL);
		if (read(fd, m->b_rptr, (size_t)sb.st_size) != sb.st_size) {
			fprintf(stderr, "%s: read error %s", m->b_rptr,
			    strerror(errno));
			uni_msg_destroy(m);
			(void)close(fd);
			break;
		}
		m->b_wptr = m->b_rptr + sb.st_size;
		(void)close(fd);
		sscfu_saalsig(sscf, SAAL_DATA_request, m);
		break;
	}
	PROMPT("<- ");
}

/*
 * User input. Allocate and read message and dispatch signal.
 */
static void
user_infunc(int fd, int mask __unused, void *uap)
{
	struct uni_msg *m;
	size_t size;
	struct sscfu *sscf = uap;

	if ((m = uni_msg_alloc(MAXMSG)) == NULL)
		err(1, NULL);

	if ((ssize_t)(size = read(fd, m->b_rptr, MAXUSRMSG)) < 0)
		err(1, "read user");

	if (size == 0) {
		poll_unregister(user_h);
		user_h = -1;
		uni_msg_destroy(m);
		VERBOSE(("EOF on user connection"));
		if (end_at_eof)
			sscfu_saalsig(sscf, SAAL_RELEASE_request, NULL);
		return;
	}
	if (m->b_rptr[size-1] == '\n')
		if (--size == 0) {
			uni_msg_destroy(m);
			PROMPT("<- ");
			return;
		}
	m->b_wptr = m->b_rptr + size;

	uinput(sscf, m);
}

int
main(int argc, char *argv[])
{
	int opt;
	struct sscfu *sscf;
	struct sscop *sscop;
	struct sigaction sa;
	struct sscop_param param;
	u_int mask;

	memset(&param, 0, sizeof(param));
	param.maxk = MAXUSRMSG;
	param.maxj = 0;
	param.maxstat = 7;
	param.maxcc = 4;
	mask = SSCOP_SET_MAXK | SSCOP_SET_MAXJ | SSCOP_SET_MAXSTAT
	    | SSCOP_SET_MAXCC;

	sscop_fd = 0;
	user_fd = 1;
	while ((opt = getopt(argc, argv, "hbeFrxa:o:S:t:v:V:W:")) != -1)
		switch (opt) {

		  case 'h':
			fprintf(stderr, usgtxt);
			exit(0);

		  case 'e':
			end_at_eof = 0;
			break;

		  case 'F':
			sscopframe = 1;
			break;

		  case 'o':
			outf = optarg;
			break;

		  case 'r':
			sscop_fd = 1;
			user_fd = 0;
			break;

		  case 'S':
			sscf_vflag = strtoul(optarg, NULL, 16);
			break;

		  case 'v':
			sscop_vflag = strtoul(optarg, NULL, 16);
			break;

		  case 'V':
			verbose = strtoul(optarg, NULL, 16);
			break;

		  case 'a':
		  case 't':
		  case 'b':
		  case 'x':
		  case 'W':
			parse_param(&param, &mask, opt, optarg);
			break;
		}

	/*
	 * Catch SIGINT and SIGUSR1
	 */
	if (sigaction(SIGINT, NULL, &sa))
		err(1, "can't get current SIGINT");
	if (sa.sa_handler != SIG_IGN) {
		sa.sa_handler = onsig;
		sa.sa_flags = SA_RESTART;
		sigemptyset(&sa.sa_mask);
		if (sigaction(SIGINT, &sa, NULL))
			err(1, "can't catch SIGINT");
	}
	sa.sa_handler = onsig;
	sa.sa_flags = SA_RESTART;
	sigemptyset(&sa.sa_mask);
	if (sigaction(SIGUSR1, &sa, NULL))
		err(1, "can't catch SIGUSR1");

	/*
	 * Construct the protocol stack
	 */
	if ((sscf = sscfu_create(&glue, &sscfu_funcs)) == NULL)
		err(1, NULL);
	sscfu_setdebug(sscf, sscf_vflag);

	if ((sscop = sscop_create(&glue, &sscop_funcs)) == NULL)
		err(1, NULL);
	if ((errno = sscop_setparam(sscop, &param, &mask)) != 0)
		err(1, "can't set sscop parameters: 0x%x", mask);
	sscop_setdebug(sscop, sscop_vflag);

	glue.sscf = sscf;
	glue.sscop = sscop;

	sscop_h = poll_register(sscop_fd, proto_infunc, sscop, POLL_IN);
	user_h = poll_register(user_fd, user_infunc, sscf, POLL_IN);

	/*
	 * Run the protocols
	 */
	PROMPT("<- ");
	while (!siged && sscop_fd >= 0)
		poll_dispatch(1);

	if (siged)
		VERBOSE(("... signal catched - exiting"));
	else
		VERBOSE(("... files close - exiting"));

	sscop_destroy(sscop);
	sscfu_destroy(sscf);

	return (0);
}

/*
 * Send AAL data 
 */
static void
sscop_send_lower(struct sscop *sscop __unused, void *arg __unused,
    struct uni_msg *m)
{
	proto_msgout(m);
}

/*
 * User SSCF signal handler
 */
static void
sscfu_send_upper(struct sscfu *sscf __unused, void *arg __unused,
    enum saal_sig sig, struct uni_msg *m)
{
	size_t ret;
	char fbuf[1000];
	int fd;

	static int cnt;

	VERBOSE(("delivering signal to USER %s", sscfu_signame(sig)));

	switch (sig) {

	  case SAAL_ESTABLISH_confirm:
		PROMPT("-> establish_confirm(");
		if (m) {
			write(user_fd, m->b_rptr, uni_msg_len(m));
			uni_msg_destroy(m);
		}
		PROMPT(")\n<- ");
		break;

	  case SAAL_ESTABLISH_indication:
		PROMPT("-> establish_indication(");
		if (m) {
			write(user_fd, m->b_rptr, uni_msg_len(m));
			uni_msg_destroy(m);
		}
		PROMPT(")\n<- ");
		break;

	  case SAAL_RELEASE_confirm:
		PROMPT("-> release_confirm\n<- ");
		if (end_at_eof) {
			VERBOSE(("... exiting"));
			poll_unregister(sscop_h);
			sscop_h = -1;

			(void)close(sscop_fd);
			sscop_fd = -1;
		}
		break;

	  case SAAL_RELEASE_indication:
		PROMPT("-> release_indication(");
		if (m) {
			write(user_fd, m->b_rptr, uni_msg_len(m));
			uni_msg_destroy(m);
		}
		PROMPT(")\n<- ");
		if (end_at_eof) {
			VERBOSE(("... exiting"));
			poll_unregister(sscop_h);
			sscop_h = -1;

			(void)close(sscop_fd);
			sscop_fd = -1;
		}
	  	break;

	  case SAAL_UDATA_indication:
		PROMPT("-> u");
		assert(m != NULL);
		write(user_fd, m->b_rptr, uni_msg_len(m));
		PROMPT("\n<- ");
		uni_msg_destroy(m);
		break;

	  case SAAL_DATA_indication:
		PROMPT("-> d");
		assert(m != NULL);
		if (outf == NULL) {
			write(user_fd, m->b_rptr, uni_msg_len(m));
			PROMPT("\n<- ");
		} else {
			sprintf(fbuf, "%s.%d", outf, cnt++);
			PROMPT(" <");
			PROMPT(fbuf);
			PROMPT(">\n<- ");
			if ((fd = open(fbuf, O_WRONLY | O_CREAT | O_TRUNC,
			    0666)) < 0) {
				fprintf(stderr, "%s: %s\n", fbuf,
				    strerror(errno));
				uni_msg_destroy(m);
				PROMPT("<- ");
				break;
			}

			ret = write(fd, m->b_rptr, uni_msg_len(m));
			if ((ssize_t)ret == -1 || ret != uni_msg_len(m)) {
				fprintf(stderr, "%s: write error %s\n", fbuf,
					strerror(errno));
				uni_msg_destroy(m);
				PROMPT("<- ");
				break;
			}

			(void)close(fd);
		}
		uni_msg_destroy(m);
		break;

	  default:
		errx(1, "bad SAAL signal");
	}
}

static void
sscop_send_manage(struct sscop *sscop __unused, void *arg __unused,
    enum sscop_maasig sig, struct uni_msg *m, u_int errcode, u_int cnt)
{
	VERBOSE(("delivering management signal to USER %s",
		sscop_msigname(sig)));

	switch (sig) {

	  case SSCOP_MDATA_indication:
		PROMPT("-> m");
		assert(m != NULL);
		write(user_fd, m->b_rptr, uni_msg_len(m));
		uni_msg_destroy(m);
		PROMPT("\n");
		break;

	  case SSCOP_MERROR_indication:
		VERBOSE(("MAAL-ERROR.indication '%c' %u", errcode, cnt));
		break;

	  case SSCOP_MDATA_request:
		abort();
	}
}

/*
 * Upper layer signal from SSCOP - relay to SSCF.
 */
static void
sscop_send_upper(struct sscop *sscop __unused, void *arg, enum sscop_aasig sig,
    struct uni_msg *m, u_int p)
{
	sscfu_input(((struct glue *)arg)->sscf, sig, m, p);
}

/*
 * Lower layer signal from SSCF - relay to SSCOP.
 */
static void
sscfu_send_lower(struct sscfu *sscf __unused, void *arg, enum sscop_aasig sig,
    struct uni_msg *m, u_int p)
{
	sscop_aasig(((struct glue *)arg)->sscop, sig, m, p);
}

/*
 * Signal from SSCF to move SSCOP window.
 */
static void
sscfu_window(struct sscfu *sscf __unused, void *arg, u_int n)
{
	(void)sscop_window(((struct glue *)arg)->sscop, n);
}

/*
 * This function get's called from SSCF to put out verbose messages
 */
static void 
sscfu_verbose(struct sscfu *sscfu __unused, void *u __unused,
    const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	fprintf(stderr, "\n");
	va_end(ap);
}
