/*
 * Copyright (c) 2001-2003
 *	Fraunhofer Institute for Open Communication Systems (FhG Fokus).
 * 	All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Author: Harti Brandt <harti@freebsd.org>
 *
 * ng_sscop test program
 *
 * $Begemot: atmsupport/test_sscop/sscoptest.c,v 1.5 2004/12/20 10:45:00 brandt Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <vis.h>
#include <err.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <limits.h>
#include <netgraph.h>
#include <rpoll.h>
#include <netnatm/saal/sscopdef.h>
#include <netgraph/atm/ng_sscop.h>
#include "testlib.h"

static int	csock, dsock;
static char	prefix[100];
static u_int	filenum;
static int 	trace;
static int	doprompt;		/* set to 1 to get a prompt */
static const char *device = "fatm0:";
static u_int	sscop_verbose;

/*                [atmtest]
 *		      |
 *                    |(uni5)
 *                    +----------------+
 *                    |(man5) (manage) |(upper)
 *                    +--------------[sscop5]
 *                    |                |(lower)
 *           (atm_out)|(right2left)    |(right)
 *                    +--------------[tee]
 *            (atm_in) (left2right)    |(left)
 *                                     |(sig5)
 *                                  [fatm]
 */
static struct stack stack[] = {
	{ STACK_MKPEER, NULL,  	  "tee",     "sig5",  "left",  "ttee" },
	{ STACK_MKPEER, "ttee:",  "sscop",   "right", "lower", "sscop5" },
	{ STACK_CONNECT, ".",     "sscop5:", "man5",  "manage", NULL },
	{ STACK_CONNECT, ".",     "sscop5:", "uni5",  "upper",  NULL },
	{ STACK_CONNECT, ".",     "ttee:", "atm_out", "right2left", NULL },
	{ STACK_CONNECT, ".",     "ttee:", "atm_in",  "left2right", NULL },
	{ STACK_END, NULL, NULL, NULL, NULL, NULL },
};

static const char help[] = "\
h			help\n\
a			toggle tracing\n\
c			recover-response\n\
d [data]		data-request\n\
D file			data-request from file\n\
e [uu-data]		establish-request\n\
E [uu-data]		establish-response\n\
m [data]		mdata-request\n\
M file			mdata-request from file\n\
p [name]		set/clear data save prefix\n\
r [uu-data]		release-request\n\
s [data]		resynch-request\n\
S			resynch-response\n\
t N			retrieve-request\n\
u [data]		udata-request\n\
U file			udata-request from file\n\
v X			set debugging level\n\
";

static const char usgtxt[] = "\
SSCOP test program\n\
usage: sscoptest [-h] [-tv] [-d dev:] [-s X]\n\
where:\n\
  -d dev:	use ATM node dev: instead of default fatm0:\n\
  -h		print this help text\n\
  -s X		set SSCOP verbosity to hex X\n\
  -t		print a trace of messages\n\
  -v		increment netgraph verbosity level\n\
";

/*
 * Something happend on the control socket. Try to read a message and to
 * dump it.
 */
static void
ready_cmsg(int fd __unused, int mask __unused, void *uap __unused)
{

	read_cmsg(csock);
	doprompt = 1;
}

/* 
 * Dump a message at the upper interface of SSCOP
 */
static void
dump_uni(u_char *buf, size_t n)
{
	struct sscop_arg *a = (struct sscop_arg *)(void *)buf;

	if (n < sizeof(*a)) {
		printf("uni message too short\n");
		return;
	}
	n -= sizeof(*a);

# define D(X) 								\
	  case X:							\
		printf("%s: arg=%u data=%zu\n", #X, a->arg, n);		\
		if (n > 0) {						\
			dump_hex(a->data, n);				\
			save_data(prefix, "d", &filenum, a->data, n);	\
		}							\
		break;

	switch (a->sig) {

	  default:
		printf("bad signal: %u", a->sig);
		break;

	  D(SSCOP_ESTABLISH_request)
	  D(SSCOP_ESTABLISH_indication)
	  D(SSCOP_ESTABLISH_response)
	  D(SSCOP_ESTABLISH_confirm)
	  D(SSCOP_RELEASE_request)
	  D(SSCOP_RELEASE_indication)
	  D(SSCOP_RELEASE_confirm)
	  D(SSCOP_DATA_request)
	  D(SSCOP_DATA_indication)
	  D(SSCOP_UDATA_request)
	  D(SSCOP_UDATA_indication)
	  D(SSCOP_RECOVER_indication)
	  D(SSCOP_RECOVER_response)
	  D(SSCOP_RESYNC_request)
	  D(SSCOP_RESYNC_indication)
	  D(SSCOP_RESYNC_response)
	  D(SSCOP_RESYNC_confirm)
	  D(SSCOP_RETRIEVE_request)
	  D(SSCOP_RETRIEVE_indication)
	  D(SSCOP_RETRIEVE_COMPL_indication)
	}
# undef D
}

/*
 * Dump a SSCOP management message
 */
static void
dump_man(u_char *buf, size_t n)
{
	struct sscop_marg *a = (struct sscop_marg *)(void *)buf;
	struct sscop_merr *e = (struct sscop_merr *)(void *)buf;

	if (n < sizeof(a->sig)) {
		printf("management message too short\n");
		return;
	}

	switch (a->sig) {

	  case SSCOP_MDATA_indication:
		n -= sizeof(*a);
		printf("SSCOP_MDATA_indication: data=%zu\n", n);
		if (n > 0) {
			dump_hex(a->data, n);
			save_data(prefix, "m", &filenum, a->data, n);
		}
		break;

	  case SSCOP_MDATA_request:
		n -= sizeof(*a);
		printf("SSCOP_MDATA_request: data=%zu\n", n);
		if (n > 0) {
			dump_hex(a->data, n);
			save_data(prefix, "m", &filenum, a->data, n);
		}
		break;

	  case SSCOP_MERROR_indication:
		if (n != sizeof(*e)) {
			fprintf(stderr, "bad size of error: %zu\n", n);
			break;
		}
		printf("SSCOP_MERROR_indication: err='%c' cnt=%u\n",
		    e->err, e->cnt);
		break;

	  default:
		printf("bad signal: %u", a->sig);
		break;
	}
}

/*
 * Dump a message at the lower SSCOP interface if tracing is enabled.
 */
static void
dump_atm(u_char *buf, size_t n, int dir)
{
	u_int pdu, pl, s;

	printf("MESSAGE %s SSCOP (%zu bytes):\n", dir ? "FROM" : "TO", n);

	if (n < 4) {
		dump_hex(buf, n);
		return;
	}
	pdu = buf[n-4] & 0x0f;
	pl = (buf[n-4] & 0xc0) >> 6;
	s = (buf[n-4] & 0x10) >> 4;

	switch (pdu) {

	  case 0x00:
		printf("ILL PDU 0x0");
		break;

	  case 0x01:
		printf("BGN pl=%u", pl);
		break;

	  case 0x02:
		printf("BGAK pl=%u", pl);
		break;

	  case 0x03:
		printf("END pl=%u s=%u", pl, s);
		break;

	  case 0x04:
		printf("ENDAK");
		break;

	  case 0x05:
		printf("RS pl=%u", pl);
		break;

	  case 0x06:
		printf("RSAK");
		break;

	  case 0x07:
		printf("BGREJ pl=%u", pl);
		break;

	  case 0x08:
		printf("SD pl=%u", pl);
		break;

	  case 0x09:
		printf("ER");
		break;

	  case 0x0a:
		printf("POLL");
		break;

	  case 0x0b:
		printf("STAT");
		break;

	  case 0x0c:
		printf("USTAT");
		break;

	  case 0x0d:
		printf("UD pl=%u", pl);
		break;

	  case 0x0e:
		printf("MD pl=%u", pl);
		break;

	  case 0x0f:
		printf("ERAK");
		break;
	}

	printf("\n");

	dump_hex(buf, n);
}

/*****************************************************************
 *
 * Input from the data socket
 */
static void
ready_dmsg(int fd __unused, int mask __unused, void *uap __unused)
{
	u_char buf[MAXMSG];
	char hook[NG_HOOKLEN + 1];
	ssize_t n;

	if ((n = NgRecvData(dsock, buf, sizeof(buf), hook)) < 0)
		err(1, "NgRecvData");

	if (n == 0)
		errx(1, "EOF from '%s'\n", hook);

	if (strcmp(hook, "uni5") == 0) {
		dump_uni(buf, (size_t)n);
		doprompt = 1;

	} else if (strcmp(hook, "man5") == 0) {
		dump_man(buf, (size_t)n);
		doprompt = 1;

	} else if (strcmp(hook, "atm_in") == 0) {
		if (trace) {
			dump_atm(buf, (size_t)n, 0);
			doprompt = 1;
		}

	} else if (strcmp(hook, "atm_out") == 0) {
		if (trace) {
			dump_atm(buf, (size_t)n, 1);
			doprompt = 1;
		}

	} else {
		printf("%zd bytes from '%s'\n", n, hook);
		doprompt = 1;
	}
}

/*
 * Called on shutdown via atexit(3) to remove the stack
 */
static void
shut(void)
{

	shutdown_stack(csock, stack);
}

/*
 * SIGINT handler
 */
static void
onintr(int sig)
{

	sig = sig;
	exit(0);
}

/*
 * Initialize the nodes.
 */
static void
init_stack(void)
{
	if (NgSendAsciiMsg(csock, device,
	    "cpcsinit {name=\"sig5\" flags=0x4 traffic=0 aal=5 vci=5}") < 0)
		err(1, "failed to init VCI");

	if (NgSendAsciiMsg(csock, "sscop5:",
	    "setparam {mask=0x800 param={flags=0x01}}") < 0)

	read_cmsg(csock);

	if (NgSendAsciiMsg(csock, "sscop5:",
	    "enable") < 0)
		err(1, "failed to enable SSCOP");

	if (NgSendAsciiMsg(csock, "sscop5:",
	    "setdebug %#x", sscop_verbose) < 0)
		err(1, "failed to set sscop debug level");
}

/*****************************************************************
 *
 * User interaction.
 */
/*
 * Send the signal with (optional) data to either the upper or the management
 * interface.
 */
static void
do_msg(int mng, u_int sig, char *uu, u_int rn)
{
	char obuf[LINE_MAX + 1];
	struct sscop_arg *arg = (struct sscop_arg *)(void *)obuf;
	int n;

	if (uu != NULL) {
		if ((n = strunvis((char *)arg->data, uu)) == -1) {
			fprintf(stderr, "bad string '%s'\n", uu);
			return;
		}
	} else
		n = 0;

	arg->sig = sig;
	arg->arg = rn;

	if (NgSendData(dsock, mng ? "man5" : "uni5", (u_char *)arg,
	    n + sizeof(*arg)) < 0)
		err(1, "NgSendData");
}

/*
 * Send the signal without data to either the upper or the management
 * interface.
 */
static void
do_dmsg(int mng, u_int sig, u_char *uu, size_t n, u_int rn)
{
	char buf[2 * LINE_MAX];
	struct sscop_arg *arg = (struct sscop_arg *)(void *)buf;

	memcpy(arg->data, uu, n);
	arg->sig = sig;
	arg->arg = rn;

	if (NgSendData(dsock, mng ? "man5" : "uni5", (u_char *)arg,
	    n + sizeof(*arg)) < 0)
		err(1, "NgSendData");
}

/*
 * Read from the user and do the right thing
 */
static void
ready_user(int fd __unused, int mask __unused, void *uap __unused)
{
	char line[LINE_MAX+1];
	u_char buf[65536];
	int argc;
	u_int q;
	ssize_t n;
	char *av[ARG_MAX], *end;

	if ((argc = read_user(line, av)) == -1)
		exit(0);

	doprompt = 1;
	if (argc == 0)
		return;

	if (strlen(av[0]) != 1) {
		printf("%s", help);
		return;
	}
	switch (av[0][0]) {

	  default:
	  case 'h':
	  case '?':
		printf("%s", help);
		return;

	  case 'E':
		do_msg(0, SSCOP_ESTABLISH_response, av[1], 0);
		break;

	  case 'e':
		do_msg(0, SSCOP_ESTABLISH_request, av[1], 0);
		break;

	  case 'r':
		do_msg(0, SSCOP_RELEASE_request, av[1], 0);
		break;

	  case 'd':
		do_msg(0, SSCOP_DATA_request, av[1], 0);
		break;

	  case 'D':
		if ((n = read_file(av[1], buf, sizeof(buf))) >= 0)
			do_dmsg(0, SSCOP_DATA_request, buf, (size_t)n, 0);
		break;

	  case 'u':
		do_msg(0, SSCOP_UDATA_request, av[1], 0);
		break;

	  case 'U':
		if ((n = read_file(av[1], buf, sizeof(buf))) >= 0)
			do_dmsg(0, SSCOP_UDATA_request, buf, (size_t)n, 0);
		break;

	  case 'm':
		do_msg(1, SSCOP_MDATA_request, av[1], 0);
		break;

	  case 'M':
		if ((n = read_file(av[1], buf, sizeof(buf))) >= 0)
			do_dmsg(1, SSCOP_MDATA_request, buf, (size_t)n, 0);
		break;

	  case 'c':
		do_msg(0, SSCOP_RECOVER_response, NULL, 0);
		break;

	  case 's':
		do_msg(0, SSCOP_RESYNC_request, av[1], 0);
		break;

	  case 'S':
		do_msg(0, SSCOP_RESYNC_response, NULL, 0);
		break;

	  case 't':
		if (av[1] == NULL) {
			fprintf(stderr, "need sequence number");
			break;
		}
		if (strcmp(av[1], "u") == 0) {
			q = SSCOP_RETRIEVE_UNKNOWN;
		} else if (strcmp(av[1], "t") == 0) {
			q = SSCOP_RETRIEVE_TOTAL;
		} else {
			q = strtoul(av[1], &end, 10);
			if (*end != '\0') {
				fprintf(stderr, "bad sequence number '%s'\n", av[1]);
				break;
			}
		}
		do_msg(0, SSCOP_RETRIEVE_request, NULL, q);
		break;

	  case 'p':
		if (av[1] == NULL)
			prefix[0] = '\0';
		else
			snprintf(prefix, sizeof(prefix), "%s", av[1]);
		break;

	  case 'a':
		trace = !trace;
		break;

	  case 'v':
		if (av[1] == NULL) {
			fprintf(stderr, "need level");
			break;
		}
		sscop_verbose = strtoul(av[1], NULL, 16);
		if (NgSendAsciiMsg(csock, "sscop5:",
		    "setdebug %#x", sscop_verbose) < 0)
			err(1, "failed to set sscop debug level");
		break;
	}
}

/*
 * The program
 */
int
main(int argc, char *argv[])
{
	int opt;
	int debug = 0;

	while ((opt = getopt(argc, argv, "htvd:s:")) != -1)
		switch (opt) {

		  case 'd':
			device = optarg;
			break;

		  case 'h':
			fprintf(stderr, usgtxt);
			exit(0);

		  case 's':
			sscop_verbose = strtoul(optarg, NULL, 16);
			break;

		  case 't':
			trace++;
			break;

		  case 'v':
			NgSetDebug(++debug);
			break;
		}

	argc -= optind;
	argv += optind;

	stack[0].path = device;

	atexit(shut);
	signal(SIGINT, onintr);

	build_stack(&csock, &dsock, stack);

	(void)poll_register(fileno(stdin), ready_user, NULL, POLL_IN);
	(void)poll_register(csock, ready_cmsg, NULL, POLL_IN);
	(void)poll_register(dsock, ready_dmsg, NULL, POLL_IN);

	init_stack();

	ev_loop(&doprompt, "sscop");
}
