/*
 * Copyright (c) 2001-2003
 *	Fraunhofer Institute for Open Communication Systems (FhG Fokus).
 * 	All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Author: Harti Brandt <harti@freebsd.org>
 *
 * $Begemot: atmsupport/test_sscop_user/sscop_umain.c,v 1.5 2004/12/20 10:45:02 brandt Exp $
 *
 * User space SSCOP test program.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <unistd.h>
#include <assert.h>
#include <err.h>
#include <rpoll.h>

#include <netnatm/unimsg.h>
#include <netnatm/saal/sscop.h>
#include "common.h"

static int end_at_eof = 1;	/* send RELEASE_request at user EOF */
static int sigusr1;
static u_int window;		/* how many messages have not been avanced */
static int wstop;		/* window avancement stopped */

static volatile int ready;	/* flag if connection is established */

static const char usgtxt[] = "\
SSCOP(U) test program\n\
Usage: sscop_u [-h] [-beFrwx] [-ap=v] [-tt=m] [-v X] [-V X] [-W N]\n\
Flags:\n\
  -a p=v  set parameter 'p' to 'v'\n\
  -b	  enable robustness enhancement\n\
  -e	  don't RELEASE_request on user EOF\n\
  -F	  use framing for sscop also\n\
  -h	  print this info\n\
  -r	  reverse user and sscop file descriptors\n\
  -t t=m  set timer 't' to 'm' milliseconds\n\
  -v X	  set verbose flag to hex X\n\
  -V X	  set sscop verbose flag to hex X\n\
  -w	  don't start conversation\n\
  -W N	  set initial window to N\n\
  -x	  enable POLL after retransmission\n\
Timers are cc, poll, ka, nr or idle; parameters are j, k, cc, pd or stat.\n";

static void uprintf(const char *fmt, ...) __printflike(1, 2);

#define PROMPT(S) write(user_fd, (S), strlen((S)))

static void sscop_send_upper(struct sscop *, void *, enum sscop_aasig,
    struct uni_msg *, u_int);
static void sscop_send_lower(struct sscop *, void *, struct uni_msg *);
static void sscop_send_manage(struct sscop *, void *, enum sscop_maasig,
    struct uni_msg *, u_int, u_int);

static const struct sscop_funcs sscop_funcs = {
	sscop_send_manage,
	sscop_send_upper,
	sscop_send_lower,
	sscop_verbose,
	sscop_start_timer,
	sscop_stop_timer
};

static void
onusr1(int s __unused)
{
	sigusr1++;
}

/*
 * SSCOP file descriptor is ready. Allocate and read one message
 * and dispatch a signal.
 */
static void
proto_infunc(int fd, int mask __unused, void *uap)
{
	struct uni_msg *m;

	if ((m = proto_msgin(fd)) != NULL)
		sscop_input((struct sscop *)uap, m);
}

/*
 * Got packet from user. Interpret it.
 */
static void
uinput(struct sscop *sscop, struct uni_msg *m)
{
	static const char hlptxt[] =
		"h|?	- this help page\n"
		"e	- establish request\n"
		"E	- establish response\n"
		"r	- release request\n"
		"d...	- data request\n"
		"u...	- unitdata request\n"
		"m...	- managment data request\n"
		"c	- recover response\n"
		"s	- resync request\n"
		"S	- resync response\n"
		"W	- stop window avance\n"
		"wN	- avance window by N messages\n"
		"x	- exit\n";

	if (uni_msg_len(m) == 0) {
		uprintf("ignoring zero-byte message");
		uni_msg_destroy(m);
		return;
	}

	switch (*m->b_rptr++) {

	  default:
		uprintf("unknown command '%c'", m->b_rptr[-1]);
		uni_msg_destroy(m);
		PROMPT("<- ");
		return;

	  case '?':
	  case 'h':
		uprintf("%s", hlptxt);
		uni_msg_destroy(m);
		PROMPT("<- ");
		return;

	  case 'e':
		sscop_aasig(sscop, SSCOP_ESTABLISH_request, m, 1);
		break;

	  case 'E':
		sscop_aasig(sscop, SSCOP_ESTABLISH_response, m, 1);
		break;

	  case 'r':
		sscop_aasig(sscop, SSCOP_RELEASE_request, m, 0);
		break;

	  case 'd':
		sscop_aasig(sscop, SSCOP_DATA_request, m, 0);
		break;

	  case 'm':
		sscop_maasig(sscop, SSCOP_MDATA_request, m);
		break;

	  case 'u':
		sscop_aasig(sscop, SSCOP_UDATA_request, m, 0);
		break;

	  case 'c':
		uni_msg_destroy(m);
		sscop_aasig(sscop, SSCOP_RECOVER_response, NULL, 0);
		break;

	  case 's':
		sscop_aasig(sscop, SSCOP_RESYNC_request, m, 0);
		break;

	  case 'S':
		uni_msg_destroy(m);
		sscop_aasig(sscop, SSCOP_RESYNC_response, NULL, 0);
		break;

	  case 'x':
		exit(0);
		break;

	  case 'W':
		uni_msg_destroy(m);
		wstop = !wstop;
		uprintf("window avancing %s\n", wstop ? "stopped" : "started");
		window = 0;
		break;

	  case 'w':
		if (wstop) {
			if (uni_msg_len(m) == 0) {
				if (window != 0) {
					uprintf("advancing by %d msgs\n",
					    window);
					sscop_window(sscop, +window);
					window = 0;
				}
			} else {
				u_int w;

				*m->b_wptr = '\0';
				if (sscanf((char *)m->b_rptr, "%u", &w) == 1 &&
				    w > 0) {
					if (w > window)
						w = window;
					uprintf("advancing by %d msgs\n", w);
					sscop_window(sscop, +w);
					window -= w;
				} else {
					uprintf("bad number '%s'\n", m->b_rptr);
				}
			}
		} else {
			uprintf("window automatically avanced\n");
		}
		uni_msg_destroy(m);
		break;

	}
	PROMPT("<- ");
}


/*
 * User input. Allocate and read message and dispatch signal.
 */
static void
user_infunc(int fd, int mask __unused, void *uap)
{
	struct sscop *sscop = uap;
	struct uni_msg *m;
	size_t size;

	if ((m = uni_msg_alloc(MAXMSG)) == NULL)
		err(1, NULL);

	if ((ssize_t)(size = read(fd, m->b_rptr, MAXUSRMSG)) == -1)
		err(1, "read user");

	if (size == 0) {
		poll_unregister(user_h);
		user_h = -1;
		uni_msg_destroy(m);
		VERBOSE(("EOF on user connection"));
		if (end_at_eof)
			sscop_aasig(sscop, SSCOP_RELEASE_request, NULL, 0);
		return;
	}
	if (m->b_rptr[size-1] == '\n')
		if (--size == 0) {
			uni_msg_destroy(m);
			PROMPT("-> ");
			return;
		}
	m->b_wptr = m->b_rptr + size;

	uinput(sscop, m);
}


int
main(int argc, char *argv[])
{
	int opt;
	struct sscop *sscop;
	struct sigaction sa;
	struct sscop_param param;
	u_int mask;
	int wait = 0;

	/*
	 * Default is to have the USER on stdin and SSCOP on stdout
	 */
	sscop_fd = 0;
	user_fd = 1;

	memset(&param, 0, sizeof(param));
	param.maxk = MAXUSRMSG;
	param.maxj = 0;
	param.maxstat = 7;
	param.maxcc = 4;
	mask = SSCOP_SET_MAXK | SSCOP_SET_MAXJ | SSCOP_SET_MAXSTAT
	    | SSCOP_SET_MAXCC;

	while ((opt = getopt(argc, argv, "hbeFrwxa:t:v:V:W:")) != -1)
		switch(opt) {

		  case 'h':
			fprintf(stderr, usgtxt);
			exit(0);

		  case 'e':
			end_at_eof = 0;
			break;

		  case 'F':
			sscopframe = 1;
			break;

		  case 'r':
			sscop_fd = 1;
			user_fd = 0;
			break;

		  case 'w':
			wait = 1;
			break;

		  case 'v':
			sscop_vflag = strtoul(optarg, NULL, 16);
			break;

		  case 'V':
			verbose = strtoul(optarg, NULL, 16);
			break;

		  case 'a':
		  case 't':
		  case 'b':
		  case 'x':
		  case 'W':
			parse_param(&param, &mask, opt, optarg);
			break;
		}

	/*
	 * Catch USR1
	 */
	sa.sa_handler = onusr1;
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = SA_RESTART;
	if(sigaction(SIGUSR1, &sa, NULL))
		err(1, "sigaction(SIGUSR1)");

	/*
	 * Allocate and initialize SSCOP
	 */
	if ((sscop = sscop_create(NULL, &sscop_funcs)) == NULL)
		err(1, NULL);
	sscop_setdebug(sscop, sscop_vflag);
	if((errno = sscop_setparam(sscop, &param, &mask)) != 0)
		err(1, "can't set sscop parameters: 0x%x", mask);

	/*
	 * Select on file descriptors
	 */
	sscop_h = poll_register(sscop_fd, proto_infunc, sscop, POLL_IN);
	user_h = poll_register(user_fd, user_infunc, sscop, POLL_IN);

	PROMPT("<- ");
	while (!sigusr1 && sscop_fd >= 0)
		poll_dispatch(1);

	VERBOSE(("SSCOP file descriptor closed - exiting"));
	sscop_destroy(sscop);

	return (0);
}

/*
 * AAL OUTPUT
 */
static void
sscop_send_lower(struct sscop *sscop __unused, void *arg __unused,
    struct uni_msg *m)
{
	proto_msgout(m);
}


/*
 * Write the message to the user and move the window
 */
static void
uoutput(struct sscop *sscop __unused, struct uni_msg *m)
{
	size_t ret;

	if (m == NULL)
		return;

	if ((ssize_t)(ret = write(user_fd, m->b_rptr, uni_msg_len(m))) == -1
	    || ret != uni_msg_len(m))
		err(1, "write user");

	uni_msg_destroy(m);
}

/*
 * SSCOP AA-SIGNALS
 */
static void
sscop_send_upper(struct sscop *sscop, void *arg __unused, enum sscop_aasig sig,
    struct uni_msg *m, u_int p __unused)
{
	VERBOSE(("--> got aa %d(%s) + %zu bytes", sig, sscop_signame(sig),
		m ? uni_msg_len(m) : (size_t)0));

	switch (sig) {

	  case SSCOP_RELEASE_indication:
		PROMPT("-> release_indication(");
		uoutput(sscop, m);
		PROMPT(")\n");
		if (end_at_eof) {
			VERBOSE((" ... exiting"));
			poll_unregister(sscop_h);
			sscop_h = -1;

			(void)close(sscop_fd);
			sscop_fd = -1;
		}
		break;

	  case SSCOP_RELEASE_confirm:
		PROMPT("-> release_confirm(");
		uoutput(sscop, m);
		PROMPT(")\n");
		if (end_at_eof) {
			VERBOSE((" ... exiting"));
			poll_unregister(sscop_h);
			sscop_h = -1;

			(void)close(sscop_fd);
			sscop_fd = -1;
		}
		break;

	  case SSCOP_ESTABLISH_indication:
		PROMPT("-> establish_indication(");
		uoutput(sscop, m);
		PROMPT(")\n");
		break;

	  case SSCOP_ESTABLISH_confirm:
		PROMPT("-> establish_confirm(");
		uoutput(sscop, m);
		PROMPT(")\n");
		ready++;
		break;

	  case SSCOP_DATA_indication:
		PROMPT("-> d");
		assert(m != NULL);
		uoutput(sscop, m);
		if (wstop)
			window++;
		else
			sscop_window(sscop, +1);
		PROMPT("\n");
		break;

	  case SSCOP_UDATA_indication:
		PROMPT("-> u");
		assert(m != NULL);
		uoutput(sscop, m);
		PROMPT("\n");
		break;

	  case SSCOP_RECOVER_indication:
		PROMPT("-> recover_indication\n");
		break;

	  case SSCOP_RESYNC_indication:
		PROMPT("-> resync_indication(");
		uoutput(sscop, m);
		PROMPT(")\n");
		break;

	  case SSCOP_RESYNC_confirm:
		PROMPT("-> resync_confirm\n");
		break;

	  case SSCOP_RETRIEVE_indication:
	  case SSCOP_RETRIEVE_COMPL_indication:
		abort();

	  case SSCOP_ESTABLISH_request:
	  case SSCOP_RELEASE_request:
	  case SSCOP_ESTABLISH_response:
	  case SSCOP_DATA_request:
	  case SSCOP_UDATA_request:
	  case SSCOP_RECOVER_response:
	  case SSCOP_RESYNC_request:
	  case SSCOP_RESYNC_response:
	  case SSCOP_RETRIEVE_request:
		abort();
	}
}

/*
 * This get's called for MAAL
 */
static void
sscop_send_manage(struct sscop *sscop, void *a __unused, enum sscop_maasig sig,
    struct uni_msg *m, u_int errcode, u_int cnt)
{
	VERBOSE(("--> got maa %d(%s) + %zu bytes",
		sig, sscop_msigname(sig),
		(sig == SSCOP_MDATA_indication) ? uni_msg_len(m) : (size_t)0));

	switch(sig) {

	  case SSCOP_MDATA_indication:
		PROMPT("-> m");
		assert(m != NULL);
		uoutput(sscop, m);
		PROMPT("\n");
		break;

	  case SSCOP_MERROR_indication:
		VERBOSE(("MAAL-ERROR.indication '%c' %d", errcode, cnt));
		break;

	  case SSCOP_MDATA_request:
		abort();
	}
}

static void
uprintf(const char *fmt, ...)
{
	va_list ap;
	static char buf[1024];

	va_start(ap, fmt);
	vsnprintf(buf, sizeof(buf), fmt, ap);

	(void)write(user_fd, buf, strlen(buf));
}
