/*
 * Copyright (c) 2001-2003
 *	Fraunhofer Institute for Open Communication Systems (FhG Fokus).
 * 	All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Author: Hartmut Brandt <harti@freebsd.org>
 *
 * $Begemot: atmsupport/test_uni/event.c,v 1.5 2004/12/20 10:45:04 brandt Exp $
 *
 * User commands.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stddef.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <err.h>

#define	index foo_index
#define	err foo_err
#include <tcl.h>
#undef index
#undef err

#include "unitest.h"


static void
dump_uni_msg(const char *dir, u_char *buf, size_t len)
{
	char str[20000];
	int error;
	struct uni_all all;

	struct uni_msg msg;

	msg.b_buf = buf;
	msg.b_lim = buf + len;
	msg.b_rptr = buf;
	msg.b_wptr = buf + len;

	(void)memset(&all, 0, sizeof(all));
	if ((error = uni_decode(&msg, &all, unicx)) != 0)
		printf("uni_decode = %d\n", error);
	if (unicx->errcnt != 0) {
		uni_print_cx(str, sizeof(str), unicx);
		printf("%s\n", str);
	}
	uni_print(str, sizeof(str), &all, unicx);
	printf("%s: %s\n", dir, str);
}

void
input_saal_in(u_char *buf, size_t len)
{
	struct sscfu_arg *a;

	if (len < sizeof(*a)) {
		fprintf(stderr, "bogus %zu bytes from saal_in\n", len);
		return;
	}
	a = (struct sscfu_arg *)(void *)buf;

	if (a->sig == SAAL_DATA_indication)
		dump_uni_msg("RECV", a->data,
		    len - offsetof(struct sscfu_arg, data));
}

void
input_saal_out(u_char *buf, size_t len)
{
	struct sscfu_arg *a;

	if (len < sizeof(*a)) {
		fprintf(stderr, "bogus %zu bytes from saal_out\n", len);
		return;
	}
	a = (struct sscfu_arg *)(void *)buf;

	if (a->sig == SAAL_DATA_request)
		dump_uni_msg("SEND", a->data,
		    len - offsetof(struct sscfu_arg, data));
}

/*
 * Signal from UNI upper interface. Decode it and invoke the 
 * event function.
 */
void
input_uni(const u_char *buf, size_t len)
{
	const struct uni_arg *a = (const struct uni_arg *)(const void *)buf;
	const void *extra = a + 1;
	Tcl_DString ev;
	const char *str;
	char sbuf[100];

	Tcl_DStringInit(&ev);

	if (len < sizeof(*a)) {
		printf("uni message too short\n");
		Tcl_DStringFree(&ev);
		return;
	}

	Tcl_DStringAppendElement(&ev, "uniev");

	len -= sizeof(*a);

	switch (a->sig) {

	  case UNIAPI_ERROR:
	    {
		const struct uniapi_error *error;

		Tcl_DStringAppendElement(&ev, "error");
		if (len != sizeof(*error)) {
			printf("wrong size %zu in error\n", len);
			goto err;
		}
		error = extra;
		sprintf(sbuf, "%u", error->reason);
		Tcl_DStringAppendElement(&ev, sbuf);
		sprintf(sbuf, "%u", a->cookie);
		Tcl_DStringAppendElement(&ev, sbuf);
		break;
	    }

	  case UNIAPI_LINK_ESTABLISH_confirm:
		Tcl_DStringAppendElement(&ev, "link-establish.confirm");
		if (len != 0)
			printf("%zu excess bytes in link-establish.confirm\n",
			    len);
		break;

	  case UNIAPI_LINK_RELEASE_confirm:
		Tcl_DStringAppendElement(&ev, "link-release.confirm");
		if (len != 0)
			printf("%zu excess bytes in link-release.confirm\n",
			    len);
		break;

	  case UNIAPI_RESET_confirm:
	    {
		const struct uniapi_reset_confirm *arg;

		Tcl_DStringAppendElement(&ev, "reset.confirm");
		if (len != sizeof(*arg)) {
			printf("wrong size %zu in reset.confirm\n", len);
			goto err;
		}
		arg = extra;
		fmt_restart(tcl_interp, &ev, &arg->restart);
		fmt_connid(tcl_interp, &ev, &arg->connid);
		break;
	    }

	  case UNIAPI_RESET_indication:
	    {
		const struct uniapi_reset_indication *arg;

		Tcl_DStringAppendElement(&ev, "reset.indication");
		if (len != sizeof(*arg)) {
			printf("wrong size %zu in reset.indication\n", len);
			goto err;
		}
		arg = extra;
		fmt_restart(tcl_interp, &ev, &arg->restart);
		fmt_connid(tcl_interp, &ev, &arg->connid);
		break;
	    }

	  case UNIAPI_RESET_ERROR_indication:
	    {
		const struct uniapi_reset_error_indication *error;

		Tcl_DStringAppendElement(&ev, "reset-error.indication");
		if (len != sizeof(*error)) {
			printf("wrong size %zu in reset-error.indication\n",
			    len);
			goto err;
		}
		error = extra;
		sprintf(sbuf, "%s", error->source ? "response" : "start");
		Tcl_DStringAppendElement(&ev, sbuf);
		sprintf(sbuf, "%u", error->reason);
		Tcl_DStringAppendElement(&ev, sbuf);
		break;
	    }

	  case UNIAPI_RESET_STATUS_indication:
	    {
		const struct uniapi_reset_status_indication *arg;

		Tcl_DStringAppendElement(&ev, "reset-status.indication");
		if (len != sizeof(*arg)) {
			printf("wrong size %zu in reset-status.indication\n",
			    len);
			goto err;
		}
		arg = extra;
		fmt_cref(&ev, &arg->cref);
		fmt_callstate(tcl_interp, &ev, &arg->callstate);
		fmt_cause(tcl_interp, &ev, &arg->cause);
		break;
	    }

	  case UNIAPI_CALL_CREATED:
	    {
		const struct uniapi_call_created *arg;

		Tcl_DStringAppendElement(&ev, "call.created");
		if (len != sizeof(*arg)) {
			printf("wrong size %zu in call.created\n",
			    len);
			goto err;
		}
		arg = extra;
		fmt_cref(&ev, &arg->cref);
		break;
	    }

	  case UNIAPI_CALL_DESTROYED:
	    {
		const struct uniapi_call_destroyed *arg;

		Tcl_DStringAppendElement(&ev, "call.destroyed");
		if (len != sizeof(*arg)) {
			printf("wrong size %zu in call.destroyed\n",
			    len);
			goto err;
		}
		arg = extra;
		fmt_cref(&ev, &arg->cref);
		break;
	    }

	  case UNIAPI_PARTY_CREATED:
	    {
		const struct uniapi_party_created *arg;

		Tcl_DStringAppendElement(&ev, "party.created");
		if (len != sizeof(*arg)) {
			printf("wrong size %zu in party.created\n",
			    len);
			goto err;
		}
		arg = extra;
		fmt_cref(&ev, &arg->cref);
		sprintf(sbuf, "%u", arg->epref.epref);
		Tcl_DStringAppendElement(&ev, sbuf);
		sprintf(sbuf, "%u", arg->epref.flag);
		Tcl_DStringAppendElement(&ev, sbuf);
		break;
	    }

	  case UNIAPI_PARTY_DESTROYED:
	    {
		const struct uniapi_party_destroyed *arg;

		Tcl_DStringAppendElement(&ev, "party.destroyed");
		if (len != sizeof(*arg)) {
			printf("wrong size %zu in party.destroyed\n",
			    len);
			goto err;
		}
		arg = extra;
		fmt_cref(&ev, &arg->cref);
		sprintf(sbuf, "%u", arg->epref.epref);
		Tcl_DStringAppendElement(&ev, sbuf);
		sprintf(sbuf, "%u", arg->epref.flag);
		Tcl_DStringAppendElement(&ev, sbuf);
		break;
	    }

	  case UNIAPI_SETUP_indication:
	    {
		const struct uniapi_setup_indication *arg;

		Tcl_DStringAppendElement(&ev, "setup.indication");
		if (len != sizeof(*arg)) {
			printf("wrong size %zu in setup.indication\n",
			    len);
			goto err;
		}
		arg = extra;
		fmt_cref(&ev, &arg->setup.hdr.cref);
		if (fmt_msgact(tcl_interp, &ev, arg->setup.hdr.act) != TCL_OK ||
		    fmt_msg_setup(tcl_interp, &ev, &arg->setup) != TCL_OK) {
			if (*(str = Tcl_GetStringResult(tcl_interp)) != 0)
				printf("formatting setup: %s\n", str);
			goto err;
		}

		break;
	    }

	  case UNIAPI_PROCEEDING_indication:
	    {
		const struct uniapi_proceeding_indication *arg;

		Tcl_DStringAppendElement(&ev, "proceeding.indication");
		if (len != sizeof(*arg)) {
			printf("wrong size %zu in proceeding.indication\n",
			    len);
			goto err;
		}
		arg = extra;
		fmt_cref(&ev, &arg->call_proc.hdr.cref);
		if (fmt_msgact(tcl_interp, &ev, arg->call_proc.hdr.act) != TCL_OK ||
		    fmt_msg_call_proc(tcl_interp, &ev, &arg->call_proc) != TCL_OK) {
			if (*(str = Tcl_GetStringResult(tcl_interp)) != 0)
				printf("formatting call_proc: %s\n", str);
			goto err;
		}

		break;
	    }

	  case UNIAPI_ALERTING_indication:
	    {
		const struct uniapi_alerting_indication *arg;

		Tcl_DStringAppendElement(&ev, "alerting.indication");
		if (len != sizeof(*arg)) {
			printf("wrong size %zu in alerting.indication\n",
			    len);
			goto err;
		}
		arg = extra;
		fmt_cref(&ev, &arg->alerting.hdr.cref);
		if (fmt_msgact(tcl_interp, &ev, arg->alerting.hdr.act) != TCL_OK ||
		    fmt_msg_alerting(tcl_interp, &ev, &arg->alerting) != TCL_OK) {
			if (*(str = Tcl_GetStringResult(tcl_interp)) != 0)
				printf("formatting alerting: %s\n", str);
			goto err;
		}

		break;
	    }

	  case UNIAPI_SETUP_confirm:
	    {
		const struct uniapi_setup_confirm *arg;

		Tcl_DStringAppendElement(&ev, "setup.confirm");
		if (len != sizeof(*arg)) {
			printf("wrong size %zu in setup.confirm\n",
			    len);
			goto err;
		}
		arg = extra;
		fmt_cref(&ev, &arg->connect.hdr.cref);
		if (fmt_msgact(tcl_interp, &ev, arg->connect.hdr.act) != TCL_OK ||
		    fmt_msg_connect(tcl_interp, &ev, &arg->connect) != TCL_OK) {
			if (*(str = Tcl_GetStringResult(tcl_interp)) != 0)
				printf("formatting connect: %s\n", str);
			goto err;
		}

		break;
	    }

	  case UNIAPI_SETUP_COMPLETE_indication:
	    {
		const struct uniapi_setup_complete_indication *arg;

		Tcl_DStringAppendElement(&ev, "setup.complete_indication");
		if (len != sizeof(*arg)) {
			printf("wrong size %zu in setup.complete_indication\n",
			    len);
			goto err;
		}
		arg = extra;
		fmt_cref(&ev, &arg->connect_ack.hdr.cref);
		if (fmt_msgact(tcl_interp, &ev, arg->connect_ack.hdr.act) != TCL_OK ||
		    fmt_msg_connect_ack(tcl_interp, &ev, &arg->connect_ack) != TCL_OK) {
			if (*(str = Tcl_GetStringResult(tcl_interp)) != 0)
				printf("formatting connect_ack: %s\n", str);
			goto err;
		}

		break;
	    }

	  case UNIAPI_RELEASE_indication:
	    {
		const struct uniapi_release_indication *arg;

		Tcl_DStringAppendElement(&ev, "release.indication");
		if (len != sizeof(*arg)) {
			printf("wrong size %zu in release.indication\n",
			    len);
			goto err;
		}
		arg = extra;
		fmt_cref(&ev, &arg->release.hdr.cref);
		if (fmt_msgact(tcl_interp, &ev, arg->release.hdr.act) != TCL_OK ||
		    fmt_msg_release(tcl_interp, &ev, &arg->release) != TCL_OK) {
			if (*(str = Tcl_GetStringResult(tcl_interp)) != 0)
				printf("formatting release: %s\n", str);
			goto err;
		}
		break;
	    }

	  case UNIAPI_RELEASE_confirm:
	    {
		const struct uniapi_release_confirm *arg;

		Tcl_DStringAppendElement(&ev, "release.confirm");
		if (len != sizeof(*arg)) {
			printf("wrong size %zu in release.confirm\n",
			    len);
			goto err;
		}
		arg = extra;
		fmt_cref(&ev, &arg->release.hdr.cref);
		if (fmt_msgact(tcl_interp, &ev, arg->release.hdr.act) != TCL_OK ||
		    fmt_msg_release(tcl_interp, &ev, &arg->release) != TCL_OK) {
			if (*(str = Tcl_GetStringResult(tcl_interp)) != 0)
				printf("formatting release: %s\n", str);
			goto err;
		}

		break;
	    }

	  case UNIAPI_NOTIFY_indication:
	    {
		const struct uniapi_notify_indication *arg;

		Tcl_DStringAppendElement(&ev, "notify.indication");
		if (len != sizeof(*arg)) {
			printf("wrong size %zu in notify.indication\n",
			    len);
			goto err;
		}
		arg = extra;
		fmt_cref(&ev, &arg->notify.hdr.cref);
		if (fmt_msgact(tcl_interp, &ev, arg->notify.hdr.act) != TCL_OK ||
		    fmt_msg_notify(tcl_interp, &ev, &arg->notify) != TCL_OK) {
			if (*(str = Tcl_GetStringResult(tcl_interp)) != 0)
				printf("formatting notify: %s\n", str);
			goto err;
		}
		break;
	    }

	  case UNIAPI_STATUS_indication:
	    {
		const struct uniapi_status_indication *arg;

		Tcl_DStringAppendElement(&ev, "status.indication");
		if (len != sizeof(*arg)) {
			printf("wrong size %zu in status.indication\n",
			    len);
			goto err;
		}
		arg = extra;
		fmt_cref(&ev, &arg->cref);
		sprintf(sbuf, "%u", arg->my_state);
		Tcl_DStringAppendElement(&ev, sbuf);
		sprintf(sbuf, "%u", arg->my_cause);
		Tcl_DStringAppendElement(&ev, sbuf);
		if (fmt_callstate(tcl_interp, &ev, &arg->his_state) != TCL_OK) {
			if (*(str = Tcl_GetStringResult(tcl_interp)) != 0)
				printf("formatting callstate: %s\n", str);
			goto err;
		}
		if (fmt_cause(tcl_interp, &ev, &arg->his_cause) != TCL_OK) {
			if (*(str = Tcl_GetStringResult(tcl_interp)) != 0)
				printf("formatting cause: %s\n", str);
			goto err;
		}
		break;
	    }

	  case UNIAPI_DROP_PARTY_indication:
	    {
		const struct uniapi_drop_party_indication *arg;

		Tcl_DStringAppendElement(&ev, "drop-party.indication");
		if (len != sizeof(*arg)) {
			printf("wrong size %zu in drop-party.indication\n",
			    len);
			goto err;
		}
		arg = extra;
		fmt_cref(&ev, &arg->drop.hdr.cref);
		if (fmt_msgact(tcl_interp, &ev, arg->drop.hdr.act) != TCL_OK ||
		    fmt_msg_drop_party(tcl_interp, &ev, &arg->drop) != TCL_OK) {
			if (*(str = Tcl_GetStringResult(tcl_interp)) != 0)
				printf("formatting drop_party: %s\n", str);
			goto err;
		}
		break;
	    }

	  case UNIAPI_DROP_PARTY_ACK_indication:
	    {
		const struct uniapi_drop_party_ack_indication *arg;

		Tcl_DStringAppendElement(&ev, "drop-party-ack.indication");
		if (len != sizeof(*arg)) {
			printf("wrong size %zu in drop-party-ack.indication\n",
			    len);
			goto err;
		}
		arg = extra;
		fmt_cref(&ev, &arg->drop.hdr.cref);
		if (fmt_msgact(tcl_interp, &ev, arg->drop.hdr.act) != TCL_OK ||
		    fmt_msg_drop_party(tcl_interp, &ev, &arg->drop) != TCL_OK) {
			if (*(str = Tcl_GetStringResult(tcl_interp)) != 0)
				printf("formatting drop_party: %s\n", str);
			goto err;
		}
		if (fmt_crankback(tcl_interp, &ev, &arg->crankback) != TCL_OK) {
			if (*(str = Tcl_GetStringResult(tcl_interp)) != 0)
				printf("formatting crankback: %s\n", str);
			goto err;
		}
		break;
	    }

	  case UNIAPI_ADD_PARTY_ACK_indication:
	    {
		const struct uniapi_add_party_ack_indication *arg;

		Tcl_DStringAppendElement(&ev, "add-party-ack.indication");
		if (len != sizeof(*arg)) {
			printf("wrong size %zu in add_party_ack.indication\n",
			    len);
			goto err;
		}
		arg = extra;
		fmt_cref(&ev, &arg->ack.hdr.cref);
		if (fmt_msgact(tcl_interp, &ev, arg->ack.hdr.act) != TCL_OK ||
		    fmt_msg_add_party_ack(tcl_interp, &ev, &arg->ack) != TCL_OK) {
			if (*(str = Tcl_GetStringResult(tcl_interp)) != 0)
				printf("formatting add_party_ack: %s\n", str);
			goto err;
		}
		break;
	    }

	  case UNIAPI_ADD_PARTY_REJ_indication:
	    {
		const struct uniapi_add_party_rej_indication *arg;

		Tcl_DStringAppendElement(&ev, "add-party-rej.indication");
		if (len != sizeof(*arg)) {
			printf("wrong size %zu in add_party_rej.indication\n",
			    len);
			goto err;
		}
		arg = extra;
		fmt_cref(&ev, &arg->rej.hdr.cref);
		if (fmt_msgact(tcl_interp, &ev, arg->rej.hdr.act) != TCL_OK ||
		    fmt_msg_add_party_rej(tcl_interp, &ev, &arg->rej) != TCL_OK) {
			if (*(str = Tcl_GetStringResult(tcl_interp)) != 0)
				printf("formatting add_party_rej: %s\n", str);
			goto err;
		}
		break;
	    }

	  case UNIAPI_PARTY_ALERTING_indication:
	    {
		const struct uniapi_party_alerting_indication *arg;

		Tcl_DStringAppendElement(&ev, "party-alerting.indication");
		if (len != sizeof(*arg)) {
			printf("wrong size %zu in party_alerting.indication\n",
			    len);
			goto err;
		}
		arg = extra;
		fmt_cref(&ev, &arg->alert.hdr.cref);
		if (fmt_msgact(tcl_interp, &ev, arg->alert.hdr.act) != TCL_OK ||
		    fmt_msg_party_alerting(tcl_interp, &ev, &arg->alert)
		    != TCL_OK) {
			if (*(str = Tcl_GetStringResult(tcl_interp)) != 0)
				printf("formatting party_alerting: %s\n", str);
			goto err;
		}
		break;
	    }

	  case UNIAPI_ADD_PARTY_indication:
	    {
		const struct uniapi_add_party_indication *arg;

		Tcl_DStringAppendElement(&ev, "add-party.indication");
		if (len != sizeof(*arg)) {
			printf("wrong size %zu in add_party.indication\n", len);
			goto err;
		}
		arg = extra;
		fmt_cref(&ev, &arg->add.hdr.cref);
		if (fmt_msgact(tcl_interp, &ev, arg->add.hdr.act) != TCL_OK ||
		    fmt_msg_add_party(tcl_interp, &ev, &arg->add) != TCL_OK) {
			if (*(str = Tcl_GetStringResult(tcl_interp)) != 0)
				printf("formatting add_party: %s\n", str);
			goto err;
		}
		break;
	    }

	  default:
		Tcl_DStringFree(&ev);
		printf("bad signal from uni: %u data=%zu", a->sig, len);
		return;
	}

	try_event(tcl_interp, &ev);
  err:
	Tcl_DStringFree(&ev);
}

/*
 * Event from SSCOP management. Print only.
 */
void
input_man(u_char *buf, size_t len)
{
	struct sscop_marg *a = (struct sscop_marg *)(void *)buf;
	struct sscop_merr *e = (struct sscop_merr *)(void *)buf;

	if (len < sizeof(a->sig)) {
		printf("management message too short\n");
		return;
	}

	switch (a->sig) {

	  case SSCOP_MDATA_indication:
		len -= sizeof(*a);
		printf("SSCOP_MDATA_indication: data=%zu\n", len);
		if (len > 0) {
			dump_hex(a->data, len);
		}
		break;

	  case SSCOP_MDATA_request:
		len -= sizeof(*a);
		printf("SSCOP_MDATA_request: data=%zu\n", len);
		if (len > 0) {
			dump_hex(a->data, len);
		}
		break;

	  case SSCOP_MERROR_indication:
		if (len != sizeof(*e)) {
			fprintf(stderr, "bad size of error: %zu\n", len);
			break;
		}
		printf("SSCOP_MERROR_indication: err='%c' cnt=%u\n",
		    e->err, e->cnt);
		break;

	  default:
		printf("bad signal: %u", a->sig);
		break;
	}
}

/*
 * Input from the debug hook. This should be ASCII.
 */
void
input_dbg(u_char *buf, size_t len)
{
	printf("DEBUG: %.*s\n", (int)len, buf);
}
