/*
 * Copyright (c) 2001-2003
 *	Fraunhofer Institute for Open Communication Systems (FhG Fokus).
 * 	All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Author: Hartmut Brandt <harti@freebsd.org>
 *
 * $Begemot: atmsupport/test_uni/main.c,v 1.3 2004/12/20 10:45:04 brandt Exp $
 *
 * Build the UNI stack.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <err.h>

#define	index	foo_index
#define	err	foo_err
#include <tcl.h>
#undef index
#undef err

#include <sys/syslimits.h>
#include <rpoll.h>
#include <netgraph.h>

#include "testlib.h"
#include "unitest.h"

static Tcl_DString	command;
static int		errinfo;
static const char	*script = "unitest.tcl";

static const char	*tclpath = ".";
static int		tflag;
static int		csock, dsock;
static int		doprompt;
static u_int32_t 	cookie;
static const char	*device = "fatm0:";

Tcl_Interp	*tcl_interp;


/*                [atmtest]
 *		      |
 *                    |
 *                    |(uni)
 *                    +----------------+
 *                    |(dbg) (debug)   |(upper)
 *                    +--------------[uni]
 *                    |                |(lower)
 *          (saal_out)|  (right2left)  |(right)
 *                    +--------------[tee]
 *           (saal_in)|  (left2right)  |(left)
 *                    |                |(upper)
 *                    |              [sscf]
 *                    |                |(lower)
 *                    |(man) (manage)  |(upper)
 *                    +--------------[sscop]
 *                                     |(lower)
 *                                     |(sig5)
 *                                  [fatm]
 */
static struct stack stack[] = {
	{ STACK_MKPEER, NULL, "sscop", "sig5", "lower", "sscop" },
	{ STACK_MKPEER, "sscop:", "sscfu", "upper", "lower", "sscf" },
	{ STACK_MKPEER, "sscf:", "tee", "upper", "left", "saaltee" },
	{ STACK_MKPEER, "saaltee:", "uni", "right", "lower", "uni" },
	{ STACK_CONNECT, ".", "saaltee:", "saal_in", "left2right", NULL },
	{ STACK_CONNECT, ".", "saaltee:", "saal_out", "right2left", NULL },
	{ STACK_CONNECT, ".", "sscop:", "man", "manage", NULL },
	{ STACK_CONNECT, ".", "uni:", "uni", "upper", NULL },
	{ STACK_CONNECT, ".", "uni:", "dbg", "debug", NULL },
	{ STACK_END, NULL, NULL, NULL, NULL, NULL }
};

static const char usgtxt[] = "\
UNI test program\n\
usage: unitest [-h] [-s file] [-p path] [-e] [-t] [args ...]\n\
where:\n\
  -d dev:	use device dev for ATM\n\
  -e		print errInfo and errCode variables on errors\n\
  -h		print this help text\n\
  -p path	use 'path' to search for the start-up script instead of '.'\n\
  -s file	read 'file' as start-up script instead of unitest.tcl\n\
  -t		trace messages between UNI and SAAL\n\
Any arguments are passed to the TCL code in the global array 'argv'.\n\
The script and path names are available in 'script' and 'tclpath'.\n\
";

/*
 * Initialize SSCOP and the VCI
 */
static void
init_stack(void)
{
	if (NgSendAsciiMsg(csock, "sscop:",
	    "setparam {mask=0x800 param={flags=0x01}}") < 0)
		err(1, "failed to set robust mode");
	read_cmsg(csock);

	if (NgSendAsciiMsg(csock, "uni:",
	    "enable") < 0)
		err(1, "failed to enable UNI");

	if (NgSendAsciiMsg(csock, "sscf:",
	    "enable") < 0)
		err(1, "failed to enable SSCF");

	if (NgSendAsciiMsg(csock, "sscop:",
	    "enable") < 0)
		err(1, "failed to enable SSCOP");

	if (NgSendAsciiMsg(csock, device,
	    "cpcsinit {name=\"sig5\" flags=0x4 traffic=0 aal=5 vci=5}") < 0)
		err(1, "failed to init VCI");
	read_cmsg(csock);
}

/*
 * Called on shutdown via atexit(3) to remove the stack
 */
static void
shut(void)
{
	shutdown_stack(csock, stack);
}

/*
 * SIGINT handler
 */
static void
onintr(int sig)
{
	sig = sig;
	exit(0);
}

/*
 * User descriptor is ready
 */
static void
ready_user(int fd __unused, int mask __unused, void *uap __unused)
{
	char line[LINE_MAX], *ptr, *end;
	ssize_t n;

	if ((n = read(0, line, sizeof(line))) == -1)
		err(1, "read from user");
	doprompt = 1;
	if (n == 0) {
		Tcl_DStringFree(&command);
		Tcl_DStringAppend(&command, "exit", 4);
		try_exec(tcl_interp);
		exit(2);
	}
	ptr = line;
	while (n >= 0) {
		for (end = ptr; end < ptr + n && *end != '\n'; end++)
			;
		if (end == ptr + n) {
			Tcl_DStringAppend(&command, ptr, n);
			break;
		}
		Tcl_DStringAppend(&command, ptr, end - ptr);
		try_exec(tcl_interp);
		n -= end - ptr + 1;
		ptr = end + 1;
	}
}

/*
 * Something happend on the control socket. Try to read a message and to
 * dump it.
 */
static void
ready_cmsg(int fd __unused, int mask __unused, void *uap __unused)
{

	read_cmsg(csock);
	doprompt = 1;
}

/*
 * Read a message from the data socket and call a handler depending
 * on the hook.
 */
static void
ready_dmsg(int fd __unused, int mask __unused, void *uap __unused)
{
	u_char buf[MAXMSG];
	char hook[NG_HOOKLEN + 1];
	ssize_t n;

	if ((n = NgRecvData(dsock, buf, sizeof(buf), hook)) < 0)
		err(1, "NgRecvData");

	if (n == 0) {
		printf("EOF from '%s'\n", hook);
		return;
	}

	if (strcmp(hook, "uni") == 0) {
		input_uni(buf, (size_t)n);
		doprompt = 1;

	} else if (strcmp(hook, "man") == 0) {
		input_man(buf, (size_t)n);
		doprompt = 1;

	} else if (strcmp(hook, "dbg") == 0) {
		input_dbg(buf, (size_t)n);
		doprompt = 1;

	} else if (strcmp(hook, "saal_in") == 0) {
		if (tflag) {
			input_saal_in(buf, (size_t)n);
			doprompt = 1;
		}

	} else if (strcmp(hook, "saal_out") == 0) {
		if (tflag) {
			input_saal_out(buf, (size_t)n);
			doprompt = 1;
		}

	} else {
		printf("%zu bytes from '%s'\n", (size_t)n, hook);
		doprompt = 1;
	}
}


int
main(int argc, char *argv[])
{
	int opt;
	char *str;
	const char *res;
	Tcl_DString args;

	while ((opt = getopt(argc, argv, "d:hs:p:et")) != -1)
		switch (opt) {

		  case 'd':
			device = optarg;
			break;

		  case 'h':
			fprintf(stderr, usgtxt);
			exit(0);

		  case 's':
			script = optarg;
			break;

		  case 'p':
			tclpath = optarg;
			break;

		  case 'e':
			errinfo++;
			break;

		  case 't':
			tflag = 1;
			break;
		}

	argc -= optind;
	argv += optind;

	stack[0].path = device;

	atexit(shut);
	signal(SIGINT, onintr);

	build_stack(&csock, &dsock, stack);
	init_stack();

	init_uni();

	tcl_interp = Tcl_CreateInterp();
	Tcl_DStringInit(&command);

	if ((str = malloc(strlen(script)+1)) == NULL)
		err(1, NULL);
	script = strcpy(str, script);
	if (Tcl_LinkVar(tcl_interp, "script", (char *)&script, 
	    TCL_LINK_STRING|TCL_LINK_READ_ONLY) != TCL_OK)
		errx(1, "Tcl_LinkVar(script): %s",
		    Tcl_GetStringResult(tcl_interp));

	if ((str = malloc(strlen(tclpath)+1)) == NULL)
		err(1, NULL);
	tclpath = strcpy(str, tclpath);
	if (Tcl_LinkVar(tcl_interp, "tclpath", (char *)&tclpath, 
	    TCL_LINK_STRING|TCL_LINK_READ_ONLY) != TCL_OK)
		errx(1, "Tcl_LinkVar(tclpath): %s",
		    Tcl_GetStringResult(tcl_interp));

	Tcl_DStringInit(&args);
	while (argc--)
		Tcl_DStringAppendElement(&args, *argv++);
	Tcl_SetVar(tcl_interp, "argv", Tcl_DStringValue(&args), 0);
	Tcl_DStringFree(&args);

	Tcl_CreateCommand(tcl_interp, "dbg", cmd_dbg, NULL, NULL);
	Tcl_CreateCommand(tcl_interp, "status", cmd_status, NULL, NULL);
	Tcl_CreateCommand(tcl_interp, "dosig", cmd_dosig, NULL, NULL);
	Tcl_CreateCommand(tcl_interp, "sleep", cmd_sleep, NULL, NULL);
	Tcl_CreateCommand(tcl_interp, "getconfig", cmd_getconfig, NULL, NULL);

	/*
	 * Run startup script
	 */
	str = malloc(strlen(tclpath) + strlen(script) + 2);
	if (str == NULL)
		err(1, NULL);
	sprintf(str, "%s/%s", tclpath, script);
	opt = Tcl_EvalFile(tcl_interp, str);
	free(str);

	if (*(res = Tcl_GetStringResult(tcl_interp)) != 0)
		printf("%s\n", res);
	if (opt != TCL_OK) {
		printf("initialisation returned %d\n", opt);
		if (errinfo) {
			res = Tcl_GetVar(tcl_interp, "errorInfo",
			    TCL_GLOBAL_ONLY);
			if (res != NULL)
				printf("errorInfo='%s'\n", res);
			res = Tcl_GetVar(tcl_interp, "errorCode",
			    TCL_GLOBAL_ONLY);
			if (res != NULL)
				printf("errorCode='%s'\n", res);
		}
		exit(1);
	}

	(void)poll_register(fileno(stdin), ready_user, NULL, POLL_IN);
	(void)poll_register(csock, ready_cmsg, NULL, POLL_IN);
	(void)poll_register(dsock, ready_dmsg, NULL, POLL_IN);

	ev_loop(&doprompt, "uni");
}


/*
 * Execute the current command if it is complete
 */
void
try_exec(Tcl_Interp *interp)
{
	int n;
	const char *info;

	if (!Tcl_CommandComplete(Tcl_DStringValue(&command)))
		return;

	n = Tcl_RecordAndEval(interp, Tcl_DStringValue(&command), 0);
	Tcl_DStringFree(&command);

	if (n != TCL_OK) {
		printf("error: %s\n", Tcl_GetStringResult(interp));
		if (errinfo) {
			info = Tcl_GetVar(interp, "errorInfo", TCL_GLOBAL_ONLY);
			if (info != NULL)
				printf("%s\n", info);
		}
	} else if (*(info = Tcl_GetStringResult(interp)) != '\0')
		printf("%s\n", info);
}

/*
 * Evaluate an event
 */
void
try_event(Tcl_Interp *interp, Tcl_DString *str)
{
	int res;
	const char *info;

	if ((res = Tcl_GlobalEval(interp, Tcl_DStringValue(str))) != TCL_OK) {
		printf("error evaluating event '%s': %s\n",
		    Tcl_DStringValue(str), Tcl_GetStringResult(interp));
		if (errinfo) {
			info = Tcl_GetVar(interp, "errorInfo", TCL_GLOBAL_ONLY);
			if (info != NULL)
				printf("%s\n", info);
		}
	}
}

void
set_error(Tcl_Interp *interp, const char *fmt, ...)
{
	va_list ap;
	char buf[1024];

	va_start(ap, fmt);
	vsnprintf(buf, sizeof(buf), fmt, ap);
	va_end(ap);
	Tcl_SetResult(interp, buf, TCL_VOLATILE);
}

/*
 * Send a control message to the UNI node.
 */
void
send_control(u_int msg, void *buf, size_t len)
{
	if (NgSendMsg(csock, "uni:", NGM_UNI_COOKIE, (int)msg, buf, len) == -1)
		err(1, "NgSendMsg");
}

/*
 * Request status from nodes.
 */
static void
do_status(const char *path)
{
	u_char buf[sizeof(struct ng_mesg) + NG_TEXTRESPONSE];
	struct ng_mesg *resp = (struct ng_mesg *)(void *)buf;

	if (NgSendMsg(csock, path, NGM_GENERIC_COOKIE, NGM_TEXT_STATUS, NULL, 0) < 0) {
		if (errno != EINVAL)
			err(1, "NGM_TEXT_STATUS");
		warnx("no status from '%s'\n", path);
		return;
	}
	if (NgRecvMsg(csock, resp, sizeof(buf), NULL) < 0 ||
	   (resp->header.flags & NGF_RESP) == 0) {
		warnx("no status from '%s'\n", path);
		return;
	}
	printf("%s", resp->data);
}

void
send_status(const char *node)
{
	char path[NG_PATHLEN];

	if (node != NULL) {
		snprintf(path, sizeof(path), "%s:", node);
		do_status(path);
	} else {
		do_status("sscf:");
		do_status("sscop:");
		do_status("uni:");
	}
}

/*
 * Request configuration information from the UNI node.
 */
void
do_getconfig(struct uni_config *config)
{
	u_char buf[sizeof(struct ng_mesg) + sizeof(*config)];
	struct ng_mesg *resp = (struct ng_mesg *)(void *)buf;

	if (NgSendMsg(csock, "uni:", NGM_UNI_COOKIE,
	    NGM_UNI_GET_CONFIG, NULL, 0) < 0)
		err(1, "uni_get_config");

	if (NgRecvMsg(csock, resp, sizeof(buf), NULL) < 0)
		err(1, "no config from 'uni:'");

	if (resp->header.arglen != sizeof(*config))
		errx(1, "uni_config length %u instead of %zu",
		    resp->header.arglen, sizeof(*config));

	memcpy(config, resp->data, sizeof(*config));
}

/*
 * Request debug levels from node.
 */
void
do_getdbg(struct ngm_uni_debug *dbg)
{
	u_char buf[sizeof(struct ng_mesg) + sizeof(*dbg)];
	struct ng_mesg *resp = (struct ng_mesg *)(void *)buf;

	if (NgSendMsg(csock, "uni:", NGM_UNI_COOKIE,
	    NGM_UNI_GETDEBUG, NULL, 0) < 0)
		err(1, "uni_getdebug");

	if (NgRecvMsg(csock, resp, sizeof(buf), NULL) < 0)
		err(1, "no debug info from 'uni:'");

	if (resp->header.arglen != sizeof(*dbg))
		errx(1, "uni_dbg length %u instead of %zu",
		    resp->header.arglen, sizeof(*dbg));

	memcpy(dbg, resp->data, sizeof(*dbg));
}


/*
 * Send a message to the UNI node.
 */
void
send_uniapi(struct uni_arg *arg, size_t len)
{
	if (++cookie == 0)
		++cookie;
	arg->cookie = cookie;
	if (NgSendData(dsock, "uni", (void *)arg, len) == -1)
		err(1, "NgSendData");
}

void
send_uniapiv(struct uni_arg *a, ...)
{
	va_list ap;
	size_t len, sum;
	void *inp;
	u_char *ptr;
	struct uni_arg *buf;

	sum = sizeof(*a);
	va_start(ap, a);
	while ((inp = va_arg(ap, void *)) != NULL)
		sum += va_arg(ap, size_t);
	va_end(ap);

	if ((buf = malloc(sum)) == NULL)
		err(1, NULL);
	ptr = (u_char *)buf;
	memcpy(ptr, a, sizeof(*a));
	ptr += sizeof(*a);
	va_start(ap, a);
	while ((inp = va_arg(ap, void *)) != NULL) {
		len = va_arg(ap, size_t);
		memcpy(ptr, inp, len);
		ptr += len;
	}
	va_end(ap);

	send_uniapi(buf, sum);

	free(buf);
}
