/*
 * Copyright (c) 2001-2003
 *	Fraunhofer Institute for Open Communication Systems (FhG Fokus).
 * 	All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Author: Hartmut Brandt <harti@freebsd.org>
 *
 * $Begemot: atmsupport/test_uni/user.c,v 1.3 2004/12/20 10:45:05 brandt Exp $
 *
 * User commands.
 */

#include <sys/types.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <err.h>
#include <rpoll.h>

#define	index foo_index
#define	err foo_err
#include <tcl.h>
#undef index
#undef err

#include "unitest.h"

static const struct {
	const char *name;
	u_int	val;
} dbgfac[] = {
	{ "timeout",	UNI_FAC_TIMEOUT },
	{ "restart",	UNI_FAC_RESTART	},
	{ "saal",	UNI_FAC_SAAL	},
	{ "parse",	UNI_FAC_PARSE	},
	{ "call",	UNI_FAC_CALL	},
	{ "warn",	UNI_FAC_WARN	},
	{ "coord",	UNI_FAC_COORD	},
	{ "api",	UNI_FAC_API	},
	{ "msg",	UNI_FAC_MSG	},
	{ "err",	UNI_FAC_ERR	},
	{ NULL, 0 }
};

struct unicx	*unicx;

void
init_uni(void)
{
	if ((unicx = malloc(sizeof(*unicx))) == NULL)
		err(1, NULL);
	uni_initcx(unicx);
	unicx->multiline = 2;
}

/*
 * Set debugging mode of the UNI node. The arguments consist of pairs of
 * a facility specified and a debugging level.
 */
int
cmd_dbg(ClientData data __unused, Tcl_Interp *interp,
    int argc, const char *argv[])
{
	int i, n;
	u_int level;
	struct ngm_uni_debug dbg;
	char *end, buf[100];
	Tcl_DString str;

	argc--;
	argv++;
	if (argc == 0) {
		do_getdbg(&dbg);
		Tcl_DStringInit(&str);
		for (i = 0; i < UNI_MAXFACILITY; i++) {
			if (dbg.level[i] != 0) {
				Tcl_DStringAppendElement(&str, dbgfac[i].name);
				sprintf(buf, "%u", dbg.level[i]);
				Tcl_DStringAppendElement(&str, buf);
			}
		}
		Tcl_DStringResult(interp, &str);
		Tcl_DStringFree(&str);
		return (0);
	}
	if (argc % 2 != 0) {
		set_error(interp, "need even number of arguments");
		return (TCL_ERROR);
	}
	for (i = 0; i < argc/2; i++) {
		level = strtoul(argv[2 * i + 1], &end, 10);
		if (*end != '\0') {
			set_error(interp, "bad integer '%s'", argv[2 * i + 1]);
			return (TCL_ERROR);
		}
		if (strcmp(argv[2*i], "all") == 0) {
			for (n = 0; n < UNI_MAXFACILITY; n++)
				dbg.level[n] = level;
		} else {
			for (n = 0; dbgfac[n].name; n++)
				if (strcmp(dbgfac[n].name, argv[2 * i]) == 0)
					break;
			if (dbgfac[n].name == NULL) {
				set_error(interp, "unknown facility '%s'",
				    argv[2 * i]);
				return (TCL_ERROR);
			}
			dbg.level[i] = level;
		}
	}

	send_control(NGM_UNI_SETDEBUG, &dbg, sizeof(dbg));
	return (TCL_OK);
}

/*
 * Request status. If no argument is given, get status from all our nodes,
 * if an argument is given this should be the node name.
 */
int
cmd_status(ClientData data __unused, Tcl_Interp *interp, int argc,
    const char *argv[] __unused)
{
	if (argc == 1) {
		send_status(NULL);
	} else if (argc == 2) {
		send_status(argv[1]);
	} else {
		set_error(interp, "usage: status [node]");
		return (TCL_ERROR);
	}
	return (TCL_OK);
}

/*
 * Get UNI configuration.
 */
int
cmd_getconfig(ClientData data __unused, Tcl_Interp *interp,
    int argc, const char *argv[])
{
	struct uni_config config;
	int first;

	if (argc != 1) {
		set_error(interp, "usage: %s", argv[0]);
		return (TCL_ERROR);
	}
	do_getconfig(&config);

	switch (config.proto) {

	  case UNIPROTO_UNI40U:
		printf("proto=UNI40U\n");
		break;
	  case UNIPROTO_UNI40N:
		printf("proto=UNI40N\n");
		break;
	  case UNIPROTO_PNNI10:
		printf("proto=PNNI10\n");
		break;
	  default:
		printf("proto=%u\n", config.proto);
		break;
	}

	printf("popt=");
	first = 1;
	if (config.popt & UNIPROTO_GFP) {
		printf("%sGPF", first ? "" : ",");
		config.popt &= ~UNIPROTO_GFP;
		first = 0;
	}
	if (config.popt)
		printf("%s0x%x", first ? "" : ",", config.popt);
	printf("\n");

	printf("option=");
	first = 1;
	if (config.option & UNIOPT_GIT_HARD) {
		printf("%shard_git", first ? "" : ",");
		config.option &= ~UNIOPT_GIT_HARD;
		first = 0;
	}
	if (config.option & UNIOPT_BEARER_HARD) {
		printf("%shard_bearer", first ? "" : ",");
		config.option &= ~UNIOPT_BEARER_HARD;
		first = 0;
	}
	if (config.option & UNIOPT_CAUSE_HARD) {
		printf("%shard_cause", first ? "" : ",");
		config.option &= ~UNIOPT_CAUSE_HARD;
		first = 0;
	}
	if (config.option)
		printf("%s0x%x", first ? "" : ",", config.option);
	printf("\n");

	printf("T301=%u\n", config.timer301);
	printf("T303=%u,%u\n", config.timer303, config.init303);
	printf("T308=%u,%u\n", config.timer308, config.init308);
	printf("T309=%u\n", config.timer309);
	printf("T310=%u\n", config.timer310);
	printf("T313=%u\n", config.timer313);
	printf("T316=%u,%u\n", config.timer316, config.init316);
	printf("T317=%u\n", config.timer317);
	printf("T322=%u,%u\n", config.timer322, config.init322);
	printf("T397=%u\n", config.timer397);
	printf("T398=%u\n", config.timer398);
	printf("T399=%u\n", config.timer399);

	return (TCL_OK);
}

static int
do_link_establish_request(Tcl_Interp *interp, int argc, const char *argv[])
{
	struct uni_arg a;

	if (argc != 2) {
		set_error(interp, "extra arguments to %s", argv[0]);
		return (TCL_ERROR);
	}
	a.sig = UNIAPI_LINK_ESTABLISH_request;
	send_uniapi(&a, sizeof(a));
	return (TCL_OK);
}

static int
do_link_release_request(Tcl_Interp *interp, int argc, const char *argv[])
{
	struct uni_arg a;

	if (argc != 2) {
		set_error(interp, "extra arguments to %s", argv[0]);
		return (TCL_ERROR);
	}
	a.sig = UNIAPI_LINK_RELEASE_request;
	send_uniapi(&a, sizeof(a));
	return (TCL_OK);
}

static int
do_reset_error_response(Tcl_Interp *interp, int argc, const char *argv[])
{
	struct uni_arg a;
	union uni_ieall ie;
	enum uni_ietype ietype;
	struct uniapi_reset_error_response resp;
	char *arg;
	int i;

	memset(&resp, 0, sizeof(resp));

	for (i = 2; i < argc; i++) {
		if ((arg = substitute_vars(interp, argv[i])) == NULL)
			return (TCL_ERROR);
		if (parse_ie(interp, arg, &ie, &ietype)) {
			free(arg);
			return (TCL_ERROR);
		}
		free(arg);

		switch (ietype) {

		  case UNI_IE_CAUSE:
			if (IE_ISPRESENT(resp.cause)) {
				set_error(interp, "double cause in %s",
				    argv[0]);
				return (TCL_ERROR);
			}
			resp.cause = ie.cause;
			break;

		  default:
			set_error(interp, "bad IE 0x%02x for %s", ietype,
			    argv[0]);
			return (TCL_ERROR);
		}
	}
	a.sig = UNIAPI_RESET_ERROR_response;
	send_uniapiv(&a, &resp, sizeof(resp), NULL);

	return (TCL_OK);
}

static int
do_reset_request(Tcl_Interp *interp, int argc, const char *argv[])
{
	struct uni_arg a;
	union uni_ieall ie;
	enum uni_ietype ietype;
	struct uniapi_reset_request req;
	char *arg;
	int i;

	memset(&req, 0, sizeof(req));

	for (i = 2; i < argc; i++) {
		if ((arg = substitute_vars(interp, argv[i])) == NULL)
			return (TCL_ERROR);
		if (parse_ie(interp, arg, &ie, &ietype)) {
			free(arg);
			return (TCL_ERROR);
		}
		free(arg);

		switch (ietype) {

		  case UNI_IE_RESTART:
			if (IE_ISPRESENT(req.restart)) {
				set_error(interp, "double restart in %s",
				    argv[0]);
				return (TCL_ERROR);
			}
			req.restart = ie.restart;
			break;

		  case UNI_IE_CONNID:
			if (IE_ISPRESENT(req.connid)) {
				set_error(interp, "double connid in %s",
				    argv[0]);
				return (TCL_ERROR);
			}
			req.connid = ie.connid;
			break;

		  default:
			set_error(interp, "bad IE 0x%02x for %s", ietype,
			    argv[0]);
			return (TCL_ERROR);
		}
	}
	a.sig = UNIAPI_RESET_request;
	send_uniapiv(&a, &req, sizeof(req), NULL);

	return (TCL_OK);
}
static int
do_reset_response(Tcl_Interp *interp, int argc, const char *argv[])
{
	struct uni_arg a;
	union uni_ieall ie;
	enum uni_ietype ietype;
	struct uniapi_reset_response resp;
	char *arg;
	int i;

	memset(&resp, 0, sizeof(resp));

	for (i = 2; i < argc; i++) {
		if ((arg = substitute_vars(interp, argv[i])) == NULL)
			return (TCL_ERROR);
		if (parse_ie(interp, arg, &ie, &ietype)) {
			free(arg);
			return (TCL_ERROR);
		}
		free(arg);

		switch (ietype) {

		  case UNI_IE_RESTART:
			if (IE_ISPRESENT(resp.restart)) {
				set_error(interp, "double restart in %s",
				    argv[0]);
				return (TCL_ERROR);
			}
			resp.restart = ie.restart;
			break;

		  case UNI_IE_CONNID:
			if (IE_ISPRESENT(resp.connid)) {
				set_error(interp, "double connid in %s",
				    argv[0]);
				return (TCL_ERROR);
			}
			resp.connid = ie.connid;
			break;

		  default:
			set_error(interp, "bad IE 0x%02x for %s", ietype,
			    argv[0]);
			return (TCL_ERROR);
		}
	}
	a.sig = UNIAPI_RESET_response;
	send_uniapiv(&a, &resp, sizeof(resp), NULL);

	return (TCL_OK);
}
static int
do_setup_request(Tcl_Interp *interp, int argc, const char *argv[])
{
	struct uniapi_setup_request req;
	struct uni_arg a;

	argc -= 2;
	argv += 2;

	(void)memset(&req, 0, sizeof(req));
	if (parse_msg_setup(interp, argc, argv, &req.setup, unicx) != TCL_OK)
		return (TCL_ERROR);

	a.sig = UNIAPI_SETUP_request;
	send_uniapiv(&a, &req, sizeof(req), NULL);

	return (TCL_OK);
}
static int
do_proceeding_request(Tcl_Interp *interp, int argc, const char *argv[])
{
	struct uniapi_proceeding_request req;
	struct uni_arg a;

	argc -= 2;
	argv += 2;

	(void)memset(&req, 0, sizeof(req));
	if (parse_cref(interp, argc, (const char **)argv,
	    &req.call_proc.hdr.cref) != TCL_OK)
		return (TCL_ERROR);
	argc -= 2;
	argv += 2;
	if (argc == 0) {
		set_error(interp, "msgact missing in call_proc.request");
		return (TCL_ERROR);
	}
	if (parse_msgact(interp, *argv, &req.call_proc.hdr) != TCL_OK)
		return (TCL_ERROR);
	argc--;
	argv++;
	if (parse_msg_call_proc(interp, argc, argv, &req.call_proc, unicx) != TCL_OK)
		return (TCL_ERROR);

	a.sig = UNIAPI_PROCEEDING_request;
	send_uniapiv(&a, &req, sizeof(req), NULL);

	return (TCL_OK);
}
static int
do_alerting_request(Tcl_Interp *interp, int argc, const char *argv[])
{
	struct uniapi_alerting_request req;
	struct uni_arg a;

	argc -= 2;
	argv += 2;

	(void)memset(&req, 0, sizeof(req));
	if (parse_cref(interp, argc, (const char **)argv,
	    &req.alerting.hdr.cref) != TCL_OK)
		return (TCL_ERROR);
	argc -= 2;
	argv += 2;
	if (argc == 0) {
		set_error(interp, "msgact missing in alerting.request");
		return (TCL_ERROR);
	}
	if (parse_msgact(interp, *argv, &req.alerting.hdr) != TCL_OK)
		return (TCL_ERROR);
	argc--;
	argv++;
	if (parse_msg_alerting(interp, argc, argv, &req.alerting, unicx) != TCL_OK)
		return (TCL_ERROR);

	a.sig = UNIAPI_ALERTING_request;
	send_uniapiv(&a, &req, sizeof(req), NULL);

	return (TCL_OK);
}
static int
do_setup_response(Tcl_Interp *interp, int argc, const char *argv[])
{
	struct uniapi_setup_response req;
	struct uni_arg a;

	argc -= 2;
	argv += 2;

	(void)memset(&req, 0, sizeof(req));
	if (parse_cref(interp, argc, (const char **)argv,
	    &req.connect.hdr.cref) != TCL_OK)
		return (TCL_ERROR);
	argc -= 2;
	argv += 2;
	if (argc == 0) {
		set_error(interp, "msgact missing in setup.response");
		return (TCL_ERROR);
	}
	if (parse_msgact(interp, *argv, &req.connect.hdr) != TCL_OK)
		return (TCL_ERROR);
	argc--;
	argv++;
	if (parse_msg_connect(interp, argc, argv, &req.connect, unicx) != TCL_OK)
		return (TCL_ERROR);

	a.sig = UNIAPI_SETUP_response;
	send_uniapiv(&a, &req, sizeof(req), NULL);

	return (TCL_OK);
}
static int
do_setup_complete_request(Tcl_Interp *interp, int argc, const char *argv[])
{
	struct uniapi_setup_complete_request req;
	struct uni_arg a;

	argc -= 2;
	argv += 2;

	(void)memset(&req, 0, sizeof(req));
	if (parse_cref(interp, argc, (const char **)argv,
	    &req.connect_ack.hdr.cref) != TCL_OK)
		return (TCL_ERROR);
	argc -= 2;
	argv += 2;
	if (argc == 0) {
		set_error(interp, "msgact missing in setup-complete.request");
		return (TCL_ERROR);
	}
	if (parse_msgact(interp, *argv, &req.connect_ack.hdr) != TCL_OK)
		return (TCL_ERROR);
	argc--;
	argv++;
	if (parse_msg_connect_ack(interp, argc, argv, &req.connect_ack, unicx) != TCL_OK)
		return (TCL_ERROR);

	a.sig = UNIAPI_SETUP_COMPLETE_request;
	send_uniapiv(&a, &req, sizeof(req), NULL);

	return (TCL_OK);
}
static int
do_release_response(Tcl_Interp *interp, int argc, const char *argv[])
{
	struct uniapi_release_response req;
	struct uni_arg a;

	argc -= 2;
	argv += 2;

	(void)memset(&req, 0, sizeof(req));
	if (parse_cref(interp, argc, (const char **)argv,
	    &req.release_compl.hdr.cref) != TCL_OK)
		return (TCL_ERROR);
	argc -= 2;
	argv += 2;
	if (argc == 0) {
		set_error(interp, "msgact missing in release.response");
		return (TCL_ERROR);
	}
	if (parse_msgact(interp, *argv, &req.release_compl.hdr) != TCL_OK)
		return (TCL_ERROR);
	argc--;
	argv++;
	if (parse_msg_release_compl(interp, argc, argv, &req.release_compl, unicx) != TCL_OK)
		return (TCL_ERROR);

	a.sig = UNIAPI_RELEASE_response;
	send_uniapiv(&a, &req, sizeof(req), NULL);

	return (TCL_OK);
}
static int
do_release_request(Tcl_Interp *interp, int argc, const char *argv[])
{
	struct uniapi_release_request req;
	struct uni_arg a;

	argc -= 2;
	argv += 2;

	(void)memset(&req, 0, sizeof(req));
	if (parse_cref(interp, argc, (const char **)argv,
	    &req.release.hdr.cref) != TCL_OK)
		return (TCL_ERROR);
	argc -= 2;
	argv += 2;
	if (argc == 0) {
		set_error(interp, "msgact missing in release.request");
		return (TCL_ERROR);
	}
	if (parse_msgact(interp, *argv, &req.release.hdr) != TCL_OK)
		return (TCL_ERROR);
	argc--;
	argv++;
	if (parse_msg_release(interp, argc, argv, &req.release, unicx) != TCL_OK)
		return (TCL_ERROR);

	a.sig = UNIAPI_RELEASE_request;
	send_uniapiv(&a, &req, sizeof(req), NULL);

	return (TCL_OK);
}
static int
do_notify_request(Tcl_Interp *interp, int argc, const char *argv[])
{
	struct uniapi_notify_request req;
	struct uni_arg a;

	argc -= 2;
	argv += 2;

	(void)memset(&req, 0, sizeof(req));
	if (parse_cref(interp, argc, (const char **)argv,
	    &req.notify.hdr.cref) != TCL_OK)
		return (TCL_ERROR);
	argc -= 2;
	argv += 2;
	if (argc == 0) {
		set_error(interp, "msgact missing in notify.request");
		return (TCL_ERROR);
	}
	if (parse_msgact(interp, *argv, &req.notify.hdr) != TCL_OK)
		return (TCL_ERROR);
	argc--;
	argv++;
	if (parse_msg_notify(interp, argc, argv, &req.notify, unicx) != TCL_OK)
		return (TCL_ERROR);

	a.sig = UNIAPI_NOTIFY_request;
	send_uniapiv(&a, &req, sizeof(req), NULL);

	return (TCL_OK);
}

static int
do_status_enquiry_request(Tcl_Interp *interp, int argc, const char *argv[])
{
	struct uniapi_status_enquiry_request req;
	struct uni_arg a;
	union uni_ieall ie;
	enum uni_ietype ietype;
	char *arg;
	int i;

	argc -= 2;
	argv += 2;

	(void)memset(&req, 0, sizeof(req));
	if (parse_cref(interp, argc, (const char **)argv,
	    &req.cref) != TCL_OK)
		return (TCL_ERROR);

	argc -= 2;
	argv += 2;

	for (i = 0; i < argc; i++) {
		if ((arg = substitute_vars(interp, argv[i])) == NULL)
			return (TCL_ERROR);
		if (parse_ie(interp, arg, &ie, &ietype)) {
			free(arg);
			return (TCL_ERROR);
		}
		free(arg);

		switch (ietype) {

		  case UNI_IE_EPREF:
			if (IE_ISPRESENT(req.epref)) {
				set_error(interp, "double epref in %s",
				    argv[0]);
				return (TCL_ERROR);
			}
			req.epref = ie.epref;
			break;

		  default:
			set_error(interp, "bad IE 0x%02x for %s", ietype,
			    argv[0]);
			return (TCL_ERROR);
		}
	}

	a.sig = UNIAPI_STATUS_ENQUIRY_request;
	send_uniapiv(&a, &req, sizeof(req), NULL);

	return (TCL_OK);
}

static int
do_add_party_request(Tcl_Interp *interp, int argc, const char *argv[])
{
	struct uniapi_add_party_request req;
	struct uni_arg a;

	argc -= 2;
	argv += 2;

	(void)memset(&req, 0, sizeof(req));
	if (parse_cref(interp, argc, (const char **)argv,
	    &req.add.hdr.cref) != TCL_OK)
		return (TCL_ERROR);
	argc -= 2;
	argv += 2;
	if (argc == 0) {
		set_error(interp, "msgact missing in add_party.request");
		return (TCL_ERROR);
	}
	if (parse_msgact(interp, *argv, &req.add.hdr) != TCL_OK)
		return (TCL_ERROR);
	argc--;
	argv++;
	if (parse_msg_add_party(interp, argc, argv, &req.add, unicx) != TCL_OK)
		return (TCL_ERROR);

	a.sig = UNIAPI_ADD_PARTY_request;
	send_uniapiv(&a, &req, sizeof(req), NULL);

	return (TCL_OK);
}

static int
do_add_party_ack_request(Tcl_Interp *interp, int argc, const char *argv[])
{
	struct uniapi_add_party_ack_request req;
	struct uni_arg a;

	argc -= 2;
	argv += 2;

	(void)memset(&req, 0, sizeof(req));
	if (parse_cref(interp, argc, (const char **)argv,
	    &req.ack.hdr.cref) != TCL_OK)
		return (TCL_ERROR);
	argc -= 2;
	argv += 2;
	if (argc == 0) {
		set_error(interp, "msgact missing in add_party_ack.request");
		return (TCL_ERROR);
	}
	if (parse_msgact(interp, *argv, &req.ack.hdr) != TCL_OK)
		return (TCL_ERROR);
	argc--;
	argv++;
	if (parse_msg_add_party_ack(interp, argc, argv, &req.ack, unicx) != TCL_OK)
		return (TCL_ERROR);

	a.sig = UNIAPI_ADD_PARTY_ACK_request;
	send_uniapiv(&a, &req, sizeof(req), NULL);

	return (TCL_OK);
}

static int
do_add_party_rej_request(Tcl_Interp *interp, int argc, const char *argv[])
{
	struct uniapi_add_party_rej_request req;
	struct uni_arg a;

	argc -= 2;
	argv += 2;

	(void)memset(&req, 0, sizeof(req));
	if (parse_cref(interp, argc, (const char **)argv,
	    &req.rej.hdr.cref) != TCL_OK)
		return (TCL_ERROR);
	argc -= 2;
	argv += 2;
	if (argc == 0) {
		set_error(interp, "msgact missing in add_party_rej.request");
		return (TCL_ERROR);
	}
	if (parse_msgact(interp, *argv, &req.rej.hdr) != TCL_OK)
		return (TCL_ERROR);
	argc--;
	argv++;
	if (parse_msg_add_party_rej(interp, argc, argv, &req.rej, unicx) != TCL_OK)
		return (TCL_ERROR);

	a.sig = UNIAPI_ADD_PARTY_REJ_request;
	send_uniapiv(&a, &req, sizeof(req), NULL);

	return (TCL_OK);
}

static int
do_party_alerting_request(Tcl_Interp *interp, int argc, const char *argv[])
{
	struct uniapi_party_alerting_request req;
	struct uni_arg a;

	argc -= 2;
	argv += 2;

	(void)memset(&req, 0, sizeof(req));
	if (parse_cref(interp, argc, (const char **)argv,
	    &req.alert.hdr.cref) != TCL_OK)
		return (TCL_ERROR);
	argc -= 2;
	argv += 2;
	if (argc == 0) {
		set_error(interp, "msgact missing in party-alerting.request");
		return (TCL_ERROR);
	}
	if (parse_msgact(interp, *argv, &req.alert.hdr) != TCL_OK)
		return (TCL_ERROR);
	argc--;
	argv++;
	if (parse_msg_party_alerting(interp, argc, argv, &req.alert, unicx) != TCL_OK)
		return (TCL_ERROR);

	a.sig = UNIAPI_PARTY_ALERTING_request;
	send_uniapiv(&a, &req, sizeof(req), NULL);

	return (TCL_OK);
}

static int
do_drop_party_request(Tcl_Interp *interp, int argc, const char *argv[])
{
	struct uniapi_drop_party_request req;
	struct uni_arg a;

	argc -= 2;
	argv += 2;

	(void)memset(&req, 0, sizeof(req));
	if (parse_cref(interp, argc, (const char **)argv,
	    &req.drop.hdr.cref) != TCL_OK)
		return (TCL_ERROR);
	argc -= 2;
	argv += 2;
	if (argc == 0) {
		set_error(interp, "msgact missing in drop_party.request");
		return (TCL_ERROR);
	}
	if (parse_msgact(interp, *argv, &req.drop.hdr) != TCL_OK)
		return (TCL_ERROR);
	argc--;
	argv++;
	if (parse_msg_drop_party(interp, argc, argv, &req.drop, unicx) != TCL_OK)
		return (TCL_ERROR);

	a.sig = UNIAPI_DROP_PARTY_request;
	send_uniapiv(&a, &req, sizeof(req), NULL);

	return (TCL_OK);
}

static int
do_drop_party_ack_request(Tcl_Interp *interp, int argc, const char *argv[])
{
	struct uniapi_drop_party_ack_request req;
	struct uni_arg a;

	argc -= 2;
	argv += 2;

	(void)memset(&req, 0, sizeof(req));
	if (parse_cref(interp, argc, (const char **)argv,
	    &req.ack.hdr.cref) != TCL_OK)
		return (TCL_ERROR);
	argc -= 2;
	argv += 2;
	if (argc == 0) {
		set_error(interp, "msgact missing in drop_party_ack.request");
		return (TCL_ERROR);
	}
	if (parse_msgact(interp, *argv, &req.ack.hdr) != TCL_OK)
		return (TCL_ERROR);
	argc--;
	argv++;
	if (parse_msg_drop_party_ack(interp, argc, argv, &req.ack, unicx) != TCL_OK)
		return (TCL_ERROR);

	a.sig = UNIAPI_DROP_PARTY_ACK_request;
	send_uniapiv(&a, &req, sizeof(req), NULL);

	return (TCL_OK);
}


static struct {
	const char *name;
	int	(*func)(Tcl_Interp *, int, const char *[]);
} functab[] = {
	{ "link-establish.request",	do_link_establish_request },
	{ "link-release.request",	do_link_release_request },
	{ "reset.request",		do_reset_request },
	{ "reset.response",		do_reset_response },
	{ "reset-error.response",	do_reset_error_response },
	{ "setup.request",		do_setup_request },
	{ "setup.response",		do_setup_response },
	{ "setup-complete.request",	do_setup_complete_request },
	{ "proceeding.request",		do_proceeding_request },
	{ "alerting.request",		do_alerting_request },
	{ "release.response",		do_release_response },
	{ "release.request",		do_release_request },
	{ "notify.request",		do_notify_request },
	{ "status-enquiry.request",	do_status_enquiry_request },
	{ "add-party.request",		do_add_party_request },
	{ "party-alerting.request",	do_party_alerting_request },
	{ "add-party-ack.request",	do_add_party_ack_request },
	{ "add-party-rej.request",	do_add_party_rej_request },
	{ "drop-party.request",		do_drop_party_request },
	{ "drop-party-ack.request",	do_drop_party_ack_request },
	{ NULL, NULL }
};

int
cmd_dosig(ClientData data __unused, Tcl_Interp *interp, int argc,
    const char *argv[])
{
	u_int i;

	if (argc == 1) {
		set_error(interp, "usage: dosig event [args...]");
		return (TCL_ERROR);
	}

	for (i = 0; functab[i].name; i++)
		if (strcmp(argv[1], functab[i].name) == 0)
			return functab[i].func(interp, argc, argv);

	set_error(interp, "bad signal '%s'", argv[1]);
	return (TCL_ERROR);
}


/*
 * Execute the given command after the number of milli-seconds given.
 * 
 */
struct sleep {
	Tcl_Interp *interp;
	Tcl_DString str;
};

static void
sleep_func(int tid __unused, void *uap)
{
	struct sleep *arg = uap;

	try_event(arg->interp, &arg->str);
	Tcl_DStringFree(&arg->str);
	free(arg);
}

int
cmd_sleep(ClientData data __unused, Tcl_Interp *interp,
    int argc, const char *argv[])
{
	char *end;
	u_long t;
	struct sleep *arg;

	if (argc != 3) {
		set_error(interp, "usage: sleep time command");
		return (TCL_ERROR);
	}
	t = strtoul(argv[1], &end, 10);
	if (*end != '\0') {
		set_error(interp, "bad integer '%s' in sleep", argv[1]);
		return (TCL_ERROR);
	}

	if ((arg = malloc(sizeof(*arg))) == NULL)
		err(1, NULL);

	Tcl_DStringInit(&arg->str);
	Tcl_DStringAppend(&arg->str, argv[2], (int)strlen(argv[2]));
	arg->interp = interp;

	(void)poll_start_timer(t, 0, sleep_func, arg);

	return (TCL_OK);
}
