/*
 * Copyright (c) 2001-2003
 *	Fraunhofer Institute for Open Communication Systems (FhG Fokus).
 * 	All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Author: Hartmut Brandt <harti@freebsd.org>
 *
 * $Begemot: atmsupport/test_uni_user/user.c,v 1.4 2004/01/02 15:15:56 hbb Exp $
 *
 * Test user space UNI stack. User interface.
 */
#include "uniu.h"
#include "unitcl.h"

static int
get_cookie(Tcl_Interp *interp, const char *arg, u_int32_t *ptr)
{
	char *end;
	u_long n;

	n = strtoul(arg, &end, 10);
	if (*end != '\0' || n == 0 || n > 0xffffffff) {
		set_error(interp, "bad cookie '%s'", arg);
		return (TCL_ERROR);
	}
	*ptr = (u_int32_t)n;
	return (TCL_OK);
}

static void
init_str(Tcl_DString *str, const char *ev, u_int32_t cookie)
{
	char buf[100];

	Tcl_DStringAppendElement(str, ev);
	sprintf(buf, "%u", cookie);
	Tcl_DStringAppendElement(str, buf);
}

/*
 * Output from the UNI stack.
 */
void
user_output(struct stk *stk, enum uni_sig sig, u_int32_t cookie,
    struct uni_msg *m)
{
	char buf[100];

	Tcl_DString str;

	Tcl_DStringInit(&str);

	switch (sig) {

	  case UNIAPI_ERROR:
	    {
		struct uniapi_error *arg =
		    uni_msg_rptr(m, struct uniapi_error *);

		init_str(&str, "uni_error", cookie);

		sprintf(buf, "%u", arg->reason);
		Tcl_DStringAppendElement(&str, buf);
		sprintf(buf, "%u", arg->state);
		Tcl_DStringAppendElement(&str, buf);
		try_event(stk->interp, &str);
		break;
	    }

	  case UNIAPI_CALL_CREATED:
	    {
		struct uniapi_call_created *arg =
		    uni_msg_rptr(m, struct uniapi_call_created *);

		init_str(&str, "uni_call_created", cookie);

		fmt_cref(&str, &arg->cref);
		try_event(stk->interp, &str);
		break;
	    }
	  case UNIAPI_CALL_DESTROYED:
	    {
		struct uniapi_call_destroyed *arg =
		    uni_msg_rptr(m, struct uniapi_call_destroyed *);

		init_str(&str, "uni_call_destroyed", cookie);

		fmt_cref(&str, &arg->cref);
		try_event(stk->interp, &str);
		break;
	    }
	  case UNIAPI_PARTY_CREATED:
	    {
		struct uniapi_party_created *arg =
		    uni_msg_rptr(m, struct uniapi_party_created *);

		init_str(&str, "uni_party_created", cookie);

		fmt_cref(&str, &arg->cref);
		if (fmt_epref(stk->interp, &str, &arg->epref) != TCL_OK)
			goto err;
		try_event(stk->interp, &str);
		break;
	    }
	  case UNIAPI_PARTY_DESTROYED:
	    {
		struct uniapi_party_destroyed *arg =
		    uni_msg_rptr(m, struct uniapi_party_destroyed *);

		init_str(&str, "uni_party_destroyed", cookie);

		fmt_cref(&str, &arg->cref);
		if (fmt_epref(stk->interp, &str, &arg->epref) != TCL_OK)
			goto err;
		try_event(stk->interp, &str);
		break;
	    }

	  case UNIAPI_LINK_ESTABLISH_confirm:
		init_str(&str, "uni_link_establish_confirm", cookie);
		try_event(stk->interp, &str);
		break;
	  case UNIAPI_LINK_RELEASE_confirm:
		init_str(&str, "uni_link_release_confirm", cookie);
		try_event(stk->interp, &str);
		break;

	  case UNIAPI_RESET_confirm:
	    {
		struct uniapi_reset_confirm *arg =
		    uni_msg_rptr(m, struct uniapi_reset_confirm *);

		init_str(&str, "uni_reset_confirm", cookie);

		if (fmt_restart(stk->interp, &str, &arg->restart) != TCL_OK)
			goto err;
		if (fmt_connid(stk->interp, &str, &arg->connid) != TCL_OK)
			goto err;
		try_event(stk->interp, &str);
		break;
	    }
	  case UNIAPI_RESET_indication:
	    {
		struct uniapi_reset_indication *arg =
		    uni_msg_rptr(m, struct uniapi_reset_indication *);

		init_str(&str, "uni_reset_indication", cookie);

		if (fmt_restart(stk->interp, &str, &arg->restart) != TCL_OK)
			goto err;
		if (fmt_connid(stk->interp, &str, &arg->connid) != TCL_OK)
			goto err;
		try_event(stk->interp, &str);
		break;
	    }
	  case UNIAPI_RESET_ERROR_indication:
	    {
		struct uniapi_reset_error_indication *arg =
		    uni_msg_rptr(m, struct uniapi_reset_error_indication *);

		init_str(&str, "uni_reset_error_indication", cookie);

		sprintf(buf, "%u", arg->source);
		Tcl_DStringAppendElement(&str, buf);
		sprintf(buf, "%u", arg->reason);
		Tcl_DStringAppendElement(&str, buf);
		try_event(stk->interp, &str);
		break;
	    }
	  case UNIAPI_RESET_STATUS_indication:
	    {
		struct uniapi_reset_status_indication *arg =
		    uni_msg_rptr(m, struct uniapi_reset_status_indication *);

		init_str(&str, "uni_reset_status_indication", cookie);

		fmt_cref(&str, &arg->cref);
		if (fmt_callstate(stk->interp, &str, &arg->callstate) != TCL_OK)
			goto err;
		if (fmt_cause(stk->interp, &str, &arg->cause) != TCL_OK)
			goto err;
		try_event(stk->interp, &str);
		break;
	    }

	  case UNIAPI_SETUP_indication:
	    {
		struct uniapi_setup_indication *arg =
		    uni_msg_rptr(m, struct uniapi_setup_indication *);

		init_str(&str, "uni_setup_indication", cookie);

		if (fmt_msg_setup(stk->interp, &str, &arg->setup) != TCL_OK)
			goto err;
		try_event(stk->interp, &str);
		break;
	    }
	  case UNIAPI_SETUP_confirm:
	    {
		struct uniapi_setup_confirm *arg =
		    uni_msg_rptr(m, struct uniapi_setup_confirm *);

		init_str(&str, "uni_setup_confirm", cookie);

		if (fmt_msg_connect(stk->interp, &str, &arg->connect) != TCL_OK)
			goto err;
		try_event(stk->interp, &str);
		break;
	    }
	  case UNIAPI_SETUP_COMPLETE_indication:
	    {
		struct uniapi_setup_complete_indication *arg =
		    uni_msg_rptr(m, struct uniapi_setup_complete_indication *);

		init_str(&str, "uni_setup_complete_indication", cookie);

		if (fmt_msg_connect_ack(stk->interp, &str, &arg->connect_ack) != TCL_OK)
			goto err;
		try_event(stk->interp, &str);
		break;
	    }
	  case UNIAPI_ALERTING_indication:
	    {
		struct uniapi_alerting_indication *arg =
		    uni_msg_rptr(m, struct uniapi_alerting_indication *);

		init_str(&str, "uni_alerting_indication", cookie);

		if (fmt_msg_alerting(stk->interp, &str, &arg->alerting) != TCL_OK)
			goto err;
		try_event(stk->interp, &str);
		break;
	    }
	  case UNIAPI_PROCEEDING_indication:
	    {
		struct uniapi_proceeding_indication *arg =
		    uni_msg_rptr(m, struct uniapi_proceeding_indication *);

		init_str(&str, "uni_proceeding_indication", cookie);

		if (fmt_msg_call_proc(stk->interp, &str, &arg->call_proc) != TCL_OK)
			goto err;
		try_event(stk->interp, &str);
		break;
	    }
	  case UNIAPI_RELEASE_indication:
	    {
		struct uniapi_release_indication *arg =
		    uni_msg_rptr(m, struct uniapi_release_indication *);

		init_str(&str, "uni_release_indication", cookie);

		if (fmt_msg_release(stk->interp, &str, &arg->release) != TCL_OK)
			goto err;
		try_event(stk->interp, &str);
		break;
	    }
	  case UNIAPI_RELEASE_confirm:
	    {
		struct uniapi_release_confirm *arg =
		    uni_msg_rptr(m, struct uniapi_release_confirm *);

		init_str(&str, "uni_release_confirm", cookie);

		if (fmt_msg_release(stk->interp, &str, &arg->release) != TCL_OK)
			goto err;
		try_event(stk->interp, &str);
		break;
	    }
	  case UNIAPI_NOTIFY_indication:
	    {
		struct uniapi_notify_indication *arg =
		    uni_msg_rptr(m, struct uniapi_notify_indication *);

		init_str(&str, "uni_notify_indication", cookie);

		if (fmt_msg_notify(stk->interp, &str, &arg->notify) != TCL_OK)
			goto err;
		try_event(stk->interp, &str);
		break;
	    }
	  case UNIAPI_STATUS_indication:
	    {
		struct uniapi_status_indication *arg =
		    uni_msg_rptr(m, struct uniapi_status_indication *);

		init_str(&str, "uni_status_indication", cookie);

		fmt_cref(&str, &arg->cref);
		sprintf(buf, "%u", arg->my_state);
		Tcl_DStringAppendElement(&str, buf);
		sprintf(buf, "%u", arg->my_cause);
		Tcl_DStringAppendElement(&str, buf);
		if (fmt_callstate(stk->interp, &str, &arg->his_state) != TCL_OK)
			goto err;
		if (fmt_cause(stk->interp, &str, &arg->his_cause) != TCL_OK)
			goto err;
		try_event(stk->interp, &str);
		break;
	    }

	  case UNIAPI_ADD_PARTY_indication:
	    {
		struct uniapi_add_party_indication *arg =
		    uni_msg_rptr(m, struct uniapi_add_party_indication *);

		init_str(&str, "uni_add_party_indication", cookie);

		if (fmt_msg_add_party(stk->interp, &str, &arg->add) != TCL_OK)
			goto err;
		try_event(stk->interp, &str);
		break;
	    }
	  case UNIAPI_PARTY_ALERTING_indication:
	    {
		struct uniapi_party_alerting_indication *arg =
		    uni_msg_rptr(m, struct uniapi_party_alerting_indication *);

		init_str(&str, "uni_party_alerting_indication", cookie);

		if (fmt_msg_party_alerting(stk->interp, &str, &arg->alert) != TCL_OK)
			goto err;
		try_event(stk->interp, &str);
		break;
	    }
	  case UNIAPI_ADD_PARTY_ACK_indication:
	    {
		struct uniapi_add_party_ack_indication *arg =
		    uni_msg_rptr(m, struct uniapi_add_party_ack_indication *);

		init_str(&str, "uni_add_party_ack_indication", cookie);

		if (fmt_msg_add_party_ack(stk->interp, &str, &arg->ack) != TCL_OK)
			goto err;
		try_event(stk->interp, &str);
		break;
	    }
	  case UNIAPI_ADD_PARTY_REJ_indication:
	    {
		struct uniapi_add_party_rej_indication *arg =
		    uni_msg_rptr(m, struct uniapi_add_party_rej_indication *);

		init_str(&str, "uni_add_party_rej_indication", cookie);

		if (fmt_msg_add_party_rej(stk->interp, &str, &arg->rej) != TCL_OK)
			goto err;
		try_event(stk->interp, &str);
		break;
	    }
	  case UNIAPI_DROP_PARTY_indication:
	    {
		struct uniapi_drop_party_indication *arg =
		    uni_msg_rptr(m, struct uniapi_drop_party_indication *);

		init_str(&str, "uni_drop_party_indication", cookie);

		if (fmt_msg_drop_party(stk->interp, &str, &arg->drop) != TCL_OK)
			goto err;
		try_event(stk->interp, &str);
		break;
	    }
	  case UNIAPI_DROP_PARTY_ACK_indication:
	    {
		struct uniapi_drop_party_ack_indication *arg =
		    uni_msg_rptr(m, struct uniapi_drop_party_ack_indication *);

		init_str(&str, "uni_drop_party_ack_indication", cookie);

		if (fmt_msg_drop_party(stk->interp, &str, &arg->drop) != TCL_OK)
			goto err;
		if (fmt_crankback(stk->interp, &str, &arg->crankback) != TCL_OK)
			goto err;
		try_event(stk->interp, &str);
		break;
	    }

	  default:
		abort();
	}

	return;

  err:
	Tcl_DStringFree(&str);
	printf("error: %s\n", Tcl_GetStringResult(stk->interp));
}

static int
cmd_uni_debug(ClientData data, Tcl_Interp *interp,
    int argc, const char *argv[])
{
	struct stk *stk = (struct stk *)data;
	Tcl_DString res;
	u_int level;
	char buf[100];
	int arg, e, ac;
	const char **av;
	char *end;

	if (argc == 1) {
		Tcl_DStringInit(&res);

# define UNI_DEBUG_DEFINE(D)						\
		Tcl_DStringStartSublist(&res);				\
		level = uni_get_debug(stk->uni, UNI_FAC_##D);		\
		Tcl_DStringAppendElement(&res, #D);			\
		sprintf(buf, "%u", level);				\
		Tcl_DStringAppendElement(&res, buf);			\
		Tcl_DStringEndSublist(&res);

		UNI_DEBUG_FACILITIES

# undef UNI_DEBUG_DEFINE

		Tcl_DStringResult(interp, &res);
		return (TCL_OK);
	}

	for (arg = 1; arg < argc; arg++) {
		if ((e = Tcl_SplitList(interp, argv[arg], &ac, &av)) != TCL_OK)
			return (e);

		if (ac == 0) {
			Tcl_Free((char *)av);
			continue;
		}

		if (ac != 2) {
			set_error(interp, "need one arg for facility '%s'", av[0]);
			Tcl_Free((char *)av);
			return (TCL_ERROR);
		}

		level = strtoul(av[1], &end, 10);
		if (*end != '\0') {
			set_error(interp, "bad debug level '%s'", av[1]);
			Tcl_Free((char *)av);
			return (TCL_ERROR);
		}

# define UNI_DEBUG_DEFINE(D)						\
		if (strcmp(av[0], #D) == 0) {				\
			uni_set_debug(stk->uni, UNI_FAC_##D, level);	\
		} else 

		UNI_DEBUG_FACILITIES

# undef UNI_DEBUG_DEFINE

		{
			set_error(interp, "unknown debug facility '%s'", av[0]);
			Tcl_Free((char *)av);
			return (TCL_ERROR);
		}
	}
	return (TCL_OK);
}

static int
cmd_uni_status(ClientData data, Tcl_Interp *interp,
    int argc, const char *argv[] __unused)
{
	struct stk *stk = (struct stk *)data;

	if (argc != 1) {
		set_error(interp, "usage: '%s'", argv[0]);
		return (TCL_ERROR);
	}
	uni_status(stk->uni, NULL);
	return (TCL_OK);
}

static int
cmd_uni_get_config(ClientData data, Tcl_Interp *interp,
    int argc, const char *argv[] __unused)
{
	struct stk *stk = (struct stk *)data;
	struct uni_config config;
	Tcl_DString res;
	char buf[100];

	if (argc != 1) {
		set_error(interp, "usage: '%s'", argv[0]);
		return (TCL_ERROR);
	}
	uni_get_config(stk->uni, &config);

	Tcl_DStringInit(&res);

	Tcl_DStringStartSublist(&res);
	Tcl_DStringAppendElement(&res, "proto");
	switch (config.proto) {

	  case UNIPROTO_UNI40U:
		Tcl_DStringAppendElement(&res, "uni4.0u");
		break;
	  case UNIPROTO_UNI40N:
		Tcl_DStringAppendElement(&res, "uni4.0n");
		break;
	  case UNIPROTO_PNNI10:
		Tcl_DStringAppendElement(&res, "pnni1.0");
		break;
	  default:
		set_error(interp, "bad UNI proto %u", config.proto);
		goto err;
	}
	Tcl_DStringEndSublist(&res);

	Tcl_DStringStartSublist(&res);
	Tcl_DStringAppendElement(&res, "popt");
	if (config.popt & UNIPROTO_GFP) {
		Tcl_DStringAppendElement(&res, "gfp");
		config.popt &= ~UNIPROTO_GFP;
	}
	Tcl_DStringEndSublist(&res);
	if (config.popt != 0) {
		set_error(interp, "bad proto option 0x%08x", config.popt);
		goto err;
	}

	Tcl_DStringStartSublist(&res);
	Tcl_DStringAppendElement(&res, "option");
	if (config.option & UNIOPT_GIT_HARD) {
		Tcl_DStringAppendElement(&res, "git_hard");
		config.option &= ~UNIOPT_GIT_HARD;
	}
	if (config.option & UNIOPT_BEARER_HARD) {
		Tcl_DStringAppendElement(&res, "bearer_hard");
		config.option &= ~UNIOPT_BEARER_HARD;
	}
	if (config.option & UNIOPT_CAUSE_HARD) {
		Tcl_DStringAppendElement(&res, "cause_hard");
		config.option &= ~UNIOPT_CAUSE_HARD;
	}
	Tcl_DStringEndSublist(&res);
	if (config.option != 0) {
		set_error(interp, "bad option 0x%08x", config.option);
		goto err;
	}

# define DO_TIMER(T)						\
	Tcl_DStringStartSublist(&res);				\
	Tcl_DStringAppendElement(&res, "T"#T);			\
	sprintf(buf, "%u", config.timer##T);			\
	Tcl_DStringAppendElement(&res, buf);			\
	Tcl_DStringEndSublist(&res);

# define DO_TIMER2(T)						\
	Tcl_DStringStartSublist(&res);				\
	Tcl_DStringAppendElement(&res, "T"#T);			\
	sprintf(buf, "%u", config.timer##T);			\
	Tcl_DStringAppendElement(&res, buf);			\
	sprintf(buf, "%u", config.init##T);			\
	Tcl_DStringAppendElement(&res, buf);			\
	Tcl_DStringEndSublist(&res);

	DO_TIMER(301);
	DO_TIMER2(303);
	DO_TIMER2(308);
	DO_TIMER(309);
	DO_TIMER(310);
	DO_TIMER(313);
	DO_TIMER2(316);
	DO_TIMER(317);
	DO_TIMER2(322);
	DO_TIMER(397);
	DO_TIMER(398);
	DO_TIMER(399);

# undef DO_TIMER
# undef DO_TIMER2

	Tcl_DStringResult(interp, &res);
	return (TCL_OK);

  err:
	Tcl_DStringFree(&res);
	return (TCL_ERROR);
}

static int
cmd_uni_set_config(ClientData data, Tcl_Interp *interp,
    int argc, const char *argv[])
{
	struct stk *stk = (struct stk *)data;
	struct uni_config config;
	u_int32_t mask, popt_mask, opt_mask;
	int e, i, arg, flg;
	int ac;
	const char **av;
	char *end;

	memset(&config, 0, sizeof(config));
	mask = popt_mask = opt_mask = 0;

	for (arg = 1; arg < argc; arg++) {
		if ((e = Tcl_SplitList(interp, argv[arg], &ac, &av)) != TCL_OK)
			return (e);

		if (ac == 0) {
			Tcl_Free((char *)av);
			continue;
		}

		if (strcmp(av[0], "proto") == 0) {
			mask |= UNICFG_PROTO;
			if (ac != 2) {
				set_error(interp, "proto needs one arg");
				Tcl_Free((char *)av);
				return (TCL_ERROR);
			}
			if (strcmp(av[1], "uni4.0u"))
				config.proto = UNIPROTO_UNI40U;
			else if (strcmp(av[1], "uni4.0n"))
				config.proto = UNIPROTO_UNI40N;
			else if (strcmp(av[1], "pnni1.0"))
				config.proto = UNIPROTO_PNNI10;
			else {
				set_error(interp, "bad proto '%s'", av[1]);
				Tcl_Free((char *)av);
				return (TCL_ERROR);
			}

		} else if(strcmp(av[0], "popt") == 0) {
			if ((ac - 1) % 2 != 0) {
				set_error(interp, "popt needs even number "
				    "of args");
				Tcl_Free((char *)av);
				return (TCL_ERROR);
			}
			for (i = 1; i < ac; i += 2) {
				if (strcmp(av[i+1], "0") == 0)
					flg = 0;
				else if(strcmp(av[i+1], "1") == 0)
					flg = 0;
				else {
					set_error(interp, "popt must "
					    "be 0 or 1: '%s'", av[i+1]);
					Tcl_Free((char *)av);
					return (TCL_ERROR);
				}
				if (strcmp(av[i], "gfp")) {
					popt_mask |= UNIPROTO_GFP;
					if (flg)
						config.popt |= UNIPROTO_GFP;
					else
						config.popt &= ~UNIPROTO_GFP;
				} else {
					set_error(interp, "unknown popt '%s'",
					    av[i]);
					Tcl_Free((char *)av);
					return (TCL_ERROR);
				}
			}

		} else if(strcmp(av[0], "option") == 0) {
			if ((ac - 1) % 2 != 0) {
				set_error(interp, "option needs even number "
				    "of args");
				Tcl_Free((char *)av);
				return (TCL_ERROR);
			}
			for (i = 1; i < ac; i += 2) {
				if (strcmp(av[i+1], "0") == 0)
					flg = 0;
				else if(strcmp(av[i+1], "1") == 0)
					flg = 0;
				else {
					set_error(interp, "option must "
					    "be 0 or 1: '%s'", av[i+1]);
					Tcl_Free((char *)av);
					return (TCL_ERROR);
				}
				if (strcmp(av[i], "git_hard")) {
					popt_mask |= UNIOPT_GIT_HARD;
					if (flg)
						config.popt |= UNIOPT_GIT_HARD;
					else
						config.popt &= ~UNIOPT_GIT_HARD;
				} else if (strcmp(av[i], "bearer_hard")) {
					popt_mask |= UNIOPT_BEARER_HARD;
					if (flg)
						config.popt |= UNIOPT_BEARER_HARD;
					else
						config.popt &= ~UNIOPT_BEARER_HARD;
				} else if (strcmp(av[i], "cause_hard")) {
					popt_mask |= UNIOPT_CAUSE_HARD;
					if (flg)
						config.popt |= UNIOPT_CAUSE_HARD;
					else
						config.popt &= ~UNIOPT_CAUSE_HARD;
				} else {
					set_error(interp, "unknown option '%s'",
					    av[i]);
					Tcl_Free((char *)av);
					return (TCL_ERROR);
				}
			}

# define DO_TIMER(T)							    \
			if (ac != 2) {					    \
				set_error(interp, "T"#T" needs one arg");   \
				Tcl_Free((char *)av);			    \
				return (TCL_ERROR);			    \
			}						    \
			config.timer##T = strtoul(av[1], &end, 10);	    \
			if (*end != '\0') {				    \
				set_error(interp, "bad number '%s'", av[1]);\
				Tcl_Free((char *)av);			    \
				return (TCL_ERROR);			    \
			}						    \
			mask |= UNICFG_TIMER##T;

# define DO_TIMER2(T)							    \
			if (ac != 3) {					    \
				set_error(interp, "T"#T" needs two args");  \
				Tcl_Free((char *)av);			    \
				return (TCL_ERROR);			    \
			}						    \
			config.timer##T = strtoul(av[1], &end, 10);	    \
			if (*end != '\0') {				    \
				set_error(interp, "bad number '%s'", av[1]);\
				Tcl_Free((char *)av);			    \
				return (TCL_ERROR);			    \
			}						    \
			config.init##T = strtoul(av[2], &end, 10);	    \
			if (*end != '\0') {				    \
				set_error(interp, "bad number '%s'", av[2]);\
				Tcl_Free((char *)av);			    \
				return (TCL_ERROR);			    \
			}						    \
			mask |= UNICFG_TIMER##T;

		} else if(strcmp(av[0], "T301") == 0) {
			DO_TIMER(301);
		} else if(strcmp(av[0], "T303") == 0) {
			DO_TIMER2(303);
		} else if(strcmp(av[0], "T308") == 0) {
			DO_TIMER2(308);
		} else if(strcmp(av[0], "T309") == 0) {
			DO_TIMER(309);
		} else if(strcmp(av[0], "T310") == 0) {
			DO_TIMER(310);
		} else if(strcmp(av[0], "T313") == 0) {
			DO_TIMER(313);
		} else if(strcmp(av[0], "T316") == 0) {
			DO_TIMER2(316);
		} else if(strcmp(av[0], "T317") == 0) {
			DO_TIMER(317);
		} else if(strcmp(av[0], "T322") == 0) {
			DO_TIMER2(322);
		} else if(strcmp(av[0], "T397") == 0) {
			DO_TIMER(397);
		} else if(strcmp(av[0], "T398") == 0) {
			DO_TIMER(398);
		} else if(strcmp(av[0], "T399") == 0) {
			DO_TIMER(399);
		} else {
			set_error(interp, "bad config suboption '%s'", av[0]);
			Tcl_Free((char *)av);
			return (TCL_ERROR);
		}

# undef DO_TIMER
# undef DO_TIMER2

		Tcl_Free((char *)av);
	}

	uni_set_config(stk->uni, &config, &mask, &popt_mask, &opt_mask);

	if (mask != 0 || popt_mask != 0 || opt_mask != 0) {
		set_error(stk->interp, "set_config failed: mask=%x popt_mask"
		    "=%x opt_mask=%x", mask, popt_mask, opt_mask);
		return (TCL_ERROR);
	}
	return (TCL_OK);
}

static int
cmd_uni_link_establish_request(ClientData data, Tcl_Interp *interp,
    int argc, const char *argv[])
{
	struct stk *stk = (struct stk *)data;
	u_int32_t cookie;
	struct uni_msg *m;

	if (argc != 2) {
		set_error(interp, "usage: '%s <cookie>'", argv[0]);
		return (TCL_ERROR);
	}
	if (get_cookie(interp, argv[1], &cookie) != TCL_OK)
		return (TCL_ERROR);
	if ((m = uni_msg_alloc(0)) == NULL)
		err(1, NULL);
	uni_uni_input(stk->uni, UNIAPI_LINK_ESTABLISH_request, cookie, m);
	uni_work(stk->uni);
	return (TCL_OK);
}

static int
cmd_uni_link_release_request(ClientData data, Tcl_Interp *interp,
    int argc, const char *argv[])
{
	struct stk *stk = (struct stk *)data;
	u_int32_t cookie;
	struct uni_msg *m;

	if (argc != 2) {
		set_error(interp, "usage: '%s <cookie>'", argv[0]);
		return (TCL_ERROR);
	}
	if (get_cookie(interp, argv[1], &cookie) != TCL_OK)
		return (TCL_ERROR);
	if ((m = uni_msg_alloc(0)) == NULL)
		err(1, NULL);
	uni_uni_input(stk->uni, UNIAPI_LINK_RELEASE_request, cookie, m);
	uni_work(stk->uni);
	return (TCL_OK);
}

static int
cmd_uni_reset_request(ClientData data, Tcl_Interp *interp,
    int argc, const char *argv[])
{
	struct stk *stk = (struct stk *)data;
	u_int32_t cookie;
	struct uni_msg *m;
	struct uniapi_reset_request *req;
	union uni_ieall ie;
	enum uni_ietype type;
	int i;

	if (argc < 2) {
		set_error(interp, "usage: '%s <cookie> [<ies>...]'", argv[0]);
		return (TCL_ERROR);
	}
	if (get_cookie(interp, argv[1], &cookie) != TCL_OK)
		return (TCL_ERROR);
	if ((m = uni_msg_alloc(sizeof(*req))) == NULL)
		err(1, NULL);
	req = uni_msg_rptr(m, struct uniapi_reset_request *);
	m->b_wptr += sizeof(*req);

	memset(req, 0, sizeof(*req));

	for (i = 2; i < argc; i++) {
		if (parse_ie(interp, argv[i], &ie, &type) != TCL_OK) {
			uni_msg_destroy(m);
			return (TCL_ERROR);
		}
		switch (type) {

		  case UNI_IE_RESTART:
			req->restart = ie.restart;
			break;

		  case UNI_IE_CONNID:
			req->connid = ie.connid;
			break;

		  default:
			set_error(interp, "bad IE for reset.request %02x",
			    type);
			uni_msg_destroy(m);
			return (TCL_ERROR);
		}
	}

	uni_uni_input(stk->uni, UNIAPI_RESET_request, cookie, m);
	uni_work(stk->uni);
	return (TCL_OK);
}

static int
cmd_uni_reset_response(ClientData data, Tcl_Interp *interp,
    int argc, const char *argv[])
{
	struct stk *stk = (struct stk *)data;
	u_int32_t cookie;
	struct uni_msg *m;
	struct uniapi_reset_response *req;
	union uni_ieall ie;
	enum uni_ietype type;
	int i;

	if (argc < 2) {
		set_error(interp, "usage: '%s <cookie> [<ies>...]'", argv[0]);
		return (TCL_ERROR);
	}
	if (get_cookie(interp, argv[1], &cookie) != TCL_OK)
		return (TCL_ERROR);
	if ((m = uni_msg_alloc(sizeof(*req))) == NULL)
		err(1, NULL);
	req = uni_msg_rptr(m, struct uniapi_reset_response *);
	m->b_wptr += sizeof(*req);

	memset(req, 0, sizeof(*req));

	for (i = 2; i < argc; i++) {
		if (parse_ie(interp, argv[i], &ie, &type) != TCL_OK) {
			uni_msg_destroy(m);
			return (TCL_ERROR);
		}
		switch (type) {

		  case UNI_IE_RESTART:
			req->restart = ie.restart;
			break;

		  case UNI_IE_CONNID:
			req->connid = ie.connid;
			break;

		  default:
			set_error(interp, "bad IE for reset.response %02x",
			    type);
			uni_msg_destroy(m);
			return (TCL_ERROR);
		}
	}

	uni_uni_input(stk->uni, UNIAPI_RESET_response, cookie, m);
	uni_work(stk->uni);
	return (TCL_OK);
}

static int
cmd_uni_reset_error_response(ClientData data, Tcl_Interp *interp,
    int argc, const char *argv[])
{
	struct stk *stk = (struct stk *)data;
	u_int32_t cookie;
	struct uni_msg *m;
	struct uniapi_reset_error_response *req;
	union uni_ieall ie;
	enum uni_ietype type;
	int i;

	if (argc < 2) {
		set_error(interp, "usage: '%s <cookie> [<ies>...]'", argv[0]);
		return (TCL_ERROR);
	}
	if (get_cookie(interp, argv[1], &cookie) != TCL_OK)
		return (TCL_ERROR);
	if ((m = uni_msg_alloc(sizeof(*req))) == NULL)
		err(1, NULL);
	req = uni_msg_rptr(m, struct uniapi_reset_error_response *);
	m->b_wptr += sizeof(*req);

	memset(req, 0, sizeof(*req));

	for (i = 2; i < argc; i++) {
		if (parse_ie(interp, argv[i], &ie, &type) != TCL_OK) {
			uni_msg_destroy(m);
			return (TCL_ERROR);
		}
		switch (type) {

		  case UNI_IE_CAUSE:
			req->cause = ie.cause;
			break;

		  default:
			set_error(interp, "bad IE for reset_error.response %02x",
			    type);
			uni_msg_destroy(m);
			return (TCL_ERROR);
		}
	}

	uni_uni_input(stk->uni, UNIAPI_RESET_ERROR_response, cookie, m);
	uni_work(stk->uni);
	return (TCL_OK);
}

static int
cmd_uni_setup_request(ClientData data, Tcl_Interp *interp,
    int argc, const char *argv[])
{
	struct stk *stk = (struct stk *)data;
	u_int32_t cookie;
	struct uni_msg *m;
	struct uniapi_setup_request *req;

	if (argc < 2) {
		set_error(interp, "usage: '%s <cookie> <act> [<ies>...]'", argv[0]);
		return (TCL_ERROR);
	}
	if (get_cookie(interp, argv[1], &cookie) != TCL_OK)
		return (TCL_ERROR);
	if ((m = uni_msg_alloc(sizeof(*req))) == NULL)
		err(1, NULL);
	req = uni_msg_rptr(m, struct uniapi_setup_request *);
	m->b_wptr += sizeof(*req);

	memset(req, 0, sizeof(*req));

	if (parse_msgact(interp, argv[2], &req->setup.hdr) != TCL_OK ||
	    parse_msg_setup(interp, argc - 3, argv + 3, &req->setup,
	    uni_context(stk->uni)) != TCL_OK) {
		uni_msg_destroy(m);
		return (TCL_ERROR);
	}

	uni_uni_input(stk->uni, UNIAPI_SETUP_request, cookie, m);
	uni_work(stk->uni);
	return (TCL_OK);
}

static int
cmd_uni_setup_response(ClientData data, Tcl_Interp *interp,
    int argc, const char *argv[])
{
	struct stk *stk = (struct stk *)data;
	u_int32_t cookie;
	struct uni_msg *m;
	struct uniapi_setup_response *req;

	if (argc < 5) {
		set_error(interp, "usage: '%s <cookie> <cref> <cref-flag> <act> [<ies>...]'", argv[0]);
		return (TCL_ERROR);
	}
	if (get_cookie(interp, argv[1], &cookie) != TCL_OK)
		return (TCL_ERROR);

	if ((m = uni_msg_alloc(sizeof(*req))) == NULL)
		err(1, NULL);
	req = uni_msg_rptr(m, struct uniapi_setup_response *);
	m->b_wptr += sizeof(*req);

	memset(req, 0, sizeof(*req));

	if (parse_cref(interp, argc - 2, (const char **)(argv + 2),
	    &req->connect.hdr.cref) != TCL_OK ||
	    parse_msgact(interp, argv[4], &req->connect.hdr) != TCL_OK ||
	    parse_msg_connect(interp, argc - 5, argv + 5, &req->connect,
	    uni_context(stk->uni)) != TCL_OK) {
		uni_msg_destroy(m);
		return (TCL_ERROR);
	}

	uni_uni_input(stk->uni, UNIAPI_SETUP_response, cookie, m);
	uni_work(stk->uni);
	return (TCL_OK);
}

static int
cmd_uni_alerting_request(ClientData data, Tcl_Interp *interp,
    int argc, const char *argv[])
{
	struct stk *stk = (struct stk *)data;
	u_int32_t cookie;
	struct uni_msg *m;
	struct uniapi_alerting_request *req;

	if (argc < 5) {
		set_error(interp, "usage: '%s <cookie> <cref> <cref-flag> <act> [<ies>...]'", argv[0]);
		return (TCL_ERROR);
	}
	if (get_cookie(interp, argv[1], &cookie) != TCL_OK)
		return (TCL_ERROR);

	if ((m = uni_msg_alloc(sizeof(*req))) == NULL)
		err(1, NULL);
	req = uni_msg_rptr(m, struct uniapi_alerting_request *);
	m->b_wptr += sizeof(*req);

	memset(req, 0, sizeof(*req));

	if (parse_cref(interp, argc - 2, (const char **)(argv + 2),
	    &req->alerting.hdr.cref) != TCL_OK ||
	    parse_msgact(interp, argv[4], &req->alerting.hdr) != TCL_OK ||
	    parse_msg_alerting(interp, argc - 5, argv + 5, &req->alerting,
	    uni_context(stk->uni)) != TCL_OK) {
		uni_msg_destroy(m);
		return (TCL_ERROR);
	}

	uni_uni_input(stk->uni, UNIAPI_ALERTING_request, cookie, m);
	uni_work(stk->uni);
	return (TCL_OK);
}

static int
cmd_uni_proceeding_request(ClientData data, Tcl_Interp *interp,
    int argc, const char *argv[])
{
	struct stk *stk = (struct stk *)data;
	u_int32_t cookie;
	struct uni_msg *m;
	struct uniapi_proceeding_request *req;

	if (argc < 5) {
		set_error(interp, "usage: '%s <cookie> <cref> <cref-flag> <act> [<ies>...]'", argv[0]);
		return (TCL_ERROR);
	}
	if (get_cookie(interp, argv[1], &cookie) != TCL_OK)
		return (TCL_ERROR);

	if ((m = uni_msg_alloc(sizeof(*req))) == NULL)
		err(1, NULL);
	req = uni_msg_rptr(m, struct uniapi_proceeding_request *);
	m->b_wptr += sizeof(*req);

	memset(req, 0, sizeof(*req));

	if (parse_cref(interp, argc - 2, (const char **)(argv + 2),
	    &req->call_proc.hdr.cref) != TCL_OK ||
	    parse_msgact(interp, argv[4], &req->call_proc.hdr) != TCL_OK ||
	    parse_msg_call_proc(interp, argc - 5, argv + 5, &req->call_proc,
	    uni_context(stk->uni)) != TCL_OK) {
		uni_msg_destroy(m);
		return (TCL_ERROR);
	}

	uni_uni_input(stk->uni, UNIAPI_PROCEEDING_request, cookie, m);
	uni_work(stk->uni);
	return (TCL_OK);
}

static int
cmd_uni_release_request(ClientData data, Tcl_Interp *interp,
    int argc, const char *argv[])
{
	struct stk *stk = (struct stk *)data;
	u_int32_t cookie;
	struct uni_msg *m;
	struct uniapi_release_request *req;

	if (argc < 5) {
		set_error(interp, "usage: '%s <cookie> <cref> <cref-flag> <act> [<ies>...]'", argv[0]);
		return (TCL_ERROR);
	}
	if (get_cookie(interp, argv[1], &cookie) != TCL_OK)
		return (TCL_ERROR);

	if ((m = uni_msg_alloc(sizeof(*req))) == NULL)
		err(1, NULL);
	req = uni_msg_rptr(m, struct uniapi_release_request *);
	m->b_wptr += sizeof(*req);

	memset(req, 0, sizeof(*req));

	if (parse_cref(interp, argc - 2, (const char **)(argv + 2),
	    &req->release.hdr.cref) != TCL_OK ||
	    parse_msgact(interp, argv[4], &req->release.hdr) != TCL_OK ||
	    parse_msg_release(interp, argc - 5, argv + 5, &req->release,
	    uni_context(stk->uni)) != TCL_OK) {
		uni_msg_destroy(m);
		return (TCL_ERROR);
	}

	uni_uni_input(stk->uni, UNIAPI_RELEASE_request, cookie, m);
	uni_work(stk->uni);
	return (TCL_OK);
}

static int
cmd_uni_release_response(ClientData data, Tcl_Interp *interp,
    int argc, const char *argv[])
{
	struct stk *stk = (struct stk *)data;
	u_int32_t cookie;
	struct uni_msg *m;
	struct uniapi_release_response *req;

	if (argc < 5) {
		set_error(interp, "usage: '%s <cookie> <cref> <cref-flag> <act> [<ies>...]'", argv[0]);
		return (TCL_ERROR);
	}
	if (get_cookie(interp, argv[1], &cookie) != TCL_OK)
		return (TCL_ERROR);

	if ((m = uni_msg_alloc(sizeof(*req))) == NULL)
		err(1, NULL);
	req = uni_msg_rptr(m, struct uniapi_release_response *);
	m->b_wptr += sizeof(*req);

	memset(req, 0, sizeof(*req));

	if (parse_cref(interp, argc - 2, (const char **)(argv + 2),
	    &req->release_compl.hdr.cref) != TCL_OK ||
	    parse_msgact(interp, argv[4], &req->release_compl.hdr) != TCL_OK ||
	    parse_msg_release_compl(interp, argc - 5, argv + 5, &req->release_compl,
	    uni_context(stk->uni)) != TCL_OK) {
		uni_msg_destroy(m);
		return (TCL_ERROR);
	}

	uni_uni_input(stk->uni, UNIAPI_RELEASE_response, cookie, m);
	uni_work(stk->uni);
	return (TCL_OK);
}

static int
cmd_uni_notify_request(ClientData data, Tcl_Interp *interp,
    int argc, const char *argv[])
{
	struct stk *stk = (struct stk *)data;
	u_int32_t cookie;
	struct uni_msg *m;
	struct uniapi_notify_request *req;

	if (argc < 5) {
		set_error(interp, "usage: '%s <cookie> <cref> <cref-flag> <act> [<ies>...]'", argv[0]);
		return (TCL_ERROR);
	}
	if (get_cookie(interp, argv[1], &cookie) != TCL_OK)
		return (TCL_ERROR);

	if ((m = uni_msg_alloc(sizeof(*req))) == NULL)
		err(1, NULL);
	req = uni_msg_rptr(m, struct uniapi_notify_request *);
	m->b_wptr += sizeof(*req);

	memset(req, 0, sizeof(*req));

	if (parse_cref(interp, argc - 2, (const char **)(argv + 2),
	    &req->notify.hdr.cref) != TCL_OK ||
	    parse_msgact(interp, argv[4], &req->notify.hdr) != TCL_OK ||
	    parse_msg_notify(interp, argc - 5, argv + 5, &req->notify,
	    uni_context(stk->uni)) != TCL_OK) {
		uni_msg_destroy(m);
		return (TCL_ERROR);
	}

	uni_uni_input(stk->uni, UNIAPI_NOTIFY_request, cookie, m);
	uni_work(stk->uni);
	return (TCL_OK);
}

static int
cmd_uni_status_enquiry_request(ClientData data, Tcl_Interp *interp,
    int argc, const char *argv[])
{
	struct stk *stk = (struct stk *)data;
	u_int32_t cookie;
	struct uni_msg *m;
	struct uniapi_status_enquiry_request *req;
	union uni_ieall ie;
	enum uni_ietype type;
	int i;

	if (argc < 2) {
		set_error(interp, "usage: '%s <cookie> <cref> <cref-flag> [<ies>...]'", argv[0]);
		return (TCL_ERROR);
	}
	if (get_cookie(interp, argv[1], &cookie) != TCL_OK)
		return (TCL_ERROR);
	if ((m = uni_msg_alloc(sizeof(*req))) == NULL)
		err(1, NULL);
	req = uni_msg_rptr(m, struct uniapi_status_enquiry_request *);
	m->b_wptr += sizeof(*req);

	memset(req, 0, sizeof(*req));

	for (i = 2; i < argc; i++) {
		if (parse_ie(interp, argv[i], &ie, &type) != TCL_OK) {
			uni_msg_destroy(m);
			return (TCL_ERROR);
		}
		switch (type) {

		  case UNI_IE_EPREF:
			req->epref = ie.epref;
			break;

		  default:
			set_error(interp, "bad IE for status_enq.request %02x",
			    type);
			uni_msg_destroy(m);
			return (TCL_ERROR);
		}
	}

	uni_uni_input(stk->uni, UNIAPI_STATUS_ENQUIRY_request, cookie, m);
	uni_work(stk->uni);
	return (TCL_OK);
}

static int
cmd_uni_add_party_request(ClientData data, Tcl_Interp *interp,
    int argc, const char *argv[])
{
	struct stk *stk = (struct stk *)data;
	u_int32_t cookie;
	struct uni_msg *m;
	struct uniapi_add_party_request *req;

	if (argc < 5) {
		set_error(interp, "usage: '%s <cookie> <cref> <cref-flag> <act> [<ies>...]'", argv[0]);
		return (TCL_ERROR);
	}
	if (get_cookie(interp, argv[1], &cookie) != TCL_OK)
		return (TCL_ERROR);

	if ((m = uni_msg_alloc(sizeof(*req))) == NULL)
		err(1, NULL);
	req = uni_msg_rptr(m, struct uniapi_add_party_request *);
	m->b_wptr += sizeof(*req);

	memset(req, 0, sizeof(*req));

	if (parse_cref(interp, argc - 2, (const char **)(argv + 2),
	    &req->add.hdr.cref) != TCL_OK ||
	    parse_msgact(interp, argv[4], &req->add.hdr) != TCL_OK ||
	    parse_msg_add_party(interp, argc - 5, argv + 5, &req->add,
	    uni_context(stk->uni)) != TCL_OK) {
		uni_msg_destroy(m);
		return (TCL_ERROR);
	}

	uni_uni_input(stk->uni, UNIAPI_ADD_PARTY_request, cookie, m);
	uni_work(stk->uni);
	return (TCL_OK);
}

static int
cmd_uni_party_alerting_request(ClientData data, Tcl_Interp *interp,
    int argc, const char *argv[])
{
	struct stk *stk = (struct stk *)data;
	u_int32_t cookie;
	struct uni_msg *m;
	struct uniapi_party_alerting_request *req;

	if (argc < 5) {
		set_error(interp, "usage: '%s <cookie> <cref> <cref-flag> <act> [<ies>...]'", argv[0]);
		return (TCL_ERROR);
	}
	if (get_cookie(interp, argv[1], &cookie) != TCL_OK)
		return (TCL_ERROR);

	if ((m = uni_msg_alloc(sizeof(*req))) == NULL)
		err(1, NULL);
	req = uni_msg_rptr(m, struct uniapi_party_alerting_request *);
	m->b_wptr += sizeof(*req);

	memset(req, 0, sizeof(*req));

	if (parse_cref(interp, argc - 2, (const char **)(argv + 2),
	    &req->alert.hdr.cref) != TCL_OK ||
	    parse_msgact(interp, argv[4], &req->alert.hdr) != TCL_OK ||
	    parse_msg_party_alerting(interp, argc - 5, argv + 5, &req->alert,
	    uni_context(stk->uni)) != TCL_OK) {
		uni_msg_destroy(m);
		return (TCL_ERROR);
	}

	uni_uni_input(stk->uni, UNIAPI_PARTY_ALERTING_request, cookie, m);
	uni_work(stk->uni);
	return (TCL_OK);
}

static int
cmd_uni_add_party_ack_request(ClientData data, Tcl_Interp *interp,
    int argc, const char *argv[])
{
	struct stk *stk = (struct stk *)data;
	u_int32_t cookie;
	struct uni_msg *m;
	struct uniapi_add_party_ack_request *req;

	if (argc < 5) {
		set_error(interp, "usage: '%s <cookie> <cref> <cref-flag> <act> [<ies>...]'", argv[0]);
		return (TCL_ERROR);
	}
	if (get_cookie(interp, argv[1], &cookie) != TCL_OK)
		return (TCL_ERROR);

	if ((m = uni_msg_alloc(sizeof(*req))) == NULL)
		err(1, NULL);
	req = uni_msg_rptr(m, struct uniapi_add_party_ack_request *);
	m->b_wptr += sizeof(*req);

	memset(req, 0, sizeof(*req));

	if (parse_cref(interp, argc - 2, (const char **)(argv + 2),
	    &req->ack.hdr.cref) != TCL_OK ||
	    parse_msgact(interp, argv[4], &req->ack.hdr) != TCL_OK ||
	    parse_msg_add_party_ack(interp, argc - 5, argv + 5, &req->ack,
	    uni_context(stk->uni)) != TCL_OK) {
		uni_msg_destroy(m);
		return (TCL_ERROR);
	}

	uni_uni_input(stk->uni, UNIAPI_ADD_PARTY_ACK_request, cookie, m);
	uni_work(stk->uni);
	return (TCL_OK);
}

static int
cmd_uni_add_party_rej_request(ClientData data, Tcl_Interp *interp,
    int argc, const char *argv[])
{
	struct stk *stk = (struct stk *)data;
	u_int32_t cookie;
	struct uni_msg *m;
	struct uniapi_add_party_rej_request *req;

	if (argc < 5) {
		set_error(interp, "usage: '%s <cookie> <cref> <cref-flag> <act> [<ies>...]'", argv[0]);
		return (TCL_ERROR);
	}
	if (get_cookie(interp, argv[1], &cookie) != TCL_OK)
		return (TCL_ERROR);

	if ((m = uni_msg_alloc(sizeof(*req))) == NULL)
		err(1, NULL);
	req = uni_msg_rptr(m, struct uniapi_add_party_rej_request *);
	m->b_wptr += sizeof(*req);

	memset(req, 0, sizeof(*req));

	if (parse_cref(interp, argc - 2, (const char **)(argv + 2),
	    &req->rej.hdr.cref) != TCL_OK ||
	    parse_msgact(interp, argv[4], &req->rej.hdr) != TCL_OK ||
	    parse_msg_add_party_rej(interp, argc - 5, argv + 5, &req->rej,
	    uni_context(stk->uni)) != TCL_OK) {
		uni_msg_destroy(m);
		return (TCL_ERROR);
	}

	uni_uni_input(stk->uni, UNIAPI_ADD_PARTY_REJ_request, cookie, m);
	uni_work(stk->uni);
	return (TCL_OK);
}

static int
cmd_uni_drop_party_request(ClientData data, Tcl_Interp *interp,
    int argc, const char *argv[])
{
	struct stk *stk = (struct stk *)data;
	u_int32_t cookie;
	struct uni_msg *m;
	struct uniapi_drop_party_request *req;

	if (argc < 5) {
		set_error(interp, "usage: '%s <cookie> <cref> <cref-flag> <act> [<ies>...]'", argv[0]);
		return (TCL_ERROR);
	}
	if (get_cookie(interp, argv[1], &cookie) != TCL_OK)
		return (TCL_ERROR);

	if ((m = uni_msg_alloc(sizeof(*req))) == NULL)
		err(1, NULL);
	req = uni_msg_rptr(m, struct uniapi_drop_party_request *);
	m->b_wptr += sizeof(*req);

	memset(req, 0, sizeof(*req));

	if (parse_cref(interp, argc - 2, (const char **)(argv + 2),
	    &req->drop.hdr.cref) != TCL_OK ||
	    parse_msgact(interp, argv[4], &req->drop.hdr) != TCL_OK ||
	    parse_msg_drop_party(interp, argc - 5, argv + 5, &req->drop,
	    uni_context(stk->uni)) != TCL_OK) {
		uni_msg_destroy(m);
		return (TCL_ERROR);
	}

	uni_uni_input(stk->uni, UNIAPI_DROP_PARTY_request, cookie, m);
	uni_work(stk->uni);
	return (TCL_OK);
}

static int
cmd_uni_drop_party_ack_request(ClientData data, Tcl_Interp *interp,
    int argc, const char *argv[])
{
	struct stk *stk = (struct stk *)data;
	u_int32_t cookie;
	struct uni_msg *m;
	struct uniapi_drop_party_ack_request *req;

	if (argc < 5) {
		set_error(interp, "usage: '%s <cookie> <cref> <cref-flag> <act> [<ies>...]'", argv[0]);
		return (TCL_ERROR);
	}
	if (get_cookie(interp, argv[1], &cookie) != TCL_OK)
		return (TCL_ERROR);

	if ((m = uni_msg_alloc(sizeof(*req))) == NULL)
		err(1, NULL);
	req = uni_msg_rptr(m, struct uniapi_drop_party_ack_request *);
	m->b_wptr += sizeof(*req);

	memset(req, 0, sizeof(*req));

	if (parse_cref(interp, argc - 2, (const char **)(argv + 2),
	    &req->ack.hdr.cref) != TCL_OK ||
	    parse_msgact(interp, argv[4], &req->ack.hdr) != TCL_OK ||
	    parse_msg_drop_party_ack(interp, argc - 5, argv + 5, &req->ack,
	    uni_context(stk->uni)) != TCL_OK) {
		uni_msg_destroy(m);
		return (TCL_ERROR);
	}

	uni_uni_input(stk->uni, UNIAPI_DROP_PARTY_request, cookie, m);
	uni_work(stk->uni);
	return (TCL_OK);
}

static int
cmd_uni_abort_call_request(ClientData data, Tcl_Interp *interp,
    int argc, const char *argv[])
{
	struct stk *stk = (struct stk *)data;
	u_int32_t cookie;
	struct uni_msg *m;
	struct uniapi_abort_call_request *req;

	if (argc != 4) {
		set_error(interp, "usage: '%s <cookie> <cref> <cref-flag>'", argv[0]);
		return (TCL_ERROR);
	}
	if (get_cookie(interp, argv[1], &cookie) != TCL_OK)
		return (TCL_ERROR);

	if ((m = uni_msg_alloc(sizeof(*req))) == NULL)
		err(1, NULL);
	req = uni_msg_rptr(m, struct uniapi_abort_call_request *);
	m->b_wptr += sizeof(*req);

	memset(req, 0, sizeof(*req));

	if (parse_cref(interp, argc - 2, (const char **)(argv + 2),
	    &req->cref) != TCL_OK) {
		uni_msg_destroy(m);
		return (TCL_ERROR);
	}

	uni_uni_input(stk->uni, UNIAPI_ABORT_CALL_request, cookie, m);
	uni_work(stk->uni);
	return (TCL_OK);
}

void
user_tcl_init(struct stk *stk)
{
	Tcl_CreateCommand(stk->interp, "uni_debug",
	    cmd_uni_debug, stk, NULL);
	Tcl_CreateCommand(stk->interp, "uni_status",
	    cmd_uni_status, stk, NULL);
	Tcl_CreateCommand(stk->interp, "uni_get_config",
	    cmd_uni_get_config, stk, NULL);
	Tcl_CreateCommand(stk->interp, "uni_set_config",
	    cmd_uni_set_config, stk, NULL);

	Tcl_CreateCommand(stk->interp, "uni_link_establish_request",
	    cmd_uni_link_establish_request, stk, NULL);
	Tcl_CreateCommand(stk->interp, "uni_link_release_request",
	    cmd_uni_link_release_request, stk, NULL);

	Tcl_CreateCommand(stk->interp, "uni_reset_request",
	    cmd_uni_reset_request, stk, NULL);
	Tcl_CreateCommand(stk->interp, "uni_reset_response",
	    cmd_uni_reset_response, stk, NULL);
	Tcl_CreateCommand(stk->interp, "uni_reset_error_response",
	    cmd_uni_reset_error_response, stk, NULL);

	Tcl_CreateCommand(stk->interp, "uni_setup_request",
	    cmd_uni_setup_request, stk, NULL);
	Tcl_CreateCommand(stk->interp, "uni_setup_response",
	    cmd_uni_setup_response, stk, NULL);
	Tcl_CreateCommand(stk->interp, "uni_alerting_request",
	    cmd_uni_alerting_request, stk, NULL);
	Tcl_CreateCommand(stk->interp, "uni_proceeding_request",
	    cmd_uni_proceeding_request, stk, NULL);
	Tcl_CreateCommand(stk->interp, "uni_release_request",
	    cmd_uni_release_request, stk, NULL);
	Tcl_CreateCommand(stk->interp, "uni_release_response",
	    cmd_uni_release_response, stk, NULL);
	Tcl_CreateCommand(stk->interp, "uni_notify_request",
	    cmd_uni_notify_request, stk, NULL);
	Tcl_CreateCommand(stk->interp, "uni_status_enquiry_request",
	    cmd_uni_status_enquiry_request, stk, NULL);

	Tcl_CreateCommand(stk->interp, "uni_add_party_request",
	    cmd_uni_add_party_request, stk, NULL);
	Tcl_CreateCommand(stk->interp, "uni_party_alerting_request",
	    cmd_uni_party_alerting_request, stk, NULL);
	Tcl_CreateCommand(stk->interp, "uni_add_party_ack_request",
	    cmd_uni_add_party_ack_request, stk, NULL);
	Tcl_CreateCommand(stk->interp, "uni_add_party_rej_request",
	    cmd_uni_add_party_rej_request, stk, NULL);
	Tcl_CreateCommand(stk->interp, "uni_drop_party_request",
	    cmd_uni_drop_party_request, stk, NULL);
	Tcl_CreateCommand(stk->interp, "uni_drop_party_ack_request",
	    cmd_uni_drop_party_ack_request, stk, NULL);

	Tcl_CreateCommand(stk->interp, "uni_abort_call_request",
	    cmd_uni_abort_call_request, stk, NULL);
}
